package edu.mit.csail.aeolus.api;

import java.io.FileNotFoundException;
import java.io.IOException;

import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
// import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;

/**
 * This class is used to create and manipulate AeolusFiles.
 */

public class AeolusFile {
	
	private UserProxyInterface proxy;
	private String hostname;
	private String path;
	
	public AeolusFile(String hostname, String path)
	{
		this.hostname = hostname;
		this.path = path;
		this.proxy = UserProxyMap.getMyProxyObject();
	}
	
	
    /**
     * Returns the secrecy label of the file
     * @return - a Label object representing the secrecy label of the file
     * @throws FileNotFoundException - if the file does not exist on hostname
     * @throws IOException - if an I/O error occurs while retrieving the label
     * @throws InfoFlowControlException - if caller labels don't allow reading the directories in the path
     * @throws AeolusException - if communication with hostname fails
     */
	public AeolusLabel getSecrecy() throws FileNotFoundException, InfoFlowControlException, IOException, AeolusException
	{
		return new AeolusLabel(proxy.getSecrecyLabel(hostname, path));
	}
	
	
    /**
     * Returns the integrity label of the file
     * @return - a Label object representing the integrity label of the file
     * @throws FileNotFoundException - if the file does not exist on hostname
     * @throws IOException - if an I/O error occurs while retrieving the label
     * @throws InfoFlowControlException - if caller labels don't allow reading the directories in the path
     * @throws AeolusException - if communication with hostname fails
     */
	public AeolusLabel getIntegrity() throws FileNotFoundException, InfoFlowControlException, IOException, AeolusException
	{
		return new AeolusLabel(proxy.getIntegrityLabel(hostname, path));
	}
	
	
    /**
     * Creates a new file if the parent directory exists and there is no file with the given name.
	 * See reference manual for constraints on labels.
	 * Returns true if file was successfully created and false if file with same name existed.
	 * Please see the reference manual for rules concerning the root directory.
     * @param sLabel - file's secrecy label
     * @param iLabel - file's integrity label
     * @return true if the file was successfully created
     * @throws InfoFlowControlException - if caller is not allowed to read and write the file's parent directory or
     * if it is not allowed to read a directory in the path.  Also the file's labels cannot be less constraining
     * than the thread's labels.  Please see the reference manual for rules concerning the root directory.
     * @throws FileNotFoundException - if the parent directory (or some part of the path) does not exist
     * @throws IOException - if an I/O error occurs while creating the file
     * @throws AeolusException - if there was an error communicating with hostname
     */
	public boolean createNewFile(AeolusLabel sLabel, AeolusLabel iLabel) throws FileNotFoundException, InfoFlowControlException, IOException, AeolusException
	{
		return proxy.createFile(hostname, path, sLabel.getLabel(), iLabel.getLabel());
	}

	
    /**
     * Creates a new directory if the parent directory exists and there is no file with the given name.
     * See reference manual for label constraints.
     * Returns true if directory was successfully created and false if file with same name existed.
     * @param sLabel - directory's secrecy label
     * @param iLabel - directory's integrity label
     * @return true if the directory was successfully created
     * @throws InfoFlowControlException - if caller is not allowed to read and write directory's parent directory
     * or if it is not allowed to read a directory in the path. Also the directory's labels cannot be less
     * constraining than the tread labels.  Please see the reference manual for rules concerning the root directory.
     * @throws FileNotFoundException - if parent directory (or some part of the path) does not exist
     * @throws IOException - if an I/O error occurs while creating the directory
     * @throws AeolusException - if there was an error communicating with hostname
     */
	public boolean mkdir(AeolusLabel sLabel, AeolusLabel iLabel) throws FileNotFoundException, InfoFlowControlException, IOException, AeolusException
	{
		return proxy.createDir(hostname, path, sLabel.getLabel(), iLabel.getLabel());
	}
	
	
    /**
     * Lists the contents of the directory as an array of file names.
     * Caller labels must allow reading the directory.
     * @return contents of the directory as an array of file names
     * @throws FileNotFoundException - if the directory does not exist and the caller labels allow reading its parent directory
     * @throws IOException - if an I/O error occurred while retrieving directory labels
     * @throws InfoFlowControlException - if caller labels don't allow reading the directory or 
     * in the case the directory is missing, reading a higher parent directory of the given directory
     * @throws AeolusException - if a failure occurs while communicating with hostname
     */
	public String[] list() throws FileNotFoundException, InfoFlowControlException, IOException, AeolusException
	{
		return proxy.listDir(hostname, path);
	}
	
	
    /**
     * Deletes the file.
     * Caller labels must allow reading and writing of the parent directory.
     * @return true if the file was successfully deleted
     * @throws FileNotFoundException - if the file does not exist and the caller labels allow reading
     * its parent directory
     * @throws IOException - if an I/O error occurred while reading directory labels
     * @throws InfoFlowControlException - if caller labels don't allow reading and writing of the parent directory
     * or reading a higher parent directory in the case the file is missing
     * @throws AeolusException - if a failure occurs while communicating with hostname
     */
	public boolean delete() throws FileNotFoundException, InfoFlowControlException, IOException, AeolusException
	{
		return proxy.delete(hostname, path);
	}
	
}
