package edu.mit.csail.aeolus.api;

import java.io.FileNotFoundException;
import java.io.IOException;

import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.filesystem.FileSystemClientInterface;
import edu.mit.csail.aeolus.filesystem.FileSystemReply;
// import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;

/**
 * An Aeolus file input stream can be used for reading data from a given file.
 * 
 * <p>The stream will only be created and opened if the requester labels
 * allow reading from the file.
 * 
 * <p> If the file contents are changed after the stream has been opened, the subsequent read
 * operation from this stream will fail.
 * 
 * <p> If the caller labels change and no longer allow the read, the stream
 * will be closed.
 * 
 * @author viq
 *
 */
public class AeolusFileInputStream {
	private int handle;
	private UserProxyInterface proxy;
	private FileSystemClientInterface fsClient;
	
	private String filePath;
	private String hostname;
	private AeolusLabel fileSLabel;
	private AeolusLabel fileILabel;

	private int position;
	private boolean isOpened;
	
    /**
     * Creates a filestream for reading from the file specified by filePath on hostname
     * @param hostname - node where the file is stored
     * @param filePath - path to the file on hostname
     * @throws InfoFlowControlException - if caller labels don't allow reading from the file 
     * or reading from a higher parent directory if the file is missing
     * @throws FileNotFoundException - if the file does not exist on the hostname
     * @throws AeolusException - if a FileSystemClient can not be created
     * @throws IOException - if an I/O error occurs while opening the file stream
     * (e.g. when retrieving file labels)
     */
	public AeolusFileInputStream(String hostname, String filePath)
		throws FileNotFoundException, IOException, AeolusException
	{	
		this.hostname = hostname;
		this.filePath = filePath;
		this.proxy = UserProxyMap.getMyProxyObject();
		this.fsClient = proxy.getFSClient(hostname);
		FileSystemReply<Integer> reply = proxy.openFileInputStream(hostname, filePath);
		
		this.handle = reply.opResult;
		this.fileSLabel = new AeolusLabel(reply.fileSLabel);
		this.fileILabel = new AeolusLabel(reply.fileILabel);
		this.position = 0;
		this.isOpened = true;
	}
	
	
	/**
	 * Reads bytes from the file into the specified buffer.
	 * Up to buffer.length bytes will be read from the file.
	 * @param buffer - array to store the data into
	 * @return the actual number of bytes read
	 * @throws IOException - if an I/O error occurred while reading from the file 
	 * (e.g. the stream has been closed) 
	 * @throws InfoFlowControlException - if labels don't allow the read
	 */
	public int read(byte[] buffer) throws IOException, AeolusException, InfoFlowControlException
	{
		return this.read(buffer, 0, buffer.length);
	}
	
	/**
	 * Reads bytes from the file into the specified buffer.
	 * Up to count bytes will be stored in the buffer starting at the given offset in the buffer.
	 * @return the actual number of bytes read or -1 if end of file has been reached
	 * @throws IOException  - if an I/O error occurred while reading from the file
	 * (e.g. the stream has been closed) 
	 * @throws InfoFlowControlException - if labels don't allow the read
	 * @throws IoException - if the stream is closed
	 */
	public int read(byte[] buffer, int offset, int count)
		throws IOException, AeolusException, InfoFlowControlException
	{
		if(!this.isOpened)
			throw new IOException("This stream has been closed");
		
		if(!proxy.canReadFromStream(handle, fileSLabel.getLabel(), fileILabel.getLabel()))
		{
			this.close();
			throw new InfoFlowControlException("Not authorized to read file");
		}
		
		if(position == -1)
			return -1; // end of file has been reached
		
		int bytesRead = fsClient.readFilestream(handle, filePath, position, buffer, offset, count);
		if(bytesRead != -1)
			position += bytesRead;
		else
			position = -1;
		
		return bytesRead;
	}

	/**
	 * Closes the stream.
	 * Any subsequent attempts to access this stream will fail with an IOException.
	 * @throws IOException - if this stream has already been closed
	 */
	public void close() throws IOException, AeolusException
	{
		if(!this.isOpened)
			throw new IOException("This stream has been closed");
		
		this.proxy.closeFilestream(handle, filePath, hostname);
		this.isOpened = false;
	}
}
