package edu.mit.csail.aeolus.api;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;

import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.filesystem.FileSystemClientInterface;
import edu.mit.csail.aeolus.filesystem.FileSystemReply;
// import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;

/**
 * An Aeolus file output stream can be used for writing data to a given file.
 * 
 * <p>The stream will only be created and opened if the requester labels
 * allow writing to the file.
 * 
 * <p> The file stream buffers all the writes. Closing the file stream causes 
 * data to be actually written to the file. 
 * 
 * <p> If the caller labels change and no longer allow the write, the stream
 * will be closed and any data previously written to the stream will be written to the file.
 * 
 * @author viq
 *
 */

public class AeolusFileOutputStream 
{	
	private boolean isOpened;
	
	
	private int handle;
	private UserProxyInterface proxy;
	private FileSystemClientInterface fsClient;
	
	private String hostname;
	private String filePath;
	private AeolusLabel fileSLabel;
	private AeolusLabel fileILabel;
	

    /**
     * Creates a filestream for writing to the file specified by filePath on hostname
     * @param hostname - node where the file is stored
     * @param filePath - path to the file on hostname
     * @param appendMode - if set to true, the bytes will be written to the end of the file
     * @throws InfoFlowControlException - if caller labels don't allow writing to the file 
     * or reading from a parent directory if the file is missing
     * @throws FileNotFoundException - if the file does not exist on the hostname
     * @throws AeolusException - if a FileSystemClient can not be created
     * @throws IOException - if an I/O error occurs while opening the file stream
     * (e.g. when retrieving file labels)
     */
	public AeolusFileOutputStream(String hostname, String filePath, boolean appendMode) throws AeolusException, FileNotFoundException, IOException
	{	
		this.hostname = hostname;
		this.filePath = filePath;
		this.proxy = UserProxyMap.getMyProxyObject();
		this.fsClient = proxy.getFSClient(hostname);
		FileSystemReply<Integer> reply = proxy.openFileOutputStream(hostname, filePath, appendMode);
		
		this.handle = reply.opResult;
		this.fileSLabel = new AeolusLabel(reply.fileSLabel);
		this.fileILabel = new AeolusLabel(reply.fileILabel);
		this.isOpened = true;
	}
	
	/**
	 * Writes buffer.length bytes to the stream from the given buffer.
	 * @param buffer - byte array storing the data to be written
	 * @throws IOException - if an I/O error occurred while writing to the stream
	 * (e.g. the stream has been closed)
	 * @throws InfoFlowControlException - if caller labels no longer allow writing to the file
	 */
	public void write(byte[] buffer) throws IOException, InfoFlowControlException, AeolusException
	{
		this.write(buffer, 0, buffer.length);
	}
	
	/**
	 * Writes count bytes into the stream from the buffer starting at the given offset.
	 * @throws IOException - if an I/O error occurred while writing to the stream
	 * (e.g. the stream has been closed)
	 * @throws InfoFlowControlException - if caller labels no longer allow writing to the file
	 */
	public void write(byte[] buffer, int offset, int count) throws IOException, InfoFlowControlException, AeolusException
	{
		if(!this.isOpened)
			throw new IOException("This stream has been closed");
		if(!proxy.canWriteToStream(handle, fileSLabel.getLabel(), fileILabel.getLabel()))
		{
			this.close();
			throw new InfoFlowControlException("Not authorized to write file");
		}
		
		if(offset < 0)
			throw new IOException("Offset cannot be negative");
		if(offset >= buffer.length)
			throw new IndexOutOfBoundsException("Offset exceeds the length of the buffer");
		if((offset+count) > buffer.length)
			throw new IndexOutOfBoundsException("Offest and count exceed the length of the buffer");
		
		this.fsClient.writeFilestream(handle, filePath, buffer, offset, count);
	}

	/**
	 * Closes the file stream.
	 * Any subsequent operations on this stream will fail with an IOException.
	 * @throws IOException - if an I/O error occurs while writing the data to the file
	 * or the file is already closed
	 */
	public void close() throws IOException, AeolusException
	{
		if(!this.isOpened)
			throw new IOException("This stream has been closed");
		
		this.proxy.closeFilestream(handle, filePath, hostname);
		
		this.isOpened = false;
	}
}
