package edu.mit.csail.aeolus.api;

// import edu.mit.csail.aeolus.api3.*;
// import edu.mit.csail.aeolus.infoflow3.*;

import java.util.List;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.concurrent.Callable;

import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.exceptions.RpcException;

import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.infoflow.PrincipalID;

import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.runtime.UserLoggingProxy;

/**
 *  This class contains utilities for accessing and manipulating the labels of the current thread;
 *  it also provides launch, shutdown, registerService, getService, fork, call, getRoot and setRoot.
 */

public final class AeolusLib {

	private AeolusLib() {}

	/**
	 * Returns caller's PID
	 */
	public static PID getPID() {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		InfoFlowState ifs = userPO.getInfoFlowState();
		return new PID(ifs.pid);
	}

	/**
	 * Returns caller's secrecy label
	 */
	public static AeolusLabel getSecrecy() {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		InfoFlowState ifs = userPO.getInfoFlowState();
		return new AeolusLabel(ifs.SLabel);
	}
	/**
	 * Returns caller's integrity label
	 */
	public static AeolusLabel getIntegrity() {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		InfoFlowState ifs = userPO.getInfoFlowState();
		return new AeolusLabel(ifs.ILabel);
	}

	/**
	 * Returns caller's basis
	 */
	public static List<Integer> getBasis() {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		InfoFlowState ifs = userPO.getInfoFlowState();
		return ifs.basis;
	}

	/**
	 * Adds AeolusTag to the caller's secrecy label
	 * @param t - AeolusTag to add to the secrecy label
	 * @throws AeolusException - if an attempt to change a label occurs
	 * while the thread is running inside a shared object
	 */
	public static void addSecrecy(AeolusTag t) throws AeolusException {
		// System.out.println("addSecrecy " + t);
		TagID tid = t.getTagID();
		// System.out.println("addSecrecy " + t + " " + tid);
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		// System.out.println("addSecrecy " + t + " " + tid + " " + userPO);
			userPO.addSecrecy(tid);
	}

	/**
	 * Adds each AeolusTag in tags to the caller's secrecy label
	 * @param tags - List of AeolusTags to add to the secrecy label
	 * @throws AeolusException - if an attempt to change a label occurs
	 * while the thread is running inside a shared object
	 */
	public static void addSecrecy(List<AeolusTag> tags) throws AeolusException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		for(AeolusTag t : tags){
			TagID tid = t.getTagID();
			userPO.addSecrecy(tid); 
		}
	}

        /**
	 * Remove AeolusTag from the caller's secrecy label
	 * @param t - AeolusTag to remove from the secrecy label
	 * @throws AuthorityException - if caller principal does not have
	 * authority for t
	 * @throws AeolusException - if an attempt to change a label occurs
	 * while the thread is running inside a shared object
	 */
	public static void declassify(AeolusTag t) throws AeolusException {
		TagID tid = t.getTagID();
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		userPO.declassify(tid);
	}

        /**
	 * Remove AeolusTags from the caller's secrecy label
	 * @param tags - List of AeolusTags to remove from the secrecy label
	 * @throws AuthorityException - if caller principal does not have
	 * authority for each tag in tags
	 * @throws AeolusException - if an attempt to change a label occurs
	 * while the thread is running inside a shared object
	 */
	public static void declassify(List<AeolusTag> tags) throws AuthorityException, AeolusException {
		for(AeolusTag t : tags){
		TagID tid = t.getTagID();
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		userPO.declassify(tid);
		}
	}

        /**
	 * Removes AeolusTag from the caller's integrity label
	 * @param t - AeolusTag to remove from integrity label
	 * @throws AeolusException - if an attempt to change a label occurs
	 * while the thread is running inside a shared object
	 */
	public static void removeIntegrity(AeolusTag t) throws AeolusException {
		TagID tid = t.getTagID();
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		// try {
			userPO.removeIntegrity(tid); // }
		// catch (Exception e) {
                //      	throw new AeolusException(e);
                // }
	}

        /**
	 * Removes AeolusTags from the caller's integrity label
	 * @param tags - List of AeolusTag to remove from integrity label
	 * @throws AuthorityException - if an attempt to use authority occurs while in shared state
	 * @throws AeolusException - if caller is inside a shared state object
	 */
	public static void removeIntegrity(List<AeolusTag> tags) throws AuthorityException, AeolusException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		for (AeolusTag t : tags) {
			TagID tid = t.getTagID();
			// try { userPO.removeIntegrity(tid); }
			userPO.removeIntegrity(tid);
			//catch (Exception e) {
                        //	throw new AeolusException(e);
                	//}
		}
	}

        /**
	 * Add AeolusTag to the caller's integrity label
	 * @param t - AeolusTag to add to the integrity label
	 * @throws AuthorityException - if caller principal does not have authority for t
	 * @throws AeolusException - if caller is inside a shared state object
	 */
	public static void endorse(AeolusTag t) throws AuthorityException, AeolusException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		TagID tid = t.getTagID();
		userPO.endorse(tid);
	}

        /**
	 * Add AeolusTags in List to the caller's integrity label
	 * @param tags - List of AeolusTags to add to the integrity label
	 * @throws AuthorityException - if caller principal does not have authority for each AeolusTag
	 * @throws AeolusException - if caller is inside a shared state object
	 */
	public static void endorse(List<AeolusTag> tags) throws AuthorityException, AeolusException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		for (AeolusTag t : tags) {
			TagID tid = t.getTagID();
			userPO.endorse(tid);
		}
	}


    /**
     * Runs code in a different user thread. The new thread runs with the authority of calleePid.
	 * Caller must be authorized to switch to calleePid.
	 * The code object, codeObj, is copied to the new process.
	 * @param codeObj - code to invoke
	 * @param calleePid - principal to run the new user thread with
	 * @throws AuthorityException - if caller principal cannot switch to calleePid
	 * @throws AeolusException - if a platform failure occurs when starting the new user thread
	 */
    public static void fork(final Runnable codeObj, final PID calleePid) 
    throws AuthorityException, AeolusException{
	UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
    	userPO.fork(codeObj, calleePid.getPrincipalID());
    }


    /**
	 * Invokes codeObj in the same user thread with the authority of calleePid.
	 * Caller principal must be authorized to switch to calleePid.
	 * After invocation, the caller PID is restored, even if the call throws an exception.
	 * @param codeObj - code to invoke
	 * @param calleePid - principal to run the code on behalf of
	 * @throws AuthorityException - if caller principal cannot switch to calleePid
	 * @throws Exception - if an exception is thrown by the invoke method of the codeObj
	 */
    public static <T> T call(final Callable<T> codeObj, final PID calleePid) 
    throws AuthorityException, Exception{
	UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
    	return userPO.call(codeObj, calleePid.getPrincipalID());    	
    }


    /**
     * Runs code in a different user thread. The new thread runs with the authority of the caller.
	 * The code object, codeObj, is copied to the new process.
	 * @param codeObj - code to invoke
	 * @throws AuthorityException - if thread can not switch to pid
	 * @throws AeolusException - if caller running in shared state
	 */
    public static void fork(final Runnable codeObj) throws AeolusException{
	UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
    	userPO.fork(codeObj, getPID().getPrincipalID()); 
    }

       /**
	 * Registers a service with Aeolus, which allows it to be accessed remotely.
	 * The service must implement an interface listing any methods that it exposes remotely.
	 * The name of this interface must be provided as the name of the service. 
	 * Any parameters and return values from the methods of this interface must be Serializable.
	 * The service name is bound to the virtual node (VN)  where the service is registered.
	 * Any invocations of the remote methods provided by this service will run in this VN 
	 * with the authority of the VN principal.  Every time a remote method of the service is
	 * invoked, a new instance of the service will be created from the specified service class
	 * and the method will be executed by this service instance.
	 * If a service by the same name exists, it will be replaced.
	 * @param serviceName - name of the interface implemented by the service
	 * @param service - class implementing the service (must implement the service interface)
	 * @throws InfoFlowControlException - if caller's secrecy label is not null
	 * @throws NullPointerException - if service is null
	 */
    public static void registerService(final String serviceName, final Class<?> service) throws InfoFlowControlException, NullPointerException {
	if (service == null) throw new NullPointerException();
	try {
        AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
                        public Void run() throws InfoFlowControlException {
			try {
				UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
                        	userPO.registerService(serviceName, service);
			} catch (InfoFlowControlException e) { throw e; }
                        return null;
                        }});
	} catch (PrivilegedActionException e) {
		Throwable cause = e.getCause();
		if (cause instanceof InfoFlowControlException) throw (InfoFlowControlException) cause;
	}

    }

      /**
	* Return service stub that implements the interface specified by serviceName.
	* Remote methods should be invoked using the service stub by casting the returned
	* object to the remote service interface; for example:
	* <p>Foo service = (Foo) getService("localhost", "path.to.Foo", serviceClass);
	* <p>Serializable z = service.methodA(x, y);
	* <p>A method invocation on the service stub will result in a remote procedure call
	* to the remote host. 
	* @param remoteHostName - node where the desired remote service was registered
	* @param serviceName - name that the service is bound to on the remote host
	* @param serviceClass - local Class which defines the interface of the service
	* @return service stub, must be cast to the interface specified by serviceName 
	* @throws NullPointerException - if serviceClass is null or any of its elements are null
	* @throws IllegalArgumentException - if any of the restrictions on the getProxyClass method
	* in Java's Proxy class are violated.
	*/
    public static Object getService(final String remoteHostName, final String serviceName, final Class<?> serviceClass) {
        try {
                return AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
                        public Object run() throws Exception {
			UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
                        return userPO.getService(remoteHostName, serviceName, serviceClass);
                        }});
        }catch(PrivilegedActionException e){
                Throwable cause = e.getCause();
                throw new RuntimeException(e);
        }

    }


    /**
     * Returns the shared state root.
     * Returns null if the root has not been set.
     */
	public static AeolusShared getRoot() {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		return userPO.getRoot();
	}	

   /**
     * Set the shared state root.
     * Only a process with an empty secrecy label can set the root.
     * @throws InfoFlowControlException - if caller's secrecy label isn't null
     */
	public static void setRoot(AeolusShared root) throws InfoFlowControlException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		userPO.setRoot(root);
	}	


    // PLATFORM INSTANCE LAUNCH/SHUTDOWN

    /**
     * Launches a new platform instance on the specified Aeolus node with the specified authority. 
     * <p>This instance starts executing the application of the given name.
     * Caller's secrecy label must be null to execute this request.
     * Caller principal must act for the specified pid.
     * @param hostname - node where the platform instance should be launched
     * @param pid - principal to launch the platform instance on behalf of
     * @param appName - application to start running in the new platform instance
     * @param appParams - parameters to the application
     * @throws InfoFlowControlException - if caller's secrecy label is not null
     * @throws AuthorityException - if caller does not act for pid
     * @throws RpcException - if failure in communicating with hostname 
     * @throws ClassNotFoundException - if application class was not found on hostname
     * @throws NoSuchMethodException - if main method was not found in the application
     * @throws InvocationTargetException - if an exception was thrown by the main method (declared or runtime exception),
     * 			this exception is a wrapper; use e.getCause() to view the actual thrown exception
     * @throws AeolusException - if any other failure while launching the application or
     * 			if an unexpected platform failure occurred
     */
    public static void launch(final String hostname, final PID pid, final String appName, final String appParams) 
    throws InfoFlowControlException, AuthorityException, AeolusException, RpcException, InvocationTargetException,
		ClassNotFoundException, NoSuchMethodException
    {
	if (!(AeolusLib.getPID().doesActFor(pid)) && !pid.equals(PID.getPublicPID())) {
		throw new AuthorityException("Thread must act for pid");
	}
    	try {
			AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
				public Void run() throws Exception {
				UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
			    	userPO.launchPlatformInstance(hostname, pid.getPrincipalID(), appName, appParams);
			    	return null;
				}
			});
		} catch (PrivilegedActionException e) {
			Throwable cause = e.getCause();
			if(cause instanceof InfoFlowControlException)
				throw (InfoFlowControlException) cause;
			if(cause instanceof AuthorityException)
				throw (AuthorityException) cause;
			if(cause instanceof RpcException)
				throw (RpcException) cause;
			if(cause instanceof InvocationTargetException)
				throw (InvocationTargetException) cause;
			if(cause instanceof ClassNotFoundException)
				throw (ClassNotFoundException) cause;
			if(cause instanceof NoSuchMethodException)
				throw (NoSuchMethodException) cause;
			if(cause instanceof AeolusException)
				throw (AeolusException) cause;
    		throw new AeolusException(e);
		}
    }
    
    /**
	 * Shuts down the virtual node of the caller
	 * <p>Any threads running in the virtual node (VN) will be terminated immediately.
	 * The VN's service will be deregistered and the VN will be shutdown.
	 * Caller's secrecy label must be null to execute this request.
	 * @throws InfoFlowControlException - if caller's seclabels are not null
	 */
    public static void shutdown() throws InfoFlowControlException {
    	try {
    		AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
    			public Void run() throws Exception {
				UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
    				userPO.shutdown();
    				return null;
    			}});
    	}catch(PrivilegedActionException e){
    		Throwable cause = e.getCause();
    		if(cause instanceof InfoFlowControlException)
    			throw (InfoFlowControlException) cause;
    	}
    }

         /**
          * Creates a new event in the Aeolus audit trails.
          * Sets the record fields to the provided arguments.
          * This event will be automatically correlated with other events 
          * in the same user thread.
          * @param desc - a String describing the event
          * @param appArgs - a list of Strings for this event chosen by the application
          * @return the EventID for the new event
          */
         public static EventID createEvent(final String desc, List<String> appArgs) throws AeolusException {
		 UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		 String[] params = new String[appArgs.size()];
		 for (int i  = 0; i< appArgs.size(); i++) params[i] = appArgs.get(i);
		 if (userPO instanceof UserLoggingProxy) {
                 	return userPO.createEvent(desc, params, null);
		 }
		 throw new AeolusException("Logging not enabled");
         }

         /**
           * Creates a new event in the Aeolus audit trails.
           * Sets the record fields to the provided arguments.
           * This event will be automatically correlated with other events 
           * in the same user thread.
           * @param desc - a String describing the event
           * @param appArgs - a list of Strings for this event chosen by the application
           * @param elist - a list of predecessors for this event
           * @return the EventID for the new event
           */
         public static EventID createEvent(final String desc, List<String> appArgs, final List<EventID> elist)
			throws AeolusException {
		 UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		 String[] params = new String[appArgs.size()];
		 for (int i  = 0; i< appArgs.size(); i++) params[i] = appArgs.get(i);
		 if (userPO instanceof UserLoggingProxy) {
                 	return userPO.createEvent(desc, elist, params, null);
		 }
		 throw new AeolusException("Logging not enabled");
         }

         /**
           * Returns the EventID of the most recent event logged for the thread.
           * @throws AeolusException - if logging is not enabled.
           */
	 public static EventID getEventID() throws AeolusException {
		 UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		 if (userPO instanceof UserLoggingProxy) {
                 	return userPO.getLastEID();
		 }
		 throw new AeolusException("Logging not enabled");
	 }

         /**
           * Creates a new block for the storage of pids, tags, and labels
           * @throws InfoFlowControlException if the thread's secrecy label is not null
           */
	public static void createBlock() throws InfoFlowControlException {
		 UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		 InfoFlowState ifs = userPO.getInfoFlowState();
		 if (ifs.SLabel != null) throw new InfoFlowControlException("thread's secrecy label needs to be null to create a new block");
                 userPO.createBlock();
	}

         /**
           * Makes the block that contains p the current block for the storage
           * of pids, tags, and labels
           * @throws InfoFlowControlException if the thread's secrecy label is not null
           * @throws AuthorityException if the thread does not act-for p
           */
	public static void makeCurrentBlock(PID p) throws InfoFlowControlException, AuthorityException {
		 UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		 InfoFlowState ifs = userPO.getInfoFlowState();
		 if (ifs.SLabel != null) throw new InfoFlowControlException(
			"thread's secrecy label needs to be null to create a new block");
		 PrincipalID pid = p.getPrincipalID();
		 if (!userPO.isActee(ifs.pid, pid)) throw new AuthorityException(
			"thread must act for p");
                 userPO.setCurrentBlock(pid);
	}

}
