package edu.mit.csail.aeolus.api;

import java.io.Serializable;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.auditing.AeolusApiConstants;
import edu.mit.csail.aeolus.exceptions.*;
import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;
import edu.mit.csail.aeolus.security.*;

/**
 * This behaves like the AeolusBox class but is a queue of objects
 * instead.
 * @author blanks
 *
 * @param <T>
 */
public final class AeolusQueue<T extends Serializable> extends AeolusShared {
	private static final long serialVersionUID = 3026794696634845057L;

	private static final long TIMEOUT_MS = 1000;

	/**
	 * This queue stores objects as their Base64 Serialization
	 */
	transient private BlockingQueue<T> queue;

    /**
      *  Construct an empty AeolusQueue with the specified labels.
      *  @throws InfoFlowControlException if thread's labels are more constraining than those provided.
      */
	public AeolusQueue(AeolusLabel sLabel, AeolusLabel iLabel)
				throws InfoFlowControlException {
		super(sLabel, iLabel);
		this.queue = new LinkedBlockingQueue<T>();
	}

    /**
      *  Construct an empty AeolusQueue with the current thread's labels.
      */
	public AeolusQueue() {
		super();
		this.queue = new LinkedBlockingQueue<T>();
	}

	/**
	 * Adds a copy of obj to the shared queue.
	 * requires obj != null
	 * @throws NullPointerException - if obj is null
	 * @throws InfoFlowControlException if queue's labels are more constrained that the thread's.
	 */
	public void enqueue(T obj) throws InfoFlowControlException, NullPointerException {
		if (obj == null) throw new NullPointerException();
		UserProxyInterface po = UserProxyMap.getMyProxyObject();
		InfoFlowState istate = po.aeolusStartSharedWrite(this);
		try {
			queue.put(AeolusCloner.Clone(obj));
			po.aeolusEndSharedWrite(this, istate);
		} catch (InterruptedException e) {
			po.aeolusEndSharedWrite(this, istate);
			throw new RuntimeException(e);
		}
	}
	
	/**
	 * This method returns the object at the head of the queue. (no copying is needed!)
	 * It will block until an object is available.
	 * @return dequeued element
	 * @throws InfoFlowControlException if thread's labels are not equal to the queue's labels.
	 */
	public T dequeue() throws InfoFlowControlException {
		UserProxyInterface po = UserProxyMap.getMyProxyObject();
		InfoFlowState istate = po.aeolusStartSharedCall(this);
		while (true) {
		try {
		    while (true) {
			T obj = queue.poll(TIMEOUT_MS, TimeUnit.MILLISECONDS);
			po.aeolusEndSharedCall(this, istate);
			if (obj == null) continue;
			
			return obj;
		    }
		} catch (InterruptedException e) {
			continue;
			// po.aeolusEndSharedCall(this, istate);
			// return null;
		}
		}
	}

	/**
	 * This call will wait for the lock on the queue before trying to remove a
	 * potentially empty head (it will not block if the queue is empty, though).
	 * 
	 * @return the object at the head of the shared queue or null if there is no
	 *         head. (no copying is needed!)
	 * @throws InfoFlowControlException if thread's labels are not equal to the queue's labels.
	 */
	public T dequeueNoWait() throws InfoFlowControlException {
		UserProxyInterface po = UserProxyMap.getMyProxyObject();
		InfoFlowState istate = po.aeolusStartSharedCall(this);

		T obj = queue.poll();
		po.aeolusEndSharedCall(this, istate);
		if(obj == null)
			return null;
		return obj;
	}
}
