package edu.mit.csail.aeolus.api;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;

import edu.mit.csail.aeolus.auditing.AeolusApiConstants;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;

/**
 * This class is used to create immutable (non-linked) list share-able objects in Aeolus.
 * The elements of this container object are also required to be "safe-to-share" ie immutable as well.
 * So, no copying of elements is needed.  "T extends AeolusSafe" causes no copying: i.e.,
 * ClassLoader will complain if T is not AeolusSafe.
 * 
 * @author dcurtis
 *
 * @param <T> 
 */

//
//  MAYBE ADD: remove, clone, toArray, iterator, subseq, fill, replace, e2s, concat, a2s, s2a, copy
//
public final class AeolusSequence<T extends AeolusSafe & Serializable> implements AeolusSafe {
	private static final long serialVersionUID = 8093917470092286062L;
        private final ArrayList<T> rep;

    /**
     * Construct an empty AeolusSequence.
     *
     */    
    public AeolusSequence() {
	this.rep = new ArrayList<T>();
    }

    /**
     * Construct an AeolusSequence given a Collection.
     *
     */    
    public AeolusSequence(Collection<T> al) {
	ArrayList<T> newal = new ArrayList<T>(al);
	this.rep = newal;
    }

    /**
     * The add method creates an new AeolusSequence which is a copy of this
     * with a new element added onto the end of the sequence.
     */    
    public AeolusSequence<T> add(T o) {
	ArrayList<T> newal = new ArrayList<T>();
	for (T e: rep) newal.add(e);
	newal.add(o);
	AeolusSequence<T> newseq = new AeolusSequence<T>(newal);
	return newseq;
    }

    /**
     * The set method creates an new AeolusSequence. The elements of the
     * new AeolusSequence are the same as the elements of the original
     * AeolusSequence except that the o replaces the old element
     * at index.
     * @throws IndexOutOfBoundsException if index < 0 or >= size
     */    
    public AeolusSequence<T> set(int index, T o) {
	ArrayList<T> newal = new ArrayList<T>();
	for (int i = 0; i < this.rep.size(); i++) {
		newal.set(i, this.rep.get(i));
	}
	newal.set(index, o);
	AeolusSequence<T> newseq = new AeolusSequence<T>(newal);
	return newseq;
    }


    /**
     * The remove method creates an new AeolusSequence. The elements of the
     * new AeolusSequence are the same as the elements of the original
     * AeolusSequence except there is one fewer element and the
     * original element at index is missing.
     * @throws IndexOutOfBoundsException if index < 0 or >= size
     */    
    public AeolusSequence<T> remove(int index) {
	ArrayList<T> newal = new ArrayList<T>();
	int i = 0;
	for (; i < index; i++) {
		newal.set(i, this.rep.get(i));
	}
	for (int j = index+1; j < this.rep.size(); j++, i++) {
		newal.set(i, this.rep.get(j));
	}
	AeolusSequence<T> newseq = new AeolusSequence<T>(newal);
	return newseq;
    }

    /**
     * The empty method returns true if the AeolusSequence is empty
     * and otherwise returns false
     */    
    public boolean isEmpty() {
	return (this.rep.size() == 0);
    }

    /**
     * The size method returns the number of elements in the AeolusSequence
     */    
    public int size() {
	return this.rep.size();
    }

    /**
     * The get method returns the element at index in the AeolusSequence
     * @throws IndexOutOfBoundsException if index < 0 or >= size
     */    
    public T get(int index) {
	return this.rep.get(index);
    }


    /**
     * The equals method returns true if this sequence's elements are
     * equal to the elements of sq and otherwise returns false.
     */    
    public boolean equals(AeolusSequence sq) {
	return (this.rep.equals(sq.rep));
    }
}
