package edu.mit.csail.aeolus.api;

import edu.mit.csail.aeolus.api.*;

import edu.mit.csail.aeolus.exceptions.*;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.infoflow.InfoFlowRules;
import edu.mit.csail.aeolus.runtime.*;
import edu.mit.csail.aeolus.security.*;
import edu.mit.csail.aeolus.auditing.*;

/**
 * AeolusShared objects reside in shared state. The Aeolus platform transforms
 * them at runtime so that they are effectively isolated, i.e., they can't
 * contain pointers to the outside world. All accesses to shared objects are
 * subject to a label check by the Aeolus platform.
 */
public abstract class AeolusShared implements AeolusSafe {
    private final AeolusLabel sLabel, iLabel;


	/*
	 * XXX UserLoggingProxy modifies this directly so we don't have to change
	 * the interface and do extra work when logging is turned off. It's
	 * important that user code can't access this, though!
	 * @deprecated Not for public use
	 */
	@Deprecated public EventID sharedLastEid;

	private void initLogging(UserProxyInterface po) {
		if (po instanceof UserLoggingProxy) {
			UserLoggingProxy logUserD = (UserLoggingProxy) po;
			EventID pPredEID = logUserD.getLastEID();
			this.sharedLastEid = logUserD.getEID();
			EventRecord createEvent = new EventRecord(this.sharedLastEid,
					pPredEID, AeolusApiConstants.CREATE_SHARED_OBJECT);
			logUserD.submitLogRecord(createEvent);
		}
	}


    /**
     * Create a new AeolusShared object with copies of the thread's current labels.
     */
    protected AeolusShared() {
	UserProxyInterface po = ((UserDomain)Thread.currentThread()).getProxyObject();
	this.sLabel = new AeolusLabel(po.getSLabel());
	this.iLabel = new AeolusLabel(po.getILabel());
	initLogging(po);
    }

    /**
     * Create a new AeolusShared object with copies of the specified labels.  The
     * supplied labels must be no less restrictive than the caller's.
     * @throws InfoFlowControlException - if labels are less restrictive than the caller's.
     */
    protected AeolusShared(AeolusLabel sLabel, AeolusLabel iLabel) throws InfoFlowControlException {
	UserProxyInterface po = ((UserDomain)Thread.currentThread()).getProxyObject();
	this.sLabel = new AeolusLabel(sLabel);
	this.iLabel = new AeolusLabel(iLabel);
	if (!InfoFlowRules.canFlow(po.getSLabel(), po.getILabel(), this.sLabel.getLabel(), this.iLabel.getLabel()))
	    throw new InfoFlowControlException("Attempt to instantiate an AeolusShared with a less restrictive label than the thread label");
	initLogging(po);
    }

    /*
     * XXX It's not safe to let callers outside of the platform call
     * these methods, since label objects are mutable.
     */
    /**
     * Return the object's secrecy label.
     * @deprecated - Not for public use
     */
    @Deprecated public final AeolusLabel getSLabel() {
	return sLabel;
    }

    /**
     * Return the object's secrecy label.
     * @deprecated - Not for public use
     */
    @Deprecated public final AeolusLabel getILabel() {
	return iLabel;
    }

    /**
     * Return a copy of the object's secrecy label.
     */
    public final AeolusLabel getSecrecyLabel() {
	return new AeolusLabel(sLabel);
    }

    /**
     * Return a copy of the object's integrity label.
     */
    public final AeolusLabel getIntegrityLabel() {
	return new AeolusLabel(iLabel);
    }


	/**
	 * This method is invoked by our instrumentation when a method on the shared
	 * state object is called. It is responsible for doing the appropriate
	 * logging and principal change.
	 * 
	 * @return the information flow state prior to the call
	 * @ throws InfoFlowControlException when thread's labels are not the same
	 * 	as the object's labels
	 * @deprecated Not for public use
	 */
	@Deprecated protected final InfoFlowState callStart() throws InfoFlowControlException {
		UserProxyInterface po = ((UserDomain) Thread.currentThread())
				.getProxyObject();
		return (po.aeolusStartSharedCall(this));
	}

	/**
	 * This method is invoked by our instrumentation when any shared state
	 * method call returns. It is responsible for doing the appropriate logging
	 * and principal change.
	 * 
	 * @param oldState
	 *            the information flow state returned by callStart() before the
	 *            invocation
	 * @deprecated Not for public use
	 */
	@Deprecated protected final void callEnd(InfoFlowState oldState) {
		UserProxyInterface po = ((UserDomain) Thread.currentThread())
				.getProxyObject();
		po.aeolusEndSharedCall(this, oldState);
	}

	/**
	 * This method clones an object.
	 * @deprecated Not for public use
	 */
	@Deprecated protected final <T> T aeolusClone(T obj) {
		try {
			return (AeolusCloner.Clone(obj));
		} catch (Exception e) {
			throw new CloningError(e);
		}
	}
}
