package edu.mit.csail.aeolus.api;

// import edu.mit.csail.aeolus.api3.PID;
// import edu.mit.csail.aeolus.infoflow3.PID;
//
import java.io.Serializable;

import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;

/**
 * This class provides AeolusTags.
 */

public final class AeolusTag implements Serializable, AeolusSafe, Cloneable {

	private TagID tid = null;

	/**
	 * Create a new top-level tag.
	 * Caller principal will have authority for the new tag. 
	 * @throws AuthorityException - if caller principal not authorized to create a new tag
	 * (i.e. if the caller is the public principal)
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 * @throws InfoFlowControlException - if caller's secrecy label is not null
	 */
	public AeolusTag() throws AeolusException, InfoFlowControlException, AuthorityException {
		// System.out.println("creating an AeolusTag");
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		this.tid = userPO.createTag();
	}

	/**
	 * Create a sub-tag of tag t.
	 * Tag t must be a top-level tag.
	 * Caller principal will have authority for the new sub-tag.
	 * @throws AuthorityException - if caller principal not authorized to create a new sub-tag
	 * (i.e. if the caller is the public principal or t is not a top-level tag.)
	 * @throws AeolusException - if caller is inside a shared state object
	 * @throws InfoFlowControlException - if caller's secrecy label is not null
	 */
	public AeolusTag(AeolusTag t) throws AeolusException, InfoFlowControlException, AuthorityException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		this.tid = userPO.createSubTag(t.tid);
	}

	/**
 	 * @deprecated  Not for public use.
 	 */
	@Deprecated public AeolusTag(TagID t) {
		this.tid = t;
	}

	/**
 	 * @deprecated  Not for public use.
 	 */
	@Deprecated public TagID getTagID() {
		return this.tid;
	}

	/**
 	 * Returns true if t is a subTag of this, otherwise false.
 	 * @param t possible child of this
 	 */
	public boolean isSubTag(AeolusTag t) {
		if (t.tid.getParent() == this.tid.getValue()) return true;
		else return false;
	}
	/**
 	 * Returns true if t is a superTag of this, otherwise false.
 	 * @param t possible parent of this
 	 */
	public boolean isSuperTag(AeolusTag t) {
		if (this.tid.getParent() == t.tid.getValue()) return true;
		else return false;
	}
	/**
 	 * Returns true if this is the same as tag t, otherwise false.
 	 * @param t - an AeolusTag
 	 */
	public boolean equals(AeolusTag t) {
		return this.tid.equals(t);
	}

	/**
 	 * Returns a new AeolusTag which will be equal to this.
 	 */
	public AeolusTag clone() {
		return new AeolusTag(this.tid.clone());
	}

	/**
 	 * Returns a string representation of this.
 	 */
	public String toString() {
		return this.tid.toString();
	}

	/**
	 * Assuming that the current thread acts for principal p1, and p2 isn't the public pid,
	 * and p1 is a node in the delegation graph of this, the delegate method gives authority
	 * for this to grantee p2.  The principal p1 must have authority for the tag.  The caller's
	 * secrecy label must be null.
	 * @param p1 principal that is authoritative for tag
	 * @param p2 principal that will be authoritative for tag
	 * @throws AuthorityException - if thread does not act for p1 or p1 does not have authority
	 * for this or p2 is the public pid
	 * @throws InfoFlowControlException - if caller's secrecy label isn't null
	 * @throws AeolusException if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred or p1 isn't a node in the delegation graph for
	 * this or adding the link would cause a cycle in the graph
	 */
	public void delegate(PID p1, PID p2) throws AeolusException, InfoFlowControlException, AeolusException {
		// System.out.println("Entering delegate");
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		// System.out.println("delegate tid " + this.tid);
		// System.out.println("delegate from p1 " + p1.getPrincipalID());
		// System.out.println("delegate to p2 " + p2.getPrincipalID());
		userPO.delegate(this.tid, p1.getPrincipalID(), p2.getPrincipalID());
	}
	/**
 	 * Returns true if current thread has authority for the tag, otherwise false.
 	 */
	public boolean hasAuthority() {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		return userPO.hasAuthority(AeolusLib.getPID().getPrincipalID(), this.tid);
	}

	/**
 	 * Returns true if principal p has authority for the tag, otherwise false.
 	 * @param p principal with respect to which authority checking is requested
 	 */
	public boolean hasAuthority(PID p) {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		return userPO.hasAuthority(p.getPrincipalID(), this.tid);
	}

	/**
	 * Revokes authority for this tag from grantor principal from p1 to grantee principal to p2.
	 * Current thread must act for the grantor principal p1 and the tag t should have been 
	 * previously delegated from p1 to p2. 
	 * Secrecy label must be null.
	 * @param p1 principal that is authoritative for tag and granted authority to p2 for tag
	 * @param p2 principal to remove from being authoritative for tag
	 * @throws AuthorityException - if caller principal does not act for p1 or if p1 isn't a node
	 * in the delegation graph of this
	 * @throws InfoFlowControlException - if caller's secrecy label is not null
	 * an unexpected platform failure occurred
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 */
	public void revoke(PID p1, PID p2) throws AeolusException, AuthorityException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		userPO.revokeDelegate(this.tid, p1.getPrincipalID(), p2.getPrincipalID());
	}
}
