package edu.mit.csail.aeolus.api;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.util.Date;
import java.util.List;

import edu.mit.csail.aeolus.auditing.CompletionStatus;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.exceptions.ServiceNotFoundException;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.infoflow.AuthStateConstants;
import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;

/**
 * An AeolusUser object must be used by application code to interact with the Aeolus platform. </p>
 * 
 * It exposes the following platform functionality:
 * 
 * <ul>
 * <li> modifications and uses of authority state (e.g., principal and tag creation, 
 * delegations and revocations, label manipulations) </li>
 * <li> node deletion</li>
 * <li> platform instance launch and shutdown</li>
 * <li> shared state access (i.e., shared objects, shared queues, and shared locks)</li>
 * <li> file system access (i.e., file operations and file-stream creation)</li>
 * <li> remote procedure calls (i.e. remote service registration and access)</li>
 * <li> authority closures (i.e. closure creation and invocation)</li>
 * <li> local calls and forks</li>
 * </ul>
 * 
 * @author viq
 */
public class AeolusUserInternal {

    private UserProxyInterface userPO;
    private final PrincipalID  pidPublic;

    /**
     * Constructs a new AeolusUser instance
     */
    public AeolusUserInternal()
    {
        this.userPO = UserProxyMap.getMyProxyObject();
        pidPublic = new PrincipalID(AuthStateConstants.pidPublic);
    }
 
    /**
     * Removes the specified node from the Aeolus platform if the caller is authorized.
     * After the node is removed, any communication with this node will be untrusted. 
     * @param hostname - node to be removed
     * @throws AuthorityException - if caller not authorized to delete the node (does not act for node root principal)
     */
    public void deleteNode(final String hostname) throws AuthorityException, AeolusException
    {
    	try {
    		AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
    			public Void run() throws Exception {
    				userPO.deleteNode(hostname);
    		    	return null;
    			}});
    	}catch(PrivilegedActionException e){
    		Throwable cause = e.getCause();
    		if(cause instanceof AuthorityException)
    			throw (AuthorityException) cause;
    		if(cause instanceof AeolusException)
    			throw (AeolusException) cause;
    		throw new AeolusException(e);
    	}
    }
    
    // Info flow state
    
    /**
	 * Returns a copy of the information flow state of the caller user thread
	 */
    public InfoFlowState getInfoFlowState(){
    	return userPO.getInfoFlowState();
    }
    
	// PRINCIPALS

    /**
         * @return the principalID for public.
         */
    public PrincipalID getPublicPID() {
             return pidPublic;

    }
    
	/**
	 * Creates a new principal in the caller's principal core. 
	 * Caller principal will act for the new principal.
	 * @return Id of the created principal
	 * @throws AuthorityException - if caller principal not authorized to create a new principal
	 * (i.e. if the caller is the public principal)
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public PrincipalID createPrincipal() throws AuthorityException, AeolusException {
		return userPO.createPrincipal();
    }
	
	/**
	 * Creates a new principal in a new core.
	 * Caller principal will act for the new principal 
	 * @return Id of the created principal
	 * @throws AuthorityException - if caller principal not authorized to create a new principal
	 * (i.e. if the caller is the public principal)
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public PrincipalID createPrincipalInNewCore() throws AuthorityException, AeolusException {
		return userPO.createPrincipalInNewCore();
    }
	
	/**
	 * Creates a new principal in the same core as principal p.
	 * Caller principal will act for the new principal.
	 * @return Id of the created principal
	 * @throws AuthorityException - if caller principal not authorized to create a new principal
	 * (i.e. if the caller is the public principal)
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public PrincipalID createPrincipal(final PrincipalID p) throws AuthorityException, AeolusException {
		return userPO.createPrincipal(p);
    }
	
	// TAGS
	
	/**
	 * Creates a new top-level tag.
	 * Caller principal will have authority for the new tag. 
	 * @return Id of the created tag
	 * @throws AuthorityException - if caller principal not authorized to create a new tag
	 * (i.e. if the caller is the public principal)
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public TagID createTag() throws AuthorityException, AeolusException{
		return userPO.createTag();
    }
	
	/**
	 * Creates a sub-tag of tag t.
	 * Tag t must be a top-level tag.
	 * Caller principal will have authority for the new sub-tag.
	 * @return Id of the created sub-tag
	 * @throws AuthorityException - if caller principal not authorized to create a new sub-tag
	 * (i.e. if the caller is the public principal) [...]
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred	  
	 */
	public TagID createSubTag(final TagID t) throws AuthorityException, AeolusException{
		return userPO.createSubTag(t);
    }
	
	/**
	 * Creates a sub-tag of tag t and stores the tag info in the core
	 * of principal p.
	 * Tag t must be a top-level tag.
	 * Caller principal will have authority for the new sub-tag.
	 * @return Id of the created sub-tag
	 * @throws AuthorityException - if caller principal not authorized to create a new sub-tag
	 * (i.e. if the caller is the public principal) [...]
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred	
	 */
	public TagID createSubTag(final TagID t, final PrincipalID p) throws AuthorityException, AeolusException{
		return userPO.createSubTag(t, p);    	
    }
	
	
	// DELEGATIONS AND REVOCATIONS API
	
	/**
	 * Adds an act-for link from principal p2 to principal p1.
	 * Caller principal must act for principal p1.
	 * @throws AuthorityException - if caller principal not authorized to allow p2 to act-for p1
	 * (if the caller does not act for p1) [...]
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public void actFor(final PrincipalID p1, final PrincipalID p2) throws AuthorityException, AeolusException{
		userPO.actFor(p1, p2);
    }
	
	/**
	 * Removes the act-for link from principal p2 to principal p1.
	 * Caller principal must act for principal p1.
	 * @throws AuthorityException - if caller principal not authorized to disallow p2 to act-for p1
	 * (if the caller does not act for p1) [...]
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public void revokeActFor(final PrincipalID p1, final PrincipalID p2) throws AuthorityException, AeolusException{
		userPO.revokeActFor(p1, p2);    	
    }
	
	/**
	 * Gives authority for tag t from grantor principal p1 to grantee principal p2.
	 * Caller principal must act for the grantor principal p1 and the grantor principal p1 
	 * must have authority for tag t.
	 * @throws AuthorityException - if caller principal does not act for p1 or p1 does not have authority for t
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public void delegate(final TagID t, final PrincipalID p1, final PrincipalID p2) 
	throws AuthorityException, AeolusException{
		userPO.delegate(t, p1, p2);
    }
	
	/**
	 * Revokes authority for tag t from grantor principal p1 to grantee principal p2.
	 * Caller principal must act for the grantor principal p1 and the tag t should have been 
	 * previously delegated from p1 to p2. 
	 * @throws AuthorityException - if caller principal does not act for p1
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
	 */
	public void revokeDelegate(final TagID t, final PrincipalID p1, final PrincipalID p2) 
	throws AuthorityException, AeolusException{
		userPO.revokeDelegate(t, p1, p2);
    }

	
    // API LABEL CHANGES //

	/**
	 * Adds tag tid to the caller's secrecy label
	 * @param tid - tag id to add to the secrecy label
	 */
    public void addSecrecy(final TagID tid) throws AuthorityException, AeolusException {
	try {
    		userPO.addSecrecy(tid);
	} catch (Exception e){
                        if(e instanceof AuthorityException)
                                throw (AuthorityException) e;
                        else
                                throw new AeolusException(e);

	}
    }

    /**
	 * Removes tag tid from the caller's integrity label
	 * @param tid - tag id to remove from integrity label
	 */
    public void removeIntegrity(final TagID tid) throws AuthorityException, AeolusException {
	try {
    		userPO.removeIntegrity(tid);
	} catch (Exception e){
                        if(e instanceof AuthorityException)
                                throw (AuthorityException) e;
                        else
                                throw new AeolusException(e);

        }
    }

    /**
	 * Remove tag tid from the caller's secrecy label
	 * @param tid - tag id to remove from the secrecy label
	 * @throws AuthorityException - if caller principal does not have authority for tag tid
	 * @throws AeolusException - if an unexpected platform failure occurred
	 */
    public void declassify(final TagID tid) throws AuthorityException, AeolusException{
    	userPO.declassify(tid);
    }

    /**
	 * Add tag tid to the caller's integrity label
	 * @param tid - tag id to add to the integrity label
	 * @throws AuthorityException - if caller principal does not have authority for tag tid
	 * @throws AeolusException - if an unexpected platform failure occurred
	 */
    public void endorse(final TagID tid) throws AuthorityException, AeolusException{
    	userPO.endorse(tid);    	
    }
    
    /**
	 * Adds each tag from the given list of tags to the caller's secrecy label
	 * @param tids - list of tag ids to add to the secrecy label
	 */
    public void addSecrecy(final List<TagID> tids) throws AuthorityException, AeolusException {
	try {
    		userPO.addSecrecy(tids);
	} catch (Exception e){
                        if(e instanceof AuthorityException)
                                throw (AuthorityException) e;
                        else
                                throw new AeolusException(e);

        }

    }

    /**
	 * Remove each tag from the given list of tags from the caller's integrity label
	 * @param tids - list of tag ids to remove from the integrity label
	 */
    public void removeIntegrity(final List<TagID> tids) throws AuthorityException, AeolusException {
	try {
    		userPO.removeIntegrity(tids);
	} catch (Exception e){
                        if(e instanceof AuthorityException)
                                throw (AuthorityException) e;
                        else
                                throw new AeolusException(e);

        }

    }

    /**
	 * Remove each tag from the given list of tags from the caller's secrecy label
	 * @param tids - list of tag ids to remove from the secrecy label
	 * @throws AuthorityException - if caller principal does not have authority for all the tags
	 * @throws AeolusException - if an unexpected platform failure occurred
	 */
    public void declassify(final List<TagID> tids) throws AuthorityException, AeolusException{
    	userPO.declassify(tids);
    }

    /**
	 * Adds each tag from the given list of tags to the caller's integrity label
	 * @param tids - list of tag ids to add to the integrity label
	 * @throws AuthorityException - if caller principal does not have authority for all the tags
	 * @throws AeolusException - if an unexpected platform failure occurred
	 */
    public void endorse(final List<TagID> tids) throws AuthorityException, AeolusException{
    	userPO.endorse(tids);
    }

    // API SHAREDSTATE //

    /**
     * Returns the copy of the content stored in the shared state root.
	 * Only an uncontaminated process can read the root.
	 * @throws InfoFlowControlException - if caller labels are non-null
	 * @throws AeolusException - if an unexpected platform failure occurs
	 */
//	public  AeolusSharedObject<Serializable> getSharedRoot() throws InfoFlowControlException, AeolusException{
//		return userPO.getSharedRoot();
//	}	

    // API FILE //

    /**
     * Starts a file server on this node if the server is not already running.
     * Data can now be stored on this node (this node effectively becomes a storage node). 
     */
    public void createFilesystem()
    {
        //userPO.createFilesystem();
    }

    /**
     * Reads data from the file specified by filePath on hostname into the provided byte array.
     * Will read up to contentBuffer.length bytes into the byte array. 
     * Caller labels must allow reading from the file.
     * @param hostname - node where the file is stored
     * @param filePath - path to the file on hostname
     * @param contentBuffer - byte array into which to store the file content
     * @return the number of bytes read or -1 if end of file is reached
     * @throws FileNotFoundException - if file does not exist and the caller labels allow reading its parent directory
     * @throws IOException - if an I/O error occurred while reading from the file or retrieving file labels
     * @throws InfoFlowControlException - if caller labels don't allow reading from the file or reading
     * a higher parent directory of the file, in the case the file is missing
     * @throws AeolusException - if a failure occurs while communicating with hostname
     */
    public int readFile(final String hostname, final String filePath, final byte[] contentBuffer) 
    throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException{
        try {
        	return AccessController.doPrivileged(new PrivilegedExceptionAction<Integer>() {
    			public Integer run() throws Exception {
    				return userPO.readFile(hostname, filePath, contentBuffer);
    			}});
    	}catch(PrivilegedActionException e){
    		Throwable cause = e.getCause();
    		if(cause instanceof FileNotFoundException)
    			throw (FileNotFoundException) cause;
    		if(cause instanceof IOException)
    			throw (IOException) cause;
    		if(cause instanceof InfoFlowControlException)
    			throw (InfoFlowControlException) cause;
    		if(cause instanceof AeolusException)
    			throw (AeolusException) cause;
    		throw new AeolusException(e);
    	}
    }

    /**
     * Writes data from the specified byte array to the file specified by filePath on hostname. 
     * Caller labels must allow writing to the file.
     * @param hostname - node where the file is stored
     * @param filePath - path to the file on hostname
     * @param fileContent - byte array storing data to be written to the file
     * @param appendMode - if set to true, will write bytes to the end of the file
     * @throws FileNotFoundException - if file does not exist and the caller labels allow reading its parent directory
     * @throws IOException - if an I/O error occurs while writing to the file or retrieving file labels
     * @throws InfoFlowControlException - if caller labels don't allow writing to the file or reading
     * a higher parent directory of the file, in the case the file is missing
     * @throws AeolusException - if a failure occurs while communicating with hostname
     */
    public void writeFile(final String hostname, final String filePath, final byte[] fileContent, final boolean appendMode) 
    throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException{
        try {
        	AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
    			public Void run() throws Exception {
    		        userPO.writeFile(hostname, filePath, fileContent, appendMode);
    				return null;
    			}});
    	}catch(PrivilegedActionException e){
    		Throwable cause = e.getCause();
    		if(cause instanceof FileNotFoundException)
    			throw (FileNotFoundException) cause;
    		if(cause instanceof IOException)
    			throw (IOException) cause;
    		if(cause instanceof InfoFlowControlException)
    			throw (InfoFlowControlException) cause;
    		if(cause instanceof AeolusException)
    			throw (AeolusException) cause;
    		throw new AeolusException(e);
    	}
    }
    
    // RPC
    
   /**
	* Return service stub that implements the interface specified by serviceName.
	* Remote methods should be invoked using the service stub by casting the returned
	* object to the remote service interface; for example:
	* <p>Foo service = (Foo) getService("localhost", "path.to.Foo");
	* <p>Serializable z = service.methodA(x, y);
	* <p>A method invocation on the service stub will result in a remote procedure call
	* to the remote host. 
	* @param remoteHostName - node where the desired remote service was registered
	* @param serviceName - name that the service is bound to on the remote host
	* @param serviceClass - local Class which defines the interface of the service
	* @return service stub, must be cast to the interface specified by serviceName 
	* @throws ClassNotFoundException - if interface with serviceName cannot be found on the client
	*/
	public Object getService(final String remoteHostName, final String serviceName, final Class<?> serviceClass) 
	throws ClassNotFoundException, ServiceNotFoundException{
    	try {
        	return AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
    			public Object run() throws Exception {
    		    	return userPO.getService(remoteHostName, serviceName, serviceClass);
    			}});
    	}catch(PrivilegedActionException e){
    		Throwable cause = e.getCause();
    		if(cause instanceof ClassNotFoundException)
    			throw (ClassNotFoundException) cause;
    		if(cause instanceof ServiceNotFoundException)
    			throw (ServiceNotFoundException) cause;
    		throw new RuntimeException(e);
    	}
    }
	
	// I/O
	
	 public void readFromIODevice() throws InfoFlowControlException{
		try {
        	AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
    			public Void run() throws Exception {
    				userPO.readFromIODevice(); 
    		    	return null;
    			}});
    	}catch(PrivilegedActionException e){
    		Throwable cause = e.getCause();
    		if(cause instanceof InfoFlowControlException)
    			throw (InfoFlowControlException) cause;
    		throw new RuntimeException(e);
    	}
	 }
	 
	 public void writeToIODevice() throws InfoFlowControlException{
		try {
        	AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
    			public Void run() throws Exception {
    				userPO.writeToIODevice();
    		    	return null;
    			}});
    	}catch(PrivilegedActionException e){
    		Throwable cause = e.getCause();
    		if(cause instanceof InfoFlowControlException)
    			throw (InfoFlowControlException) cause;
    		throw new RuntimeException(e);
    	}
	 }
	 
	 // AUDITING

	 /**
	  * Creates a new event in the Aeolus audit trails.
	  * Sets the record fields to the provided arguments.
	  * This event will be automatically correlated with other events 
	  * in the same user thread.
	  * @param methodName - the name of the event
	  * @param parameters - event parameters
	  * @param result - the return value associated with an event
	  */
	 public EventID createEvent(final String methodName, final Serializable[] parameters, final Serializable result){
		 return userPO.createEvent(methodName, parameters, result);
	 }
	 
	 /**
	  * Creates a new event in the Aeolus audit trails.
	  * Sets the record fields to the provided arguments.
	  * This event will be automatically correlated with other events 
	  * in the same user thread.
	  * @param methodName - the name of the event
	  * @param parameters - event parameters
	  * @param result - the return value associated with an event
	  * @param timestamp - time of the event
	  */
	 public EventID createEvent(final String methodName, final Serializable[] parameters, 
			 final Serializable result, final Date timestamp){
		 return userPO.createEvent(methodName, parameters, result, timestamp);
	 }
	 
	 /**
	  * Creates a new event in the Aeolus audit trails.
	  * Sets the record fields to the provided arguments.
	  * This event will be automatically correlated with other events 
	  * in the same user thread.
	  * @param methodName - the name of the event
	  * @param preds - a list of predecessors for this event
	  * @param parameters - event parameters
	  * @param result - the return value associated with an event
	  */
	 public EventID createEvent(final String methodName, final List<EventID> preds, 
			 final Serializable[] parameters, final Serializable result){
		 return userPO.createEvent(methodName, preds, parameters, result);
	 }
	 
	 /**
	  * Creates a new event in the Aeolus audit trails.
	  * Sets the record fields to the provided arguments.
	  * This event will be automatically correlated with other events 
	  * in the same user thread.
	  * @param methodName - the name of the event
	  * @param preds - a list of predecessors for this event
	  * @param parameters - event parameters
	  * @param result - the return value associated with an event
	  * @param timestamp - time of the event
	  */
	 public EventID createEvent(final String methodName, final List<EventID> preds, final Serializable[] parameters, 
			 final Serializable result, final Date timestamp){
		 return userPO.createEvent(methodName, preds, parameters, result, timestamp);
	 }
	 
	 /**
	  * Creates a new event in the Aeolus audit trails.
	  * Sets the record fields to the provided arguments.
	  * This event will be automatically correlated with other events 
	  * in the same user thread.
	  * @param methodName - the name of the event
	  * @param preds - a  list of predecessor events
	  * @param parameters - event parameters
	  * @param result - the return value associated with an event
	  * @param status - success or failure status of the event
	  * @param timestamp - time of the event
	  */
	 public EventID createEvent(final String methodName, final List<EventID> preds, final Serializable[] parameters, 
			 final Serializable result, final CompletionStatus status, final Date timestamp){
		 return userPO.createEvent(methodName, preds, parameters, result, status, timestamp);
	 }
	 
	 public EventID createEvent(final int methodName, final EventID pred, final Serializable[] parameters, 
			 final Serializable result, final CompletionStatus status){
		 return userPO.createEvent(methodName, pred, parameters, result, status);
	 }
	 
	 public EventID createEvent(final int methodName, final EventID pred, final Serializable[] parameters, 
			 final Serializable result){
		 return userPO.createEvent(methodName, pred, parameters, result);
	 }
}
