package edu.mit.csail.aeolus.api;

import java.io.Serializable;

import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.runtime.UserProxyInterface;
import edu.mit.csail.aeolus.runtime.UserProxyMap;
import edu.mit.csail.aeolus.infoflow.AuthStateConstants;

import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;

/**
 * Objects of this class represent principals.
 */

public final class PID  implements Serializable, Cloneable, AeolusSafe {
	private PrincipalID pid;
	// private UserProxyInterface userPO = null;
	private static PrincipalID principalIDPublic = new PrincipalID(AuthStateConstants.pidPublic);
	private static PID pidPublic = new PID(new PrincipalID(AuthStateConstants.pidPublic));

	/**
 	 * Creates a new principal.
         * Caller principal will act for the new principal.
         * @throws AuthorityException - if caller principal not authorized to create a new principal
         * (i.e. if the caller is the public principal)
         * @throws AeolusException - if failure in communicating with the authority server or 
         * an unexpected platform failure occurred
         * @throws InfoFlowControlException - if caller's secrecy label is not null
         */
	public PID () throws AuthorityException, InfoFlowControlException, AeolusException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		this.pid = userPO.createPrincipal();
	}

	/**
         * Creates a new PID given a PrincipalID
         * @deprecated Not for public use
         */
	@Deprecated public PID (PrincipalID pid) {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		this.pid = pid;
	}


	/**
	 * Returns the PID for the current principal
	 * @deprecated Not for public use
	 */
	@Deprecated public PrincipalID getPrincipalID() {
		return this.pid;
	}

	/**
	 * Returns the public PID
	 */
        public static PID getPublicPID() {
             return pidPublic;

        }
	/**
	 * Returns the principalID for the public principalID.
	 * @deprecated Not for public use
	 */
        public static PrincipalID getPublicPrincipalID() {
             return principalIDPublic;

        }


	/**
	 * Adds an act-for link from this to principal p.
	 * Caller's principal must act for this and p must not be the public pid.
	 * @throws AuthorityException - if caller principal does not act for this
	 * or if p is the public pid
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred or would create a cycle in the principal hierarchy
         * @throws InfoFlowControlException - if caller's secrecy label is not null
	 */
	public void actFor(PID  p) throws InfoFlowControlException, AuthorityException, AeolusException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		userPO.actFor(this.pid, p.getPrincipalID());
	}

	/**
	 * Returns true if and only if p acts for this
	 */
	public boolean doesActFor(PID  p) {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		return userPO.isActee(this.getPrincipalID(), p.getPrincipalID());
	}


	/**
	 * Removes the act-for link from this to principal p.
	 * Current principal must act for principal p.
	 * @throws AuthorityException - if caller principal doesn't act-for this or
	 * there is no link from this to p
	 * @throws AeolusException - if failure in communicating with the authority server or 
	 * an unexpected platform failure occurred
         * @throws InfoFlowControlException - if caller's secrecy label is not null
	 */
	public void revoke(PID  p) throws AuthorityException, AeolusException, InfoFlowControlException {
		UserProxyInterface userPO = UserProxyMap.getMyProxyObject();
		userPO.revokeActFor(p.getPrincipalID(), this.pid);
	}

	/**
	 * Returns true if the PID p equals this, otherwise returns false
	 */
	public boolean equals(PID  p) {
		return this.pid.equals(p.getPrincipalID());
	}

        /**
 	 * Returns a new PID which will be equal to this.
         */

	public PID clone() {
		return new PID(this.pid);
	}

        /**
 	 * Returns a string representation of this.
         */

	public String toString() {
		return this.pid.toString();
	}
}
