package edu.mit.csail.aeolus.auditing;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.List;

/**
 * Stores information associated with an Aeolus event
 * 
 * @author viq
 * 
 */
public class EventRecord implements Serializable {
	private static final long serialVersionUID = -7605699857902785485L;

	public static enum Pred_Type {
		PROCESS_PRED, COMM_PRED, O_PRED;
	}

	// Fields

	/** Unique event Id */
	private EventID eventId;

	/** Id of the predecessor event */
	private EventID processPred;
	private EventID commPred;
	private List<EventID> otherPreds;

	/** Information about the invoked operation */
	private String methodName;	// method name for closure/shared obj
	private int methodNumber;	// really the event type
	private String infoString;
	private List<Serializable> paramValues;

	/** Return value after method is invoked */
	private Serializable returnValue;

	/** Status */
	private CompletionStatus requestStatus;

	/** Timestamp of the request */
	private Date timestamp;
	
	/** File System file modification times (0 for non-FS events) */
	private long fsLastModifiedBefore;
	private long fsLastModifiedAfter;
	
	private int platformID;
	private long userD;

	// public EventRecord(){}

	/**
	 * Constructs an event record
	 */
	public EventRecord(EventID eventId, EventID processPred, int methodNumber, String methodName) {
		this(eventId, processPred, null, methodNumber, methodName, new Serializable[0]);
	}

	public EventRecord(EventID eventId, EventID processPred, int methodNumber) {
		this(eventId, processPred, null, methodNumber, new Serializable[0]);
	}

	public EventRecord(EventID eventId, EventID pPred, EventID cPred,
			   int methodNumber, String methodName,
			   Serializable[] parameterValues)
	{
		if (eventId == null)
			throw new RuntimeException();
		this.eventId = eventId;
		this.processPred = pPred;
		this.commPred = cPred;
		this.otherPreds = Collections.emptyList();
		this.methodName = methodName;
		this.methodNumber = methodNumber;
		if(parameterValues == null)
			this.paramValues = Collections.emptyList();
		else
			this.paramValues = Arrays.asList(parameterValues);
		this.requestStatus = CompletionStatus.SUCCEEDED;
	}

	public EventRecord(EventID eventId, EventID pPred, EventID cPred,
			int methodNumber, Serializable[] parameterValues) {
		if (eventId == null)
			throw new RuntimeException();
		this.eventId = eventId;
		this.processPred = pPred;
		this.commPred = cPred;
		this.otherPreds = Collections.emptyList();
		this.methodNumber = methodNumber;
		if(parameterValues == null)
			this.paramValues = Collections.emptyList();
		else
			this.paramValues = Arrays.asList(parameterValues);
		this.requestStatus = CompletionStatus.SUCCEEDED;
	}

	public EventID getEventId() {
		return this.eventId;
	}

	public EventID getProcessPred() {
		return this.processPred;
	}

	public EventID getCommunicationPred() {
		return this.commPred;
	}

	public List<EventID> getOtherPreds() {
		return this.otherPreds;
	}

	public int getMethod() {
		return this.methodNumber;
	}

	public String getMethodName() {
		String s = AeolusApiConstants.getMethodName(this.methodNumber);
		if (this.methodName != null)
			return s + "(" + this.methodName + ")";
		else
			return s;
	}

	public String getInfoString() {
		return this.infoString;
	}

	public List<Serializable> getParamValues() {
		return this.paramValues;
	}

	public Date getTimestamp() {
		return this.timestamp;
	}
	
	public void setFsAfterTime(long fsLastModifiedAfter) {
		this.fsLastModifiedAfter = fsLastModifiedAfter;
	}

	public long getFsAfterTime() {
		return fsLastModifiedAfter;
	}

	public void setFsBeforeTime(long fsLastModifiedBefore) {
		this.fsLastModifiedBefore = fsLastModifiedBefore;
	}

	public long getFsBeforeTime() {
		return fsLastModifiedBefore;
	}

	public Serializable getReturnValue() {
		return this.returnValue;
	}

	public CompletionStatus getCompletionStatus() {
		return this.requestStatus;
	}

	public void setEventID(EventID eventID) {
		this.eventId = eventID;
	}

	public void setMethod(int method) {
		this.methodNumber = method;
	}

	public void setMethodName(String methodName) {
		this.methodName = methodName;
	}

	public void setInfoString(String info) {
		this.infoString = info;
	}

	public void setParams(Serializable[] params) {
		if (params==null)
			return;
		this.paramValues = Arrays.asList(params);
	}

	public void setTimestamp(Date t) {
		this.timestamp = t;
	}

	public void setReturnValue(Serializable value) {
		this.returnValue = value;
	}

	public void setCompletionStatus(CompletionStatus status) {
		this.requestStatus = status;
	}

	/**
	 * String representation of this LogRecord
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		String delimiter = " ";

		sb.append("EventID:=" + this.eventId);
		sb.append(delimiter);

		sb.append("preds:=" + this.otherPreds.toString());

		sb.append("Method:=" + this.getMethodName());
		sb.append(delimiter);

		if (this.paramValues != null) {
			sb.append("params:=" + this.paramValues.toString());
			sb.append(delimiter);
		}

		if (this.returnValue != null) {
			sb.append(delimiter);
			sb.append("ReturnValue:=" + this.returnValue.toString());
		}
		if (this.requestStatus != null) {
			sb.append(delimiter);
			sb.append("Status:=" + this.requestStatus.name());
		}

		return sb.toString();
	}

	public void setCommunicationPred(EventID predecessorEventId) {
		if (predecessorEventId == null)
			return;
		this.commPred = predecessorEventId;
	}

	public void setOtherPreds(List<EventID> preds) {
		this.otherPreds = preds;
	}

	public void setPlatformID(int platformID) {
		this.platformID = platformID;
	}

	public void setUserD(long userd) {
		this.userD = userd;
	}

	public int getPlatformID() {
		return platformID;
	}

	public long getUserD() {
		return userD;
	}
}
