package edu.mit.csail.aeolus.auditing;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Stores lists of events based on the id of the appDomain that submitted them
 * Uses Java ConcurrentHashMap to provide thread-safety
 * Each entry of the map has an associated lock to synchronize the addEvent and getAllAndResent
 * operations
 * Note: this implementation assumes that the events for a given appDomain id can 
 * be added only by the thread with that id (that is, there are no concurrent addEvent operations 
 * to the same entry of the table)s
 * 
 * @author viq
 *
 */

public class SynchronizedEventsTable<K, E> {

	/** Table of events based on User-D appDomain */
	private ConcurrentHashMap<K, EventsTableEntry<E>> eventsTable;
	
	public SynchronizedEventsTable()
	{
		this.eventsTable = new ConcurrentHashMap<K, EventsTableEntry<E>>();
	}
	
	/**
	 * Adds a new event record to the events table
	 * 
	 * @assumes only one thread at a time can call this method with the same id
	 * @param id - Used-D appDomain id
	 * @param event - event record
	 */
	public void addEvent(K id, E event)
	{
		EventsTableEntry<E> entry = eventsTable.get(id);
		if(entry == null)
		{
			// create new entry
			entry = new EventsTableEntry<E>();
			entry.eventsList.add(event);
			eventsTable.put(id, entry);
		}
		else
		{
			// existing entry: lock the entry separately
			entry.entryLock.lock();
			entry.eventsList.add(event);
			entry.entryLock.unlock();
		}
	}
	
	/**
	 * Returns all the event records currently stored in the table
	 * Resets all the event lists to empty lists
	 * Adding events to existing key id values can proceed 
	 * concurrently with the processing of unrelated entries
	 * (each table entry is locked separately)
	 * Adding new entries (new keys) to the table is handled
	 * by the underlying Java concurrent hash map
	 * @return contents of the events table
	 */
	public HashMap<K, ArrayList<E>> getAllAndReset()
	{
		HashMap<K, ArrayList<E>> contents = new HashMap<K, ArrayList<E>>();
		
		Set<Map.Entry<K, EventsTableEntry<E>>> entries = eventsTable.entrySet();
		Iterator<Map.Entry<K, EventsTableEntry<E>>> it = entries.iterator();
		while(it.hasNext())
		{
			Map.Entry<K, EventsTableEntry<E>> mapEntry = it.next();
			K id = mapEntry.getKey();
			EventsTableEntry<E> entry = mapEntry.getValue();
			
			entry.entryLock.lock();
			// store events
			contents.put(id, entry.eventsList);
			// reset entry
			entry.eventsList = new ArrayList<E>();
			entry.entryLock.unlock();
			
		}

		return contents;
	}
	
}
