package edu.mit.csail.aeolus.auditing;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

/**
 * Vector timestamps are used to find a partial order
 * among all the events occurring in the Aeolus distributed system.
 * 
 * For simplicity, this implementation assumes that the number
 * of processes (n) is fixed (i.e. new processes are not dynamically created).
 * 
 * A timestamp is represented by an array of n elements; each element is an integer
 * clock value of every process in the system.
 * 
 * For example, a timestamp [a b c d] for event x on process 1, means that the local
 * clock value of event x is a and that b is the last known clock value for process 2.
 * 
 * @author viq
 *
 */
public class Timestamp implements Serializable {

	private static final long serialVersionUID = 1L;

	/** Local clocks for each process */
	private int[] clocks;
	
	/** Total number of processes in the system */
	private int numProcs;
	
	/**
	 * Constructs a timestamp with all clocks set to 0
	 */
	public Timestamp(int numProcessors)
	{
		numProcs = numProcessors;
		clocks = new int[numProcs];
		for (int i = 0; i < numProcs; i++)
		{
			clocks[i] = 0;
		}
	}
	
	/**
	 * Constructs a copy of the provided timestamp t
	 */
	public Timestamp(Timestamp t)
	{
		numProcs = t.getNumProcs();
		clocks = new int[numProcs];
		
		for (int i = 0; i < numProcs; i++)
		{
			clocks[i] = t.getClockValue(i);
		}
	}
	
	/**
	 * Constructs a timestamp from a list of integers
	 * (each integer in the list becomes a clock value)
	 * Meant to be used for RPCs
	 */
	public Timestamp (List<Integer> list)
	{
		numProcs = list.size();
		
		clocks = new int[numProcs];
		
		int i = 0;
		
		for (int clock : list)
		{
			clocks[i] = clock;
			i++;
		}
	}
	
	/** Returns the total number of processes in the timestamp */
	public int getNumProcs()
	{
		return this.numProcs;
	}
	
	/**
	 * Returns the local time corresponding to processor with procID
	 */
	public int getClockValue(int procID)
	{
		return clocks[procID];
	}
	
	/**
	 * Sets the local clock value of process with procID to time
	 */
	public void setTime(int procID, int time)
	{
		clocks[procID] = time;
	}
	
	/**
	 * Increases the clock value of proc with procID by 1
	 */
	public void localEventUpdate(int procID)
	{
		clocks[procID]++;
	}
	
	/**
	 * Updates all clock values upon a communication 
	 * event with another process
	 * Will pick the maximum of local clock values
	 */
	public void rpcEventUpdate(int receiverID, Timestamp recTimestamp)
	{
		if(this.clocks[receiverID] <= recTimestamp.getClockValue(receiverID))
		{
			this.clocks[receiverID] = 1 + recTimestamp.getClockValue(receiverID);
		}
		
		for (int i = 0; i < this.numProcs; i ++)
		{
			this.clocks[i] = Math.max(this.clocks[i], recTimestamp.getClockValue(i));
		}
	}
	
	/**
	 * String representarion: [x y z]
	 */
	public String toString()
	{
		String ret = "[ ";
		for(int i = 0; i < this.numProcs; i++)
		{
			ret += this.clocks[i] + " ";
		}
		
		ret += "]";
		return ret;
	}
	
	public static Timestamp valueOf(String s){
		s = s.substring(1, s.length() - 1);
		String[] clocks = s.split(" ");
		ArrayList<Integer> clockInts = new ArrayList<Integer>();
		for(String sClock : clocks)
			clockInts.add(Integer.valueOf(sClock));
		return new Timestamp(clockInts);
	}
	
	/**
	 * List of values representation (to be used for RPCs)
	 */
	public List<Integer> toList()
	{
		ArrayList<Integer> list = new ArrayList<Integer>();
		for (int i = 0; i < this.numProcs; i++)
		{
			list.add(this.clocks[i]);
		}
		
		return list;
	}
}
