package edu.mit.csail.aeolus.auditing.graphdb;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Queue;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import edu.mit.csail.aeolus.auditing.CausalityLink;
import edu.mit.csail.aeolus.auditing.EventRecord;
import edu.mit.csail.aeolus.auditing.LogReceiver;
import edu.mit.csail.aeolus.auditing.LoggingServiceInterface;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;

/**
 * This class receives (in a non-blocking fashion) 
 * logging information from Aeolus nodes and forwards those log records
 * to a LogReceiver class using a BlockingQueue.
 * @author blanks
 */
public class AeolusGraphLoggingService implements LoggingServiceInterface{

	private Registry registry;
	private Boolean initialized = false;
	private LogReceiver logger;
	private BlockingQueue<Runnable> bq = new LinkedBlockingQueue<Runnable>();

	private RuntimeConfig config = RuntimeConfig.get(); 
	public void initialize() throws RemoteException {
		if (initialized)
			return;
		registry = LocateRegistry.createRegistry(config.getLoggingPort());

		LoggingServiceInterface serviceStub = (LoggingServiceInterface) UnicastRemoteObject
					.exportObject(this, 0);
		registry.rebind(config.getAeolusLoggingServiceName(), serviceStub);

		initializeDBLogger();

		initializeWatcherThread();

		initialized = true;
	}

	private boolean shutdown = false;
	
	private void initializeWatcherThread() {
		Runnable r = new Runnable() {
			public void run() {
				while (!shutdown) {
					try {
						Runnable lp = bq.poll(1, TimeUnit.SECONDS);
						if (lp != null)
							lp.run();
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		};
		Thread t = new Thread(r);
		t.start();
	}

	public void shutdownService(){
		if(registry != null){
			try{
				UnicastRemoteObject.unexportObject(registry ,true);
			} catch (NoSuchObjectException e){
				e.printStackTrace();
			}  
		}
		shutdown = true;
		logger.shutdown();
	}

	private void initializeDBLogger() {
		System.out.println("Initializing Logger");

		logger = LogProcessor.getInstance();

		System.out.println("FINISHED initializing Logger");
	}

	/**
	 * Receive log records via Java serialization mechanism
	 */

	private class LogProcess implements Runnable {
		Integer platformid;
		Date timestamp;
		HashMap<Long, ArrayList<EventRecord>> eventsTable;

		LogProcess(final Integer platformid, final Date timestamp,
				final HashMap<Long, ArrayList<EventRecord>> eventsTable) {
			this.platformid = platformid;
			this.timestamp = timestamp;
			this.eventsTable = eventsTable;
		}

		public void run() {
			processAndStoreLogs(platformid, timestamp, eventsTable);
		}
	}

	private class AuthLogProcess implements Runnable {
		ArrayList<EventRecord> eventList;
		Date timestamp;

		AuthLogProcess(final Date timestamp,
				final ArrayList<EventRecord> eventList) {
			this.timestamp = timestamp;
			this.eventList = eventList;
		}

		public void run() {
			logger.writeEntries(-1, -1, eventList, timestamp);
		}
	}

	private class FileLogProcess implements Runnable {
		List<EventRecord> eventList;
		Date timestamp;

		FileLogProcess(final String hostname, final Date timestamp,
				final List<EventRecord> eventList) {
			this.timestamp = timestamp;
			this.eventList = eventList;
		}

		public void run() {
			logger.writeEntries(-1, -1, eventList, timestamp);
		}
	}

	public void receiveLogs(final Integer platformid,  final Date timestamp, 
			final HashMap<Long, ArrayList<EventRecord>> eventsTable) 
	throws RemoteException
	{
		bq.add(new LogProcess(platformid, timestamp, eventsTable));
	}

	public void receiveFileSystemLogs(String hostname,  Date timestamp, Queue<EventRecord> events) 
	throws RemoteException{
		List<EventRecord> l = Arrays.asList(events.toArray(new EventRecord[0]));
		bq.add(new FileLogProcess(hostname, timestamp, l));
	}

	public void receiveLogs(String xmlString) throws RemoteException{
		System.out.println("received logs");
	}

	/**
	 * Process the log record information and store into the database
	 */
	long count = 0;
	long thresh = 0;
	long time = 0;
	public void processAndStoreLogs(Integer platformid, Date timestamp, HashMap<Long, ArrayList<EventRecord>> eventsTable){
		long s = System.currentTimeMillis();
		for(Entry<Long,ArrayList<EventRecord>> e : eventsTable.entrySet()){
		    if (e.getValue().size() == 0)
			continue;
			System.out.println("Pushing " + e.getValue().size());
			logger.writeEntries(platformid, e.getKey(), e.getValue(), timestamp);
			count += e.getValue().size();
			if(count > thresh){
				System.out.println("Log Count: " + count);
				thresh = count + 200;
			}
		}
		long e = System.currentTimeMillis();
		time += (e - s);
	}
	
	public void printStats(){
		System.out.println("Count: \t" + count);
		System.out.println("Time: \t" + time);
	}

	public int getSessionUid() throws RemoteException{
		return 1;
	}

	public ArrayList<CausalityLink> getCausalityLinks() throws SQLException{
		return null;
	}

	public void receiveAuthLogs(Date timestamp,
			ArrayList<EventRecord> eventsList) throws RemoteException {
		bq.add(new AuthLogProcess(timestamp, eventsList));
	}

	public List<EventRecord> getAllRecords() throws RemoteException,
			SQLException {
		return null;
	}
}
