package edu.mit.csail.aeolus.auditing.graphdb;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import edu.mit.csail.aeolus.auditing.EventCounter;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.auditing.EventQueueReader;
import edu.mit.csail.aeolus.auditing.EventRecord;
import edu.mit.csail.aeolus.auditing.LogReceiver;
import edu.mit.csail.aeolus.auditing.graphdb.wrappers.AbstractEvent;
import edu.mit.csail.aeolus.auditing.graphdb.wrappers.AbstractEventCache;
import edu.mit.csail.aeolus.auditing.graphdb.wrappers.EventFactory;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;

/**
 * This class handles the analysis and storage of logging data.
 * Events are received in the form of EventRecord objects. This class uses
 * the EventFactory class to translate EventRecord objects into AbstractEvent 
 * objects.
 * 
 * To add event types, you need to do a lot of stuff:
 *  (1) Add a wrapper class for the event -- examples are in the wrappers package
 *  and add this wrapper to EventFactory and the EventVisitor
 *  (2) To store the event in the SQL database -- the visitor in the SQLEventStore
 *  store needs to know how to store the event, the examples there should be fairly clear.
 *  (3) If you need to perform some analysis look at the Analyse interface and the
 *  AnalyseFull class.
 * @author blanks
 *
 */
public class LogProcessor  implements LogReceiver{
	private AbstractEventCache eventCache = AbstractEventCache.getInstance();
	private BlockingQueue<AbstractEvent> outQ = new LinkedBlockingQueue<AbstractEvent>();
	private long evCounter = 0;
	
	// map from predecessor to eventrecord waiting for that predecessor
	private Map<EventID, List<EventRecord>> waiting 
		= new HashMap<EventID, List<EventRecord>>();
	
	private static LogProcessor singleton;
	
	private RuntimeConfig config = RuntimeConfig.get();
	
	private LogProcessor(){	
		SQLEventStore.get(config.isLogProcessorActive());
		AnalyseFull p = new AnalyseFull();
		EventQueueReader eqr = new EventQueueReader(p, outQ);
		Thread t = new Thread(eqr);
		t.start();
	}
	
	public static LogProcessor getInstance(){
		if(singleton == null)
			singleton = new LogProcessor();
		return singleton;
	}

	public synchronized 
	void writeEntries(int platformID, long userd, List<EventRecord> elist, Date timestamp){
		for(EventRecord l : elist){
			// Uniqueness!!!
			if( eventCache.getEvent(l.getEventId()) != null )
				continue;

			if(timestamp != null)
				l.setTimestamp(timestamp);
			if(userd != -1)
				l.setUserD(userd);
			if(platformID != -1)
				l.setPlatformID(platformID);
			
			// Check to see if the predecessors have already been through the LP.
			EventID check = l.getProcessPred();
			if(check != null && eventCache.getEvent(check) == null){
				if(!waiting.containsKey(check))
					waiting.put(check, new ArrayList<EventRecord>());
				waiting.get(check).add(l);
				System.out.println("1ignoring: " + l + "\n" + 
									"because of " + check);
				continue;
			}
			check = l.getCommunicationPred();
			if(check != null && eventCache.getEvent(check) == null){
				if(!waiting.containsKey(check))
					waiting.put(check, new ArrayList<EventRecord>());
				waiting.get(check).add(l);
				System.out.println("2ignoring: " + l + "\n" + 
						"because of " + check);				continue;
			}
			for(EventID otherPred : l.getOtherPreds()){
				if(otherPred != null && eventCache.getEvent(otherPred) == null){
					if(!waiting.containsKey(otherPred))
						waiting.put(otherPred, new ArrayList<EventRecord>());
					waiting.get(otherPred).add(l);
					System.out.println("3ignoring: " + l + "\n" + 
							"because of " + check);
					continue;
				}	
			}
			
			AbstractEvent ev = EventFactory.getEvent(l);
			EventID myID = l.getEventId();
			eventCache.addEvent(myID, ev);
			ev.setEventCounter(new EventCounter(evCounter++));
			outQ.add(ev);
			
			// Lookup + process any waiting descendants!
			if(waiting.containsKey(myID)){
				List<EventRecord> descendants = waiting.remove(myID);
				System.out.println(descendants);
				writeEntries(-1, -1, descendants, null);
			}
		}

	}

	public void shutdown() {	
	}
}
