package edu.mit.csail.aeolus.authority.api;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.List;

import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.stream.client.InvalidationPacket;
import edu.mit.csail.aeolus.stream.server.api.RegisterReturn;
import edu.mit.csail.aeolus.stream.server.api.SnapshotNumAndName;
import edu.mit.csail.aeolus.utilities.Pair;
import edu.mit.csail.aeolus.exceptions.AuthorityException;

public interface InterfaceAuthorityServer extends Remote {

	/**
     *	Generates a new principal ID in the core of the creator.
     *	RESULTS: Creator principal has ownership of the created principal.
     *	(Add Owns-link in PH; Record pidCreator as creator of pidCreated)
     * @return Pair of integers: 
     * 		First: the new PID 
     *  	Second: the core of the new PID
     *  Returns NULL on error
	 * @throws AuthorityException if pidCreator is the public PID 
     */
	AuthReply<PrincipalReply> createPrincipal(PrincipalID pidCreator, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<PrincipalReply> createPrincipalGI(PrincipalID pidCreator, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;
    
    /**
     * Generates a new principal ID in a new core
     * @return Pair: First = the new PID, Second = the core of the new PID
     * @throws AuthorityException if pidCreator is the public PID 
     */
    AuthReply<PrincipalReply> createPrincipalInNewCore(PrincipalID pidCreator, EventID pred)  throws RemoteException, AuthorityException;
    AuthReply<PrincipalReply> createPrincipalInNewCoreGI(PrincipalID pidCreator, int seqNumFirstEntry, EventID pred)  throws RemoteException, AuthorityException;

    /**
     * Creates a new Taint
     *	RESULTS: Creator principal is explicitly granted privilege for taint.
	 *	(Record pidCreator as creator of tidCreated)
     * @return Integer ID of the new taint
     */
    AuthReply<TagID> createTag(PrincipalID pidCreator, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<TagID> createTaintGI(PrincipalID pidCreator, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;

    /**
     * Creates a new closure.
     * Results: Generate a new closure bound to a new principal in a new core.
     * @return int[] containing the closureId of the new closure, pidClosure of the new principle.
     */
    AuthReply<ClosureReply> createClosure(String className, EventID pred) throws RemoteException;
    AuthReply<ClosureReply> createClosureGI(String className, int seqNumFirstEntry, EventID pred) throws RemoteException;

    /**
     * Allows a closure to act-for pidActee
     * @param pidRequestor The principal that requested this link (this principal's authority is checked.)
     * @param cid The closure that will act-for
     * @throws AuthorityException if pidRequestor does not have authority to assign act-for links for pidActee
     */
    AuthReply<Void> closureActsFor(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidActee, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<Void> closureActsForGI(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidActee, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;

    /**
     * Grants authority to closure cid for taint tid through pidGrantor's authority.
     * @param pidRequestor Principal must act-for pidGrantor
     * @param pidGrantor Principal upon whose authority this grant rests
     * @param tid Taint to grant
     * @param cid closure to be granted authority
     * @throws AuthorityException if pidRequestor does not act-for pidGrantor or pidGrantor is not authoritative for the taint.
     */
    AuthReply<Void> closureGrant(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidGrantor, TagID tid, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<Void> closureGrantGI(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidGrantor, TagID tid, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;
    
    /**
     * Grants authority to principal for taint through pidGrantor's authority.
     * @param pidRequestor Principal must act-for pidGrantor
     * @param pidGrantor Principal upon whose authority this grant rests
     * @param pidGrantee Principal receiving the grant
     * @param tid Taint being granted
     * @throws AuthorityException if pidRequestor does not act-for pidGrantor or pidGrantor is not authoritative for the taint.
     */
    AuthReply<Void> grant(PrincipalID pidRequestor, PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<Void> grantGI(PrincipalID pidRequestor, PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;

    /**
     *	Principal pidRequestor allows principal pidFrom to act for pidTo.
     *	Authority on taints is transitive. Principal pidFrom will have all authority on taints that
     *	principal pidTo has.
     *
     *	@throws AuthorityException if requestor does not own principal pidTo, or if this acts-for link
     *	would create a cycle in the acts-for graph.
     */
    AuthReply<Void> delegateActsFor(PrincipalID pidRequestor, PrincipalID pidFrom, PrincipalID pidTo, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<Void> delegateActsForGI(PrincipalID pidRequestor, PrincipalID pidFrom, PrincipalID pidTo, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;

    /**
     * Revokes a grant from principal pidGrantor to principal pidGrantee on taint tid.
     * @throws AuthorityException if pidRevoker does not act-for pidGrantor.
     */
    AuthReply<Void> revokeGrant(PrincipalID pidRevoker, PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<Void> revokeGrantGI(PrincipalID pidRevoker, PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;

    /**
     * Revokes a the act-for link from pidActor to pidActee and removes any transitive grants.
     * pidActor acts-for pidActee before, and does not after.
     * @throws AuthorityException if pidRevoker does not act-for pidActee.
     */
    AuthReply<Void> revokeActsFor(PrincipalID pidRevoker, PrincipalID pidActor, PrincipalID pidActee, EventID pred) throws RemoteException, AuthorityException;
    AuthReply<Void> revokeActsForGI(PrincipalID pidRevoker, PrincipalID pidActor, PrincipalID pidActee, int seqNumFirstEntry, EventID pred) throws RemoteException, AuthorityException;

    /*
     * Snapshot
     */
    /**
     * Creates a new snapshot of the authority state
     * @param snapshotVersion version number of this snapshot (TODO: this really shouldn't be exposed here...)
     * @return the sequence number and the snapshot name of the new snapshot.
     */
    SnapshotNumAndName createSnapshot(int snapshotVersion) throws RemoteException;

    /**
     * Removes old snapshots from the server
     */
    void removeOldSnapshots() throws RemoteException;
    
    /**
     * Warms up snapshotDbName // 
     * @param snapshotDbName
     * @throws RemoteException
     */
    void warmupDB(String snapshotDbName) throws RemoteException;

    /*int LoadCoresFromSnapshot(String snapshotDbName, List<Integer> coreIds,
        out List<TaintTableEntry> taintEntries,
        out List<PrincipalTableEntry> principalEntries,
        out List<GrantTableEntry> grantEntries,
        out List<LinkTableEntry> linkEntries,
        out List<ClosureTableEntry> closureEntries);*/

    /**
     * Loads the contents of cores listed in coreIds from the snapshot snapshotDbName
     */
    CoreContents loadCoresFromSnapshot(String snapshotDbName, List<Integer> coreIds)  throws RemoteException;

    /**
     * Loads the contents of the core containing the principal pid from snapshot snapshotDbName
     * @return Pair: First = Integer Core ID of the principal. 
     * 	Second = the contents of the core.
     */
    CoreContents loadCoreFromSnapshot(String snapshotDbName, PrincipalID pid)  throws RemoteException;
    CoreContents loadCoreFromSnapshot(String snapshotDbName, ClosureID closure) throws RemoteException;
    
	RegisterReturn registerClient(String clientRMIString) throws RemoteException;
	Integer getNextPlatformInstanceID() throws RemoteException;
	void createAuthorityServer() throws RemoteException;
	void removeAllSnapshots() throws RemoteException;

	/**
     * Returns the re-do log entries between seqNumFirstEntry and seqNumLastEntry
     * @return a packet of invalidations which detail changes in the authority state
     */
	InvalidationPacket getInvalidationPacket(int seqNumFirstEntry, int seqNumLastEntry) throws RemoteException;
	/**
     * Returns the re-do log entries between seqNumFirstEntry and the most recent invalidation
     * @return a packet of invalidations which detail changes in the authority state
	 */
	InvalidationPacket getInvalidationPacket(int seqNumFirstEntry) throws RemoteException;
	boolean hasAuthority(PrincipalID pid, TagID tid) throws RemoteException;

	public AuthReply<TagID> createSubTagGI(TagID parent, PrincipalID creator,
			int seqNumFirstEntry, EventID predecessor)
			throws RemoteException, AuthorityException;

	public AuthReply<TagID> createSubTag(TagID parent, PrincipalID creator,
			EventID predecessor) throws RemoteException, AuthorityException;
	public ClosureTableEntry getClosureInfo(ClosureID closure) throws RemoteException;
	public boolean isActee(int pidActor, int pidActee) throws RemoteException;
	public Pair<Boolean, List<Integer>> isActeeBasis(int pidActor, int pidActee) throws RemoteException;
	public boolean isSystemNodeRegistered(int value) throws RemoteException;
}
