package edu.mit.csail.aeolus.authority.client;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import edu.mit.csail.aeolus.api.SystemNodeID;
import edu.mit.csail.aeolus.authority.api.ClosureTableEntry;
import edu.mit.csail.aeolus.authority.api.CompoundTagTableEntry;
import edu.mit.csail.aeolus.authority.api.GrantTableEntry;
import edu.mit.csail.aeolus.authority.api.LinkTableEntry;
import edu.mit.csail.aeolus.authority.api.NodeTableEntry;
import edu.mit.csail.aeolus.authority.api.PrincipalTableEntry;
import edu.mit.csail.aeolus.authority.api.TaintTableEntry;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.infoflow.AuthStateConstants;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.utilities.Pair;

public abstract class AbstractCoreCache {

	private static final int BAD_INT = -1;
	// In-memory cached data tables

    LinkedList<TaintTableEntry> taintTable;
    ReadWriteLock taintLock = new ReentrantReadWriteLock();
    LinkedList<CompoundTagTableEntry> compoundTable; 
    ReadWriteLock compoundLock = new ReentrantReadWriteLock();
    LinkedList<PrincipalTableEntry> principalTable;
    ReadWriteLock principalLock = new ReentrantReadWriteLock();
    LinkedList<GrantTableEntry> grantTable;
    ReadWriteLock grantLock = new ReentrantReadWriteLock();
    LinkedList<LinkTableEntry> linkTable;
    ReadWriteLock linkLock = new ReentrantReadWriteLock();
//    BlockingQueue<ClosureTableEntry> closureTable;
    HashMap<Integer, ClosureTableEntry> closureTable;
    ReadWriteLock closureLock = new ReentrantReadWriteLock();
    HashMap<NodeTableEntry, Boolean> nodeTable;
    ReadWriteLock nodeLock = new ReentrantReadWriteLock();
    
    
    
    public AbstractCoreCache(){
        // In-memory data tables
        taintTable = new LinkedList<TaintTableEntry>(); // initialCapacity*avgEntryPerCore
        principalTable = new LinkedList<PrincipalTableEntry>();
        grantTable = new LinkedList<GrantTableEntry>();
        linkTable = new LinkedList<LinkTableEntry>();
//        closureTable = new LinkedBlockingQueue<ClosureTableEntry>();
        closureTable = new HashMap<Integer, ClosureTableEntry>();
        compoundTable = new LinkedList<CompoundTagTableEntry>();
        nodeTable = new HashMap<NodeTableEntry, Boolean>();
    }
        
    /**
     * @param dependentCores will be modified to add dependent cores to the list.
     */
    public boolean hasAuthority(int pid, int tid, List<Integer> dependentCores){
    	// 1. Check that we know at least the pid
    	int coreOfPid = isPidCached(pid);
        if (coreOfPid == BAD_INT){
            handleMissOnPrincipal(pid);
        }
        List<Integer> dependentPids = new LinkedList<Integer>();
        
        
        Pair<Boolean, Integer> hasAuth = hasAuthorityCompoundCheck(pid, tid, 0, dependentPids);

        if (hasAuth.first){
        	addDependentCores(dependentPids, dependentCores);
        }
        return hasAuth.first;
    }

    private static class AuthLookup implements Objective<PrincipalTableEntry>{
    	private int depPid;
		public AuthLookup(int depPid) {
			this.depPid = depPid;
		}
		public boolean match(PrincipalTableEntry entry) {
    		return (entry.pid == depPid);
    	}
    }
    
    /**
     * Shoves the cores of the dependentPids into the input list dependentCores.
     * @param dependentPids
     * @param dependentCores
     */
    private void addDependentCores(Collection<Integer> dependentPids, List<Integer> dependentCores){
		for (int depPid : dependentPids) {
			PrincipalTableEntry principalEntry = search(principalTable, principalLock, new AuthLookup(depPid));
			if (principalEntry != null) {
				// only unique entries in dependentCores
				// FIXME: this may be slow, should we use hashmap?
				if (dependentCores.contains(principalEntry.coreIdCreated) == false) {
					dependentCores.add(principalEntry.coreIdCreated);
				}
			}
		}
	}
    
    public boolean isActee(int pidFrom, int pidTo, List<Integer> dependentCores){
        // 1. Check that we know at least the pid
        int coreOfPid = isPidCached(pidFrom);
        if (coreOfPid == BAD_INT){
            handleMissOnPrincipal(pidFrom);
        }
        Stack<Integer> dependentPids = new Stack<Integer>();
        boolean r = findPidRecursively(pidFrom, pidTo, dependentPids);
        if(r){
        	addDependentCores(dependentPids, dependentCores);
        }
        return r;
    }

    public Pair<Boolean, List<Integer>> isActeeBasis(int pidActor, int pidActee){
        List<Integer> basis = new LinkedList<Integer>();

        // 1. Check that we know at least the pid
        int coreOfPid = isPidCached(pidActor);
        if (coreOfPid == BAD_INT){
            handleMissOnPrincipal(pidActor);
        }

        Stack<Integer> stackBasis = new Stack<Integer>();
        boolean yes = findPidRecursively(pidActor, pidActee, stackBasis);
        if (yes){
        	// stack will not have pidActor
        	basis.add(pidActor);
        	basis.addAll(stackBasis);
        }
        Pair<Boolean,List<Integer>> r = new Pair<Boolean, List<Integer>>(yes, basis);
        return r;   
    }
    
    public Boolean isActeeBasis(int pidActor, int pidActee, List<Integer> basisOut){

        // 1. Check that we know at least the pid
        int coreOfPid = isPidCached(pidActor);
        if (coreOfPid == BAD_INT){
            handleMissOnPrincipal(pidActor);
        }
        
        Stack<Integer> stackBasis = new Stack<Integer>();
        boolean yes = findPidRecursively(pidActor, pidActee, stackBasis);
        if (yes){
        	// stack will not have pidActor
        	basisOut.add(pidActor);
        	basisOut.addAll(stackBasis);
        }
        return yes;

    }

    public boolean isBasisValid(List<Integer> basis){
        boolean valid = true;

        int prevPid = BAD_INT;
        boolean first = true;
        for (Integer currPid : basis){
            // check that the pid is cached
            int coreOfPid = isPidCached(currPid);
            if (coreOfPid == BAD_INT){
                // pull in the core of the pid
                handleMissOnPrincipal(currPid);
            }
            
            if (! first){   
                // check whether currPid acts-for prevPid
                boolean found = null != search(linkTable, linkLock, new BasisCheck(prevPid,currPid));
                if (!found){
                    valid = false;
                    break;
                }
            }
            first = false;
            prevPid = currPid;
        }
        return valid;
    }
    
    private static class BasisCheck implements Objective<LinkTableEntry>{
    	private int prevPid, curPid;
		public BasisCheck(int prev, int cur){
			prevPid = prev;
			curPid = cur;
		}
		public boolean match(LinkTableEntry entry) {
            return ((entry.pidFrom == prevPid) && (entry.pidTo == curPid));
    	}
    }

    public ClosureTableEntry getClosureInfo(final ClosureID closure) throws AeolusException{
    	closureLock.readLock().lock();
    	try{
    		ClosureTableEntry foundEntry = closureTable.get(closure.getValue());
    		if(foundEntry != null)
    			return foundEntry;
    	}finally{
    		closureLock.readLock().unlock();
    	}
        // It's not cached, we need to fetch another core...
    	handleMissOnClosure(closure);
    	// check if its in the cache again
    	closureLock.readLock().lock();
    	try{
    		ClosureTableEntry foundEntry = closureTable.get(closure.getValue());
    		if(foundEntry == null)
    			throw new AeolusException("No such closure " + closure + " in this snapshot.");
        
    		return foundEntry; // if null, the closure does not exist
    	}finally{
    		closureLock.readLock().unlock();
    	}
    }


    public void handleRegisterNode(NodeTableEntry entry){
    	nodeLock.writeLock().lock();
    	try{
    		nodeTable.put(entry, new Boolean(true));
    	}finally{
    		nodeLock.writeLock().unlock();
    	}
    }
    
    public void handleRemoveNode(final SystemNodeID node){
    	final int compareToID = node.getValue();
    	removeAll(nodeTable, nodeLock, new Objective<NodeTableEntry>() {
    		public boolean match(NodeTableEntry entry) {
    			return entry.nodeId == compareToID;
					}
			});	
    }
    
	public void handleCreateCompound(CompoundTagTableEntry compEntry) {
		compoundLock.writeLock().lock();
		try{
			compoundTable.add(compEntry);
		}finally{
			compoundLock.writeLock().unlock();
		}
	}
    
	public void handleCreatePrincipal(PrincipalTableEntry p, List<Integer> coreIds){
		// 1. Insert Principal Table(pidCreated, pidCreator, coreId)

		if (checkCores(coreIds,p.coreIdCreated)){
			principalLock.writeLock().lock();
			try{
				principalTable.add(p);
			}finally{
				principalLock.writeLock().unlock();
			}
		}

		if(checkCores(coreIds,p.coreIdCreator)){
			linkLock.writeLock().lock();
			// 2. Insert LinkTable (pidCreator Owns pidCreated)
			try{
				LinkTableEntry entryLink = new LinkTableEntry();
				entryLink.pidFrom = p.creatorPid;
				entryLink.pidTo = p.pid;
				entryLink.linkType = LinkTableEntry.LinkType.L_OWNS;
				entryLink.coreId = p.coreIdCreator;
				entryLink.coreIdTo = p.coreIdCreated;
				linkTable.add(entryLink);
			}finally{
				linkLock.writeLock().unlock();
			}
		}
	}

    public void handleCreateTaint(TaintTableEntry tt, List<Integer> coreIds){
        // FAST FILTER
        if (!checkCores(coreIds,tt.coreId)) return;

        // 1. Insert TaintTable(tidCreated, pidCreator)
        taintLock.writeLock().lock();
        try{
        	taintTable.add(tt);
        }finally{
        	taintLock.writeLock().unlock();
        }
        // 2. Insert GrantTable
        GrantTableEntry entryGrant = new GrantTableEntry(AuthStateConstants.pidSystem, tt.creatorPid, tt.tid, tt.coreId);
        grantLock.writeLock().lock();
        try{
        	grantTable.add(entryGrant);
        }finally{
        	grantLock.writeLock().unlock();
        }
    }

    public void handleCreateClosure(ClosureTableEntry c, List<Integer> coreIds){
        // FAST FILTER
        if (!checkCores(coreIds,c.coreId)) return;

        // 1. Insert ClosureTable
        closureLock.writeLock().lock();
        try{
        	closureTable.put(c.cid , c);
        }finally{
        	closureLock.writeLock().unlock();
        }
    }

    public void handleGrant(GrantTableEntry g, List<Integer> coreIds){
        // FAST FILTER
        if (!checkCores(coreIds,g.coreId)) return;

        // 1. Insert GrantTable
        grantLock.writeLock().lock();
        try{
        	grantTable.add(g);
        }finally{
        	grantLock.writeLock().unlock();
        }
    }

    public void handleDelegateActsFor(int pidFrom, int pidTo, int coreFrom, int coreTo, List<Integer> coreIds){
        // FAST FILTER
        if (!checkCores(coreIds,coreFrom)) return;

        // 1. Insert LinkTable(pidFrom, pidTo, L_ACTSFOR)
        LinkTableEntry entryLink = new LinkTableEntry();
        entryLink.pidFrom = pidFrom;
        entryLink.pidTo = pidTo;
        entryLink.linkType = LinkTableEntry.LinkType.L_ACTSFOR;
        entryLink.coreId = coreFrom;
        entryLink.coreIdTo = coreTo;
        linkLock.writeLock().lock();
        try{
        	linkTable.add(entryLink);
        }finally{
        	linkLock.writeLock().unlock();
        }
    }

    public void handleRevokeGrant(final int pidGrantor, final int pidGrantee, final int tid, int coreId, List<Integer> coreIds){
        // FAST FILTER
        if (!checkCores(coreIds,coreId)) return;

        // 1. Delete from GrantTable (all rows where <pidGrantor, pidGrantee, tid> match
        removeAll(grantTable, grantLock, new Objective<GrantTableEntry>(){
        	public boolean match(GrantTableEntry entry) {
                return ((entry.grantorPid == pidGrantor) && (entry.granteePid == pidGrantee) && (entry.tid == tid));
        	}
        });  
    }

    public void handleRevokeActsFor(int coreId, final int pidActor, final int pidActee, List<Integer> coreIds)
    {
        // FAST FILTER
        if (!checkCores(coreIds,coreId)) 
        	return;
        // 1. Delete from LinkTable
        removeAll(linkTable, linkLock, new Objective<LinkTableEntry>(){ 
        	public boolean match(LinkTableEntry entry) {
        		return (entry.pidFrom == pidActor && entry.pidTo == pidActee && entry.linkType == LinkTableEntry.LinkType.L_ACTSFOR);}
        });
    }

    /**
     * 
     * @param pid
     * @param tid
     * @param currentDepth
     * @param dependentPids -- this method appends any dependent pids to this list
     * @return Pair.first = boolean hasAuthority
     * 	Pair.second = depth at which the answer was found.
     */
    private Pair<Boolean, Integer> hasAuthorityCompoundCheck(final int pid, int tid, int currentDepth, List<Integer> dependentPids){
        // 1. Check whether tid is a child of a compound tag.
    	TagID parentTag = hasParentTag(tid);
    	if(parentTag != null){
    		Pair<Boolean, Integer> rv = hasAuthorityCompoundCheck(pid, parentTag.getValue(), currentDepth, dependentPids);
    		if(rv.first)
    			return rv;
    	}
    	
    	return hasAuthorityRecursively(pid, tid, currentDepth, dependentPids);
    }
    
    private Pair<Boolean,Integer> hasAuthorityRecursively(final int pid, int tid, int currentDepth, List<Integer> dependentPids){
    	
    	// 1. Check whether pid has explicit auth for taint
        if (hasExplicitAuthority(pid, tid)){
            // definite YES
            // add core of this pid (the 'child')
            dependentPids.add(pid);
            return new Pair<Boolean, Integer>(true, currentDepth);
        }

        // 2. Ask Children
        currentDepth++;
        List<LinkTableEntry> children = searchAll(linkTable, linkLock, new Objective<LinkTableEntry>(){
        	public boolean match(LinkTableEntry e){
        		return e.pidFrom == pid;
        	}
        });

        int maxChildDepthNotFound = currentDepth;
        List<LinkTableEntry> childrenWithMiss = new LinkedList<LinkTableEntry>();
        for (LinkTableEntry childEntry : children){

        	int childCore = childEntry.coreIdTo;
            if (isCached(childCore)){

                Pair<Boolean, Integer> childAuth = hasAuthorityRecursively(childEntry.pidTo, tid, currentDepth, dependentPids);
                if (childAuth.first){
                    dependentPids.add(pid);
                    return childAuth; // propagate the depth info up the tree
                }else{
                	// check the depth that the child went to.
                    if (childAuth.second > maxChildDepthNotFound) 
                    	maxChildDepthNotFound = childAuth.second;
                }
            }else{
                // record the miss and try later.
                childrenWithMiss.add(childEntry);
            }
        }
        // Try the ones that missed (lazy refill)
        for (LinkTableEntry childEntry : childrenWithMiss){

        	// load core into cache
            handleMissOnCore(childEntry.coreIdTo);

            // Core is now cached (proceed)
            Pair<Boolean, Integer> childAuth = hasAuthorityRecursively(childEntry.pidTo, tid, currentDepth, dependentPids);
            if (childAuth.first){
                dependentPids.add(pid);
                return childAuth; // propagate the depth info up the tree
            }else{
            	// check the depth that the child went to.
                if (childAuth.second > maxChildDepthNotFound) 
                	maxChildDepthNotFound = childAuth.second;
            }
        }
        return new Pair<Boolean, Integer>(false, maxChildDepthNotFound);
    }
    
    private boolean findPidRecursively(final int pidFrom, int pidTo, Stack<Integer> stackBasis)
    {
        if (pidTo == pidFrom){
        	if(stackBasis != null)
        		stackBasis.push(pidFrom);
        	return true;
        }

        // 1. Ask Children of pidTarget
        List<LinkTableEntry> children = searchAll(linkTable, linkLock, new Objective<LinkTableEntry>(){
        	public boolean match(LinkTableEntry e){
        		return e.pidFrom == pidFrom;
        	}
        });
        
        List<LinkTableEntry> childrenWithMiss = new LinkedList<LinkTableEntry>();
        for (LinkTableEntry childEntry : children){
            int childCore = childEntry.coreIdTo;
            // MINOR OPT: we try the cores that are cached first
            if (isCached(childCore)){
                // check this core
                boolean childAuth = findPidRecursively(childEntry.pidTo, pidTo, stackBasis);
                if (childAuth){
                	if(stackBasis != null)
                		stackBasis.push(childEntry.pidTo);
                    return true;
                }else{
                    // ignore
                }
            }else{
                // record the index and try later
                childrenWithMiss.add(childEntry);
            }
        }

        // Try the ones that missed (lazy refill)
        for (LinkTableEntry childEntry : childrenWithMiss){
//            LinkTableEntry childEntry = children[childrenWithMiss[i]];

            // load core into cache
        	handleMissOnCore(childEntry.coreIdTo);

            // Core is now cached (proceed)
            boolean childAuth = findPidRecursively(childEntry.pidTo, pidTo, stackBasis);
            if (childAuth){
            	if(stackBasis != null)
            		stackBasis.push(childEntry.pidTo);
                return true;
            }else{
                // ignore
            }
        }
        return false;
    }

    private TagID hasParentTag(final int tagId){
    	CompoundTagTableEntry e = search(compoundTable, compoundLock, new Objective<CompoundTagTableEntry>(){
			public boolean match(CompoundTagTableEntry entry) {
				return (entry.child == tagId);
			}
    	});
    	if(e == null)
    		return null;
    	return new TagID(e.parent);
    }
    
	private boolean hasExplicitAuthority(final int pid, final int tid){
        // filterPid *independently*
        GrantTableEntry found =
        	search(grantTable, grantLock, new Objective<GrantTableEntry>(){
        		public boolean match(GrantTableEntry entry) {
                    return ((entry.granteePid == pid) && (entry.tid == tid));
        		}
        	});

        return found != null;
    }

	public boolean isSystemNodeRegistered(final int systemNode){
		NodeTableEntry found = search(nodeTable, nodeLock, new Objective<NodeTableEntry>() {
			public boolean match(NodeTableEntry entry) {
				return entry.nodeId == systemNode;
			}
		});
		return found != null;
	}
	
    // IsPidCached
    // Returns the pid if in cache.
    //  BAD_INT otherwise.
    protected int isPidCached(final int pid){
        PrincipalTableEntry foundEntry = search(principalTable, principalLock, new Objective<PrincipalTableEntry>(){
        	public boolean match(PrincipalTableEntry e){
        		return e.pid == pid;
        	}
        } );
        if (foundEntry == null) 
        	return BAD_INT;
        return foundEntry.coreIdCreated;
    }
    
    protected boolean isCached(final int coreId){
    	return true;
    }
    
    private static interface Objective<T>{
    	boolean match(T entry);
    }
    
    private <T> T search(Collection<T> l, ReadWriteLock lock, Objective<T> comp){
    	lock.readLock().lock();
    	try{
    		for(T cur : l){
    			if(comp.match(cur))
    				return cur;
    		}
    		return null;
    	}finally{
    		lock.readLock().unlock();
    	}
    }

    private <T> T search(Map<T,?> l, ReadWriteLock lock, Objective<T> comp){
    	lock.readLock().lock();
    	try{
    		for(T cur : l.keySet()){
    			if(comp.match(cur))
    				return cur;
    		}
    		return null;
    	}finally{
    		lock.readLock().unlock();
    	}
    }
    
    private <T> List<T> searchAll(Collection<T> l, ReadWriteLock lock, Objective<T> comp){
    	lock.readLock().lock();
    	try{
	    	List<T> res = new LinkedList<T>();
	    	for(T cur : l){
	    		if(comp.match(cur))
	    			res.add(cur);
	    	}
	    	return res;
    	}finally{
    		lock.readLock().unlock();
    	}
    }
    
    private <T> void removeAll(Collection<T> l, ReadWriteLock lock, Objective<T> comp){
    	lock.writeLock().lock();
    	try{
    		Iterator<T> iter = l.iterator();
    		while(iter.hasNext()){
    			T cur = iter.next();
    			if(comp.match(cur))
    				iter.remove();
    		}
    	}finally{
    		lock.writeLock().unlock();
    	}
    }
    
    private <T> void removeAll(Map<T, ?> l, ReadWriteLock lock, Objective<T> comp){
    	lock.writeLock().lock();
    	try{

    		Iterator<T> iter = l.keySet().iterator();
    		while(iter.hasNext()){
    			T cur = iter.next();
    			if(comp.match(cur))
    				iter.remove();
    		}
    	}finally{
    		lock.writeLock().unlock();
    	}
    }

    abstract public void handleMissOnPrincipal(int pid);
    abstract public void handleMissOnClosure(ClosureID closure);
    abstract public void handleMissOnCore(int coreIdTo);
    /**
     * If this function returns true, this cache should handle the CoreId.
     * In implementation this is used to check whether or not a given stream message should
     * be handled.
     */
    abstract public boolean checkCores(List<Integer> CoreIds, int CoreId);
}
