package edu.mit.csail.aeolus.authority.client;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.mit.csail.aeolus.authority.client.api.EvictionMode;

public class ActsForCache {
	// Key: <pid,tid>
	// Value: <coresAffectedList>
	private Map<ActedKey, ActedValue> queryCache;

	// CoreDepInfo
	// Purpose: Fast determination of which QueryCache entries to invalidate
	// Key: <coreId>
	// Value: <HashBitMap of <pid,tid>> that are dependent
	private Map<Integer, CoreDepKeys> coreDepInfo;
	private EvictionMgr<ActedKey> evictionMgr;
	private int capacity;
	private final int evictionSize = 1;

	public ActsForCache(int capacity, EvictionMode evictionMode) {
		this.capacity = capacity;
		switch (evictionMode) {
		case RANDOM:
			evictionMgr = new EvictionRandom<ActedKey>();
			break;
		case FIFO:
			evictionMgr = new EvictionFifo<ActedKey>();
			break;
		case LRU:
			evictionMgr = new EvictionLru<ActedKey>();
			break;
		}
		queryCache = new HashMap<ActedKey, ActedValue>();
		coreDepInfo = new HashMap<Integer, CoreDepKeys>();
	}

	/**
	 * @returns true if the pid_actor acts-for pid_acted relationship is stored
	 *          in the cache.
	 */
	public boolean lookup(int acted, int actor) {
		boolean found = false;
		ActedKey lookup = new ActedKey(acted, actor);
		ActedValue value = queryCache.get(lookup);
		if (value != null) {
			found = true;
			evictionMgr.notifyLookup(lookup);
		}
		return found;
	}

	public void add(int acted, int actor, List<Integer> coresAffected) {

		// 1. Check whether we need to evict any to make room
		if (queryCache.size() >= capacity) {
			List<ActedKey> evictCandidates = evictionMgr
					.evictAndRemove(evictionSize);

			// 2. Remove suggested candidates from query cache
			for (ActedKey keyToRemove : evictCandidates) {
				ActedValue value = queryCache.get(keyToRemove);

				// update core dependency info
				List<Integer> coresList = value.coresAffected;
				for (Integer coreId : coresList) {
					CoreDepKeys keysMap = coreDepInfo.get(coreId);
					// if (coreDepInfo.tryGetValue(coreId, out keysMap))
					if (keysMap != null) {
						// remove key from map (update core bitmap)
						keysMap.dependents.remove(keyToRemove);
					}
				}

				// remove entry from query cache
				// NOTE: eviction metadata already updated
				queryCache.remove(keyToRemove);
			}
		}

		// 3. Add new entries to query cache
		ActedKey keyToAdd = new ActedKey(acted, actor);
		ActedValue valueToAdd = new ActedValue();
		valueToAdd.coresAffected = coresAffected;
		queryCache.put(keyToAdd, valueToAdd);

		// 4. Notify Eviction Manager
		// evictionMgr.NotifyAdded(keyToAdd);
		evictionMgr.notifyAdded(keyToAdd);

		// 5. Update cores dep info
		for (Integer coreId : coresAffected) {
			CoreDepKeys keysMap = coreDepInfo.get(coreId);
			if (keysMap != null) {
				keysMap.dependents.add(keyToAdd);
			} else {
				// create a new entry
				keysMap = new CoreDepKeys();
				keysMap.dependents.add(keyToAdd);
				coreDepInfo.put(coreId, keysMap);
			}
		}
	}

	private static class CoreDepKeys {
		Set<ActedKey> dependents = new HashSet<ActedKey>();
	}

	private static class ActedValue {
		List<Integer> coresAffected;
	}

	private static class ActedKey {
		public int acted;
		public int actor;

		public ActedKey(int acted, int actor) {
			this.acted = acted;
			this.actor = actor;
		}

		public boolean equals(Object obj) {
			if(!(obj instanceof ActedKey))
				return false;
			ActedKey a2 = (ActedKey) obj;
			return (acted == a2.acted && actor == a2.actor);
		}

		public int hashCode() {
			return 37 * acted + actor;
		}

		public String toString() {
			return acted + ", " + actor;
		}
	}

	public void invalidate(List<Integer> coresChanged) {
		for (int coreId : coresChanged){
    		CoreDepKeys coreDep = coreDepInfo.get(coreId);
    		if (coreDep != null){
    			for (ActedKey keyToInvalidate : coreDep.dependents){
    				// NOTE: it is possible for the keyPair to not be cached in L1
    				// It's ok for this bitmap to have some 'garbage' since we only 
    				// look at it when we are actually invalidating the core
    				if (queryCache.remove(keyToInvalidate) != null){
    					evictionMgr.notifyRemoved(keyToInvalidate);
    				}
    			}
    			// Remove the core entry
    			coreDepInfo.remove(coreId);
    		}
    	}
	}
}
