package edu.mit.csail.aeolus.authority.client;

import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.List;

import edu.mit.csail.aeolus.api.SystemNodeID;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.authority.api.AuthReply;
import edu.mit.csail.aeolus.authority.api.AuthorityConfiguration;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.authority.api.ClosureReply;
import edu.mit.csail.aeolus.authority.api.ClosureTableEntry;
import edu.mit.csail.aeolus.authority.api.InterfaceAuthorityServer;
import edu.mit.csail.aeolus.authority.api.PrincipalReply;
import edu.mit.csail.aeolus.authority.client.api.EvictionMode;
import edu.mit.csail.aeolus.authority.client.api.InterfaceAuthorityClient;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;
import edu.mit.csail.aeolus.stream.client.InvalidationPacket;
import edu.mit.csail.aeolus.stream.client.StreamClient;
import edu.mit.csail.aeolus.stream.server.api.InterfaceStreamServer;
import edu.mit.csail.aeolus.stream.server.api.RegisterReturn;
import edu.mit.csail.aeolus.stream.server.api.SnapshotNumAndName;
import edu.mit.csail.aeolus.utilities.Logger;
import edu.mit.csail.aeolus.utilities.LoggerLevel;
import edu.mit.csail.aeolus.utilities.Utils;

public class AuthorityClient implements InterfaceAuthorityClient{
	
	private static final int QUERY_CACHE_SIZE = 10000;

	private final boolean OPERATIONS_BLOCKING;
	private final boolean LOGGING_ENABLED;
	
	// Main Interfaces of Managed Objs
	CacheObjectManager cacheObjMgr;
	InterfaceAuthorityServer authServer;
	StreamClient streamClient;
	InterfaceStreamServer streamServer;

	// Settings
	int queryCacheSize;
	EvictionMode queryCacheEvictionMode;
	Logger logger;

	public AuthorityClient(int queryCacheSize, EvictionMode queryCacheEvictionMode, boolean blockingOperations){
		OPERATIONS_BLOCKING = blockingOperations;
		LOGGING_ENABLED = RuntimeConfig.get().isEnableLogging();
		
		logger = new Logger("AuthorityClient");

		cacheObjMgr = new CacheObjectManager();
		this.streamClient = new StreamClient(this.cacheObjMgr);

		this.queryCacheSize = queryCacheSize;
		this.queryCacheEvictionMode = queryCacheEvictionMode;
	}
	public void printStats(){
		cacheObjMgr.printStats();
	}
	
	public AuthorityClient(int queryCacheSize, EvictionMode queryCacheEvictionMode){
		this(queryCacheSize, queryCacheEvictionMode, true);
	}
	
	public AuthorityClient(){
		this(QUERY_CACHE_SIZE, EvictionMode.LRU);
	}

	public RegisterReturn register() throws AeolusException{

		try {
			Registry registry = LocateRegistry.getRegistry(AuthorityConfiguration.AuthorityHostName);
			authServer = (InterfaceAuthorityServer) registry.lookup(AuthorityConfiguration.AuthorityServiceName);
		} catch (Exception e) {
			e.printStackTrace();
		}

		SnapshotNumAndName lastSnapshot = new SnapshotNumAndName(0,"");

		RegisterReturn ret = streamClient.register();
		lastSnapshot = streamClient.getLastSnapshot();

		cacheObjMgr.initialize(authServer, queryCacheSize, 
				lastSnapshot.getSnapshotSeqNum(), lastSnapshot.getSnapshotDbName());
		
		return ret;
	}

	/*
	 * Public Methods
	 */

	//    public void forceStreamMessage(int coreId, int requestType, String streamMessage)
	//    {
	//        int seqNumDefault = -1;
	//        cacheObjMgr.handleStreamMessage(seqNumDefault, coreId, StreamMessageOpcode.values()[requestType], streamMessage);
	//    }

	/**
	 * Method will request invalidations from the server to see the results of 
	 * authority changes that took place in this client during the current epoch without
	 * triggering a new snapshot.
	 */
	public void getInvalidations(){
		try {
			InvalidationPacket ip = authServer.getInvalidationPacket(getCurrentInvalidationNumber() + 1);
			System.out.println("Manually received " + ip.invalidations.size() + " invalidations.");
			handleInvalidations(ip);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
	}
	
	public int getCurrentInvalidationNumber(){
		return cacheObjMgr.lastAppliedSeqNum;
	}
	
	/**
	 * @return the EventID of the last authority modification event that this
	 * client has witnessed.
	 */
	public EventID getCurrentPredecessor(){
		if(!LOGGING_ENABLED)
			return null;
		return cacheObjMgr.getCurrentPredecessor();
	}
	
	/**
	 * Moves this AuthorityClient forward in the stream to the given invalidation number. This method blocks while
	 * the authority state progresses.
	 * @param invalidationNumber
	 * @return true if the authorityclient moved forward in the stream, false if the authorityclient was already
	 * at or ahead of the given invalidation.
	 * @throws RemoteException 
	 */
	public boolean getUpToInvalidation(int invalidationNumber) throws RemoteException{
		if(invalidationNumber <= getCurrentInvalidationNumber()){
			return false;
		}
//		List<RedoLogEntry> redoLogEntries = authServer.getRedoLogEntries(getCurrentInvalidationNumber() + 1, 
//				invalidationNumber);
		InvalidationPacket ip = authServer.getInvalidationPacket(getCurrentInvalidationNumber() + 1,
				invalidationNumber);
		handleInvalidations(ip);
		return true;
	}
	
	private void handleInvalidations(InvalidationPacket ip){
		streamClient.handleInvalidationPacketBlocking(ip);
	}
	
	public void forceStreamNow()
	{
		System.out.println("Invoking ForceStreamNow");
		streamClient.requestStream();
		
		int lastSeqNum = 0; // streamServer.streamNow();
		if (lastSeqNum < 0)
		{
			System.out.println("ERROR in ForceStreamNow");
		}
		else
		{
			// check that we have processed the lastSeqNum
			Boolean correct = false;
			int appliedSeqNum = -1;
			int numTries = 0;
			for (int i = 0; i < 10000; i++)
			{
				numTries++;
				appliedSeqNum = cacheObjMgr.getLastAppliedSeqNum();
				if (appliedSeqNum >= lastSeqNum)
				{
					correct = true;
					break;
				}
				// try again .. but sleep for sometime first
				try {
					Thread.sleep(1);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			if (!correct)
			{
				System.out.println("ERROR in synching ForceStreamNow (Expected: " + lastSeqNum + ", Got: " + appliedSeqNum);

			}
			else
			{
				System.out.println("NUM RE-SYNC: " + numTries);
			}
		}
	}

	public AuthReply<Boolean> hasAuthority(PrincipalID pid, TagID tid){
		Boolean b = cacheObjMgr.hasAuthority(pid.getValue(), tid.getValue());
		return new AuthReply<Boolean>(getCurrentPredecessor(), b);
	}

	public AuthReply<Boolean> isActee(PrincipalID pidActor, PrincipalID pidActee){
		Boolean b = cacheObjMgr.isActee(pidActor.getValue(), pidActee.getValue());
		return new AuthReply<Boolean>(getCurrentPredecessor(), b);

	}

	public AuthReply<Boolean> isBasisValid(List<Integer> basis){
		Boolean b = cacheObjMgr.isBasisValid(basis);
		return new AuthReply<Boolean>(getCurrentPredecessor(), b);
	}

	public AuthReply<ClosureTableEntry> getClosureInfo(ClosureID closure) throws AeolusException{
		ClosureTableEntry c = cacheObjMgr.getClosureInfo(closure);
		if(LOGGING_ENABLED)
			return new AuthReply<ClosureTableEntry>(getCurrentPredecessor(), c);
		else
			return new AuthReply<ClosureTableEntry>(null, c);
	}
	
	public AuthReply<PrincipalID> createPrincipal(PrincipalID pidCreator, EventID predecessor) throws RemoteException, AuthorityException{
		if(OPERATIONS_BLOCKING){
			AuthReply<PrincipalReply> resp = 
				authServer.createPrincipalGI(pidCreator, getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return new AuthReply<PrincipalID>(resp.event, resp.value.principal);
		}else{
			AuthReply<PrincipalReply> resp = authServer.createPrincipal(pidCreator, predecessor);
			return new AuthReply<PrincipalID>(resp.event, resp.value.principal);
		}
	}

	public AuthReply<PrincipalID> createPrincipalInNewCore(PrincipalID pidCreator, EventID predecessor) throws RemoteException, AuthorityException{
		if(OPERATIONS_BLOCKING){
			AuthReply<PrincipalReply> resp = 
				authServer.createPrincipalInNewCoreGI(pidCreator, getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return new AuthReply<PrincipalID>(resp.event, resp.value.principal);
		}else{
			AuthReply<PrincipalReply> resp = authServer.createPrincipalInNewCore(pidCreator, predecessor);
			return new AuthReply<PrincipalID>(resp.event, resp.value.principal);
		}
	}

	public AuthReply<TagID> createTaint(PrincipalID pidCreator, EventID predecessor) throws RemoteException, AuthorityException{
		if(OPERATIONS_BLOCKING){
			AuthReply<TagID> resp =
				authServer.createTaintGI(pidCreator, getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{
			return authServer.createTag(pidCreator, predecessor);
		}
	}

	public AuthReply<Void> grant(PrincipalID pidRequestor, PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, EventID predecessor) 
	throws AuthorityException, RemoteException{
		if(OPERATIONS_BLOCKING){
			AuthReply<Void> resp = authServer.grantGI(pidRequestor, pidGrantor, pidGrantee, tid, 
					getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{
			return authServer.grant(pidRequestor, pidGrantor, pidGrantee, tid, predecessor);
		}
	}

	public AuthReply<Void> delegateActsFor(PrincipalID pidRequestor, PrincipalID pidFrom, PrincipalID pidTo, EventID predecessor) throws AuthorityException, RemoteException{
		if(OPERATIONS_BLOCKING){
			AuthReply<Void> resp = authServer.delegateActsForGI(pidRequestor, pidFrom, pidTo, 
					getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{
			return authServer.delegateActsFor(pidRequestor, pidFrom, pidTo, predecessor);			
		}
	}

	public AuthReply<Void> revokeGrant(PrincipalID pidRevoker, PrincipalID pidGrantor, PrincipalID pidGrantee,
			TagID tid, EventID predecessor) throws RemoteException, AuthorityException{
		if(OPERATIONS_BLOCKING){
			AuthReply<Void> resp = authServer.revokeGrantGI(pidRevoker, pidGrantor, pidGrantee, tid,
					getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{
			return authServer.revokeGrant(pidRevoker, pidGrantor, pidGrantee, tid, predecessor);
		}
	}

	public AuthReply<Void> revokeActsFor(PrincipalID pidRevoker, PrincipalID pidActor, PrincipalID pidActee, EventID predecessor) throws RemoteException, AuthorityException{
		if(OPERATIONS_BLOCKING){
			AuthReply<Void> resp = authServer.revokeActsForGI(pidRevoker, pidActor, pidActee, 
					getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{
			return authServer.revokeActsFor(pidRevoker, pidActor, pidActee, predecessor);
		}
	}
	
    public AuthReply<ClosureReply> createClosure(String className, EventID predecessor) throws RemoteException{
    	if(OPERATIONS_BLOCKING){
			AuthReply<ClosureReply> resp = 
				authServer.createClosureGI(className, getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{
			return authServer.createClosure(className, predecessor);
		}
    }

    public AuthReply<Void> closureActsFor(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidActee, EventID predecessor) throws RemoteException, AuthorityException{
    	if(OPERATIONS_BLOCKING){
			AuthReply<Void> resp = authServer.closureActsForGI(pidRequestor, cid, pidActee, 
					getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{ 
			return authServer.closureActsFor(pidRequestor, cid, pidActee, predecessor);
		}
    }

    public AuthReply<Void> closureGrant(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidGrantor, TagID tid, EventID predecessor) throws RemoteException, AuthorityException{
    	if(OPERATIONS_BLOCKING){
			AuthReply<Void> resp = authServer.closureGrantGI(pidRequestor, cid, pidGrantor, tid,
					getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return resp;
		}else{	
			return authServer.closureGrant(pidRequestor, cid, pidGrantor, tid, predecessor);
		}
    }

    public AuthReply<Boolean> canSwitch(String currBasisList, PrincipalID fromPid, PrincipalID toPid){
        List<Integer> currBasis = Utils.stringToListInt(currBasisList);
        return canSwitch(currBasis, fromPid, toPid);
    }

    public AuthReply<Boolean> canSwitch(List<Integer> currBasis, PrincipalID fromPid, PrincipalID toPid){
    	if (! isBasisValid(currBasis).value) 
            return new AuthReply<Boolean>(getCurrentPredecessor(), false);

        if (! isActee(fromPid, toPid).value){
            logger.log(LoggerLevel.DEBUG, "CanSwitch unable to call Authority Client");
            return new AuthReply<Boolean>(getCurrentPredecessor(), false);
        }

        return new AuthReply<Boolean>(getCurrentPredecessor(), true);
    }
      
	public AuthReply<Boolean> hasAuthority(List<Integer> currBasis, PrincipalID pid,
			List<TagID> tids) {
        if (!isBasisValid(currBasis).value) 
            return new AuthReply<Boolean>(getCurrentPredecessor(), false);

        for(TagID tid : tids){
            if (! hasAuthority(pid, tid).value) 
                return new AuthReply<Boolean>(getCurrentPredecessor(), false);
        }
        return new AuthReply<Boolean>(getCurrentPredecessor(), true);
	}

	public AuthReply<Boolean> isSystemNodeRegistered(SystemNodeID nodeID) throws AeolusException {
		boolean isRegistered = cacheObjMgr.isSystemNodeRegistered(nodeID);
		return new AuthReply<Boolean>(getCurrentPredecessor(), isRegistered);
	}

	public int getNextPlatformInstanceID() throws AeolusException {
		try {
			return authServer.getNextPlatformInstanceID();
		} catch (RemoteException e) {
			throw new AeolusException(e);
		}
	}
	public void resetStats() {
		cacheObjMgr.resetStats();
	}

	public AuthReply<TagID> createSubTag(TagID parent, PrincipalID pidCreator,
			EventID predecessor) throws RemoteException, AuthorityException {
		if (OPERATIONS_BLOCKING) {
			AuthReply<TagID> resp = authServer
					.createSubTagGI(parent, pidCreator,
							getCurrentInvalidationNumber() + 1, predecessor);
			this.handleInvalidations(resp.piggyback);
			return new AuthReply<TagID>(resp.event, resp.value);
		} else {
			AuthReply<TagID> resp = authServer.createSubTag(parent, pidCreator,
					predecessor);
			return new AuthReply<TagID>(resp.event, resp.value);
		}
	}
}
