package edu.mit.csail.aeolus.authority.client;

import java.rmi.RemoteException;
import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.ReentrantLock;

import org.junit.Assert;

import edu.mit.csail.aeolus.api.SystemNodeID;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.authority.api.ClosureTableEntry;
import edu.mit.csail.aeolus.authority.api.CompoundTagTableEntry;
import edu.mit.csail.aeolus.authority.api.InterfaceAuthorityServer;
import edu.mit.csail.aeolus.authority.api.NodeTableEntry;
import edu.mit.csail.aeolus.authority.api.PrincipalTableEntry;
import edu.mit.csail.aeolus.authority.api.RedoLogEntry;
import edu.mit.csail.aeolus.authority.api.TaintTableEntry;
import edu.mit.csail.aeolus.authority.client.api.EvictionMode;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.profiling.CacheManagerProfile;
import edu.mit.csail.aeolus.serversinfo.DelegateActsForMsg;
import edu.mit.csail.aeolus.serversinfo.ReturnStatus;
import edu.mit.csail.aeolus.serversinfo.RevokeActsForMsg;
import edu.mit.csail.aeolus.serversinfo.RevokeGrantMsg;
import edu.mit.csail.aeolus.serversinfo.StreamMessageParser;
import edu.mit.csail.aeolus.stream.server.api.StreamMessageOpcode;
import edu.mit.csail.aeolus.utilities.Logger;
import edu.mit.csail.aeolus.utilities.LoggerLevel;
import edu.mit.csail.aeolus.utilities.Pair;

public class CacheObjectManager  {
	private static final boolean QUERY_CACHE_OFF = false;

	private static final boolean CORE_CACHE_OFF = false;

	private LinkedList<RedoLogEntry> invalidationsSinceLastSnapshot;

	// L1: Query Cache
	private QueryCache queryCache;
	private ActsForCache actsForCache;
	private EvictionMode queryCacheEvictionMode;

	// L2: Core Cache
	private CoreCache coreCache;

	// Alert List: info on act-for dependency that app wants to know about
	private AlertList alertList;

	// Cache Stats (High-Level)
	private CacheManagerProfile profiler;
	private boolean profiling;

	// Stream Message State
	private ReentrantLock streamMessageLock = new ReentrantLock();


	int lastAppliedSeqNum = -1;
	String lastStreamSnapshotDbName = "";

	public final Date creationTime;

	private Logger logger;

	private EventID currentPredecessor;

	private InterfaceAuthorityServer authServer;

	public CacheObjectManager() {
		creationTime = new Date();
	}

	public int isAlive(){
		return ReturnStatus.opSuccess;
	}

	public Date getCreationTime()
	{
		return creationTime;
	}

	public void initialize(InterfaceAuthorityServer authServer, int queryCacheSize, 
			int lastStreamSnapshotSeqNum, String lastStreamSnapshotDbName){
		// save settings
		this.queryCacheEvictionMode = EvictionMode.FIFO;
		this.lastAppliedSeqNum = lastStreamSnapshotSeqNum;
		this.lastStreamSnapshotDbName = lastStreamSnapshotDbName;
		this.invalidationsSinceLastSnapshot = new LinkedList<RedoLogEntry>();

		this.alertList = new AlertList();

		this.logger = new Logger("CacheObjMgr");

		// instantiate cache objects
		queryCache = new QueryCache(queryCacheSize, queryCacheEvictionMode);
		actsForCache = new ActsForCache(queryCacheSize, queryCacheEvictionMode);

		coreCache = new CoreCache(authServer, lastStreamSnapshotDbName, this);
		this.authServer = authServer;
	}

	long streamMsgHandling = 0;
	
	public void printStats(){
		System.out.println(queryCache.getSummary());
		System.out.println("Invalidation processing: " + streamMsgHandling);
	}
	
	// ---- Performance Profiling ----
	public void turnOnProfiling(){
		profiling = true;
		profiler.startTimer();
	}

	public void turnOffProfiling()
	{
		profiling = false;
		profiler.stopTimer();
	}

	// ResetStats()
	//   Resets any internal counters in various cache objects.
	//   Does NOT flush the caches.
	public void resetStats(){
		streamMsgHandling = 0;
		profiler = new CacheManagerProfile();

		// reset cache obj counters
		queryCache.resetStats();
		coreCache.resetStats();
	}

	// --------------------------- LEVEL 1 Cache ------------------------------------
	private void addToC1(int pid, int tid, Boolean yes, List<Integer> coresAffected){
		// ASSUMPTION: <pid, tid> is not in L1
		queryCache.add(pid, tid, yes, coresAffected);
	}

	// --------------------------- LEVEL 2 Cache ------------------------------------

	// ------------------------- INTERFACE ICacheObjMgr -----------------------------

	public ClosureTableEntry getClosureInfo(ClosureID closure) throws AeolusException{
		if(CORE_CACHE_OFF){
			try {
				return authServer.getClosureInfo(closure);
			} catch (RemoteException e) {
				e.printStackTrace();
				return null;
			}
		}
		return coreCache.getClosureInfo(closure);
	}

	// Normal HasAuthority check
	public boolean hasAuthority(int pid, int tid) {		
		// 1. Check in Cache 1
		if (profiling) {
			profiler.countAuthReq();
			profiler.addC1Check();
		}
		
		if(!QUERY_CACHE_OFF){
			Pair<Boolean,Boolean> queryRes = queryCache.lookup(pid, tid);

			if(queryRes.first){ // first = query cache HIT.
				if (profiling) {
					profiler.addC1Hit();
					if (queryRes.second) 
						profiler.addHasAuth();
				}

				return queryRes.second; // second = hasAuth?
			}
		}

		// 2. Check in Cache 2
		if (profiling)
			profiler.addC2Check();
		// Core Cache
		
		List<Integer> coresAffected = new LinkedList<Integer>();
		boolean hasAuth = false;
		try {
			if(CORE_CACHE_OFF)
				hasAuth = authServer.hasAuthority(new PrincipalID(pid), new TagID(tid));
			else
				hasAuth = coreCache.hasAuthority(pid, tid, coresAffected);
		} catch (RemoteException e) {
			e.printStackTrace();
		}

		// Fill L1 with answer from L2
		if(!QUERY_CACHE_OFF)
			addToC1(pid, tid, hasAuth, coresAffected);

		if (profiling) {
			profiler.addC2Hit();
			if(hasAuth)
				profiler.addHasAuth();
		}

		return hasAuth;

	}

	// Normal IsActee Check
	// doesn't profile Cache-1 information
	public boolean isActee(int pidActor, int pidActee) {
		if(CORE_CACHE_OFF){
			try {
				return authServer.isActee(pidActor,pidActee);
			} catch (RemoteException e) {
				return false;
			}
		}
		
		// C1
		if(!QUERY_CACHE_OFF){
			boolean cacheHit = actsForCache.lookup(pidActee, pidActor);
			if(cacheHit)
				return true;
		}
		
		// Check in Cache 2
		if (profiling) 
			profiler.addC2Check();

		// Core Cache
		List<Integer> coresAffected = new LinkedList<Integer>();
		boolean r = coreCache.isActee(pidActor, pidActee, coresAffected);

		if(r && !QUERY_CACHE_OFF)
			actsForCache.add(pidActee, pidActor, coresAffected);			
		
		if (profiling)
			profiler.addC2Hit();

		return r;
	}

	public Boolean isActeeBasis(int pidActor, int pidActee, List<Integer> basisOut) {		
		if(CORE_CACHE_OFF){
			try {
				Pair<Boolean, List<Integer>> rev = authServer.isActeeBasis(pidActor, pidActee);
				if(rev.first){
					basisOut.addAll(basisOut);
					return true;
				}else{
					return false;
				}
			} catch (RemoteException e) {
				e.printStackTrace();
				return false;
			}
		}
			
		if (profiling) 
			profiler.addC2Check();
		// Core Cache
		Boolean yes = coreCache.isActeeBasis(pidActor, pidActee, basisOut);
		if (profiling)
			profiler.addC2Hit();
		return yes;

	}

	public boolean isBasisValid(List<Integer> basis) {
		int len = basis.size();
		if(len < 2)
			return true;
		if( ! QUERY_CACHE_OFF){
			if(actsForCache.lookup(basis.get(0), basis.get(basis.size() - 1))){
				return true;
			}
		}
		return coreCache.isBasisValid(basis);
	}
	// -------------------- Registering a running Pid ---------------------------------
	/**
	 * @return first = alertID ; second = Basis
	 */
	public Pair<Integer, List<Integer>> runAsPid(int pid){
		List<Integer> basis = new LinkedList<Integer>(Collections.singleton(pid));
		int alertId = alertList.addEntry(basis);
		return new Pair<Integer, List<Integer>>(alertId, basis);
	}

	public Pair<Boolean, Integer> canSwitchPid(int callerAlertId, int callerPid, 
			int calleePid, List<Integer> calleeBasis) {

		// Check that callerAlertId is still valid
		List<Integer> callerBasis = alertList.getEntry(callerAlertId);
		if (callerBasis == null){
			// THROW EXCEPTION: basis no longer valid
			logger.log(LoggerLevel.IFCERR, "Basis no longer valid");
			throw new InvalidBasisException();
		}

		// Since callerAlertId is still valid, we just check whether callerPid act-for pid
		Boolean isActee = isActeeBasis(callerPid, calleePid, calleeBasis);

		if (! isActee){
			logger.log(LoggerLevel.IFCERR, "CanSwitchPid no authority to switch");
			return new Pair<Boolean,Integer>(isActee, -1);
		}

		// Register alert
		int alertId = alertList.addEntry(calleeBasis);
		return new Pair<Boolean,Integer>(isActee, alertId);
	}

	// Deregistering is piggybacked on "HasAuthority" checks since
	// we don't want an explicit IPC at call return to deregister
	public boolean hasAuthorityPiggybacked(List<Integer> completedAlertIds, int alertId, int pid, int tid) {
		// Handle the piggybacked de-registration
		for (Integer i : completedAlertIds){
			alertList.removeEntry(i);
		}

		// Check if basis for pid has been invalidated
		if (! alertList.isBasisValid(alertId)){
			// THROW EXCEPTION: basis no longer valid
			throw new InvalidBasisException();
		}

		// Issue HasAuthority
		return hasAuthority(pid, tid);
	}

	// IF possible, can also deregister explicitly
	public void deregisterAlert(int alertId){
		alertList.removeEntry(alertId);
	}

	public void deregisterAlert(List<Integer> alertIds){
		for (Integer i : alertIds){
			alertList.removeEntry(i);
		}
	}

	public String displayAlertList(){
		return alertList.displayAlertList();
	}

	public int getLastAppliedSeqNum(){
		return this.lastAppliedSeqNum;
	}
	public EventID getCurrentPredecessor(){
		return this.currentPredecessor;
	}

	/**
	 * 
	 * @param queue This queue needs to hold the invalidations 
	 * in the order that they should be processed. When this method returns, 
	 * the top of the queue will be removed.
	 */
	
	
	public void handleStreamMessage(ConcurrentLinkedQueue<RedoLogEntry> queue, EventID predecessor){
		streamMessageLock.lock();
		long s = System.currentTimeMillis();
		RedoLogEntry m = queue.remove();
		handleStreamMessage(m, predecessor);
		long e = System.currentTimeMillis();
		streamMsgHandling += (e - s);
		streamMessageLock.unlock();
	}

	private void handleStreamMessage(RedoLogEntry m, EventID predecessor){

		if(m.requestType != StreamMessageOpcode.S_SNAPSHOT_REMOVE)
			invalidationsSinceLastSnapshot.add(m);

//		debugWriteLine("HandleStreamMessage[" + m.seqNum + "]: " + m.requestType + ", " + m.message);
		if(profiling)
			profiler.countStreamMsg();

		int expectedSeqNum = lastAppliedSeqNum + 1;

		if (m.seqNum < expectedSeqNum){
//			System.out.println("Skipping update...");
			invalidationsSinceLastSnapshot.remove(m);
			return;
		}

		if (m.seqNum > expectedSeqNum){
			System.err.println("Missed updates: Expect " + expectedSeqNum + ", Got " + m.seqNum);
			invalidationsSinceLastSnapshot.remove(m);
			return;
		}

		if(profiling)
			profiler.startStream();

		applyStreamMessage(m.seqNum, m.coreId, m.requestType, m.message, null);

		if(profiling)
			profiler.stopStream();

		// Update state after processing
		lastAppliedSeqNum = Math.max(m.seqNum, lastAppliedSeqNum);
		currentPredecessor = predecessor;
//		System.out.println("LastAppliedSeqNum = " + lastAppliedSeqNum);
	}

	/**
	 * @return first = Double duration of the fetch. second = Integer number of cores fetched.
	 */
	public Pair<Double,Integer> getFetchStats(){
		return coreCache.getFetchStats();
	}

	protected void reapplyStreamMessages(List<Integer> coreIds){
//		System.out.println("Reapplying stream messages...");
		streamMessageLock.lock();
		for(RedoLogEntry m : invalidationsSinceLastSnapshot){
			applyStreamMessage(m.seqNum, m.coreId, m.requestType, m.message, coreIds);
		}
		streamMessageLock.unlock();
	}

	/**
	 * 
	 * @param seqNum
	 * @param coreId
	 * @param requestType
	 * @param streamMessage
	 * @param coreIds - The coreIDs that should be modified. 
	 * @return
	 */
	private boolean applyStreamMessage(int seqNum, int coreId, StreamMessageOpcode requestType, String streamMessage, List<Integer> coreIds){
		// big demux
		switch (requestType)
		{
		case S_SNAPSHOT_ADD:
		{
			if(coreIds != null)
				Assert.fail();
			// EFFECT on L2
			String snapshotDbName = StreamMessageParser.unpackCreateSnapshot(streamMessage);

			coreCache.handleCreateSnapshot(snapshotDbName);

			this.invalidationsSinceLastSnapshot.clear();
		}	break;
		case S_PRINCIPAL_ADD:
		{
			PrincipalTableEntry tt = StreamMessageParser.unpackCreatePrincipal(streamMessage);
			coreCache.handleCreatePrincipal(tt, coreIds);
			if(coreIds == null || coreIds.contains(tt.coreIdCreator)){
				queryCache.invalidate(Collections.singletonList(tt.coreIdCreator));
				actsForCache.invalidate(Collections.singletonList(coreId));
			}
		}	break;
		case S_TAINT_ADD:
		{	
			TaintTableEntry tt = StreamMessageParser.unpackCreateTaint(streamMessage, coreId);

			// EFFECT on L2
			coreCache.handleCreateTaint(tt, coreIds);

			// EFFECT on L1
			if(coreIds == null || coreIds.contains(coreId)){
				queryCache.invalidate(Collections.singletonList(coreId));
				actsForCache.invalidate(Collections.singletonList(coreId));
			}
		}
		break;
		case S_CLOSURE_ADD:
		{
			// EFFECT on L2
			coreCache.handleCreateClosure(StreamMessageParser.unpackCreateClosure(streamMessage, coreId), coreIds);
		}
		break;
		case S_GRANT_ADD:
		{
			// EFFECT on L2
			coreCache.handleGrant(StreamMessageParser.unpackGrant(streamMessage, coreId), coreIds);

			// EFFECT on L1
			if(coreIds == null || coreIds.contains(coreId)){
				queryCache.invalidate(Collections.singletonList(coreId));
				actsForCache.invalidate(Collections.singletonList(coreId));
			}

		}
		break;
		case S_ACTSFOR_ADD:
		{
			// EFFECT on L2
			DelegateActsForMsg m = StreamMessageParser.unpackDelegateActsFor(streamMessage);
			coreCache.handleDelegateActsFor(m.pidFrom, m.pidTo, m.coreFrom, m.coreTo, coreIds);

			// EFFECT on L1
			if(coreIds == null || coreIds.contains(coreId)){
				queryCache.invalidate(Collections.singletonList(coreId));
				actsForCache.invalidate(Collections.singletonList(coreId));
			}
		}
		break;
		case S_GRANT_REMOVE:
		{
			RevokeGrantMsg m = StreamMessageParser.unpackRevokeGrant(streamMessage);
			// EFFECT on L2
			coreCache.handleRevokeGrant(m.pidGrantor, m.pidGrantee, m.tid, coreId, coreIds);

			// EFFECT on L1
			if(coreIds == null || coreIds.contains(coreId)){
				queryCache.invalidate(Collections.singletonList(coreId));
				actsForCache.invalidate(Collections.singletonList(coreId));
			}
		}
		break;
		case S_ACTSFOR_REMOVE:
		{
			RevokeActsForMsg m = StreamMessageParser.unpackRevokeActsFor(streamMessage);

			// EFFECT on L2
			coreCache.handleRevokeActsFor(coreId,m.actorId, m.acteeId, coreIds);

			// EFFECT on L1
			if(coreIds == null || coreIds.contains(coreId)){
				queryCache.invalidate(Collections.singletonList(coreId));
				actsForCache.invalidate(Collections.singletonList(coreId));
			}
			// EFFECT on alertList
			alertList.handleRevokeActFor(m.actorId, m.acteeId);

		}
		break;
		case S_COMPOUNDLINK_ADD:
		{
			CompoundTagTableEntry compEntry = 
				StreamMessageParser.unpackCreateCompoundLink(streamMessage, coreId);
			coreCache.handleCreateCompound(compEntry);
			
			if(coreIds == null || coreIds.contains(coreId)){
				queryCache.invalidate(Collections.singletonList(coreId));
				actsForCache.invalidate(Collections.singletonList(coreId));
			}
		}
		break;
		case NODE_REGISTERED:
		{
			NodeTableEntry nodeEntry = StreamMessageParser.unpackRegisterNode(streamMessage);
			coreCache.handleRegisterNode(nodeEntry);
		}
		break;
		case NODE_REMOVED:
		{
			SystemNodeID node = StreamMessageParser.unpackRemoveNode(streamMessage);
			coreCache.handleRemoveNode(node);
		}
		break;
		}
		return true;
	}

	public boolean isSystemNodeRegistered(SystemNodeID nodeId) {
		if(CORE_CACHE_OFF)
			try {
				return authServer.isSystemNodeRegistered(nodeId.getValue());
			} catch (RemoteException e) {
				e.printStackTrace();
				return false;
			}
		return coreCache.isSystemNodeRegistered(nodeId.getValue());
	}
}
