package edu.mit.csail.aeolus.authority.client;

import java.rmi.RemoteException;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.locks.ReadWriteLock;

import edu.mit.csail.aeolus.authority.api.ClosureTableEntry;
import edu.mit.csail.aeolus.authority.api.CoreContents;
import edu.mit.csail.aeolus.authority.api.InterfaceAuthorityServer;
import edu.mit.csail.aeolus.authority.api.NodeTableEntry;
import edu.mit.csail.aeolus.authority.api.TaintTableEntry;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.utilities.CanalTimer;
import edu.mit.csail.aeolus.utilities.Pair;
import edu.mit.csail.aeolus.utilities.StopWatch;

public class CoreCache extends AbstractCoreCache {

	private static final int BAD_INT = -1;
	// In-memory cached data tables
    final int initialCapacity = 100000;   // in number of cores
    final int thresholdNumCores = 100000;   // will try to remove when above this number

    final int removalNumCores = 1;
    final int avgEntryPerCore = 10;

    CoreLru evictionMgr;

    // Authority Snapshot
    String snapshotDbName;
    InterfaceAuthorityServer authServer;

    // stats
    int numCoreLookup = 0;
    int numCoreAnswer = 0;
    int numCoreAdded = 0;
    int numCoreEvicted = 0;

    // fetch stats
    Double durationFetch = 0.0;

    final int systemCoreId = -1;
    final int pidDefunct = 1;

    CacheObjectManager parentRef;

    CanalTimer timer;
    
    public CoreCache(InterfaceAuthorityServer authServer, String snapshotDbName, CacheObjectManager parentRef){
       super();
       
       this.snapshotDbName = snapshotDbName;
       this.parentRef = parentRef;
       this.authServer = authServer;
        
        // Cache management data structures
        evictionMgr = new CoreLru();
        
        // Timer
        timer = new StopWatch();

        // Bootstrap
        bootstrap();
    }
    
    public void linkAuthServer(InterfaceAuthorityServer authServer){
    	this.authServer = authServer;
    }
    
    
    // dependentCores is supposed to be returned...
    /**
     * @param dependentCores will be modified to add dependent cores to the list.
     */

    @Override
    protected int isPidCached(final int pid){
    	int coreId = super.isPidCached(pid);
    	if(coreId != -1){
            evictionMgr.NotifyLookup(coreId);
            numCoreAnswer++;
    	}
    	return coreId;
    }
    
    @Override
    protected boolean isCached(final int coreId){
    	numCoreLookup++;
    	boolean r = evictionMgr.IsCoreCached(coreId);
    	if(r){
    		evictionMgr.NotifyLookup(coreId);
    		numCoreAnswer++;
    	}
    	return r;
    }
    
    public boolean hasAuthority(int pid, int tid, List<Integer> dependentCores){
    	numCoreLookup++;
    	return super.hasAuthority(pid, tid, dependentCores);
    }
    
    public boolean isActee(int pidFrom, int pidTo, List<Integer> coresAffected){
        numCoreLookup++;
        return super.isActee(pidFrom, pidTo, coresAffected);
    }

    public Pair<Boolean, List<Integer>> isActeeBasis(int pidActor, int pidActee){
    	numCoreLookup++;
        return super.isActeeBasis(pidActor, pidActee);
    }
    
    public Boolean isActeeBasis(int pidActor, int pidActee, List<Integer> basisOut){
    	numCoreLookup++;
        return super.isActeeBasis(pidActor, pidActee, basisOut);
    }

    /**
     * @return first = Duration of Fetch, second = Number of Cores Fetched
     */
    public Pair<Double, Integer> getFetchStats(){
        return new Pair<Double, Integer>(this.durationFetch, this.numCoreAdded);
    }

    public void resetStats(){
        numCoreLookup = 0;
        numCoreAnswer = 0;
        numCoreAdded = 0;
        numCoreEvicted = 0;

        durationFetch = 0.0;
    }

    public void handleCreateSnapshot(String snapshotDbName){
        this.snapshotDbName = snapshotDbName;
        if(this.snapshotDbName == null)
        	return;
        
        // do a null connect (to preload some server data structures for the new snapshot)
        try {
			authServer.warmupDB(snapshotDbName);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
    }

    @Override
    public boolean checkCores(List<Integer> CoreIds, int CoreId){
    	if(CoreIds != null)
    		return CoreIds.contains(CoreId);
    	else
    		return evictionMgr.IsCoreCached(CoreId);
    }

    private void bootstrap(){
        // Load some values
        loadCore(systemCoreId);
    }

    private void loadCore(int coreId){
        // sanity check: double check that it is indeed missing
        if (evictionMgr.IsCoreCached(coreId)){
            numCoreAnswer++;
            evictionMgr.NotifyLookup(coreId);
            return;
        }

        // load one core into cache
        numCoreAdded++;
        load(Collections.singletonList(coreId)); 

        // update cache state
        markCached(coreId);
    }
    
    // --------------------------- Cache Management -------------------------------
    
    // MarkCached
    //   - this should be called to mark a core as being cached
    //   - even if cached, this should be called to update the timer
    private void markCached(int coreId)
    {
        // Update Hashtable
        if (evictionMgr.IsCoreCached(coreId) == false)
        {
            // new entry
            evictionMgr.NotifyAdded(coreId);
        }
        else
        {
            evictionMgr.NotifyLookup(coreId);
        }
    }

    public void checkAndEvict(){
        if (evictionMgr.GetNumCoreCached() >= thresholdNumCores){
            List<Integer> removedCores = evictionMgr.EvictAndRemove(removalNumCores);
            for(int removedCoreId : removedCores){
                unload(removedCoreId);
                numCoreEvicted++;
            }
        }
    }

    private int loadCoreOfClosure(final ClosureID closure){
        // 1. Check and Evict if necessary
        checkAndEvict();
        
        // 2. Add core entry

        timer.start();
        CoreContents r;
		try {
				r = AccessController.doPrivileged(new PrivilegedExceptionAction<CoreContents>() {
					public CoreContents run() throws Exception {
						return authServer.loadCoreFromSnapshot(snapshotDbName, closure);
					}});
		} catch (Exception e) {
			e.printStackTrace();
			return BAD_INT;
		}
        timer.stop();
        durationFetch += timer.getDuration();

        // Fill the cache
        loadCoreContents(r);
        
        // 3. Update metadata as result of add
        numCoreAdded++;
        markCached(r.coreID);
        
        // 4. Reapply Streams...
        parentRef.reapplyStreamMessages(Collections.singletonList(r.coreID));
        
        return r.coreID;
    }    
    
    /**
     * 
     * @param pid
     * @return the int coreId of the input pid
     */
    private int loadCoreOfPid(final int pid)
    {
        // 1. Check and Evict if necessary
        checkAndEvict();
        
        // 2. Add core entry

        timer.start();
        CoreContents r;
		try {
			r = AccessController.doPrivileged(new PrivilegedExceptionAction<CoreContents>() {
				public CoreContents run() throws Exception {
					return authServer.loadCoreFromSnapshot(snapshotDbName, new PrincipalID(pid));
				}});
		} catch (Exception e) {
			e.printStackTrace();
			return BAD_INT;
		}
        timer.stop();
        durationFetch += timer.getDuration();

        // Fill the cache
        loadCoreContents(r);
        
        // 3. Update metadata as result of add
        numCoreAdded++;
        markCached(r.coreID);
        
        // 4. Reapply Streams...
        parentRef.reapplyStreamMessages(Collections.singletonList(r.coreID));
        
        return r.coreID;
    }

    private <T> void addAll(List<T> table, List<T> adding, ReadWriteLock lock){
    	lock.writeLock().lock();
        try{
        	table.addAll(adding);
        }finally{
        	lock.writeLock().unlock();
        }
    }
    
    private void loadCoreContents(CoreContents core){
    	addAll(taintTable, core.getTaintEntries(), taintLock);
    	addAll(principalTable, core.getPrincipalEntries(), principalLock);
        addAll(grantTable, core.getGrantEntries(), grantLock);
        addAll(linkTable, core.getLinkEntries(), linkLock);
        addAll(compoundTable, core.getCompoundEntries(), compoundLock);
        
        closureLock.writeLock().lock();
        try{
        	for(ClosureTableEntry e : core.getClosureEntries()){
        		closureTable.put(e.cid, e);
        	}
        }finally{
        	closureLock.writeLock().unlock();
        }
//        closureTable.addAll(core.getClosureEntries());
        nodeLock.writeLock().lock();
        try{
	       	for(NodeTableEntry e : core.getNodeEntries()){
	        	nodeTable.put(e, new Boolean(true));
	        }
        }finally{
        	nodeLock.writeLock().unlock();
        }
    }
    
    private void load(final List<Integer> coreIds){
        // 1. Check if we need to evict before adding
        checkAndEvict();

        // 2. Add the new cores

        timer.start();
        CoreContents core;
		try {
			core = AccessController.doPrivileged(new PrivilegedExceptionAction<CoreContents>() {
				public CoreContents run() throws Exception {
					return authServer.loadCoresFromSnapshot(snapshotDbName, coreIds);
				}
			});
		} catch (Exception e){			
			e.printStackTrace();
			return;
		}
        timer.stop();
        durationFetch += timer.getDuration();

        // Fill the cache
        loadCoreContents(core);

        // 3. Update metadata as result of add
        for(int coreId : coreIds) 
        	markCached(coreId);
        
        // 4. Re-apply stream messages
        parentRef.reapplyStreamMessages(coreIds);
    }

    private void unload(int coreId)
    {
//        taintTable.RemoveAll(delegate(TaintTableEntry victim) { return (victim.coreId == coreId); });
//        principalTable.RemoveAll(delegate(PrincipalTableEntry victim) { return (victim.coreId == coreId); });
//        grantTable.RemoveAll(delegate(GrantTableEntry victim) { return (victim.coreId == coreId); });
//        linkTable.RemoveAll(delegate(LinkTableEntry victim) { return (victim.coreId == coreId); });
//        closureTable.RemoveAll(delegate(ClosureTableEntry victim) { return (victim.coreId == coreId); });
    }

	@Override
	public void handleMissOnPrincipal(int pid) {
		loadCoreOfPid(pid);
	}

	@Override
	public void handleMissOnClosure(ClosureID closure) {
		loadCoreOfClosure(closure);
	}

	@Override
	public void handleMissOnCore(int coreIdTo) {
		loadCore(coreIdTo);
	}
}
