package edu.mit.csail.aeolus.authority.client;

import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

import edu.mit.csail.aeolus.authority.client.api.EvictionMode;
import edu.mit.csail.aeolus.utilities.Pair;

/**
 * QueryCache Purpose: Fast answer HasAuthority checks caches that ones with
 * 'positive' authority
 */
public class QueryCache {
	// Key: <pid,tid>
	// Value: <coresAffectedList>
	Dictionary<AuthLookupKey_t, QueryCacheValue_t> queryCache;

	// CoreDepInfo
	// Purpose: Fast determination of which QueryCache entries to invalidate
	// Key: <coreId>
	// Value: <HashBitMap of <pid,tid>> that are dependent
	Dictionary<Integer, CoreDepValue_t> coreDepInfo;

	// MeasureInvalidationImprecision
	Dictionary<AuthLookupKey_t, Boolean> penaltyBox;
	Boolean measurePenalty = true; // this can slow things down

	EvictionMode evictionMode;
	int evictionSize = 1;

	int capacity;

	// stats
	int numLookupRequests = 0;
	int numLookupFound = 0;
	int numAddRequests = 0;
	int numEvictRequests = 0;
	int numEvictions = 0;
	int numInvalidations = 0;
	int numPenalties = 0;

	EvictionMgr<AuthLookupKey_t> evictionMgr;

	public QueryCache(int capacity, EvictionMode evictionMode) {
		this.capacity = capacity;
		this.evictionMode = evictionMode;

		initialize();
	}

	private void initialize() {
		// Data Cache
		queryCache = new Hashtable<AuthLookupKey_t, QueryCacheValue_t>(capacity);
		coreDepInfo = new Hashtable<Integer, CoreDepValue_t>();

		// Measurement
		penaltyBox = new Hashtable<AuthLookupKey_t, Boolean>();

		// Decide on which eviction scheme to use
		switch (evictionMode) {
		case RANDOM:
			evictionMgr = new EvictionRandom<AuthLookupKey_t>();
			break;
		case FIFO:
			evictionMgr = new EvictionFifo<AuthLookupKey_t>();
			break;
		case LRU:
			evictionMgr = new EvictionLru<AuthLookupKey_t>();
			break;
		}
	}

	public void resetStats() {
		numLookupRequests = 0;
		numLookupFound = 0;
		numAddRequests = 0;
		numEvictRequests = 0;
		numEvictions = 0;
		numInvalidations = 0;
		numPenalties = 0;
	}

	public void add(int pid, int tid, Boolean hasAuth,
			List<Integer> coresAffected) {
		if (!hasAuth)
			return;

		numAddRequests++;

		// 1. Check whether we need to evict any to make room
		if (queryCache.size() >= capacity) {
			List<AuthLookupKey_t> evictCandidates;
			evictCandidates = evictionMgr.evictAndRemove(evictionSize);

			// 2. Remove suggested candidates from query cache
			numEvictRequests++;
			numEvictions += evictCandidates.size();
			for (int i = 0; i < evictCandidates.size(); i++) {
				AuthLookupKey_t keyToRemove = evictCandidates.get(i);
				QueryCacheValue_t queryCacheValue = queryCache.get(keyToRemove);

				// update core dependency info
				List<Integer> coresList = queryCacheValue.coresAffected;
				for (int c = 0; c < coresList.size(); c++) {
					int coreId = coresList.get(c);
					CoreDepValue_t keysMap;
					// if (coreDepInfo.tryGetValue(coreId, out keysMap))
					if ((keysMap = coreDepInfo.get(coreId)) != null) {
						// remove key from map (update core bitmap)
						keysMap.bitMap.remove(keyToRemove);
					}
				}

				// remove entry from query cache
				// NOTE: eviction metadata already updated
				queryCache.remove(keyToRemove);
			}
		}

		// 3. Add new entries to query cache
		AuthLookupKey_t keyToAdd = new AuthLookupKey_t();
		keyToAdd.pid = pid;
		keyToAdd.tid = tid;
		QueryCacheValue_t valueToAdd = new QueryCacheValue_t();
		valueToAdd.coresAffected = coresAffected;
		queryCache.put(keyToAdd, valueToAdd);

		// 4. Notify Eviction Manager
		// evictionMgr.NotifyAdded(keyToAdd);
		evictionMgr.notifyAdded(keyToAdd);

		// 5. Update cores dep info
		for (int i = 0; i < coresAffected.size(); i++) {
			int coreId = coresAffected.get(i);
			CoreDepValue_t keysMap;
			// if (coreDepInfo.TryGetValue(coreId, out keysMap))
			if ((keysMap = coreDepInfo.get(coreId)) != null) {
				// simply update the bitmap
				try {
					keysMap.bitMap.put(keyToAdd, true);
				} catch (Exception e) {
					// ignore duplicates
				}
			} else {
				// create a new entry
				keysMap = new CoreDepValue_t();
				keysMap.bitMap = new Hashtable<AuthLookupKey_t, Boolean>();
				keysMap.bitMap.put(keyToAdd, true);
				coreDepInfo.put(coreId, keysMap);
			}
		}

	}

	/**
	 * @param pid
	 * @param tid
	 * @return first = found entry? second = has authority?
	 */
	public Pair<Boolean, Boolean> lookup(int pid, int tid) {
		Boolean foundEntry = false;
		Boolean hasAuth = false;

		try {
			AuthLookupKey_t keyToLookup = new AuthLookupKey_t();
			keyToLookup.pid = pid;
			keyToLookup.tid = tid;
			QueryCacheValue_t value = queryCache.get(keyToLookup);
			if (value != null) {
				hasAuth = true;
				foundEntry = true;

				numLookupFound++;

				// notify the eviction scheme
				evictionMgr.notifyLookup(keyToLookup);
			}
		} catch (Exception e) {
			System.err.println("Lookup: on <" + pid + ", " + tid + "> " + e);
		} finally {
			numLookupRequests++;
		}
		return new Pair<Boolean, Boolean>(foundEntry, hasAuth);
	}

	public void invalidate(List<Integer> coresChanged) {
		for (int coreId : coresChanged) {
			CoreDepValue_t coreDepValue = coreDepInfo.get(coreId);
			if (coreDepValue != null) {
				for (Enumeration<AuthLookupKey_t> en = coreDepValue.bitMap
						.keys(); en.hasMoreElements();) {
					AuthLookupKey_t keyToInvalidate = en.nextElement();
					// NOTE: it is possible for the keyPair to not be cached in
					// L1
					// It's ok for this bitmap to have some 'garbage' since we
					// only
					// look at it when we are actually invalidating the core
					if (queryCache.remove(keyToInvalidate) != null) {
						numInvalidations++;
						evictionMgr.notifyRemoved(keyToInvalidate);

					}
				}
				// Remove the core entry
				coreDepInfo.remove(coreId);
			}
		}
	}

	// -----------------------------------------------------------

	public String getSummary() {
		StringBuilder sb = new StringBuilder();

		sb.append("QueryCache\n");
		sb.append("  Cache Capacity: " + capacity + "\n");
		sb.append("  Num of Cached Entries: " + queryCache.size() + "\n");
		sb.append("  Num of Lookup Requests: " + numLookupRequests + "\n");
		sb.append("    Found: " + numLookupFound + "\n");
		sb.append("  Num of Add Requests: " + numAddRequests + "\n");
		sb.append("  Num of Evict Requests: " + numEvictRequests + "\n");
		sb.append("    Num of Evictions: " + numEvictions + "\n");
		sb.append("  Num of Invalidations: " + numInvalidations + "\n");
		sb.append("  Num of Penalties: " + numPenalties + "\n");

		return sb.toString();
	}

	/*
	 * out int numCacheEntries, out int numCacheLookup, out int numCacheAnswer,
	 * out int numCacheAdd, out int numCacheEvict, out int numCacheInvalidate,
	 * out int numCachePenalty
	 */
	public int[] getCacheStats() {
		int[] tmp = new int[7];
		/*
		 * numCacheEntries = queryCache.size(); numCacheLookup =
		 * numLookupRequests; numCacheAnswer = numLookupFound; numCacheAdd =
		 * numAddRequests; numCacheEvict = numEvictions; numCacheInvalidate =
		 * numInvalidations; // may be caused by L2 eviction numCachePenalty =
		 * numPenalties;
		 */
		tmp[0] = queryCache.size();
		tmp[1] = numLookupRequests;
		tmp[2] = numLookupFound;
		tmp[3] = numAddRequests;
		tmp[4] = numEvictions;
		tmp[5] = numInvalidations; // may be caused by L2 eviction
		tmp[6] = numPenalties;
		return tmp;
	}

	private static class CoreDepValue_t {
		public Dictionary<AuthLookupKey_t, Boolean> bitMap;
	}

	private static class QueryCacheValue_t {
		public List<Integer> coresAffected;
	}
}
