package edu.mit.csail.aeolus.authority.client.api;

import java.rmi.RemoteException;
import java.util.List;

import edu.mit.csail.aeolus.api.SystemNodeID;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.authority.api.AuthReply;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.authority.api.ClosureReply;
import edu.mit.csail.aeolus.authority.api.ClosureTableEntry;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.stream.server.api.RegisterReturn;

public interface InterfaceAuthorityClient {

	/**
	 * @return the invalidation number that the authority client is currently seeing.
	 */
	public int getCurrentInvalidationNumber();
	
	/**
	 * stalls the authority client until it gets to invalidation number given.
	 * @param invalidationNumber
	 * @return true if an update occurred, false if the current invalidation number >= the parameter.
	 * @throws RemoteException
	 */
	public boolean getUpToInvalidation(int invalidationNumber) throws RemoteException;
	
	/**
	 * Registers this authority client with the authority server
	 * @return root Principal for this Platform Instance
	 * @throws AeolusException
	 */
	public RegisterReturn register() throws AeolusException;
	
	/**
	 * Creates a new principal in the core of pidCreator. pidCreator will act for the newly created principal
	 * @param pidCreator
	 * @param predecessor
	 * @return AuthReply containing the new PrincipalID
	 * @throws AuthorityException
	 * @throws RemoteException
	 */
	public AuthReply<PrincipalID> createPrincipal(PrincipalID pidCreator, EventID predecessor) throws AuthorityException, RemoteException;

	/**
	 * Creates a new principal in a new core. pidCreator will act for the newly created principal
	 * @param pidCreator
	 * @param predecessor
	 * @return AuthReply containing the new PrincipalID
	 * @throws AuthorityException
	 * @throws RemoteException
	 */
	public AuthReply<PrincipalID> createPrincipalInNewCore(PrincipalID pidCreator, EventID predecessor) throws AuthorityException, RemoteException;

	/**
	 * Creates a new tag in the core of pidCreator. pidCreator is authoritative for the new tag.
	 * @param pidCreator
	 * @param predecessor
	 * @return AuthReply containing the new TagID
	 * @throws AuthorityException
	 * @throws RemoteException
	 */
	public AuthReply<TagID> createTaint(PrincipalID pidCreator, EventID predecessor) throws AuthorityException, RemoteException;


	/**
	 * Grants authority from pidGrantor to pidGrantee on Tag tid. pidRequestor must act for pidGrantor and pidGrantor must
	 * have authority for tid.
	 * @param pidRequestor
	 * @param pidGrantor
	 * @param pidGrantee
	 * @param tid
	 * @param predecessor
	 * @return
	 * @throws AuthorityException - if pidRequestor does not act for pidGrantor OR pidGrantor is not authoritative for tid
	 * @throws RemoteException
	 */
	public AuthReply<Void> grant(PrincipalID pidRequestor, PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, EventID predecessor) throws AuthorityException, RemoteException;

    /**
     *	Principal pidRequestor allows principal pidFrom to act for pidTo.
     *	Authority on taints is transitive. Principal pidFrom will have all authority on taints that
     *	principal pidTo has.
     *
     *	@throws AuthorityException if requestor does not own principal pidTo, or if this acts-for link
     *	would create a cycle in the acts-for graph.
     */
	public AuthReply<Void> delegateActsFor(PrincipalID pidRequestor, PrincipalID pidFrom, PrincipalID pidTo, EventID predecessor) throws AuthorityException, RemoteException;

	/**
	 * Revokes authority on tag tid from pidGrantee
	 * @param pidRevoker - the principal requesting the revoke. This principal must act for pidGrantor
	 * @param pidGrantor - the authority that pidGrantee's authority was based on
	 * @param pidGrantee
	 * @param tid
	 * @param predecessor
	 * @return
	 * @throws AuthorityException
	 * @throws RemoteException
	 */
	public AuthReply<Void> revokeGrant(PrincipalID pidRevoker, PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, EventID predecessor) throws AuthorityException, RemoteException;

	/**
	 * Revokes pidFrom's ability to act-for pidTo.
	 * @param pidRevoker - must act for pidTo
	 * @param pidFrom
	 * @param pidTo
	 * @param predecessor
	 * @return
	 * @throws AuthorityException
	 * @throws RemoteException
	 */
	public AuthReply<Void> revokeActsFor(PrincipalID pidRevoker, PrincipalID pidFrom, PrincipalID pidTo, EventID predecessor) throws AuthorityException, RemoteException;
    
	/**
	 * Checks whether principal pid has authority for tag tid.
	 * @param pid
	 * @param tid
	 * @return
	 */
	public AuthReply<Boolean> hasAuthority(PrincipalID pid, TagID tid);
	
	/**
	 * Checks whether pidFrom acts for pidTo.
	 * @param pidFrom
	 * @param pidTo
	 * @return
	 */
	public AuthReply<Boolean> isActee(PrincipalID pidFrom, PrincipalID pidTo);

	/**
	 * Returns info (key and pid mapping) about the given closure.
	 * @param closure
	 * @return
	 * @throws AeolusException
	 */
    public AuthReply<ClosureTableEntry> getClosureInfo(ClosureID closure) throws AeolusException;

    /**
     * Checks the current basis for validity and then checks whether fromPid acts for toPid.
     * @param currBasis
     * @param fromPid
     * @param toPid
     * @return
     */
    public AuthReply<Boolean> canSwitch(List<Integer> currBasis, PrincipalID fromPid, PrincipalID toPid);

    /**
     * Checks the basis is still valid (all act-for links in the basis still exist.)
     * @param basis
     * @return
     */
    public AuthReply<Boolean> isBasisValid(List<Integer> basis);

    /**
     * Checks the current basis for validity and then checks whether pid is authoritative for
     * all tags in tids.
     * @param currBasis
     * @param pid
     * @param tids
     * @return
     */
    public AuthReply<Boolean> hasAuthority(List<Integer> currBasis, PrincipalID pid, List<TagID> tids);
     
    /**
     * Creates a new closure.
     * @param className
     * @param predecessor
     * @return
     * @throws RemoteException
     */
    public AuthReply<ClosureReply> createClosure(String className, EventID predecessor) throws RemoteException;
    
    /**
     * Allows closure cid to act for pidActee. pidRequestor must act for pidActee.
     * @param pidRequestor
     * @param cid
     * @param pidActee
     * @param predecessor
     * @return
     * @throws AuthorityException
     * @throws RemoteException
     */
    public AuthReply<Void> closureActsFor(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidActee, EventID predecessor) throws AuthorityException,  RemoteException;

    /**
     * Grants authority on tag tid to closure cid based on pidGrantor's authority. pidGrantor must be authoritative for 
     * tid and pidRequestor must act for pidGrantor.
     * @param pidRequestor
     * @param cid
     * @param pidGrantor
     * @param tid
     * @param predecessor
     * @return
     * @throws AuthorityException
     * @throws RemoteException
     */
    public AuthReply<Void> closureGrant(PrincipalID pidRequestor, ClosureID cid, PrincipalID pidGrantor, TagID tid, EventID predecessor) throws AuthorityException, RemoteException;

    /**
     * Gets a unique platform instance id from the Authority Server.
     * @return
     * @throws AeolusException
     */
	public int getNextPlatformInstanceID() throws AeolusException;
	public AuthReply<TagID> createSubTag(TagID parent, PrincipalID pidCreator,
			EventID predecessor) throws RemoteException, AuthorityException;

	public AuthReply<Boolean> isSystemNodeRegistered(SystemNodeID node) throws AeolusException;
}
