package edu.mit.csail.aeolus.authority.server;

import java.rmi.server.ExportException;

import junit.framework.Assert;

import org.junit.Before;
import org.junit.Test;

import edu.mit.csail.aeolus.api.SystemNodeID;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.authority.api.AuthReply;
import edu.mit.csail.aeolus.authority.client.AuthorityClient;
import edu.mit.csail.aeolus.authority.client.api.EvictionMode;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;
import edu.mit.csail.aeolus.stream.server.StreamServer;
import edu.mit.csail.aeolus.stream.server.api.RegisterReturn;

public class AuthorityIntTest {

	AuthorityManager mgr;
	StreamServer ss;
	AuthorityClient ac;

	@Test
	public void testServerBlockingClient() throws Exception{
		//		AuthorityConfiguration config = new AuthorityConfiguration("localhost/", "aeolus", "aeolus", "b1b1");

		this.mgr = new AuthorityManager();
		this.mgr.removeAllSnapshots();
		this.mgr.createAuthorityServer();
		mgr.startServer();

		ss = new StreamServer();
		ss.initialize();

		ac = new AuthorityClient(100, EvictionMode.LRU);
		RegisterReturn rr = ac.register();
		PrincipalID root = rr.rootPID;
		
		Assert.assertTrue(ac.isSystemNodeRegistered(rr.systemNodeID).value);
		Assert.assertFalse(ac.isSystemNodeRegistered(new SystemNodeID(rr.systemNodeID.getValue() + 1)).value);
		
		ss.streamNow();
		
		AuthReply<PrincipalID> pRepl = ac.createPrincipal(root, new EventID("1"));
		PrincipalID p = pRepl.value;
		Assert.assertEquals(new EventID("auth-4"), pRepl.event);
		
		Assert.assertTrue(ac.isActee(root, p).value);

		AuthReply<TagID> repl = ac.createTaint(p, new EventID("2"));
		TagID t = repl.value;
		Assert.assertEquals(new EventID("auth-5"), repl.event);

		ss.streamNow();

		Assert.assertTrue(ac.hasAuthority(p, t).value);

		repl = ac.createTaint(root, new EventID("3"));
		t = repl.value;
		Assert.assertEquals(new EventID("auth-6"), repl.event);
		ac.grant(root, root, p, t, new EventID("4"));
		Assert.assertTrue(ac.hasAuthority(p, t).value);
		
		repl = ac.createSubTag(t, root, new EventID("5"));
		TagID subt = repl.value;

		Assert.assertTrue(ac.hasAuthority(p, subt).value);

		ac.revokeGrant(root, root, p, t, new EventID("6"));
		Assert.assertFalse(ac.hasAuthority(p, t).value);
		Assert.assertFalse(ac.hasAuthority(p, subt).value);

	}

	@Before
	public void startRMI() throws Exception{
		RuntimeConfig c = RuntimeConfig.get();
		c.setEnableLogging(true);
		c.setEnableLogShipment(false);
		RuntimeConfig.get().finalize();
		try{
			java.rmi.registry.LocateRegistry.createRegistry(1099);
		}catch(ExportException e){
		}
	}	

	private void waitB(){
		try{
			Thread.sleep(100);
		}catch(Exception e){
		}
	}

	@Test
	public void testServerNonBlockingClient() throws Exception{		
		//		AuthorityConfiguration config = new AuthorityConfiguration("localhost/", "aeolus", "aeolus", "b1b1");
		this.mgr = new AuthorityManager();
		this.mgr.removeAllSnapshots();
		this.mgr.createAuthorityServer();
		mgr.startServer();

		Assert.assertNull(this.mgr.getError());

		ss = new StreamServer();
		ss.initialize();

		ac = new AuthorityClient(100, EvictionMode.LRU, false);

		PrincipalID root = ac.register().rootPID;

		ss.streamNow();
		waitB();
		PrincipalID P3 = mgr.createPrincipal(root, new EventID("1")).value.principal;
		AuthReply<TagID> repl = mgr.createTag(root,new EventID("2"));
		TagID T1 = repl.value;
		Assert.assertEquals(new EventID("auth-5"), repl.event);
		Assert.assertNull(this.mgr.getError());
		ss.streamNow();
		waitB();


		Assert.assertFalse(ac.hasAuthority(P3, T1).value);
		AuthReply<Void> replV = mgr.grant(root, root, P3, T1, new EventID("3"));
		Assert.assertNull(this.mgr.getError());
		Assert.assertEquals(new EventID("auth-6"), replV.event);

		Assert.assertFalse( "Changes should not have been pushed via stream yet.", ac.hasAuthority(P3, T1).value);
		Assert.assertEquals(new EventID("auth-5"), ac.getCurrentPredecessor());
		ss.streamNow();
		waitB();
		Assert.assertEquals(new EventID("auth-6"), ac.getCurrentPredecessor());
		Assert.assertTrue(ac.hasAuthority(P3, T1).value);
		Assert.assertNull(this.mgr.getError());

		replV = mgr.revokeGrant(root, root, P3, T1, new EventID("4"));
		Assert.assertTrue( "Changes should not have been pushed via stream yet.", ac.hasAuthority(P3, T1).value);
		Assert.assertEquals(new EventID("auth-6"), ac.getCurrentPredecessor());
		ss.streamNow();
		waitB();
		Assert.assertFalse( "Changes should have been pushed via stream.", ac.hasAuthority(P3, T1).value);
		Assert.assertEquals(new EventID("auth-7"), replV.event);
		Assert.assertEquals(new EventID("auth-7"), ac.getCurrentPredecessor());

		PrincipalID P4 = mgr.createPrincipal(root, null).value.principal;
		Assert.assertFalse( "P4 should not have authority for t1 yet.", ac.hasAuthority(P4, T1).value);	
		mgr.grant(root, root, P3, T1, null);

		ss.streamNow();
		waitB();
		Assert.assertTrue( "Changes should be pushed.", ac.hasAuthority(P3, T1).value);
		Assert.assertFalse( "P4 should not have authority for t1 yet.", ac.hasAuthority(P4, T1).value);	

		mgr.delegateActsFor(root, P4, P3, null);
		Assert.assertFalse( "P4 should not have authority for t1 yet.", ac.hasAuthority(P4, T1).value);	
		ss.streamNow();
		waitB();
		Assert.assertTrue( "Changes should be pushed.", ac.hasAuthority(P4, T1).value);

		Assert.assertNull(this.mgr.getError());

		mgr.revokeGrant(root, root, P3, T1, null);
		ss.streamNow();
		waitB();
		Assert.assertFalse( "Changes should have been pushed via stream.", ac.hasAuthority(P3, T1).value);
		Assert.assertFalse( "Changes should have been pushed via stream.", ac.hasAuthority(P4, T1).value);
		mgr.grant(root, root, P3, T1, null);
		ss.streamNow();
		waitB();
		Assert.assertTrue( "Changes should be pushed.", ac.hasAuthority(P3, T1).value);

		mgr.revokeActsFor(root, P4, P3, null);
		Assert.assertTrue( "Changes should not have been pushed.", ac.hasAuthority(P4, T1).value);
		ss.streamNow();
		waitB();
		Assert.assertFalse( "P4 should not have authority for t1 now.", ac.hasAuthority(P4, T1).value);	

		Assert.assertNull(this.mgr.getError());

	}

	//	@After
	//	public void testCleanServer() throws Exception{
	//		this.mgr.removeAllSnapshots();
	//	}
}
