package edu.mit.csail.aeolus.authority.server;

import java.beans.PropertyVetoException;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ExportException;
import java.rmi.server.UnicastRemoteObject;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.postgresql.ds.PGConnectionPoolDataSource;

import edu.mit.csail.aeolus.api.SystemNodeID;
import edu.mit.csail.aeolus.auditing.AeolusApiConstants;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.auditing.EventRecord;
import edu.mit.csail.aeolus.authority.api.AuthReply;
import edu.mit.csail.aeolus.authority.api.AuthorityConfiguration;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.authority.api.ClosureReply;
import edu.mit.csail.aeolus.authority.api.ClosureTableEntry;
import edu.mit.csail.aeolus.authority.api.CompoundTagTableEntry;
import edu.mit.csail.aeolus.authority.api.CoreContents;
import edu.mit.csail.aeolus.authority.api.GrantTableEntry;
import edu.mit.csail.aeolus.authority.api.InterfaceAuthorityServer;
import edu.mit.csail.aeolus.authority.api.LinkTableEntry;
import edu.mit.csail.aeolus.authority.api.NodeTableEntry;
import edu.mit.csail.aeolus.authority.api.LinkTableEntry.LinkType;
import edu.mit.csail.aeolus.authority.api.PrincipalReply;
import edu.mit.csail.aeolus.authority.api.PrincipalTableEntry;
import edu.mit.csail.aeolus.authority.api.RedoLogEntry;
import edu.mit.csail.aeolus.authority.api.TaintTableEntry;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;
import edu.mit.csail.aeolus.serversinfo.StreamMessageEntry;
import edu.mit.csail.aeolus.serversinfo.StreamMessageParser;
import edu.mit.csail.aeolus.stream.client.InvalidationPacket;
import edu.mit.csail.aeolus.stream.server.StreamServer;
import edu.mit.csail.aeolus.stream.server.api.RegisterReturn;
import edu.mit.csail.aeolus.stream.server.api.SnapshotNumAndName;
import edu.mit.csail.aeolus.stream.server.api.StreamMessageOpcode;
import edu.mit.csail.aeolus.utilities.Logger;
import edu.mit.csail.aeolus.utilities.LoggerLevel;
import edu.mit.csail.aeolus.utilities.Pair;

public class AuthorityManager implements InterfaceAuthorityServer {

	// fast verified values
	private final int coreIdSystem = 1;

	private final int coreIdInvalid = -1;
	private final int seqInvalid = -1;
	
	private final int numSnapshotToBuffer = 5;

	// blanks: Should these really be compile-time values?
	private final PrincipalID pidSystem = new PrincipalID(1);// AuthStatefinal.pidSystem;
	private final PrincipalID  pidPublic = new PrincipalID(2);// AuthStatefinal.pidPublic;
	private final PrincipalID  pidDefunct = new PrincipalID(3);// AuthStatefinal.pidDefunct;
	private final TagID  tidInvalid = new TagID(1);// AuthStatefinal.tidInvalid;

	private static final EventID CREATE_AUTHORITY_ROOT_PID_EVENTID = new EventID("1");

	// perf measure
	private final boolean INVALIDATION_CACHING = true;

	private Logger logger;

	// --------------------- DATABASE IMPLEMENTATION SCHEMAS -------------------
	public final static String CoreIdAttribName = "CoreId"; 
    
	// 1. Core table
	// <CoreId> <RefCount>
	public final static String CI_RelationName = "CoreTable";
	public final static String CI_CoreIdAttribName = CoreIdAttribName;
	public final static String CI_CoreIdAttribType = "SERIAL PRIMARY KEY";
	public final static String CI_RefCountAttribName = "RefCount";
	public final static String CI_RefCountAttribType = "int";

	// 2. Taint table
	// <Tid unique> <creatorPid>
	public final static String T_RelationName = "TaintTable";
	public final static String T_CoreIdAttribName = CoreIdAttribName;
	public final static String T_CoreIdAttribType = "int";
	public final static String T_TidAttribName = "Tid";
	public final static String T_TidAttribType = "SERIAL PRIMARY KEY ";
	public final static String T_CreatorPidAttribName = "CreatorPid";
	public final static String T_CreatorPidAttribType = "int";

	// 3. Principal table
	// <Pid unique> <creatorPid> <coreId>
	public final static String P_RelationName = "PrincipalTable";
	public final static String P_CoreIdAttribName = CoreIdAttribName;
	public final static String P_CoreIdAttribType = "int";
	public final static String P_PidAttribName = "Pid";
	public final static String P_PidAttribType = "SERIAL PRIMARY KEY ";
	public final static String P_CreatorPidAttribName = "CreatorPid";
	public final static String P_CreatorPidAttribType = "int";

	// 4. Grant table
	// <granteePid> <grantorPid> <tid> <coreId>
	public final static String G_RelationName = "GrantTable";
	public final static String G_CoreIdAttribName = CoreIdAttribName; // core of
	// Grantee
	public final static String G_CoreIdAttribType = "int";
	public final static String G_GrantorPidAttribName = "GrantorPid";
	public final static String G_GrantorPidAttribType = "int";
	public final static String G_GranteePidAttribName = "GranteePid";
	public final static String G_GranteePidAttribType = "int";
	public final static String G_TidAttribName = "Tid";
	public final static String G_TidAttribType = "int";
//	public final static String G_Constraints = "PRIMARY KEY("
//		+ G_GrantorPidAttribName + ", " + G_GranteePidAttribName + ", "
//		+ G_TidAttribName + ")";

	// 5. Links table (linkType 0: owns, 1: acts-for)
	// <PidNode> <PidTo> <linkType> <coreId>
	public final static String L_RelationName = "LinkTable";
	public final static String L_CoreIdFromAttribName = CoreIdAttribName; // coreId
	// of
	// PidFrom
	public final static String L_CoreIdFromAttribType = "int";
	public final static String L_PidFromAttribName = "PidFrom";
	public final static String L_PidFromAttribType = "int";
	public final static String L_PidToAttribName = "PidTo";
	public final static String L_PidToAttribType = "int";
	public final static String L_LinkTypeAttribName = "LinkType";
	public final static String L_LinkTypeAttribType = "int";
	public final static String L_CoreIdToAttribName = "CoreIdTo";
	public final static String L_CoreIdToAttribType = "int";
//	public final static String L_Constraints = "PRIMARY KEY("
//		+ L_PidFromAttribName + ", " + L_PidToAttribName + ")";
	public final int L_OWNS = 0;
	public final int L_ACTSFOR = 1;

	// 6. Closure Info
	// <closure-id> <anon-principal> <key>
	public final static String C_RelationName = "ClosureTable";
	public final static String C_CoreIdAttribName = CoreIdAttribName;
	public final static String C_CoreIdAttribType = "int";
	public final static String C_CidAttribName = "Cid";
	public final static String C_CidAttribType = "SERIAL PRIMARY KEY";
	public final static String C_PidClosureAttribName = "PidClosure";
	public final static String C_PidClosureAttribType = "int";
	public final static String C_ClassNameAttribName = "ClassName";
	public final static String C_KeyAttribType = "VARCHAR(200)";

	// 7. Redo Log (Server Use Only)
	public final static String R_RelationName = "RedoLog";
	public final static String R_SeqNumAttribName = "SeqNum";
	public final static String R_SeqNumAttribType = "SERIAL PRIMARY KEY ";
	public final static String R_CoreIdAttribName = "CoreId";
	public final static String R_CoreIdAttribType = "int";
	public final static String R_RequestTypeAttribName = "RequestType";
	public final static String R_RequestTypeAttribType = "int";
	public final static String R_MessageAttribName = "Message";
	public final static String R_MessageAttribType = "VARCHAR(512)";
	public final static String R_EventIDName = "EventID";
	public final static String R_EventIDType = "VARCHAR(128)";

	// Compound Tag Table
	public final static String Comp_RelationName = "CompoundTagTable";
	public final static String Comp_CoreIdAttribName = CoreIdAttribName;
	public final static String Comp_CoreIdAttribType = "int";
	public final static String Comp_ChildTidAttribName = "ChildTag";
	public final static String Comp_ChildTidAttribType = "int";
	public final static String Comp_ParTidAttribName = "ParTag";
	public final static String Comp_ParTidAttribType = "int";

	
	// 8. Server State
	// This is common practice in database system to have a key-value store for
	// attributes or internal state (or debugging)
	public final static String S_RelationName = "ServerState";
	public final static String S_KeyAttribName = "StateId";
	public final static String S_KeyAttribType = "VARCHAR(100) PRIMARY KEY";
	public final static String S_ValueAttribName = "Value";
	public final static String S_ValueAttribType = "VARCHAR(100)";	

	private final static String NodeTable_Name = "PITable";
	private final static String NodeTable_NodeID = "nodeid";
	private final static String NodeTable_NodeIDAttr = "SERIAL PRIMARY KEY";
	private final static String PINodeTable_Host = "hostname";
	private final static String PINodeTable_HostAttr = "VARCHAR(256) UNIQUE NOT NULL";
	private final static String PINodeTable_PI = "PrincipalID"; 
	private final static String PINodeTable_PIAttr = "INT NOT NULL";
	
	private final static String PlatformID_Name = "PIIDTable";
	private final static String PlatformID_PIID = "PIID";
	private final static String PlatformID_PIIDAttr = "SERIAL PRIMARY KEY";
	private final static String PlatformID_Holder = "holder";
	private final static String PlatformID_HolderAttr = "SMALLINT";


		
	private final static String[] relNames = { CI_RelationName, T_RelationName,
		P_RelationName, G_RelationName, L_RelationName, C_RelationName,
		R_RelationName, S_RelationName, NodeTable_Name, PlatformID_Name, Comp_RelationName };

	private final static String[] snapCopy = {
		T_RelationName, P_RelationName, G_RelationName, L_RelationName, C_RelationName,
		Comp_RelationName
	};

	static final EventID START_EID = new EventID("auth-start");

	
	private Exception error = null;

	private boolean started;

	private StreamServer ss;

	private List<RedoLogEntry> invalidsSinceLastSnapshot;

	private InvalidationsCache invalidationsCache = new InvalidationsCache();

	private EventID lastEID = null;
	
	private AuthLogger authLog = new AuthLogger();
	
	/* To GET YOUR DATABASE STARTED...
	 * 
	 *
	 * CREATE DATABASE AeolusAuthority;
	 *
	 * CONNECT AeolusAuthority;
	 * 
	 * GRANT ALL ON * TO 'aeolus'@'localhost' IDENTIFIED BY 'PFlate762!';
	 *
	 */

        PGConnectionPoolDataSource cpds;
    
	private boolean enableLogging;
	
	public AuthorityManager(){
		logger = new Logger("AuthMgr");
		invalidsSinceLastSnapshot = new LinkedList<RedoLogEntry>();
		cpds = new PGConnectionPoolDataSource();
		cpds.setDatabaseName(AuthorityConfiguration.DB_Name);
		cpds.setPortNumber(5432);
		cpds.setServerName("localhost");
		cpds.setDefaultAutoCommit(true);
		
		RuntimeConfig c = RuntimeConfig.get();
		enableLogging = c.isEnableLogging();
		if(c.isEnableLogShipment())
			authLog.startScheduler();
	}

	/**
	 * Instantiates and initializes a streamserver to handle invalidations for this
	 * authority manager.
	 * 
	 * Right now, the stream server sends an invalidation every 15 seconds.
	 * 
	 * The startServer() MUST be called before calling this method.
	 */
	public void startStreamServer() {
		assert started;
		ss = new StreamServer();
		ss.initialize();

//		Timer t = new Timer();
//		t.scheduleAtFixedRate(new TimerTask(){
//			public void run() {
//				ss.streamNow();
//			}
//
//		}, 15 * 1000, 15 * 1000);
	}

	public void startServer() {
		InterfaceAuthorityServer export;
		try {
			export = (InterfaceAuthorityServer) UnicastRemoteObject
			.exportObject(this, 0);

			// Bind the remote object's stub in the registry
			Registry registry;
			try{
				registry = LocateRegistry.createRegistry(1099);
			}catch(ExportException e){
				registry = LocateRegistry.getRegistry(1099);
			}
			registry.rebind(AuthorityConfiguration.AuthorityServiceName, export);
			logger.log(LoggerLevel.STATUS, "AuthServer started.");
		} catch (RemoteException e) {
			e.printStackTrace();
			this.error = e;
			logger.log(LoggerLevel.WARNING,
			"Remote exception while starting server.");
		} 
		started = true;
	}
	
	private Connection getConnection() throws SQLException {
		try{
		    return cpds.getConnection();
		}catch(SQLException e){
			e.printStackTrace();
			throw e;
		}
	}
	
	public void createAuthorityServer() {
		// Make sure old tables are deleted
		
		lastEID = START_EID;
		EventRecord startEvent = new EventRecord(lastEID,  null, AeolusApiConstants.AUTHORITY_START);
		
		authLog.push(startEvent);
		
		deleteAuthorityServer(null);

		// Create the necessary tables
		createTables(null);

		// create system principal and core
		PrincipalReply pid_coreid;
		pid_coreid = createSystemPrincipal(CREATE_AUTHORITY_ROOT_PID_EVENTID);

		if (!pid_coreid.principal.equals(pidSystem)) {
			logger.log(LoggerLevel.ASSERT, "PidSystem is not as expected");
		}
		if (pid_coreid.coreID != coreIdSystem) {
			logger.log(LoggerLevel.ASSERT, "CoreIdSys is not as expected");
		}

		try{
			// create public principal
			AuthReply<PrincipalReply> rep = createPrincipal(pidSystem, lastEID);
			pid_coreid = rep.value;
			lastEID = rep.event;
			if (!pid_coreid.principal.equals(pidPublic)) {
				logger.log(LoggerLevel.ASSERT, "PidPublic is not as expected");
			}

			// create defunct principal
			rep = createPrincipal(pidSystem, lastEID);
			pid_coreid = rep.value;
			lastEID = rep.event;
			
			if (!pid_coreid.principal.equals(pidDefunct)) {
				logger.log(LoggerLevel.ASSERT, "PidDefunct is not as expected");
			}

			// create defunct taint
			AuthReply<TagID> repTag = createTag(pidSystem, lastEID);
			TagID tidGenerated = repTag.value;
			lastEID = repTag.event;
			if (! tidGenerated.equals(tidInvalid)) {
				logger.log(LoggerLevel.ASSERT, "TidInvalid is not as expected");
			}
		}catch(AuthorityException e){
			e.printStackTrace();
		}
	}

	public void deleteAuthorityServer(String postFix) {
		if(postFix == null)
			postFix = "";
		try {
			Connection conn = getConnection();
			try {
				Statement cmd = conn.createStatement();
				deleteAuthorityServer(postFix, cmd);
			} catch (SQLException e) {
				this.error = e;
				e.printStackTrace();
				logger.log(LoggerLevel.DEBUG,
						"DeleteAuthorityServer Drop Exception ["
						+ e.getMessage() + "]");
			} finally {
				close(conn);
			}
		} catch (SQLException e) {
			this.error = e;
			e.printStackTrace();
		}
	}
	
	public void deleteAuthorityServer(String postFix, Statement cmd) throws SQLException {

				// NOTE: Not using transactions since some of these
				// statements may fail (raise exceptions)
				for (String rel : relNames) {
					String dropStmt = "DROP TABLE IF EXISTS " + rel + postFix;
					cmd.execute(dropStmt);
				}

				cmd.clearBatch();

	}

	// WarmupAuthorityServer
	// This function is used to read all tables to warm up the sqlserver cache
	public void warmupAuthorityServer() {
		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			try {
				for (String relName : relNames) {
					ResultSet rs = cmd.executeQuery("SELECT * FROM " + relName);
					while (rs.next()) {
					}
					;
				}
			} catch (SQLException e) {
				this.error = e;
				logger.log(LoggerLevel.DEBUG, "Warmup Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "Warmup Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
	}

	private List<Integer> getActors(int pidActor) {
		// simple query traversal
		Connection conn = null;
		try {
			conn = getConnection();
			return getActorPidsRecursively(pidActor, conn);
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"GetActors Connection Exception on pidActor " + pidActor
					+ " [" + e.getMessage() + "]");
			return null;
		} finally {
			close(conn);
		}
	}

	private List<Integer> getActorPidsRecursively(int pidTo, Connection conn)
	throws SQLException {
		// Base case: empty list
		List<Integer> rList = new LinkedList<Integer>();

		// Get inlinks (any links)
		String query = "SELECT " + L_PidFromAttribName + " FROM "
		+ L_RelationName + " WHERE " + L_PidToAttribName + "=" + pidTo;
		Statement cmd = conn.createStatement();
		ResultSet rs = cmd.executeQuery(query);

		// construct list of inlinks
		int count = 0;
		List<Integer> inlinkPids = new LinkedList<Integer>();
		while (rs.next()) {
			int pidFound = rs.getInt(1);
			if (pidFound != pidSystem.getValue()) {
				inlinkPids.add(pidFound);
				count++;
			}
		}
		cmd.close();

		// Recurse (merge results)
		for (Integer par : inlinkPids) {
			// add the parent to the return list
			rList.add(par);

			// merge the children/sublist with return list
			List<Integer> subList = getActorPidsRecursively(par, conn);
			rList.addAll(subList);
		}

		return rList;
	}

	public List<Integer> getActees(PrincipalID actee) {
		// simple query traversal
		int pidActee = actee.getValue();

		Connection conn = null;
		try {
			conn = getConnection();
			return getActeePidsRecursively(pidActee, conn);
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"GetActees Connection Exception on pidActee " + pidActee
					+ " [" + e.getMessage() + "]");
			return null;
		} finally {
			close(conn);
		}
	}

	private List<Integer> getActeePidsRecursively(int pidFrom, Connection conn)
	throws SQLException {
		// Base case: empty list
		List<Integer> rList = new LinkedList<Integer>();

		// Get outlinks (any links)
		String query = "SELECT " + L_PidToAttribName + " FROM "
		+ L_RelationName + " WHERE " + L_PidFromAttribName + "="
		+ pidFrom;
		Statement cmd = conn.createStatement();
		ResultSet rs = cmd.executeQuery(query);

		// construct list of outlinks
		int count = 0;
		List<Integer> outlinkPids = new LinkedList<Integer>();
		while (rs.next()) {
			int pidFound = rs.getInt(1);
			outlinkPids.add(pidFound);
			count++;
		}
		cmd.close();

		// Recurse (merge results)
		for (Integer par : outlinkPids) {
			// add the parent to the return list
			rList.add(par);

			// merge the children/sublist with return list
			List<Integer> subList = getActeePidsRecursively(par, conn);
			rList.addAll(subList);
		}

		return rList;
	}

	private List<Integer> grantedTo(int pidGrantor, int tid) {
		List<Integer> listofGrantees = new LinkedList<Integer>();

		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			try {
				String grantQuery = "SELECT " + G_GranteePidAttribName
				+ " FROM " + G_RelationName + " WHERE "
				+ G_GrantorPidAttribName + "=" + pidGrantor + " AND "
				+ G_TidAttribName + "=" + tid;
				ResultSet rs = cmd.executeQuery(grantQuery);
				while (rs.next()) {
					listofGrantees.add(rs.getInt(1));
				}
				rs.close();
			} catch (SQLException e) {
				this.error = e;
				logger.log(LoggerLevel.DEBUG, "HasGranted Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "HasGranted Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}

		return listofGrantees;
	}



	public InvalidationPacket getInvalidationPacket(int seqNumFirstEntry) {
		if(INVALIDATION_CACHING){
			InvalidationPacket l = this.invalidationsCache.getInvalidations(seqNumFirstEntry);
			if(l != null)
				return l;
		}
		return getInvalidationPacket(seqNumFirstEntry, -1);
	}

	public InvalidationPacket getInvalidationPacket(int seqNumFirstEntry,
			int seqNumLastEntry) {

		List<RedoLogEntry> entries = new LinkedList<RedoLogEntry>();
		Connection conn = null;

		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();

			try {
				String query = "SELECT * FROM "
				+ R_RelationName + " WHERE " + R_SeqNumAttribName
				+ " >= " + seqNumFirstEntry;
				if(seqNumLastEntry != -1){
					query += " AND " + R_SeqNumAttribName + " <= " + seqNumLastEntry;
				}
				query += " ORDER BY " + R_SeqNumAttribName + " ASC";
				ResultSet rs = cmd.executeQuery(query);
				String lastEvent = null;
				while (rs.next()) {
					RedoLogEntry entry = new RedoLogEntry();
					entry.seqNum = rs.getInt(R_SeqNumAttribName);
					entry.coreId = rs.getInt(R_CoreIdAttribName);
					entry.requestType = StreamMessageOpcode.values()[rs
					                                                 .getInt(R_RequestTypeAttribName)];
					entry.message = rs.getString(R_MessageAttribName);
					lastEvent = rs.getString(R_EventIDName);
					entries.add(entry);
				}
				rs.close();
				
				return new InvalidationPacket(entries, new EventID(lastEvent));

			} catch (SQLException e) {
				this.error = e;
				logger.log(LoggerLevel.DEBUG, "CreateTaintTable Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"CreateTaintTable Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}
		return null;
	}

	public void warmupDB(String snapshotDbName) {
	}

	public CoreContents loadCoresFromSnapshot(String snapshotDbName,
			List<Integer> coreIds) {
		return loadCores(snapshotDbName, coreIds);
	}

	public CoreContents loadCoreFromSnapshot(
			String snapshotDbName, ClosureID closure) throws RemoteException {
		// Get coreId
		// System.out.println("FFFFF loadCoreFromSnapshot 1 " + closure + " " + snapshotDbName);
		int coreId = getCoreOfClosure(closure);
		// System.out.println("FFFFF loadCoreFromSnapshot 2 " + coreId);
		CoreContents cc = loadCores(snapshotDbName, Collections
				.singletonList(coreId));
		// System.out.println("FFFFF loadCoreFromSnapshot 3 " + cc);
		return cc;
	}

	public CoreContents loadCoreFromSnapshot(
			String snapshotDbName, PrincipalID pid) {
		int coreId = getCoreOfPrincipal(pid.getValue());
		CoreContents cc = loadCores(snapshotDbName, Collections
				.singletonList(coreId));
		return cc;
	}

	private CoreContents loadCores(String snapshotPostfix, List<Integer> coreIds) {
		List<TaintTableEntry> taintEntries = new LinkedList<TaintTableEntry>();
		List<PrincipalTableEntry> principalEntries = new LinkedList<PrincipalTableEntry>();
		List<GrantTableEntry> grantEntries = new LinkedList<GrantTableEntry>();
		List<LinkTableEntry> linkEntries = new LinkedList<LinkTableEntry>();
		List<ClosureTableEntry> closureEntries = new LinkedList<ClosureTableEntry>();
		List<CompoundTagTableEntry> compoundEntries = new LinkedList<CompoundTagTableEntry>();
		List<NodeTableEntry> nodeEntries = new LinkedList<NodeTableEntry>();
		
		// Construct condition string
		if (coreIds.isEmpty())
			return null;

		StringBuilder coreConditionBuilder = new StringBuilder();
		boolean first = true;
		for (Integer core : coreIds) {
			if (!first) {
				coreConditionBuilder.append(" OR ");
			}
			coreConditionBuilder.append(CoreIdAttribName + "=" + core);
			first = false;
		}

		String coreConditionClause = coreConditionBuilder.toString();
		Connection conn;
		try{
			conn = getConnection();
		}catch(SQLException e){
			e.printStackTrace();
			return null;
		}
		
		try {
			Statement cmd = conn.createStatement();
			String loadQuery = "";

				// Taint Table: <tid> <creatorPid> <coreId>
				loadQuery = "SELECT * FROM " + T_RelationName + snapshotPostfix + " WHERE ("
				+ coreConditionClause + ")";
				ResultSet rs = cmd.executeQuery(loadQuery);
				while (rs.next()) {
					TaintTableEntry entry = new TaintTableEntry(rs
							.getInt(T_CreatorPidAttribName), rs
							.getInt(T_TidAttribName), rs
							.getInt(T_CoreIdAttribName));
					taintEntries.add(entry);
				}

				// Principal Table: <pid> <creatorPid> <coreId>
				loadQuery = "SELECT * FROM " + P_RelationName + snapshotPostfix + " WHERE ("
				+ coreConditionClause + ")";
				rs = cmd.executeQuery(loadQuery);

				while (rs.next()) {
					PrincipalTableEntry entry = new PrincipalTableEntry(rs
							.getInt(P_PidAttribName), rs
							.getInt(P_CreatorPidAttribName), -1, rs
							.getInt(P_CoreIdAttribName));
					principalEntries.add(entry);
				}

				// Grant Table: <grantorPid> <granteePid> <tid> <coreId>
				loadQuery = "SELECT * FROM " + G_RelationName + snapshotPostfix + " WHERE ("
				+ coreConditionClause + ")";
				 rs = cmd.executeQuery(loadQuery);
				while (rs.next()) {
					GrantTableEntry entry = new GrantTableEntry(rs
							.getInt(G_GrantorPidAttribName), rs
							.getInt(G_GranteePidAttribName), rs
							.getInt(G_TidAttribName), rs
							.getInt(G_CoreIdAttribName));
					grantEntries.add(entry);
				}

				// Link Table: <pidFrom> <pidTo> <linkType>
				loadQuery = "SELECT * FROM " + L_RelationName + snapshotPostfix + " WHERE ("
				+ coreConditionClause + ")";
				rs = cmd.executeQuery(loadQuery);
				while (rs.next()) {
					LinkTableEntry entry = new LinkTableEntry();
					entry.pidFrom = rs.getInt(L_PidFromAttribName);
					entry.pidTo = rs.getInt(L_PidToAttribName);
					entry.linkType = LinkType.values()[rs
					                                   .getInt(L_LinkTypeAttribName)];
					entry.coreId = rs.getInt(L_CoreIdFromAttribName);
					entry.coreIdTo = rs.getInt(L_CoreIdToAttribName);
					linkEntries.add(entry);
				}

				// Closure Table: <cid> <pid> <key> <coreId>
				loadQuery = "SELECT * FROM " + C_RelationName + snapshotPostfix + " WHERE ("
				+ coreConditionClause + ")";
				rs = cmd.executeQuery(loadQuery);
				while (rs.next()) {
					ClosureTableEntry entry = new ClosureTableEntry(rs
							.getInt(C_CidAttribName), rs
							.getInt(C_PidClosureAttribName), rs
							.getString(C_ClassNameAttribName), rs
							.getInt(C_CoreIdAttribName));
					closureEntries.add(entry);
				}

				// Compound tag table: <child_tagID> <par_tagID> <coreID>
				loadQuery = "SELECT * FROM " + Comp_RelationName + snapshotPostfix + " WHERE ("
				+ coreConditionClause + ")";
				rs = cmd.executeQuery(loadQuery);
				while(rs.next()){
					CompoundTagTableEntry entry = new CompoundTagTableEntry(
							rs.getInt(Comp_ChildTidAttribName),
							rs.getInt(Comp_ParTidAttribName),
							rs.getInt(Comp_CoreIdAttribName));
					compoundEntries.add(entry);
				}
				loadQuery = "SELECT * FROM " + NodeTable_Name;
				rs = cmd.executeQuery(loadQuery);
				while(rs.next()){
					NodeTableEntry entry = new NodeTableEntry(
							rs.getInt(NodeTable_NodeID),
							rs.getString(PINodeTable_Host));
					nodeEntries.add(entry);
				}

		} catch (SQLException e) {
			this.error = e;
			System.err.println("LoadCore Exception ["
					+ e.getMessage() + "]");
			return null;
		} finally {
			close(conn);
		}
		CoreContents r = new CoreContents(coreIds.get(0));
		r.setClosureEntries(closureEntries);
		r.setGrantEntries(grantEntries);
		r.setLinkEntries(linkEntries);
		r.setPrincipalEntries(principalEntries);
		r.setTaintEntries(taintEntries);
		r.setCompoundTagEntries(compoundEntries);
		r.setNodeTableEntries(nodeEntries);
		return r;
	}

	public boolean hasAuthority(PrincipalID pid, TagID tid) {
		return hasAuthoritySingleTransaction(pid, tid).first;
	}

	// HasAuthoritySingleTransaction
	// One implementation of HasAuthority	
	/** 
	 * @return Boolean for authoratative, int for found depth.
	 */
	private Pair<Boolean, Integer> hasAuthoritySingleTransaction(PrincipalID pid,
			TagID tid) {
		// , out Boolean yes, out int foundDepth
		Connection conn = null;

		try {
			conn = getConnection();
			return hasAuthorityRecursively(pid, tid, conn, 0);
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"HasAuthorityFast Connection Exception [" + e.getMessage()
					+ "]");
			return null;
		} finally {
			close(conn);
		}
	}

	// HasAuthorityRecursively
	// HasAuthority
	private Pair<Boolean, Integer> hasAuthorityRecursively(PrincipalID pid, TagID tid,
			Connection conn, int currentDepth) throws SQLException {
		String authQuery = "SELECT COUNT(*) FROM " + G_RelationName + " WHERE "
		+ G_GranteePidAttribName + "=" + pid.getValue() + " AND "
		+ G_TidAttribName + "=" + tid.getValue();
		String childQuery = "SELECT " + L_PidToAttribName + " FROM "
		+ L_RelationName + " WHERE " + L_PidFromAttribName + "=" + pid.getValue();

		// Base case:
		// Either 1) YES has authority or 2) No more child and can return
		// definite NO

		Statement cmd = conn.createStatement();

		// check whether pid has explicit authority
		ResultSet rs = cmd.executeQuery(authQuery);
		if (!rs.next()) {
			logger.log(LoggerLevel.ASSERT,
			"HasAuthorityRecursively failed AuthQuery");
			return new Pair<Boolean, Integer>(false, currentDepth);
		}
		int numGrants = rs.getInt(1);

		if (numGrants > 0) {
			return new Pair<Boolean, Integer>(true, currentDepth);
		}

		// Otherwise: may ask other children
		rs = cmd.executeQuery(childQuery);
		List<Integer> childPids = new LinkedList<Integer>();
		// save children so
		// can dispose
		while (rs.next()) {
			int childPid = rs.getInt(1);
			childPids.add(childPid);
		}
		cmd.close();

		// recurse on children
		int maxChildDepthNotFound = currentDepth;// the parent (in case no
		// children)
		currentDepth++;
		for (int p : childPids) {
			// NOTE: since we are doing depth-first search, the returned
			// child-found
			// depth is not always the min
			// Can potentially put in some heuristics here..
			// Cache can learn a weighted tree?
			Pair<Boolean, Integer> childAuth = hasAuthorityRecursively(new PrincipalID(p), tid,
					conn, currentDepth);
			if (childAuth.first) {
				// propagate the depth info up the tree
				return childAuth;
			} else {
				// not found in the child (but see if it was the longest branch)
				if (childAuth.second > maxChildDepthNotFound)
					maxChildDepthNotFound = childAuth.second;
			}
		}
		return new Pair<Boolean, Integer>(false, maxChildDepthNotFound); // dead
		// branch
		// (returns
		// max
		// depth)
	}

	// IsActee
	// Tests whether pidActor acts-for pidActee
	public boolean isActee(int pidActor, int pidActee) {
		boolean answer = false;

		Connection conn = null;

		try {
			conn = getConnection();
			answer = searchActeeRecursive(pidActee, pidActor, conn, null);
			return answer;
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "IsActee Connection Exception ["
					+ e.getMessage() + "]");
			return false;
		} finally {
			close(conn);
		}
	}

	// SearchActeeRecursive
	// Recursive helper for IsActee
	private boolean searchActeeRecursive(int pidTarget, int pidNode,
			Connection conn, List<Integer> basis) throws SQLException {
		// Base case: found
		if (pidTarget == pidNode)
			return true;

		// 1. Find Immediate Actees (as long as there is one linke)
		String query = "SELECT " + L_PidToAttribName + " FROM "
		+ L_RelationName + " WHERE " + L_PidFromAttribName + "="
		+ pidNode;
		Statement cmd = conn.createStatement();
		ResultSet rs = cmd.executeQuery(query);

		List<Integer> outlinkPids = new LinkedList<Integer>();
		while (rs.next()) {
			int pidFound = rs.getInt(1);
			outlinkPids.add(pidFound);
		}
		cmd.close();

		// Base case: not found
		if (outlinkPids.isEmpty())
			return false;

		// 2. Search each branch
		for (int pidLink : outlinkPids) {
			if (searchActeeRecursive(pidTarget, pidLink, conn, basis)){
				if(basis != null)
					basis.add(0, pidLink);
				return true;
			}
		}

		return false;
	}

	protected int getCoreOfPrincipal(int pid) {
		Connection conn = null;
		try {
			// override by serving load cores request from the most recent version
			conn = getConnection();
			Statement cmd = conn.createStatement();
			int coreId = getCoreIdOf(cmd, pid);
			return coreId;
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.ASSERT,
			"Failed to lookup COREID in LoadCoreFromSNAPSHOT");
			return -1;
		} finally {
			close(conn);
		}
	}

	private int getCoreOfClosure(ClosureID closure) {
		Connection conn = null;
		try {
			// override by serving load cores request from the most recent
			// version
			conn = getConnection();
			Statement cmd = conn.createStatement();
			String coreCreatorQuery = "SELECT " + C_CoreIdAttribName + " FROM "
			+ C_RelationName + " WHERE " + C_CidAttribName + "=" + closure.getValue();
			int coreId = executeScalarInt(cmd, coreCreatorQuery);
			return coreId;
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.ASSERT,
			"Failed to lookup COREID in LoadCoreFromSNAPSHOT");
			return -1;
		} finally {
			close(conn);
		}
	}

	public Exception getError(){
		return error;
	}

	private int executeScalarInt(Statement cmd, String q) throws SQLException {
		ResultSet rs = cmd.executeQuery(q);
		if (rs.next()) {
			return rs.getInt(1);
		} else {
			throw new SQLException("No results found for scalar query: \n " + q);
		}
	}

    private int executeScalarInt2(Statement cmd, String t, String servar) throws SQLException {
	String qry = "SELECT CURRVAL(pg_get_serial_sequence('"+t+"','"+servar.toLowerCase()+"'))";
	return executeScalarInt(cmd, qry);
	}

	private int getCoreIdOf(Statement cmd, int pid) throws SQLException {
		String coreCreatorQuery = "SELECT " + P_CoreIdAttribName + " FROM "
		+ P_RelationName + " WHERE " + P_PidAttribName + "=" + pid;

		return executeScalarInt(cmd, coreCreatorQuery);
	}

	private void incrementReferenceCount(Statement cmd, int coreId)
	throws SQLException {
		String coreRefQuery = "SELECT " + CI_RefCountAttribName + " FROM "
		+ CI_RelationName + " WHERE " + CI_CoreIdAttribName + "="
		+ coreId;
		int coreRefCount = executeScalarInt(cmd, coreRefQuery);
		String coreRefUpdate = "UPDATE " + CI_RelationName + " SET "
		+ CI_RefCountAttribName + "=" + (coreRefCount + 1) + " WHERE "
		+ CI_CoreIdAttribName + "=" + coreId;
		cmd.execute(coreRefUpdate);
	}

	private int createCore(Statement cmd) throws SQLException {
		String coreInsert = "INSERT INTO " + CI_RelationName + "(" + 
		CI_RefCountAttribName + ") VALUES(1)";
		cmd.execute(coreInsert);

		return executeScalarInt2(cmd,CI_RelationName,CI_CoreIdAttribName);
	}

	private enum IN_CORE {
		NEW_CORE, CREATORS_CORE
	};

	private int getAnonPrinciple(Statement cmd, ClosureID closureId)
	throws SQLException {
		String cidQuery = "SELECT " + C_PidClosureAttribName + " FROM "
		+ C_RelationName + " WHERE " + C_CidAttribName + "="
		+ closureId.getValue();
		return executeScalarInt(cmd, cidQuery);
	}



	private static void close(Connection conn) {
		if (conn != null)
			try {
				conn.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
	}

	// Table Creation
	private void createTable(Statement cmd, String tableName, String... attrs_types) throws SQLException {
			String createStmt = "CREATE TABLE " + tableName + " (";
			for (int i = 0; i < attrs_types.length; i += 2) {
				if (i == attrs_types.length - 1) {
					createStmt += attrs_types[i];
				} else {
					createStmt += attrs_types[i] + " " + attrs_types[i + 1];
					if (i != attrs_types.length - 2) {
						createStmt += ",";
					}
				}
			}
			createStmt += ")";

			cmd.execute(createStmt);
	}

	
	 /* 
	  *The following methods all change the authority state.
	  */
	private int appendToRedoLog(Statement cmd, StreamMessageEntry e, int coreId, EventID eid) throws SQLException{
		String logStmt = "INSERT INTO " + R_RelationName 
		+ "(" + R_CoreIdAttribName + ", " + R_RequestTypeAttribName + "," + R_MessageAttribName + ", " + R_EventIDName + ")" 
		+ " VALUES("
		+ coreId + ", " + e.getRequestType() + ", " + "'"
		+ e.getStreamMessage() + "', '" + eid + "')";
		cmd.execute(logStmt);
		int seqNum = executeScalarInt2(cmd,R_RelationName,R_SeqNumAttribName);
		RedoLogEntry r = new RedoLogEntry(seqNum, coreId, 
				StreamMessageOpcode.values()[e.getRequestType()], e.getStreamMessage());
		invalidationsCache.addInvalidation(r, eid);
		return seqNum;
	}

	private PrincipalReply createSystemPrincipal(EventID predecessor) {
		Connection conn = null;
		try {
			conn = getConnection();

			Statement cmd = conn.createStatement();

			int pidCreator = pidSystem.getValue();

			// Set a savepoint for rollback.
			conn.setAutoCommit(false);
			try {

				// 1. Create a new core
				String coreInsert = "INSERT INTO " + CI_RelationName + "(" + CI_RefCountAttribName +")"
				    + " VALUES(1)";
				cmd.execute(coreInsert);

				int coreIdCreated = executeScalarInt2(cmd,CI_RelationName,CI_CoreIdAttribName);

				// 2. Create new principal
				String pidInsert = "INSERT INTO " + P_RelationName + 
				"(" + P_CreatorPidAttribName + ", " + P_CoreIdAttribName + ")" 
				+ " VALUES("
				+ pidCreator + ", " + coreIdCreated + ")";
				cmd.execute(pidInsert);

				int pidCreated = executeScalarInt2(cmd, P_RelationName, P_PidAttribName);

				// Log event, get EID.
				
				// 3. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser
				.generateCreatePrincipal(pidCreator, pidCreated,
						coreIdSystem, coreIdCreated);

				appendToRedoLog(cmd, e, coreIdCreated, predecessor);

				// COMMIT
				conn.commit();
				
				return new PrincipalReply(new PrincipalID(pidCreated), coreIdCreated);
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				e.printStackTrace();
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "CreateSystemPid Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"CreateSystemPid Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}
		return null;
	}

	private AuthReply<PrincipalReply> createPrincipal(PrincipalID creator, IN_CORE core, EventID predecessor) 
	throws AuthorityException {

		int coreIdCreated = coreIdInvalid;

		if (creator.equals(pidPublic)) {
			logger.log("CreatePrincipalInNewCore - pidCreator is pidPublic");
			throw new AuthorityException();
		}
		Connection conn = null;
		try {
			conn = getConnection();

			Statement cmd = conn.createStatement();
			// Set a savepoint for rollback.
			conn.setAutoCommit(false);

			try {
				// 1. Find coreId of creator
				int pidCreator = creator.getValue();
				int coreIdCreator = getCoreIdOf(cmd, pidCreator);

				// 2. Create a new core
				if (core == IN_CORE.NEW_CORE) {
					coreIdCreated = createCore(cmd);
				} else if (core == IN_CORE.CREATORS_CORE) {
					incrementReferenceCount(cmd, coreIdCreator);
					coreIdCreated = coreIdCreator;
				} else {
					throw new RuntimeException();
				}

				// 3. Create new principal
				String pidInsert = "INSERT INTO " + P_RelationName +
				joinComma(P_CreatorPidAttribName, P_CoreIdAttribName)
				+ " VALUES " + joinComma(""+ pidCreator, ""+ coreIdCreated);
				cmd.execute(pidInsert);

				int pidCreated = executeScalarInt2(cmd,P_RelationName,P_PidAttribName);

				// 4. Add Link
				if (pidCreated != pidCreator) {
					String linkStmt = "INSERT INTO " + L_RelationName +
					joinComma(L_PidFromAttribName, L_PidToAttribName,
					L_LinkTypeAttribName, L_CoreIdFromAttribName,
					L_CoreIdToAttribName)					
					+ " VALUES(" + pidCreator + ", " + pidCreated
					+ ", " + L_OWNS + ", " + coreIdCreator + ", "
					+ coreIdCreated + ")";
					cmd.execute(linkStmt);
				} else {
					throw new RuntimeException(
					"CreatePrincipalInNewCore pidCreator and pidCreated are same pid");
				}

				// 5. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser
				.generateCreatePrincipal(pidCreator, pidCreated,
						coreIdCreator, coreIdCreated);
				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params = {creator, core};
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.CREATE_PRINCIPAL, params);
					myEvent.setReturnValue(new PrincipalID(pidCreated));
					authLog.push(myEvent);
					lastEID = myEid;
				}
				
				appendToRedoLog(cmd, e, coreIdCreated, myEid);
				
				// COMMIT
				conn.commit();
				return new AuthReply<PrincipalReply>( myEid , 
						new PrincipalReply(new PrincipalID(pidCreated), coreIdCreated));
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				e.printStackTrace();
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "CreatePrinciple Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"CreatePrinciple Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}
		return null;
	}

	private EventID getNextEid(EventID lastEid) {
		if(lastEid.equals(START_EID))
			return new EventID("auth-0");
		else{
			int lastInt = Integer.parseInt(lastEid.toString().substring(5));
			return new EventID("auth-" + (lastInt + 1));
		}
	}

	private EventID getLastEid(Statement cmd) throws SQLException {
		if(lastEID != null)
			return lastEID;
		String sql = "SELECT " + R_EventIDName + " FROM " + R_RelationName + 
		" ORDER BY " + R_SeqNumAttribName + " DESC LIMIT 1";
		ResultSet rs = cmd.executeQuery(sql);
		rs.first();
		lastEID = new EventID(rs.getString(1));
		return lastEID;
	}

	public AuthReply<PrincipalReply> createPrincipalInNewCore(PrincipalID pidCreator, EventID predecessor) throws AuthorityException {
		return createPrincipal(pidCreator, IN_CORE.NEW_CORE, predecessor);
	}

	public AuthReply<PrincipalReply> createPrincipal(PrincipalID pidCreator, EventID predecessor) throws AuthorityException {
		return createPrincipal(pidCreator, IN_CORE.CREATORS_CORE, predecessor);
	}

	public AuthReply<TagID> createTag(PrincipalID creator, EventID predecessor) throws AuthorityException {
		int pidCreator = creator.getValue();
		
		if (creator.equals(pidPublic)) {
			logger.log("CreateTaint - pidCreator is pidPublic");
			throw new AuthorityException();
		}

		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {
				// 1. Find coreId of creator
				int coreId = getCoreIdOf(cmd, pidCreator);

				// 2. Get and update reference count of core
				incrementReferenceCount(cmd, coreId);

				// 3. Create new taint
				String insertStmt = "INSERT INTO " + T_RelationName
				+ "( " + T_CreatorPidAttribName + ", " + T_CoreIdAttribName + ")"
				+ " VALUES(" + pidCreator + ", " + coreId + ")";
				cmd.execute(insertStmt);

				int tidCreated = executeScalarInt2(cmd,T_RelationName,T_TidAttribName);

				// 4. Grant explicit authority for created taint
				// NOTE: this is mostly so that we don't have to look in taints
				// table
				// when determining authority
				String insertGrantStmt = "INSERT INTO " + G_RelationName
				+ "( " + G_CoreIdAttribName +", " + G_GrantorPidAttribName 
				+ ", " + G_GranteePidAttribName + ", " + G_TidAttribName + ")"
				+ " VALUES(" + coreId + ", " + pidSystem.getValue() + ", "
				+ pidCreator + ", " + tidCreated + ")";
				cmd.execute(insertGrantStmt);

				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				
				// 5. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser.generateCreateTaint(
						pidCreator, tidCreated);

				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params = {creator};
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.CREATE_TAG, params);
					myEvent.setReturnValue(new TagID(tidCreated));
					authLog.push(myEvent);
					lastEID = myEid;
				}
				
				appendToRedoLog(cmd, e, coreId, myEid);
				
				
				// COMMIT
				conn.commit();
				
				return new AuthReply<TagID>(myEid, new TagID(tidCreated));
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "CreateTaint Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "CreateTaint Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
		return null;
	}
	
	public AuthReply<TagID> createSubTag(TagID parent, PrincipalID creator, EventID predecessor) 
	throws AuthorityException {
		int pidCreator = creator.getValue();
		
		if (creator.equals(pidPublic)) {
			logger.log("CreateTaint - pidCreator is pidPublic");
			throw new AuthorityException();
		}

		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {
				// 1. Find coreId of creator
				int coreId = getCoreIdOf(cmd, pidCreator);

				// 2. Get and update reference count of core
				incrementReferenceCount(cmd, coreId);

				// 3. Create new taint
				String insertStmt = "INSERT INTO " + T_RelationName
				+ "( " + T_CreatorPidAttribName + ", " + T_CoreIdAttribName + ")"
				+ " VALUES(" + pidCreator + ", " + coreId + ")";
				cmd.execute(insertStmt);

				int tidCreated = executeScalarInt2(cmd,T_RelationName,T_TidAttribName);

				// 4. Grant explicit authority for created taint
				// NOTE: this is mostly so that we don't have to look in taints
				// table
				// when determining authority
				String insertGrantStmt = "INSERT INTO " + G_RelationName
				+ "( " + G_CoreIdAttribName +", " + G_GrantorPidAttribName 
				+ ", " + G_GranteePidAttribName + ", " + G_TidAttribName + ")"
				+ " VALUES(" + coreId + ", " + pidSystem.getValue() + ", "
				+ pidCreator + ", " + tidCreated + ")";
				cmd.execute(insertGrantStmt);

				// Insert link into Compound Tag table
				String insertCmpdStmt = "INSERT INTO " + Comp_RelationName +
				"( " + Comp_ChildTidAttribName + ", " + Comp_ParTidAttribName + ", " +
				Comp_CoreIdAttribName + ") VALUES( " + tidCreated + ", " + parent.getValue()
				+ ", " + coreId + ")";
				cmd.execute(insertCmpdStmt);

				EventID lastEid = enableLogging ? getLastEid(cmd) : null;

				// 5. Append to Redo Log
				StreamMessageEntry e1 = StreamMessageParser.generateCreateTaint(
						pidCreator, tidCreated);
				StreamMessageEntry e2 = StreamMessageParser.generateCreateCompoundLink(
						tidCreated, parent.getValue());

				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params = {parent, creator};
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.CREATE_SUBTAG, params);
					myEvent.setReturnValue(new TagID(tidCreated, parent.getValue()));
					authLog.push(myEvent);
					lastEID = myEid;
				}
				
				appendToRedoLog(cmd, e1, coreId, myEid);
				appendToRedoLog(cmd, e2, coreId, myEid);				
				
				// COMMIT
				conn.commit();
				
				return new AuthReply<TagID>(myEid, new TagID(tidCreated, parent.getValue()));
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "CreateTaint Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "CreateTaint Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
		return null;
	}


	private String joinComma(String ... args){
		String b = "";
		for(String s : args){
			b += s + ", ";
		}
		return " (" + b.substring(0, b.length() - 2) + ") ";
	}
	
	public AuthReply<ClosureReply> createClosure(String classname, EventID predecessor) {
		int pidClosure = pidDefunct.getValue();
		
		Connection conn = null;
		try {
			conn = getConnection();

			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);
			try {
				// 1. Create a new core
				int coreIdCreated = createCore(cmd);

				// 2. Create a new anon principal
				String pidInsert = "INSERT INTO " + P_RelationName 
				+ "( " + P_CreatorPidAttribName + ", " + P_CoreIdAttribName + ")"
				+ " VALUES("
				+ pidSystem.getValue() + ", " + coreIdCreated + ")";
				cmd.execute(pidInsert);

				pidClosure = executeScalarInt2(cmd,P_RelationName,P_PidAttribName);

				// 3. Add entry to closure table
				String closureInsert = "INSERT INTO " + C_RelationName 
				+ joinComma(C_PidClosureAttribName, C_ClassNameAttribName, C_CoreIdAttribName)
				+ " VALUES(" + pidClosure + ", " + "'" + classname + "', "
				+ coreIdCreated + ")";
				cmd.execute(closureInsert);
				int closureId = executeScalarInt2(cmd,C_RelationName,C_CidAttribName);

				// 4. Add Link (from pidSystem)
				String linkStmt = "INSERT INTO " + L_RelationName+
				joinComma(L_PidFromAttribName, L_PidToAttribName, L_LinkTypeAttribName, L_CoreIdFromAttribName, L_CoreIdToAttribName)
				+ " VALUES("
				+ pidSystem.getValue() + ", " + pidClosure + ", " + L_OWNS + ", "
				+ coreIdSystem + ", " + coreIdCreated + ")";
				cmd.execute(linkStmt);

				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				
				// 5. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser
				.generateCreateClosure(closureId, pidClosure, classname,
						coreIdCreated);
				
				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params = {closureId, pidClosure, classname};
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.CREATE_CLOSURE, params);
					authLog.push(myEvent);
					lastEID = myEid;
				}				
				appendToRedoLog(cmd, e, coreIdCreated, myEid);

				// COMMIT
				
				conn.commit();
				return new AuthReply<ClosureReply>(myEid, 
						new ClosureReply(new ClosureID(closureId), new PrincipalID(pidClosure)));
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "CreateTaint Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "CreateTaint Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
		return null;
	}

	// ClosureActFor
	// Allows closure's anon principal to act-for actee
	public AuthReply<Void> closureActsFor(PrincipalID requestor, ClosureID cid, PrincipalID actee, EventID predecessor)
	throws AuthorityException {

		int pidActee = actee.getValue();
		int pidRequestor = actee.getValue();
		
		// 1. Special Principal Check
		// - pidActee cannot be pidSystem
		// - pidActee cannot be pidPublic
		if (pidActee == pidSystem.getValue()) {
			logger.log(LoggerLevel.DEBUG,
			"ClosureActsFor - pidActee is pidSystem");
			return new AuthReply<Void>(null, null);
		}
		if (pidActee == pidPublic.getValue()) {
			logger.log(LoggerLevel.DEBUG,
			"ClosureActsFor - pidActee is pidPublic");
			return new AuthReply<Void>(null, null);
		}

		// 2. Check requestor act-for pidActee
		if (!isActee(pidRequestor, pidActee)) {
			logger.log(LoggerLevel.IFCERR, "ClosureActsFor - pidRequestor "
					+ pidRequestor + " does not act-for pidActee " + pidActee);
			throw new AuthorityException();
		}

		// NOTE: cycle not possible since no pid can act-for anon principal

		// 3. Look up anon principal and insert link
		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {
				// 1. Get anon principal
				int pidClosure = getAnonPrinciple(cmd, cid);

				// 2. Get coreId of pidClosure and pidActee
				int coreIdPidClosure = getCoreIdOf(cmd, pidClosure);
				int coreIdPidActee = getCoreIdOf(cmd, pidActee);

				// 3. Insert the link
				String insertStmt = "INSERT INTO " + L_RelationName
				+ joinComma(L_PidFromAttribName, L_PidToAttribName, L_LinkTypeAttribName, L_CoreIdFromAttribName, L_CoreIdToAttribName)
				+ " VALUES(" + pidClosure + ", " + pidActee + ", "
				+ L_ACTSFOR + ", " + coreIdPidClosure + ", "
				+ coreIdPidActee + ")";
				cmd.execute(insertStmt);

				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				// 4. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser
				.generateDelegateActsFor(pidClosure, pidActee,
						coreIdPidClosure, coreIdPidActee);
				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params = {requestor, cid, actee, pidClosure};
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.CLOSURE_ACT_FOR, params);
					authLog.push(myEvent);
					lastEID = myEid;
				}
				appendToRedoLog(cmd, e, coreIdPidClosure, myEid);

				// COMMIT
				conn.commit();

				return new AuthReply<Void>(myEid, null);
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "ClosureActsFor Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"ClosureActsFor Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}
		return null;
	}

	public AuthReply<Void> closureGrant(PrincipalID requestor, ClosureID cid, PrincipalID grantor, TagID tag, EventID predecessor)
	throws AuthorityException {
		// 1. Special Principal Check
		// - pidGrantor cannot be pidPublic (but it shouldn't have any authority
		// anyways)
		int pidGrantor = grantor.getValue();
		int pidRequestor = requestor.getValue();
		int tid = tag.getValue();
		
		if (pidGrantor == pidPublic.getValue()) {
			logger.log(LoggerLevel.DEBUG,
					"ClosureGrant - pidGrantor is pidPublic");
			return new AuthReply<Void>(null, null);
		}

		// 2. Requestor must act-for pidGrantor
		if (!isActee(pidRequestor, pidGrantor)) {
			logger.log(LoggerLevel.IFCERR, "ClosureGrant - pidRequestor "
					+ pidRequestor + " does not act-for pidGrantor "
					+ pidGrantor);
			throw new AuthorityException();
		}

		// 3. pidGrantor must have authority for tid
		if (!hasAuthority(grantor, tag)) {
			logger.log(LoggerLevel.IFCERR, "ClosureGrant - pidGrantor "
					+ pidGrantor + " does not have authority for tid " + tid);
			throw new AuthorityException();
		}

		// 4. Lookup anon principal and grant
		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {

				// 1. Lookup anon principal
				int pidClosure = getAnonPrinciple(cmd, cid);

				// 2. Get coreId of anon pid
				int coreId = getCoreIdOf(cmd, pidClosure);

				// 3. Grant (coreIdTo)
				String insertStmt = "INSERT INTO " + G_RelationName
				+ joinComma(G_CoreIdAttribName, G_GrantorPidAttribName, G_GranteePidAttribName, G_TidAttribName)
				+ " VALUES(" + coreId + ", " + pidGrantor + ", "
				+ pidClosure + ", " + tid + ")";
				cmd.execute(insertStmt);

				@SuppressWarnings("all")
				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				
				// 4. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser.generateGrant(
						pidGrantor, pidClosure, tid);

				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params =  { requestor, cid, pidClosure, grantor, tag };
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.CLOSURE_DELEGATE, params);
					authLog.push(myEvent);
					lastEID = myEid;
				}
				appendToRedoLog(cmd, e, coreId, myEid);
				// COMMIT

				conn.commit();
				
				return new AuthReply<Void>(myEid, null);
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "ClosureGrant Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "ClosureGrant Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
		return null;
	}

	public AuthReply<Void> grant(PrincipalID requestor, PrincipalID grantor, PrincipalID grantee, TagID tag, EventID predecessor)
	throws AuthorityException {
		int pidGrantor = grantor.getValue();
		int pidRequestor = grantor.getValue();
		int pidGrantee = grantee.getValue();
		int tid = tag.getValue();

		// 1. Special Principal Check
		// - Cannot grant to pidPublic
		if (pidGrantor == pidPublic.getValue()) {
			logger.log(LoggerLevel.DEBUG, "Grant - Cannot grant to Ppublic");
			return new AuthReply<Void>(null, null);
		}

		// 2. Requestor must act-for pidGrantor
		if (!isActee(pidRequestor, pidGrantor)) {
			logger.log(LoggerLevel.IFCERR, "ClosureGrant - pidRequestor "
					+ pidRequestor + " does not act-for pidGrantor "
					+ pidGrantor);
			throw new AuthorityException();
		}

		// 3. pidGrantor must have authority for tid
		if (!hasAuthority(grantor, tag)) {
			logger.log(LoggerLevel.IFCERR, "ClosureGrant - pidGrantor "
					+ pidGrantor + " does not have authority for tid " + tid);
			throw new AuthorityException();
		}

		// 4. Grant
		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {
				// 1. Get coreId of granted to pid
				int coreId = getCoreIdOf(cmd, pidGrantee);

				// 2. Grant
				String insertStmt = "INSERT INTO " + G_RelationName
				+ joinComma(G_CoreIdAttribName, G_GrantorPidAttribName, G_GranteePidAttribName, G_TidAttribName)
				+ " VALUES(" + coreId + ", " + pidGrantor + ", "
				+ pidGrantee + ", " + tid + ")";
				cmd.execute(insertStmt);

				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				
				// 3. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser.generateGrant(
						pidGrantor, pidGrantee, tid);
				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params = {tag, grantor, grantee, requestor};
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.DELEGATE, params);
					authLog.push(myEvent);
					lastEID = myEid;
				}
				appendToRedoLog(cmd, e, coreId, myEid);

				// COMMIT
				conn.commit();
				return new AuthReply<Void>(myEid,null);
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "ClosureGrant Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "ClosureGrant Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}

		return null;
	}

	public AuthReply<Void> delegateActsFor(PrincipalID requestor, PrincipalID from, PrincipalID to, EventID predecessor)
	throws AuthorityException {
		int pidFrom = from.getValue();
		int pidTo = to.getValue();
		int pidRequestor = requestor.getValue();
		
		// Sanity Check: pidFrom != pidTo
		if (pidFrom == pidTo) {
			logger.log(LoggerLevel.DEBUG,
			"DelegateActsFor - Same pidFrom and pidTo");
			return new AuthReply<Void>(null, null);
		}

		// 1. Special Principal Check
		// - pidFrom cannot be pidPublic
		// - pidTo cannot be pidSystem
		if (pidFrom == pidSystem.getValue()) {
			logger.log(LoggerLevel.DEBUG,
			"DelegateActsFor - pidFrom is pidSystem");
			return new AuthReply<Void>(null, null);

		}
		if (pidTo == pidPublic.getValue()) {
			logger.log(LoggerLevel.DEBUG,
			"DelegateActsFor - pidTo is pidPublic");
			return new AuthReply<Void>(null, null);

		}

		// 2. Requestor must act-for pidTo
		if (!isActee(pidRequestor, pidTo)) {
			logger.log(LoggerLevel.IFCERR, "ClosureGrant - pidRequestor "
					+ pidRequestor + " does not act-for pidGrantor " + pidTo);
			throw new AuthorityException();
		}

		// 3. Check for cycle
		if (isActee(pidTo, pidFrom)) {
			logger.log(LoggerLevel.DEBUG, "DelegateActsFor - Cycle detected");
			return new AuthReply<Void>(null, null);

		}

		// 4. Permit act-for
		Connection conn = null;

		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {
				// 1. Get core info of pidFrom and pidTo
				int coreFrom = getCoreIdOf(cmd, pidFrom);
				int coreTo = getCoreIdOf(cmd, pidTo);

				// 2. Insert the link entry
				String insertStmt = "INSERT INTO " + L_RelationName
				+ joinComma(L_PidFromAttribName, L_PidToAttribName, L_LinkTypeAttribName, 
						L_CoreIdFromAttribName, L_CoreIdToAttribName)
				+ " VALUES(" + pidFrom + ", " + pidTo + ", "
				+ L_ACTSFOR + ", " + coreFrom + ", " + coreTo + ")";
				cmd.execute(insertStmt);


				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				
				// 3. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser
				.generateDelegateActsFor(pidFrom, pidTo, coreFrom,
						coreTo);
				EventID myEid = null;
				if(enableLogging){
					myEid = getNextEid(lastEid);
					Serializable[] params = {requestor, from, to};
					EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, AeolusApiConstants.ACT_FOR, params);
					authLog.push(myEvent);
					lastEID = myEid;
				}
				appendToRedoLog(cmd, e, coreFrom, myEid);
				// COMMIT
				conn.commit();
				
				return new AuthReply<Void>(myEid, null);
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "ClosureGrant Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "ClosureGrant Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}

		return null;
	}

	private EventID deleteGrantTransitive(int pidGrantor, int pidGrantee, int tid, EventID myEid) {
		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();

			// NOTE: going to try to do this recursively without transaction
			// first

			try {
				Queue<Pair<Integer, Integer>> traversalQueue = new LinkedList<Pair<Integer, Integer>>();

				Pair<Integer, Integer> root = new Pair<Integer, Integer>(
						pidGrantor, pidGrantee);
				traversalQueue.add(root);

				while (traversalQueue.size() > 0) {
					// 1. Dequeue
					Pair<Integer, Integer> entry = traversalQueue.poll();
					int _grantor = entry.first;
					int _grantee = entry.second;

					// 2. Find core of grantee (also check that this entry
					// exists)
					String coreQuery = "SELECT " + G_CoreIdAttribName
					+ " FROM " + G_RelationName + " WHERE "
					+ G_GrantorPidAttribName + "=" + _grantor + " AND "
					+ G_GranteePidAttribName + "=" + _grantee + " AND "
					+ G_TidAttribName + "=" + tid;
					int coreId = executeScalarInt(cmd, coreQuery);

					// 3. Find the grantees of this grantee
					String granteeQuery = "SELECT " + G_GranteePidAttribName
					+ " FROM " + G_RelationName + " WHERE "
					+ G_GrantorPidAttribName + "=" + _grantee + " AND "
					+ G_TidAttribName + "=" + tid;
					ResultSet rs = cmd.executeQuery(granteeQuery);
					while (rs.next()) {
						Pair<Integer, Integer> depEntry = new Pair<Integer, Integer>(
								_grantee, rs.getInt(1));
						traversalQueue.add(depEntry);
					}

					// 4. Delete the current grant
					String deleteStmt = "DELETE FROM " + G_RelationName
					+ " WHERE " + G_GrantorPidAttribName + "="
					+ _grantor + " AND " + G_GranteePidAttribName + "="
					+ _grantee + " AND " + G_TidAttribName + "=" + tid;
					cmd.execute(deleteStmt);

					EventID lastEid = enableLogging ? getLastEid(cmd) : null;
					
					// 5. Append to Redo Log
					StreamMessageEntry e = StreamMessageParser
					.generateRevokeGrant(_grantor, _grantee, tid);

					appendToRedoLog(cmd, e, coreId, myEid);
					
					return lastEid;
				}
			} catch (SQLException e) {
				this.error = e;
				logger.log(LoggerLevel.DEBUG,
						"DeleteGrantTransitive Exception [" + e.getMessage()
						+ "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"DeleteGrantTransitive Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
		return null;
	}

	public AuthReply<Void> revokeGrant(PrincipalID revoker, PrincipalID grantor, PrincipalID grantee,
			TagID tag, EventID predecessor) throws AuthorityException {

		int pidRevoker = revoker.getValue();
		int pidGrantee = grantee.getValue();
		int pidGrantor = grantor.getValue();
		int tid = tag.getValue();
		
		// 1. Requestor must be authoritative for pidGrantor
		// - i.e. act-for
		if (!isActee(pidRevoker, pidGrantor)) {
			logger
			.log(LoggerLevel.IFCERR, "ClosureGrant - pidRequestor "
					+ pidRevoker + " does not act-for pidGrantor "
					+ pidGrantor);
			throw new AuthorityException();
		}

		// 2. Delete Grant Transitively
		EventID myEid = getNextEid(lastEID);
		EventID lastEid = deleteGrantTransitive(pidGrantor, pidGrantee, tid, myEid);

		// 3. Revoke act-for transitive effects
		// NOTE: Might have delegated because has authority for T via act-for
		// - find all actors of pidGrantee (and is a pidGrantor of T)
		// - for each actor, check whether they still has auth for T
		// - if yes, do nothing
		// - if no, see if they had granted T and revoke that
		List<Integer> actors = getActors(pidGrantee);

		for (Integer pidActor : actors) {
			// Check if this actor granted T
			List<Integer> listOfGrantees = grantedTo(pidActor, tid);

			if (!listOfGrantees.isEmpty()) {
				// Check if actor still has authority for T (after we have
				// revoked this grant)
				if (!hasAuthority(new PrincipalID(pidActor), tag)) {
					// No longer has authority (must be due to this revocation)
					// - revoke all these grants to its grantees
					for (Integer granteeInt : listOfGrantees) {
						lastEid = deleteGrantTransitive(pidActor, granteeInt, tid, myEid);
					}
				}
			}
		}

		Serializable[] params = {revoker, grantor, grantee, tag};
		EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, 
				AeolusApiConstants.REVOKE_DELEGATE, params);
		authLog.push(myEvent);
		lastEID = myEid;
		
		return new AuthReply<Void>(myEid, null);
	}

	private EventID deleteActForLink(int pidFrom, int pidTo, EventID myEid) {
		int linkType = L_ACTSFOR;

		Connection conn = null;
		try {
			conn = getConnection();

			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {
				// Get coreId
				String coreQuery = "SELECT " + L_CoreIdFromAttribName
				+ " FROM " + L_RelationName + " WHERE "
				+ L_PidFromAttribName + "=" + pidFrom + " AND "
				+ L_PidToAttribName + "=" + pidTo + " AND "
				+ L_LinkTypeAttribName + "=" + linkType;
				int coreId = executeScalarInt(cmd, coreQuery);

				String deleteStmt = "DELETE FROM " + L_RelationName + " WHERE "
				+ L_PidFromAttribName + "=" + pidFrom + " AND "
				+ L_PidToAttribName + "=" + pidTo + " AND "
				+ L_LinkTypeAttribName + "=" + linkType;
				cmd.execute(deleteStmt);

				StreamMessageEntry e = StreamMessageParser
				.generateRevokeActsFor(pidFrom, pidTo);

				EventID lastEid = enableLogging ? getLastEid(cmd) : null;
				appendToRedoLog(cmd, e, coreId, myEid);

				// COMMIT
				conn.commit();
				return lastEid;
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "DeleteActForLink Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"DeleteActForLink Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}
		return null;
	}

	public AuthReply<Void> revokeActsFor(PrincipalID revoker, PrincipalID actor, PrincipalID actee, EventID predecessor)
	throws AuthorityException {
		int pidRevoker = revoker.getValue();
		int pidActor = actor.getValue();
		int pidActee = actee.getValue();
		
		// 1. Requestor must be authoritative for pidActee
		// - i.e. act-for
		if (!isActee(pidRevoker, pidActee)) {
			logger.log(LoggerLevel.DEBUG, "RevokeActsFor - pidRevoker "
					+ pidRevoker + " does not act-for pidActee " + pidActee);
			throw new AuthorityException();
		}

		// 2. Get the actors of pidActor (their delegation may be affected)
		List<Integer> listOfActors = getActors(pidActor);
		// also include the current actor
		listOfActors.add(pidActor);

		// 3. Compute Eff Authority of pidActor (Auth_before)
		// TODO

		// 4. Delete link
		
		EventID myEid = getNextEid(lastEID); // TODO: this is a bug! lastEID may not have been initialized!!
	
		EventID lastEid = deleteActForLink(pidActor, pidActee, myEid);
		
		Serializable[] params = {revoker, actor, actee};
		EventRecord myEvent = new EventRecord(myEid, predecessor, lastEid, 
				AeolusApiConstants.REVOKE_ACT_FOR, params);
		authLog.push(myEvent);
		lastEID = myEid;
		
		// 5. Compute Eff Authority of pidActor (Auth_after)
		// TODO

		// 6. Get the different between Auth_before and Auth_after (Auth_Diff)

		// 7. Reduce possible actors to check
		// QUERY: <listOfActorsToCheck, T> = listofActors AND grantor of T in
		// Auth_Diff
		// may need grantee also for RevokeGrantsTransitive

		// 8. For each of these returned tuples
		// Check whether actor still has auth for T
		// If not, RevokeGrantsTransitive
		return new AuthReply<Void>(myEid, null);
	}

	public SnapshotNumAndName createSnapshot(int snapshotVersion) {
		int seqNum = seqInvalid;
		String snapshotDbName = AuthorityConfiguration.DB_SnapshotNames + snapshotVersion;
		Connection conn = null;
		invalidsSinceLastSnapshot.clear();
		System.out.println("Creating Snapshot version "+snapshotVersion);

		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			// NOTE: "Create Database" not allowed to be executed in a
			// multi-statement txn
			try {
				
				// 1. Create the tables...
				deleteAuthorityServer(snapshotDbName, cmd);
				createTables(snapshotDbName, cmd);

				// 0. Lock Tables!
				// String lockSql = "LOCK TABLES ";
				// for(int i = 0; i < snapCopy.length; i++ ){
				// 	String s = snapCopy[i];
				// 	lockSql += s + snapshotDbName + " WRITE, ";
					
				// 	if(i < snapCopy.length - 1)
				// 		lockSql += s + " READ, ";
				// 	else
				// 		lockSql += s + " READ";
				// }
				// cmd.execute(lockSql);
				
				
				// 2. SELECT INTO!
				cmd.execute("begin");
				selectFromLive(snapshotDbName, cmd);
				cmd.execute("commit");

				// 4. Unlock tables!
				// lockSql = "UNLOCK TABLES";
				// cmd.execute(lockSql);
				
				// 3. Append to Redo Log
				StreamMessageEntry e = StreamMessageParser
				.generateCreateSnapshot(snapshotDbName);
				
				if(enableLogging)
					seqNum = appendToRedoLog(cmd, e, coreIdInvalid, getLastEid(cmd));
				else	
					seqNum = appendToRedoLog(cmd, e, coreIdInvalid, null);
				
			} catch (SQLException e) {
				this.error = e;
				logger.log(LoggerLevel.DEBUG, "CreateSnapshot Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"CreateSnapshot Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}

		return new SnapshotNumAndName(seqNum, snapshotDbName);
	}

	private void selectFromLive(String postfix, Statement cmd) throws SQLException {
		for (String tableName : snapCopy){
			String sql = "INSERT INTO " + tableName + postfix + " SELECT * FROM " + tableName;
			cmd.execute(sql);
		}
	}

	private void createTables(String postFix) {
		Connection conn = null;
		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			createTables(postFix, cmd);
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"CreateTables Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}

	}
	
	private void createTables(String postFix, Statement cmd) throws SQLException{
		if(postFix == null){
			postFix = "";

			createTable(cmd, NodeTable_Name + postFix,NodeTable_NodeID, NodeTable_NodeIDAttr, PINodeTable_Host, PINodeTable_HostAttr, PINodeTable_PI,
					PINodeTable_PIAttr);
			createTable(cmd, PlatformID_Name + postFix, PlatformID_PIID, PlatformID_PIIDAttr, PlatformID_Holder, PlatformID_HolderAttr);
			createTable(cmd, CI_RelationName + postFix, CI_CoreIdAttribName, CI_CoreIdAttribType,
					CI_RefCountAttribName, CI_RefCountAttribType);
			createTable(cmd, R_RelationName + postFix, R_SeqNumAttribName, R_SeqNumAttribType,
					R_CoreIdAttribName, R_CoreIdAttribType,
					R_RequestTypeAttribName, R_RequestTypeAttribType,
					R_MessageAttribName, R_MessageAttribType,
					R_EventIDName, R_EventIDType);
			createTable(cmd, S_RelationName + postFix, S_KeyAttribName, S_KeyAttribType,
					S_ValueAttribName, S_ValueAttribType);
		}
		createTable(cmd, T_RelationName + postFix, T_TidAttribName, T_TidAttribType,
				T_CreatorPidAttribName, T_CreatorPidAttribType,
				T_CoreIdAttribName, T_CoreIdAttribType);
		createTable(cmd, P_RelationName + postFix, P_PidAttribName, P_PidAttribType,
				P_CreatorPidAttribName, P_CreatorPidAttribType,
				P_CoreIdAttribName, P_CoreIdAttribType);
		createTable(cmd, G_RelationName + postFix, G_CoreIdAttribName, G_CoreIdAttribType,
				G_GrantorPidAttribName, G_GrantorPidAttribType,
				G_GranteePidAttribName, G_GrantorPidAttribType,
				G_TidAttribName, G_TidAttribType);
		createTable(cmd, L_RelationName + postFix, L_PidFromAttribName, L_PidFromAttribType,
				L_PidToAttribName, L_PidToAttribType, L_LinkTypeAttribName,
				L_LinkTypeAttribType, L_CoreIdFromAttribName,
				L_CoreIdFromAttribType, L_CoreIdToAttribName,
				L_CoreIdToAttribType);
		createTable(cmd, C_RelationName + postFix, C_CidAttribName, C_CidAttribType,
				C_PidClosureAttribName, C_PidClosureAttribType,
				C_ClassNameAttribName, C_KeyAttribType, C_CoreIdAttribName,
				C_CoreIdAttribType);
		createTable(cmd, Comp_RelationName + postFix, 
				Comp_ChildTidAttribName, Comp_ChildTidAttribType,
				Comp_ParTidAttribName, Comp_ParTidAttribType,
				Comp_CoreIdAttribName, Comp_CoreIdAttribType);		
	}
	
	public void removeSnapshot(String snapshotDbName) {
		Connection conn = null;

		try {
			conn = getConnection();
			Statement cmd = conn.createStatement();
			// TODO make this a transaction ?
			try {
				// drop the tables!
				for(String tableName : relNames){
					String dropStmt = "DROP TABLE IF EXISTS " + tableName + snapshotDbName;
					cmd.execute(dropStmt);
				}

				// Append to Redo Log
				StreamMessageEntry e = StreamMessageParser
				.generateRemoveSnapshot(snapshotDbName);
				if(enableLogging)
					appendToRedoLog(cmd, e, coreIdInvalid, getLastEid(cmd));
				else
					appendToRedoLog(cmd, e, coreIdInvalid, null);				
			} catch (SQLException e) {
				this.error = e;
				logger.log(LoggerLevel.DEBUG, "RemoveSnapshot Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"RemoveSnapshot Connection Exception [" + e.getMessage()
					+ "]");
		} finally {
			close(conn);
		}
	}

	public void removeOldSnapshots(){
		removeSnapshots(this.numSnapshotToBuffer);
	}

	public void removeAllSnapshots() {
		this.removeSnapshots(0);
	}

	private void removeSnapshots(int snapshots) {
		// Removes all but the top few in buffer
		Connection connRead = null;
		Connection conn = null;

		try {
			connRead = getConnection();
			conn = getConnection();

			Statement cmd = conn.createStatement();
			Statement cmdR = connRead.createStatement();
			try {
				String query = "SELECT " + R_SeqNumAttribName + ", "
				+ R_MessageAttribName + " FROM " + R_RelationName
				+ " WHERE " + R_RequestTypeAttribName + "="
				+ StreamMessageOpcode.S_SNAPSHOT_ADD.ordinal()
				+ " ORDER BY " + R_SeqNumAttribName + " DESC";

				ResultSet res = cmdR.executeQuery(query);

				int count = 0;
				int smallestSeqNumToKeep = -1;
				boolean moreSnapshotThanNeeded = false;

				while (res.next()) {
					// skip the recent ones
					if (count >= snapshots) {
						moreSnapshotThanNeeded = true;
						String createStreamMessage = res
						.getString(R_MessageAttribName);

						String snapshotName = StreamMessageParser
						.unpackCreateSnapshot(createStreamMessage);

						// Drop the database snapshot
						for(String tableName : snapCopy){
							String dropStmt = "DROP TABLE IF EXISTS " + tableName + snapshotName;
							cmd.execute(dropStmt);
						}

						// Append a 'remove snapshot' to the redo log
						StreamMessageEntry e = StreamMessageParser
						.generateRemoveSnapshot(snapshotName);

						if(enableLogging)
							appendToRedoLog(cmd, e, coreIdInvalid, getLastEid(cmd));
						else
							appendToRedoLog(cmd, e, coreIdInvalid, null);							
					} else {
						smallestSeqNumToKeep = res.getInt(R_SeqNumAttribName);
					}
					count++;
				}

				if (moreSnapshotThanNeeded) {
					// Truncate the log
					String truncLogStmt = "DELETE FROM " + R_RelationName
					+ " WHERE " + R_SeqNumAttribName + "<"
					+ smallestSeqNumToKeep;
					cmd.execute(truncLogStmt);
				}
				res.close();
			} catch (SQLException e) {
				this.error = e;
				logger.log(LoggerLevel.DEBUG, "RemoveOldSnapshots Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG,
					"RemoveOldSnapshots Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
			close(connRead);
		}
	}

	public AuthReply<Void> closureActsForGI(PrincipalID pidRequestor,
			ClosureID cid, PrincipalID pidActee, int seqNumFirstEntry, EventID predecessor) throws RemoteException,
			AuthorityException {
		AuthReply<Void> r = closureActsFor(pidRequestor, cid, pidActee, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<Void> closureGrantGI(PrincipalID pidRequestor,
			ClosureID cid, PrincipalID pidGrantor, TagID tid, int seqNumFirstEntry, EventID predecessor)
			throws RemoteException, AuthorityException {
		AuthReply<Void> r = closureGrant(pidRequestor, cid, pidGrantor, tid, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<ClosureReply> createClosureGI(
			String className, int seqNumFirstEntry, EventID predecessor) throws RemoteException {
		AuthReply<ClosureReply> r = createClosure(className, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<PrincipalReply> createPrincipalGI(
			PrincipalID pidCreator, int seqNumFirstEntry, EventID predecessor) throws RemoteException, AuthorityException {
		AuthReply<PrincipalReply> r = createPrincipal(pidCreator, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<PrincipalReply> createPrincipalInNewCoreGI(
			PrincipalID pidCreator, int seqNumFirstEntry, EventID predecessor) throws RemoteException, AuthorityException {
		AuthReply<PrincipalReply> r = createPrincipalInNewCore(pidCreator, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<TagID> createTaintGI(PrincipalID pidCreator, int seqNumFirstEntry, EventID predecessor)
	throws RemoteException, AuthorityException {
		AuthReply<TagID> r = createTag(pidCreator, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}
	public AuthReply<TagID> createSubTagGI(TagID parent, PrincipalID creator, int seqNumFirstEntry, EventID predecessor) 
	throws AuthorityException {
		AuthReply<TagID> r = createSubTag(parent, creator, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}
	
	public AuthReply<Void> delegateActsForGI(PrincipalID pidRequestor,
			PrincipalID pidFrom, PrincipalID pidTo, int seqNumFirstEntry, EventID predecessor) throws RemoteException,
			AuthorityException {
		AuthReply<Void> r =delegateActsFor(pidRequestor, pidFrom, pidTo, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<Void> grantGI(PrincipalID pidRequestor,
			PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, int seqNumFirstEntry, EventID predecessor)
			throws RemoteException, AuthorityException {
		AuthReply<Void> r =grant(pidRequestor, pidGrantor, pidGrantee, tid, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<Void> revokeActsForGI(PrincipalID pidRevoker,
			PrincipalID pidActor, PrincipalID pidActee, int seqNumFirstEntry, EventID predecessor) throws RemoteException,
			AuthorityException {
		AuthReply<Void> r =revokeActsFor(pidRevoker, pidActor, pidActee, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	public AuthReply<Void> revokeGrantGI(PrincipalID pidRevoker,
			PrincipalID pidGrantor, PrincipalID pidGrantee, TagID tid, int seqNumFirstEntry, EventID predecessor)
			throws RemoteException, AuthorityException {
		AuthReply<Void> r =revokeGrant(pidRevoker, pidGrantor, pidGrantee, tid, predecessor);
		r.piggyback = getInvalidationPacket(seqNumFirstEntry);
		return r;
	}

	private static class InvalidationsCache{
		private int firstSeqNum = 0; // Also the first "ADD_SNAPSHOT"
		private int currentSnapshotIndex = -1;
		private ArrayList<RedoLogEntry> cache = new ArrayList<RedoLogEntry>();
		private EventID currentEvent = null;
		
		void addInvalidation(RedoLogEntry e, EventID entryEvent){
			if(e.seqNum == 0)
				return;
			if(firstSeqNum == 0 && e.seqNum == 1)
				firstSeqNum = 1;
			if(e.seqNum != firstSeqNum + cache.size())
				return;

			if(e.requestType == StreamMessageOpcode.S_SNAPSHOT_ADD){
				if(currentSnapshotIndex != -1){
					// Drop the backend of the cache out
					for(int i = 0; i < currentSnapshotIndex; i++)
						cache.remove(0);
					firstSeqNum += currentSnapshotIndex;
				}
				cache.add(e);
				currentSnapshotIndex = cache.size() - 1;
			}else{
				cache.add(e);
			}
			currentEvent = entryEvent;
		}

		InvalidationPacket getInvalidations(int fromSeqNum){
			if(fromSeqNum < firstSeqNum)
				return null;
			int startIndex = fromSeqNum - firstSeqNum;
			return new InvalidationPacket(new ArrayList<RedoLogEntry>(cache.subList(startIndex, cache.size())), currentEvent);
		}
	}

	public void deregisterClient(String clientRMIString) {
		try {
			Connection conn = getConnection();
			
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);
			try {
				String sql = "SELECT " + NodeTable_NodeID + " FROM " + NodeTable_Name + " WHERE " +
						PINodeTable_Host + " = '" + clientRMIString +"'";
				Integer nodeID = executeScalarInt(cmd, sql);
				sql = "DELETE FROM " + NodeTable_Name + " WHERE " 
				+ NodeTable_NodeID + " = " + nodeID;
				cmd.executeUpdate(sql);
				
				// Log the deregistration...
				StreamMessageEntry e = StreamMessageParser
					.generateRemoveNode(nodeID);
				
				if(enableLogging)
					appendToRedoLog(cmd, e, coreIdInvalid, getLastEid(cmd));
				else
					appendToRedoLog(cmd, e, coreIdInvalid, null);
				
				conn.commit();
			}catch(SQLException e){
				if(e.getMessage().startsWith("No results found for scalar query:")){
					System.out.println("Attempted to remove non-registered client: " + clientRMIString);
				}
				e.printStackTrace();
				logger.log("SQLException in getNextPIID");
				conn.rollback();
				throw new RuntimeException();
			}

		} catch (SQLException e) {
			e.printStackTrace();
			throw new RuntimeException();
		}
	}
	
	public RegisterReturn registerClient(String clientRMIString) {
		try {
			Connection conn = getConnection();
			
			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);
			try {
				String sql = "SELECT " + NodeTable_NodeID + ", " + PINodeTable_PI + " FROM " + NodeTable_Name + " WHERE " +
						PINodeTable_Host + " = '" + clientRMIString +"'";
				boolean isRegisteredBefore = true;
				SystemNodeID nodeId;
				PrincipalID nodeRoot; 
				ResultSet rs = cmd.executeQuery(sql);			
				if(rs.next()){
					nodeId = new SystemNodeID(rs.getInt(NodeTable_NodeID));
					nodeRoot = new PrincipalID(rs.getInt(PINodeTable_PI));
				}else{
					isRegisteredBefore = false;
					AuthReply<PrincipalReply> rep = createPrincipal(pidSystem, lastEID);
					lastEID = rep.event;
					nodeRoot = rep.value.principal;
					sql = "INSERT INTO " + NodeTable_Name+ " (hostname, PrincipalID) VALUES ('" + clientRMIString + "', " + nodeRoot +")";
					cmd.execute(sql);
					nodeId = new SystemNodeID(executeScalarInt2(cmd,NodeTable_Name,NodeTable_NodeID));
					
					// Log the registration...
					StreamMessageEntry e = StreamMessageParser
						.generateRegisterNode(nodeId.getValue(), clientRMIString);
					
					if(enableLogging)
						appendToRedoLog(cmd, e, coreIdInvalid, getLastEid(cmd));
					else
						appendToRedoLog(cmd, e, coreIdInvalid, null);
				}

				conn.commit();
				return new RegisterReturn(nodeId, nodeRoot, isRegisteredBefore);
			}catch(SQLException e){
				e.printStackTrace();
				logger.log("SQLException in getNextPIID");
				conn.rollback();
				throw new RuntimeException();
			}

//			
//			return createPrincipal(new PrincipalID(0)).value.principal;
		} catch (AuthorityException e) {
			throw new RuntimeException("This should be impossible.");
		} catch (SQLException e) {
			e.printStackTrace();
			throw new RuntimeException();
		}
	}
	
	public Integer getNextPlatformInstanceID(){
		Connection conn = null;
		try {
			conn = getConnection();

			Statement cmd = conn.createStatement();
			conn.setAutoCommit(false);

			try {
				String sql = "INSERT INTO " + PlatformID_Name +
				"( " + PlatformID_Holder + ") " +
				" VALUES (0)";
				cmd.execute(sql);

				int PIID = executeScalarInt2(cmd, PlatformID_Name, PlatformID_PIID);
				conn.commit();
				return PIID;
			}catch(SQLException e){
				e.printStackTrace();
				logger.log("SQLException in getNextPIID");
				conn.rollback();
				return -1;
			}
		} catch (SQLException e){
			logger.log("Connection Exception in getNextPIID");
			return -1;
		}
	}

	public void destroy() {
		this.authLog.destroy();
	}

	public ClosureTableEntry getClosureInfo(ClosureID closure)
			throws RemoteException {
		Connection conn = null;
		try {
			conn = getConnection();

			String selectStmt =
			"SELECT " + C_PidClosureAttribName + ", " + C_ClassNameAttribName + ", " + 
					C_CoreIdAttribName + " FROM " + C_RelationName +
			" WHERE " + C_CidAttribName + " = " + closure.getValue();
			
			Statement cmd = conn.createStatement();
			try {
				ResultSet rs = cmd.executeQuery(selectStmt);
				while(rs.next()){
					int pidClosure = rs.getInt(1);
					String className = rs.getString(2);
					int coreId = rs.getInt(3);
					ClosureTableEntry cte = 
						new ClosureTableEntry(closure.getValue(), pidClosure, className, coreId);								
					return cte;
				}
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				logger.log(LoggerLevel.DEBUG, "CreateTaint Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "CreateTaint Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
		return null;		
	}

	public Pair<Boolean, List<Integer>> isActeeBasis(int pidActor, int pidActee)
			throws RemoteException {
		boolean answer = false;
		LinkedList<Integer> basis = new LinkedList<Integer>();
		
		Connection conn = null;

		try {
			conn = getConnection();
			answer = searchActeeRecursive(pidActee, pidActor, conn, basis);
			return new Pair<Boolean, List<Integer>>(answer, basis);
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "IsActee Connection Exception ["
					+ e.getMessage() + "]");
			return new Pair<Boolean, List<Integer>>(false, null);
		} finally {
			close(conn);
		}	
	}
	
	public boolean isSystemNodeRegistered(int nodeID) throws RemoteException {
		Connection conn = null;
		try {
			conn = getConnection();

			String selectStmt =
			"SELECT * FROM " + NodeTable_Name +
			" WHERE " + NodeTable_NodeID + " = " + nodeID;
			
			Statement cmd = conn.createStatement();
			try {
				ResultSet rs = cmd.executeQuery(selectStmt);
				while(rs.next()){
					return true;
				}
			} catch (SQLException e) {
				this.error = e;
				// ROLLBACK
				conn.rollback();
				logger.log(LoggerLevel.DEBUG, "CreateTaint Exception ["
						+ e.getMessage() + "]");
			}
		} catch (SQLException e) {
			this.error = e;
			logger.log(LoggerLevel.DEBUG, "CreateTaint Connection Exception ["
					+ e.getMessage() + "]");
		} finally {
			close(conn);
		}
		return false;
	}
}
