package edu.mit.csail.aeolus.authority.server;

import java.rmi.server.ExportException;

import junit.framework.Assert;

import org.junit.Before;
import org.junit.Test;

import edu.mit.csail.aeolus.authority.api.AuthReply;
import edu.mit.csail.aeolus.authority.client.AuthorityClient;
import edu.mit.csail.aeolus.authority.client.api.EvictionMode;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.stream.server.StreamServer;

public class AuthorityManagerPerfTests {

	AuthorityManager mgr;
	StreamServer ss;
	AuthorityClient ac;
	
	@Before
	public void startRMI() throws Exception{
		try{
			java.rmi.registry.LocateRegistry.createRegistry(1099);
		}catch(ExportException e){
		}
	}	
	
	@Test
	public void testMissRate() throws Exception{
		this.mgr = new AuthorityManager();
		this.mgr.removeAllSnapshots();
		this.mgr.createAuthorityServer();
		mgr.startServer();

		ss = new StreamServer();
		ss.initialize();

		ac = new AuthorityClient(100, EvictionMode.LRU, false);
		
		PrincipalID root = ac.register().rootPID;
		
		int TEST_SIZE = 1000;
		int [] MISSES = {0, 0, (int) (.25 * TEST_SIZE), (int) (.5 * TEST_SIZE), (int) (.75 * TEST_SIZE), TEST_SIZE};
		for(int FORCED_MISSES : MISSES){
			// Create our required information for queries...
			// The `` hit ''
			PrincipalID hit = ac.createPrincipalInNewCore(root, null).value;
			TagID hit_t = ac.createTaint(root, null).value;
			ac.grant(root, root, hit, hit_t, null);
			
			// The misses
			int pidOff = 0;
			int tidOff = 0;
			for(int i = 0; i < FORCED_MISSES; i++){
				PrincipalID pid = ac.createPrincipalInNewCore(root, null).value;
				TagID t = ac.createTaint(root, null).value;
				ac.grant(root, root, pid, t, null);
				if(i == 0){
					pidOff = pid.getValue();
					tidOff = t.getValue();
				}
			}
			
			long s = System.currentTimeMillis();
			ac.hasAuthority(hit, hit_t);
			// Now for the hit tests.
			for(int i = 0; i < TEST_SIZE - FORCED_MISSES; i++)
				ac.hasAuthority(hit, hit_t);
			for(int i = 0; i < FORCED_MISSES; i++){
				ac.hasAuthority(new PrincipalID(pidOff + i), new TagID(tidOff + i));
			}
			long e = System.currentTimeMillis();
			
			System.out.println("FORCED MISSES = " + FORCED_MISSES + "\t TIME = " + (e - s));
		}
	}
	
	@Test
	public void testProcessingInvalidations() throws Exception{
		int TEST_SIZE = 100;

		this.mgr = new AuthorityManager();
		this.mgr.removeAllSnapshots();
		this.mgr.createAuthorityServer();
		mgr.startServer();

		ss = new StreamServer();
		ss.initialize();

		ac = new AuthorityClient(100, EvictionMode.LRU, false);
		PrincipalID root = ac.register().rootPID;

		ss.streamNow();
		for(int STREAM_SIZE = 1; STREAM_SIZE < 20; STREAM_SIZE++){
			for(int j = 0; j < TEST_SIZE; j++){
				for(int i = 0; i < STREAM_SIZE; i ++){
					ac.createPrincipalInNewCore(root, null);
				}
				ss.streamNow();
			}
			System.out.println("STREAM SIZE: " + STREAM_SIZE + " \t TESTS:" + TEST_SIZE);
			ac.printStats();
			ac.resetStats();
		}
	}

	@Test
	public void testEvictionCost() throws Exception{
		this.mgr = new AuthorityManager();
		this.mgr.removeAllSnapshots();
		this.mgr.createAuthorityServer();
		mgr.startServer();

		ss = new StreamServer();
		ss.initialize();

		int TEST_SIZE = 2000;
		int HALF_SIZE = TEST_SIZE / 2;
		
		ac = new AuthorityClient(HALF_SIZE, EvictionMode.LRU);
		PrincipalID root = ac.register().rootPID;

		int pidOffset = root.getValue() + 1;
		int tagOffset = ac.createTaint(root, null).value.getValue() + 1;
		for(int i = 0; i < TEST_SIZE; i++){
			if(i % 10 == 0)
				System.out.print(".");
			AuthReply<PrincipalID> repl = ac.createPrincipalInNewCore(root, null);
			TagID t = ac.createTaint(root, null).value;
			ac.grant(root, root, repl.value, t, null);
			Assert.assertEquals(t.getValue(), tagOffset + i);
			Assert.assertEquals(repl.value.getValue(), pidOffset + i);
		}
		System.out.println("!");
		// Now for the actual queries...
		
		long st_1 = System.currentTimeMillis();
		for(int i = 0; i < HALF_SIZE; i++){
			ac.hasAuthority(new PrincipalID(pidOffset + i), new TagID(tagOffset + i));
		}
		long en_1 = System.currentTimeMillis();
		
		long st_2 = System.currentTimeMillis();
		for(int i = HALF_SIZE; i < TEST_SIZE; i++){
			ac.hasAuthority(new PrincipalID(pidOffset + i), new TagID(tagOffset + i));
		}
		long en_2 = System.currentTimeMillis();
		
		System.out.println("" + (en_1-st_1));
		System.out.println("" + (en_2-st_2));
		
		ac.printStats();
		
	}
	
	@Test
	public void testHasAuthWarm() throws Exception{

		this.mgr = new AuthorityManager();
		this.mgr.removeAllSnapshots();
		this.mgr.createAuthorityServer();
		mgr.startServer();

		ss = new StreamServer();
		ss.initialize();

		ac = new AuthorityClient(100, EvictionMode.LRU);
		PrincipalID root = ac.register().rootPID;

		TagID tag = ac.createTaint(root, null).value;
		
		for(int depth = 1; depth <= 20; depth++){

			PrincipalID query = ac.createPrincipal(root, null).value;
			ac.grant(root, root, query, tag, null);
			for(int i = 1; i < depth; i++){
				PrincipalID newP = ac.createPrincipal(root, null).value;
				ac.delegateActsFor(root, newP, query, null);
				query = newP;
			}

			Assert.assertTrue(ac.hasAuthority(root, tag).value);
			Assert.assertTrue(ac.isActee(root, query).value);
			// Warmup the cache...
			Assert.assertTrue(ac.hasAuthority(query, tag).value);


			// Bang out some queries...
			long s = System.currentTimeMillis();
			for(int i = 0; i < 100; i++){
				ac.hasAuthority(query, tag);
			}
			long e = System.currentTimeMillis();
			System.out.println("100 Queries at depth " + depth + ": \t" + (e-s));
			try{
				Thread.sleep(500);
			}catch(Exception ex){
				
			}
		}
	}
}
