package edu.mit.csail.aeolus.filesystem;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * LRU cache of most recently accessed files on a client machine 
 * Stores entire file content
 * 
 * @author viq
 *
 */
public class FileCache {
	
	/* Map file path to file content and info */
	private Map<String, FileCacheEntry> cache;
	private int cacheSize;
	
	/* Cache statistics */
	private int numCacheLookups;
	private int numCacheEvictions;
	private int numCacheMisses;
	
	public FileCache(int cacheSize)
	{
		this.cacheSize = cacheSize;
		
		LinkedHashMap<String, FileCacheEntry> lruMap = 
			new LinkedHashMap<String, FileCacheEntry>(this.cacheSize, 0.75f, true)
			{
				private static final long serialVersionUID = -3999913693934781921L;
				// this method is invoked by the put operation - 
				// allows to automatically remove the least recently used entry from the map
				// when a new entry is added
				@Override protected boolean removeEldestEntry(Map.Entry<String, FileCacheEntry> eldest) 
				{
					FileCache.this.numCacheEvictions++;
					return this.size() > FileCache.this.cacheSize;
				}
			};
		
		// the cache is accessed concurrently by user threads - need to synchronize access
	    // note: a get() call structurally modifies the LRU map
		this.cache = Collections.synchronizedMap(lruMap);
	}

	public FileCacheEntry getFile(String filePath)
	{
		this.numCacheLookups++;
		
		if(!this.cache.containsKey(filePath))
			this.numCacheMisses++;

		return cache.get(filePath);
	}
	
	public void putFile(String filePath, FileCacheEntry file)
	{
		this.cache.put(filePath, file);
	}
	
	/**
	 * Removes the file from the cache
	 * Subsequent access to this file needs to retrieve the data from the server
	 */
	public void invalidateFile(String fileName)
	{
		this.cache.remove(fileName);
	}
	
	public void clearCache()
	{
		this.cache.clear();
	}
	
	public int getCacheStats()
	{
		return this.numCacheMisses;
	}
}
