package edu.mit.csail.aeolus.filesystem;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.EOFException;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.HashMap;

import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.infoflow.InfoFlowRules;
import edu.mit.csail.aeolus.infoflow.Label;

/**
 * Aeolus file system operations
 * 
 * Aeolus file label attributes for each file are stored in a separate hidden
 * meta-data file inside the same directory as the data file All operations
 * first access the attributes file and then the data file
 * 
 * @author viq
 * 
 */
public class FileSystem {

	private HashMap<String, FileAttributes> fileNames;
	private String rootPath;

	public FileSystem() {
		initialize();
	}

	public void initialize() {
		fileNames = new HashMap<String, FileAttributes>();
	}

	protected void setupFSRoot(String rootPath) {
		this.rootPath = rootPath;
		File aeolusRootDir = new File(rootPath);
		File aeolusRootParentDir = aeolusRootDir.getParentFile();
		fileNames.put(rootPath, new FileAttributes(new Label(), new Label(),
				rootPath, rootPath));

		aeolusRootParentDir.mkdir();
		try {
			storeAttributes(rootPath, aeolusRootDir, new FileAttributes());
		} catch (IOException e) {
			e.printStackTrace();
		}
		if (!aeolusRootDir.mkdir()) {
			try {
				crawlFS(aeolusRootDir);
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	/**
	 * Returns the attributes associated with the provided file Note: In Java 7,
	 * this function should extract attributes from the data file's meta-data
	 */
	public void parseAttrFile(File f) throws FileNotFoundException, IOException {

		// will throw I/O exception if file does not exist or cannot be read
		FileInputStream fis;
		ObjectInputStream ois;
		FileAttributes attrs = new FileAttributes();
		try {
			try {
				fis = new FileInputStream(f);
				ois = new ObjectInputStream(fis);
				while (attrs != null) {
					try {
						attrs = (FileAttributes) ois.readObject();
						fileNames.put(attrs.userName, attrs);
					} catch (EOFException eof) {
						break;
					} catch (ClassNotFoundException e) {
						e.printStackTrace();
					}
				}
				ois.close();
			} catch (IOException e) {
				// we can't forward the exception directly to the user
				// since it deals with the hidden file, instead create a general
				// I/O exceptions
				throw new IOException("Could not access the file");
			}

		} catch (FileNotFoundException e) {
			throw new FileNotFoundException("Could not find file");
		}

	}

	/**
	 * Stores the labels for a new file/directory in a separate hidden file with
	 * the same name ('.' + name - only works on Linux) In Java 7, this function
	 * should place the labels into the file's meta-data
	 * 
	 * @throws IOException
	 */
	public boolean storeAttributes(String name, File f, FileAttributes attrs)
			throws IOException {
		File attrFile = new File(getHiddenName(f));
		try {
			boolean exists = attrFile.exists();
			if (!exists)
				if (!attrFile.createNewFile())
					return false;

			FileOutputStream fos = new FileOutputStream(attrFile, true);
			ObjectOutputStream oos = null;
			if (exists)
				oos = new ObjectOutputStream(fos) {
					@Override
					protected void writeStreamHeader() throws IOException {
						// do not write a header
					}
				};
			else
				oos = new ObjectOutputStream(fos);
			oos.reset();
			oos.writeObject(attrs);
			oos.flush();
			oos.close();
			return true;
		} catch (IOException e) {
			e.printStackTrace();
			// we can't forward the exception directly to the user
			// since it deals with the hidden file, instead create a general I/O
			// exception
			throw new IOException("Could not create file " + name);
		}
	}

	/**
	 * Manually generates a hidden filename for the given file ('.' + fileName -
	 * only works on Linux)
	 */
	private String getHiddenName(File f) {
		return f.getParent() + "/.attrs";
	}

	// FILE

	/**
	 * Creates a new file if the parent directory exists and there is no file
	 * with the given name Requires read and write permission of the parent
	 * directory Returns true if file was successfully created and false if file
	 * with same name existed
	 */
	public FileSystemReply<Boolean> createFile(String filePath,
			Label sLabelFile, Label iLabelFile, Label sLabelRequestor,
			Label iLabelRequestor) throws IOException, InfoFlowControlException {
		if (fileNames.containsKey(filePath))
			return new FileSystemReply<Boolean>(false);

		File fileToCreate = new File(filePath);

		FileAttributes parentAttr = fileNames.get(fileToCreate.getParent());

		if (parentAttr == null)
			throw new FileNotFoundException(
					"Cannot create file. Invalid file pathname: "
							+ fileToCreate.getParent());

		String parentName = parentAttr.uniqueName;

		if (parentName.equals(rootPath)) {
			if (!new Label(sLabelRequestor).isEmpty())
				throw new InfoFlowControlException(
						"Writing to the root directory requires null secrecy label");

			if (!new Label(iLabelFile).isSubsetOf(new Label(iLabelRequestor)))
				throw new InfoFlowControlException(
						"Cannot create file with integrity = {"
								+ iLabelFile
								+ "} in root directory with process integrity = {"
								+ iLabelRequestor + "}");
		} else {
			if (!InfoFlowRules.canReadWrite(new Label(sLabelRequestor),
					new Label(iLabelRequestor), parentAttr.sLabel,
					parentAttr.iLabel)) {
				throw new InfoFlowControlException(
						"Cannot create file "
								+ filePath
								+ " : not authorized to read and write the parent directory "
								+ parentAttr.userName + "; process labels = {"
								+ sLabelRequestor + ", " + iLabelRequestor
								+ "}," + " file labels = {"
								+ parentAttr.sLabel.toString() + ", "
								+ parentAttr.iLabel.toString() + "}");
			}

			if (!(parentAttr.sLabel.isSubsetOf(sLabelFile) && iLabelFile
					.isSubsetOf(parentAttr.iLabel))) {
				throw new InfoFlowControlException(
						"Cannot create file with labels = {" + sLabelFile
								+ ", " + iLabelFile + "},"
								+ " in directory with labels = {"
								+ parentAttr.sLabel.toString() + ", "
								+ parentAttr.iLabel.toString() + "}");
			}
		}

		String uniqueName = parentName + "/" + Long.toString(System.nanoTime());
		fileToCreate = new File(uniqueName);

		FileAttributes fAttrs = new FileAttributes(sLabelFile, iLabelFile,
				filePath, uniqueName);
		// TODO: fix this (in case of crash attributes could exist although the
		// file doesn't)
		// store the attributes (false - if attr file already exists -> file
		// already exists)
		if (!storeAttributes(filePath, fileToCreate, fAttrs)) {
			return new FileSystemReply<Boolean>(false);
		}
		fileNames.put(filePath, fAttrs);

		// atomic
		if (!fileToCreate.createNewFile()) {
			return new FileSystemReply<Boolean>(false);
		}
		fileToCreate.setReadable(false, false);
		fileToCreate.setWritable(false, false);

		fileToCreate.setReadable(true, true);
		fileToCreate.setWritable(true, true);

		return new FileSystemReply<Boolean>(true);
	}

	// DIR

	public FileSystemReply<Boolean> createDir(String filePath,
			Label sLabelFile, Label iLabelFile, Label sLabelRequestor,
			Label iLabelRequestor) throws InfoFlowControlException, IOException {
		if (fileNames.containsKey(filePath))
			return new FileSystemReply<Boolean>(false);
		File fileToCreate = new File(filePath);
		FileAttributes parentAttr = fileNames.get(fileToCreate.getParent());

		if (parentAttr == null)
			throw new FileNotFoundException(
					"Cannot create file. Invalid file pathname: "
							+ fileToCreate.getParent());

		if (parentAttr.uniqueName.equals(rootPath)) {
			if (!sLabelRequestor.isEmpty())
				throw new InfoFlowControlException(
						"Writing to the root directory requires null secrecy label");

			if (!iLabelFile.isSubsetOf(iLabelRequestor))
				throw new InfoFlowControlException(
						"Cannot create file with integrity = {"
								+ iLabelFile
								+ "} in root directory with process integrity = {"
								+ iLabelRequestor + "}");
		} else {
			if (!InfoFlowRules.canReadWrite(sLabelRequestor, iLabelRequestor,
					parentAttr.sLabel, parentAttr.iLabel)) {
				throw new InfoFlowControlException(
						"Cannot create file "
								+ filePath
								+ " : not authorized to read and write the parent directory "
								+ parentAttr.userName + "; process labels = {"
								+ sLabelRequestor + ", " + iLabelRequestor
								+ "}," + " file labels = {"
								+ parentAttr.sLabel.toString() + ", "
								+ parentAttr.iLabel.toString() + "}");
			}

			if (!(parentAttr.sLabel.isSubsetOf(sLabelFile) && iLabelFile
					.isSubsetOf(parentAttr.iLabel))) {
				throw new InfoFlowControlException(
						"Cannot create file with labels = {" + sLabelFile
								+ ", " + iLabelFile + "},"
								+ " in directory with labels = {"
								+ parentAttr.sLabel.toString() + ", "
								+ parentAttr.iLabel.toString() + "}");
			}
		}

		String uniqueName = parentAttr.uniqueName + "/"
				+ Long.toString(System.nanoTime());
		fileToCreate = new File(uniqueName);
		FileAttributes fAttrs = new FileAttributes(sLabelFile, iLabelFile,
				filePath, uniqueName);
		// TODO: fix this (in case of crash attributes could exist although the
		// file doesn't)
		// store the attributes (false - if attr file already exists -> file
		// already exists)
		if (!storeAttributes(filePath, fileToCreate, fAttrs)) {
			return new FileSystemReply<Boolean>(false);
		}
		fileNames.put(filePath, fAttrs);

		// atomic
        if(!fileToCreate.mkdir())
        {
        	return new FileSystemReply<Boolean>(false);
        }
        
        // Not sure if this is what we wanna do to prevent an exception when listing an empty directory
        // hence I commented out the following line.
//        storeAttributes(filePath, fileToCreate, new FileAttributes());
		        
    	fileToCreate.setReadable(false, false);
    	fileToCreate.setWritable(false, false);
    	
    	fileToCreate.setReadable(true, true);
    	fileToCreate.setWritable(true, true);
        return new FileSystemReply<Boolean> (true);
	}

	/**
	 * Deletes a directory - directory must be empty Returns true if directory
	 * was successfully deleted
	 * 
	 * @throws InfoFlowControlException
	 */
	public FileSystemReply<Boolean> delete(String filePath,
			Label sLabelRequestor, Label iLabelRequestor)
			throws InfoFlowControlException, IOException {

		FileAttributes fileAttr = fileNames.get(filePath);

		if (fileAttr != null) {
			// get the labels of the parent directory (throws IOError no such
			// file / cannot read file)
			FileAttributes parentAttr = fileNames.get(new File(filePath)
					.getParent());
			File fileToRemove = new File(fileAttr.uniqueName);

			if (parentAttr.uniqueName.equals(rootPath)) {
				if (!new Label(sLabelRequestor).isEmpty())
					throw new InfoFlowControlException(
							"Writing to the root directory requires null secrecy label");

				if (!fileAttr.iLabel.isSubsetOf(new Label(iLabelRequestor)))
					throw new InfoFlowControlException(
							"Cannot delete file with integrity = {"
									+ fileAttr.iLabel
									+ "} in root directory with process integrity = {"
									+ iLabelRequestor + "}");
			} else {
				if (!InfoFlowRules.canReadWrite(new Label(sLabelRequestor),
						new Label(iLabelRequestor), parentAttr.sLabel,
						parentAttr.iLabel))
					throw new InfoFlowControlException(
							"Not authorized to read and write the parent directory");
			}
			// delete labels file and data file
			String[] children = fileToRemove.list();

			if (children != null && children.length == 1) {
				new File(fileToRemove.getPath() + "/.attrs").delete();
			} else if (children != null && children.length > 1) {
				return new FileSystemReply<Boolean>(false);
			}

			if (fileToRemove.delete()) {
				// this should include the case if the file doesn't exist
				fileNames.remove(filePath);
				rewriteAttrsFile(fileAttr.uniqueName);
				return new FileSystemReply<Boolean>(true);
			}
			return new FileSystemReply<Boolean>(false);
		} else {
			throw new FileNotFoundException(
					"Cannot remove file - parent directory does not exist");
		}
	}

	private void rewriteAttrsFile(String fileToRemove) {
		File attrs = new File(getHiddenName(new File(fileToRemove)));
		ArrayList<String> entries = new ArrayList<String>();
		fileToRemove = fileToRemove.replace(rootPath, "");
		BufferedReader reader;
		BufferedWriter writer;
		String entry;
		try {
			reader = new BufferedReader(new FileReader(attrs));
			entry = reader.readLine();
			while (entry != null) {
				if (!entry.contains(fileToRemove))
					entries.add(entry);
				entry = reader.readLine();
			}
			reader.close();

			writer = new BufferedWriter(new FileWriter(attrs, false));
			for (String e : entries) {
				writer.write(e + "\n");
			}
			writer.flush();
			writer.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void crawlFS(File dir) throws FileNotFoundException, IOException {
		File[] files = dir.listFiles(new FileFilter() {
			public boolean accept(File f) {
				return (f.isDirectory() || f.getName().equals(".attrs"));
			}
		});

		for (File f : files) {
			if (f.isDirectory())
				crawlFS(f);
			else
				parseAttrFile(f);
		}
	}

}
