package edu.mit.csail.aeolus.filesystem;

import java.io.BufferedReader;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.HashMap;

import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.infoflow.InfoFlowRules;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;
import edu.mit.csail.aeolus.utilities.RMIUtil;
import edu.mit.csail.aeolus.utilities.Utils;

/**
 * Runs on a compute node acting as a proxy for an NFS mounted, remote
 * filesystem
 * 
 * @author pmckee
 * 
 */
public class FileSystemClient implements FileSystemClientInterface {

	private FileSystemRemoteInterface fileServer;
	private String mountLocation;
	private String clientServiceName;
	private boolean isAlive;
	private Registry registry;
	private HashMap<String, FileAttributes> remoteFileAttrs;
	private int handleCounter = 1;
	private HashMap<Integer, FileInputStream> openInputStreams;
	private HashMap<Integer, FileOutputStream> openOutputStreams;

	/* Initialization */
	public FileSystemClient(String fileServerHostname) throws AeolusException {
		this.clientServiceName = fileServerHostname; // only one client per file
														// server per node
		// this lock allows file invalidations to block all other requests

		this.isAlive = false;

		RuntimeConfig c = RuntimeConfig.get();

		// export this client
		try {
			FileSystemClientInterface client = (FileSystemClientInterface) UnicastRemoteObject
					.exportObject(this, 0);
			registry = RMIUtil.getOrCreate(c.getFsClientPort());
			registry.rebind(this.clientServiceName, client);

		} catch (RemoteException e) {
			throw new AeolusException("Could not setup file system client", e);
		}

		// get server stub
		try {
			Registry registry = LocateRegistry.getRegistry(fileServerHostname,
					c.getFsServicePort());
			fileServer = (FileSystemRemoteInterface) registry.lookup(c
					.getFsServiceName());
		} catch (Exception e) {
			throw new AeolusException("Could not find the file system server",
					e);
		}
		mountLocation = c.getFsClientMountLocation() + "/" + fileServerHostname;
		remoteFileAttrs = new HashMap<String, FileAttributes>();

		remoteFileAttrs.put(mountLocation, new FileAttributes(new Label(),
				new Label(), mountLocation, mountLocation));
		remoteFileAttrs.put(mountLocation + "/", new FileAttributes(new Label(), new Label(), mountLocation + "/", mountLocation));

		openInputStreams = new HashMap<Integer, FileInputStream>();
		openOutputStreams = new HashMap<Integer, FileOutputStream>();
		System.out.println("File Client Running");

	}

	public void shutdown() {
		if (registry != null) {
			try {
				UnicastRemoteObject.unexportObject(registry, true);
			} catch (NoSuchObjectException e) {
				// e.printStackTrace();
			}
		}
	}

	public boolean isAlive() throws RemoteException {
		return this.isAlive;
	}

	public void register() throws AeolusException {
		try {
			this.fileServer.registerFSClient(this.clientServiceName);
			this.isAlive = true;
		} catch (RemoteException e) {
			e.printStackTrace();
			throw new AeolusException(e);
		}
	}

	/* File Operations */

	private void parseMappingsFile(String uniqueDirectoryName)
			throws FileNotFoundException, IOException {
		FileInputStream fis;
		ObjectInputStream ois;
		FileAttributes attrs = new FileAttributes();
		String attrsFile = uniqueDirectoryName + "/.attrs";
		try {
			try {
				fis = new FileInputStream(attrsFile);
				ois = new ObjectInputStream(fis);
				while (attrs != null){
					try{
						attrs = (FileAttributes) ois.readObject();
						remoteFileAttrs.put(attrs.userName, attrs);
					} catch (EOFException eof){
						break;
					}
				}
				ois.close();
			} catch (IOException e) {
				// we can't forward the exception directly to the user
				// since it deals with the hidden file, instead create a general
				// I/O exceptions
				throw new IOException("Could not access the file "
						+ uniqueDirectoryName);
			} catch (ClassNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

		} catch (FileNotFoundException e) {
			throw new FileNotFoundException("Could not find file "
					+ uniqueDirectoryName);
		}
	}

	private String getRealPath(String filePath) throws FileNotFoundException {
		if (remoteFileAttrs.containsKey(filePath)) {
			return remoteFileAttrs.get(filePath).uniqueName;
		} else {
			String parentName = new File(filePath).getParent();
			if (parentName == null) {
				return filePath;
			}

			parentName = getRealPath(parentName);

			try {
				// TODO: handle exceptions properly
				parseMappingsFile(parentName);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}

			if (remoteFileAttrs.containsKey(filePath)) {
				return remoteFileAttrs.get(filePath).uniqueName;
			}

			throw new FileNotFoundException(filePath);

		}
	}

	public FileSystemReply<Boolean> createFile(String filePath,
			Label sLabelFile, Label iLabelFile, Label sLabelRequestor,
			Label iLabelRequestor, EventID opEid) throws IOException,
			AeolusException {
		FileSystemReply<Boolean> reply = fileServer.createFile(filePath,
				sLabelFile, iLabelFile,
				sLabelRequestor, iLabelRequestor, opEid);
		return reply;
	}

	public FileSystemReply<Integer> readFile(String filePath, byte[] buffer,
			Label sLabelRequestor, Label iLabelRequestor, EventID opEid)
			throws IOException, FileNotFoundException, AeolusException {

		String path = getRealPath(mountLocation + filePath);
		FileAttributes fileAttr = remoteFileAttrs.get(mountLocation + filePath);

		if (!InfoFlowRules.canRead(sLabelRequestor, iLabelRequestor,
				fileAttr.sLabel, fileAttr.iLabel)) {
			throw new InfoFlowControlException(
					"Cannot read file: not authorized; process labels = {"
							+ sLabelRequestor + ", " + iLabelRequestor + "},"
							+ " file labels = {" + fileAttr.sLabel.toString()
							+ ", " + fileAttr.iLabel.toString() + "}");
		}

		FileSystemReply<Integer> reply = new FileSystemReply<Integer>();
		File f = new File(path);

		FileInputStream stream = new FileInputStream(f);
		reply.beforeTime = f.lastModified();
		reply.opResult = stream.read(buffer);
		stream.close();
		reply.afterTime = f.lastModified();
		return reply;

	}

	public FileSystemReply<Void> writeFile(String filePath, byte[] fileContent,
			boolean appendMode, Label sLabelRequestor, Label iLabelRequestor,
			EventID opEid) throws FileNotFoundException, IOException,
			AeolusException {

		String path = getRealPath(mountLocation + filePath);
		FileAttributes fileAttr = remoteFileAttrs.get(mountLocation + filePath);

		if (!InfoFlowRules.canReadWrite(sLabelRequestor, iLabelRequestor,
				fileAttr.sLabel, fileAttr.iLabel)) {
			throw new InfoFlowControlException(
					"Cannot write file: not authorized; process labels = {"
							+ sLabelRequestor + ", " + iLabelRequestor + "},"
							+ " file labels = {" + fileAttr.sLabel.toString()
							+ ", " + fileAttr.iLabel.toString() + "}");
		}

		FileSystemReply<Void> reply = new FileSystemReply<Void>();
		File f = new File(path);

		reply.beforeTime = f.lastModified();
		FileOutputStream stream = new FileOutputStream(f, appendMode);
		stream.write(fileContent);
		stream.close();
		reply.afterTime = f.lastModified();
		return reply;
	}

	public FileSystemReply<Boolean> createDir(String filePath,
			Label sLabelFile, Label iLabelFile, Label sLabelRequestor,
			Label iLabelRequestor, EventID opEid) throws IOException,
			AeolusException {
		FileSystemReply<Boolean> reply = fileServer.createDir(filePath,
				sLabelFile, iLabelFile,
				sLabelRequestor, iLabelRequestor, opEid);
		return reply;
	}

	public FileSystemReply<Boolean> delete(String filePath,
			Label sLabelRequestor, Label iLabelRequestor, EventID opEid)
			throws FileNotFoundException, IOException, AeolusException {
		FileSystemReply<Boolean> reply = fileServer.delete(filePath,
				sLabelRequestor, iLabelRequestor, opEid);
		return reply;
	}

	public FileSystemReply<String[]> listDir(String filePath,
			Label sLabelRequestor, Label iLabelRequestor, EventID opEid)
			throws FileNotFoundException, IOException, AeolusException {
		String path = getRealPath(mountLocation + filePath);
		FileAttributes fileAttr = remoteFileAttrs.get(mountLocation + filePath);

		if (!fileAttr.uniqueName.equals(mountLocation)
				&& !InfoFlowRules.canRead(new Label(sLabelRequestor),
						new Label(iLabelRequestor), fileAttr.sLabel,
						fileAttr.iLabel)) {
			throw new InfoFlowControlException(
					"Not authorized to read the directory; process labels = {"
							+ sLabelRequestor + ", " + iLabelRequestor + "},"
							+ " file labels = {" + fileAttr.sLabel.toString()
							+ ", " + fileAttr.iLabel.toString() + "}");
		}

		String attrsFile = path + "/.attrs";
		FileInputStream fis;
		ObjectInputStream ois;
		FileAttributes attrs = new FileAttributes();
		long beforeTime, afterTime;
		ArrayList<String> entries = new ArrayList<String>();
		File f = new File(attrsFile);
		try {
			fis = new FileInputStream(attrsFile);
			ois = new ObjectInputStream(fis);
			beforeTime = f.lastModified();
			while (attrs != null){
				try{
					attrs = (FileAttributes) ois.readObject();
					String[] splitPath = attrs.userName.split("/");
					entries.add(splitPath[splitPath.length-1]);
				} catch (EOFException eof){
					break;
				} catch (ClassNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			ois.close();
			afterTime = f.lastModified();
		} catch (IOException e) {
			// we can't forward the exception directly to the user
			// since it deals with the hidden file, instead create a general I/O
			// exceptions
			throw new IOException("Could not access the file " + filePath);
		}

		FileSystemReply<String[]> reply = new FileSystemReply<String[]>(
				(String[]) entries.toArray(new String[entries.size()]));
		reply.beforeTime = beforeTime;
		reply.afterTime = afterTime;
		return reply;
	}

	public FileSystemReply<Label> getSecrecyLabel(String filePath,
			Label sLabelRequestor, Label iLabelRequestor, EventID opEid)
			throws FileNotFoundException, IOException, AeolusException {

		FileAttributes parentAttrs = remoteFileAttrs.get(new File(mountLocation
				+ filePath).getParent());
		FileAttributes fileAttrs;

		if (parentAttrs == null)
			getRealPath(mountLocation + filePath);

		parentAttrs = remoteFileAttrs.get(new File(mountLocation + filePath)
				.getParent());

		if (!parentAttrs.uniqueName.equals(mountLocation)
				&& !InfoFlowRules.canRead(new Label(sLabelRequestor),
						new Label(iLabelRequestor), parentAttrs.sLabel,
						parentAttrs.iLabel)) {
			throw new InfoFlowControlException(
					"Cannot read file: not authorized to read the parent directory; process labels = {"
							+ sLabelRequestor
							+ ", "
							+ iLabelRequestor
							+ "},"
							+ " file labels = {"
							+ parentAttrs.sLabel.toString()
							+ ", "
							+ parentAttrs.iLabel.toString() + "}");
		}

		fileAttrs = remoteFileAttrs.get(mountLocation + filePath);

		return new FileSystemReply<Label>(fileAttrs.sLabel);

	}

	public FileSystemReply<Label> getIntegrityLabel(String filePath,
			Label sLabelRequestor, Label iLabelRequestor, EventID opEid)
			throws FileNotFoundException, IOException, AeolusException {

		FileAttributes parentAttrs = remoteFileAttrs.get(new File(mountLocation
				+ filePath).getParent());
		FileAttributes fileAttrs;

		if (parentAttrs == null)
			getRealPath(mountLocation + filePath);

		parentAttrs = remoteFileAttrs.get(new File(mountLocation + filePath)
				.getParent());

		if (!parentAttrs.uniqueName.equals(mountLocation)
				&& !InfoFlowRules.canRead(new Label(sLabelRequestor),
						new Label(iLabelRequestor), parentAttrs.sLabel,
						parentAttrs.iLabel)) {
			throw new InfoFlowControlException(
					"Cannot read file: not authorized to read the parent directory; process labels = {"
							+ sLabelRequestor
							+ ", "
							+ iLabelRequestor
							+ "},"
							+ " file labels = {"
							+ parentAttrs.sLabel.toString()
							+ ", "
							+ parentAttrs.iLabel.toString() + "}");
		}

		fileAttrs = remoteFileAttrs.get(mountLocation + filePath);

		return new FileSystemReply<Label>(fileAttrs.iLabel);

	}

	// File Streams

	// File Input Stream

	/**
	 * Reads file and labels from sever if not in cache Registers file stream
	 * with the file Returns the file stream handle
	 */
	public FileSystemReply<Integer> openFileInputStream(String filePath,
			Label sLabelRequestor, Label iLabelRequestor, EventID opEid)
			throws FileNotFoundException, IOException, InfoFlowControlException {
		int handle = this.getNextFilestreamHandle();
		String path = getRealPath(mountLocation + filePath);
		FileAttributes fileAttr = remoteFileAttrs.get(mountLocation + filePath);
		File f = new File(path);

		if (!InfoFlowRules.canRead(sLabelRequestor, iLabelRequestor,
				fileAttr.sLabel, fileAttr.iLabel)) {
			throw new InfoFlowControlException(
					"Cannot read file: not authorized; process labels = {"
							+ sLabelRequestor + ", " + iLabelRequestor + "},"
							+ " file labels = {" + fileAttr.sLabel.toString()
							+ ", " + fileAttr.iLabel.toString() + "}");
		}
		
		FileInputStream stream = new FileInputStream(f);
		long timeStamp = f.lastModified();
		openInputStreams.put(handle, stream);

		FileSystemReply<Integer> reply = new FileSystemReply<Integer>(handle,
				fileAttr.sLabel, fileAttr.iLabel, null);
		reply.beforeTime = timeStamp;
		return reply;

	}

	public int readFilestream(int handle, String filePath, int position,
			byte[] buffer, int offset, int count) throws IOException {

		int bytes = openInputStreams.get(handle).read(buffer, offset, count);
		return bytes;
	}

	public FileSystemReply<Void> closeFileInputStream(int handle,
			String filePath) throws IOException {
		FileAttributes fileAttr = remoteFileAttrs.get(mountLocation + filePath);
		FileSystemReply<Void> reply = new FileSystemReply<Void>();

		if (fileAttr == null)
			throw new FileNotFoundException();

		File f = new File(fileAttr.uniqueName);

		openInputStreams.get(handle).close();
		reply.afterTime = f.lastModified();
		openInputStreams.remove(handle);
		return reply;
	}

	// File Output Stream

	/**
	 * Reads the file's labels from server and checks if can open for write
	 * Returns the file stream handle
	 */
	public FileSystemReply<Integer> openFileOutputStream(String filePath,
			boolean appendMode, Label sLabelRequestor, Label iLabelRequestor,
			EventID opEid) throws FileNotFoundException, IOException,
			AeolusException {
		int handle = this.getNextFilestreamHandle();
		String path = getRealPath(mountLocation + filePath);
		File temp = new File(filePath);
		FileAttributes fileAttr = remoteFileAttrs.get(mountLocation + filePath);
		FileAttributes parentAttr = remoteFileAttrs.get(mountLocation + temp.getParent());
		if(!InfoFlowRules.canRead(sLabelRequestor, iLabelRequestor, parentAttr.sLabel, parentAttr.iLabel)){
			throw new InfoFlowControlException(
					"Cannot read parent directory, process labels = {"
							+ sLabelRequestor + ", " + iLabelRequestor + "},"
							+ " file labels = {" + parentAttr.sLabel.toString()
							+ ", " + parentAttr.iLabel.toString() + "}");
		}
		if (!InfoFlowRules.canWrite	(sLabelRequestor, iLabelRequestor,
				fileAttr.sLabel, fileAttr.iLabel)) {
			throw new InfoFlowControlException(
					"Cannot write to file: not authorized; process labels = {"
							+ sLabelRequestor + ", " + iLabelRequestor + "},"
							+ " file labels = {" + fileAttr.sLabel.toString()
							+ ", " + fileAttr.iLabel.toString() + "}");
		}
		
		File f = new File(path);
		FileOutputStream stream = new FileOutputStream(f, appendMode);
		long timeStamp = f.lastModified();
		openOutputStreams.put(handle, stream);

		FileSystemReply<Integer> reply = new FileSystemReply<Integer>(handle,
				fileAttr.sLabel, fileAttr.iLabel, null);
		reply.afterTime = timeStamp;
		return reply;

	}

	public void writeFilestream(int handle, String filePath, byte[] buffer,
			int offset, int count) throws IOException {

		openOutputStreams.get(handle).write(buffer, offset, count);
	}

	public FileSystemReply<Void> closeFileOutputStream(int handle,
			String filePath) throws RemoteException, FileNotFoundException,
			IOException, AeolusException {
		FileAttributes fileAttr = remoteFileAttrs.get(mountLocation + filePath);
		FileSystemReply<Void> reply = new FileSystemReply<Void>();

		if (fileAttr == null)
			throw new FileNotFoundException();

		File f = new File(fileAttr.uniqueName);

		openOutputStreams.get(handle).close();
		reply.afterTime = f.lastModified();
		openOutputStreams.remove(handle);
		return reply;
	}

	private synchronized int getNextFilestreamHandle() {
		return this.handleCounter++;
	}

}
