package edu.mit.csail.aeolus.filesystem;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.NoSuchObjectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ServerNotActiveException;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import edu.mit.csail.aeolus.auditing.AeolusApiConstants;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.auditing.EventRecord;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;
import edu.mit.csail.aeolus.utilities.RMIUtil;

/**
 * File system manager executes file and filestream operation requests from file system clients 
 * by accessing the data stored in the underlying file system (each storage node runs its
 * own file manager).
 * 
 * File system manager keeps track of any remote or local clients. It sends periodic invalidation
 * packages to the clients to invalidate entries in their local file caches (the invalidation
 * package includes names of files that have been modified since the last invalidation
 * package sent to the given client).
 * Each file manager reply also includes a list of files to invalidate. Clients process the 
 * invalidations before returning any file data to the users.
 * 
 * 
 * @author viq
 * 
 */
public class FileSystemManager extends UnicastRemoteObject implements FileSystemRemoteInterface {

	private static final long serialVersionUID = 46512314443626910L;

//	private static final int FSCLIENT_CONNECTION_TIMEOUT_MS = 5000;
	
	/** Map of registered clients (by hostname) to their list of latest file invalidations */
	private ConcurrentHashMap<String, ConcurrentLinkedQueue<String>> registeredClients;
	private ConcurrentHashMap<String, FileClientInfo> clientInfo; // TODO: add this data to one map
	
	/** Underlying file system */
	private FileSystem fs;
	private String aeolusRootPath;

	/** Table of recently accessed files (need to make sure a file is always represented by same name) */
	private HashMap<String, ReadWriteLock> accessedFiles;
	private Lock accessedFilesTableLock;
	/** Map of file names to the latest write events on this files */
	private ConcurrentHashMap<String, EventID> latestFileUpdates;
	
	/** Audit trail management */
	private long fileEventCounter;
	private FileLogCollector collector;
	private String hostname;
	@SuppressWarnings("unused")
	private EventID fsLaunchEid;

	private Registry registry;

	private int fsClientPort;

	private boolean enableFSLogging;

	private int fsServicePort;

	private String fsServiceName;
	
	public enum FileMode {
		READ, WRITE;
	}

	/* Initialization */

	public FileSystemManager() throws RemoteException 
	{
		fs = new FileSystem();
		RuntimeConfig c = RuntimeConfig.get();
		
		aeolusRootPath = c.getFileSystemRootPath();
		fsClientPort = c.getFsClientPort();
		enableFSLogging = c.isEnableFSLogging();
		fsServicePort = c.getFsServicePort();
		fsServiceName = c.getFsServiceName();
		
		fs.setupFSRoot(aeolusRootPath);	
		accessedFiles = new HashMap<String, ReadWriteLock>();
		accessedFilesTableLock = new ReentrantLock();
		// writes to the same file are serialized
		this.latestFileUpdates = new ConcurrentHashMap<String, EventID>();

		this.registeredClients = new ConcurrentHashMap<String, ConcurrentLinkedQueue<String>>();
		this.clientInfo = new ConcurrentHashMap<String, FileClientInfo>();
		
		fileEventCounter = 1;
		collector = new FileLogCollector();
		try 
		{
			hostname = InetAddress.getLocalHost().getHostName();
		} 
		catch (UnknownHostException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Exports this service to Java RMI
	 */
	public void start() 
	{
		try {
			Registry r = RMIUtil.getOrCreate(fsServicePort);
			r.rebind(fsServiceName, this);
			
			System.out.println("FS Server is running");
			
			// log server startup
			if(enableFSLogging)
			{
				//fsLaunchEid = this.getNewEventID();
				//collector.submitLogRecord(new EventRecord(fsLaunchEid, null, null, AeolusApiConstants.FS_LAUNCH, null));
			}
			
		} 
		catch (Exception e) 
		{
			e.printStackTrace();
		}
	}
	
	public void shutdown()
	{
		if(registry != null)
    	{
    		try 
    		{
				UnicastRemoteObject.unexportObject(registry ,true);
			} 
    		catch (NoSuchObjectException e) 
    		{
				e.printStackTrace();
			}  
    	}
	}

	/** FileSystem Client Management **/

	/**
	 * Registers a new file client
	 * @param clientServiceName - name bound on the client node to this file client service
	 * in the Java RMI registry
	 */
	public void registerFSClient(String clientServiceName) throws RemoteException
	{
		System.out.println("REGISTER CLIENT");
		
		String clientHostname;
		try 
		{
			clientHostname = getClientHost();
		} 
		catch (ServerNotActiveException e) 
		{
			e.printStackTrace();
			return;
		}
		
		// check if this client is already registered
	    if(this.registeredClients.containsKey(clientHostname))
	    	return;
	    
	    // add the new client 
	    FileClientInfo clientInfo = new FileClientInfo(clientHostname, clientServiceName);
	    this.registeredClients.put(clientHostname, new ConcurrentLinkedQueue<String>());
	    this.clientInfo.put(clientHostname, clientInfo);
	    
	    // check
	    Registry registry = LocateRegistry.getRegistry(clientHostname, fsClientPort);
		clientInfo = this.clientInfo.get(clientHostname);
		
		System.out.println("SENDING CHECK HOSTNAME = " + clientHostname + " with name = " + clientInfo.clientServiceName);
		
		FileSystemClientInterface fsClient;
		try {
			fsClient = (FileSystemClientInterface) registry.lookup(clientInfo.clientServiceName);
			fsClient.isAlive();
		} catch (NotBoundException e) {
			
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}
	
	/**
	 * Removes a client from this service (e.g. due to communication timeout)
	 * Any subsequent requests from client will fail
	 */
	public void removeFSClient(String clientHostname)
	{
		this.registeredClients.remove(clientHostname);
		this.clientInfo.remove(clientHostname);
	}
	
	/** Handle Client Request */

	private void initiateClientRequest(String filePath, FileMode accessMode) 
	{
		accessedFilesTableLock.lock();
		ReadWriteLock fileLock = accessedFiles.get(filePath);
		if (fileLock == null) {
			fileLock = new ReentrantReadWriteLock();
			accessedFiles.put(filePath, fileLock);
			//this.latestFileUpdates.put(filePath, this.fsLaunchEid);
		}
		accessedFilesTableLock.unlock();

		if (accessMode == FileMode.READ)
			fileLock.readLock().lock();
		else
			fileLock.writeLock().lock();
	}

	private void terminateClientRequest(String filePath, FileMode accessedMode) 
	{
		ReadWriteLock fileLock = accessedFiles.get(filePath);
		if (fileLock != null) {
			if (accessedMode == FileMode.READ)
				fileLock.readLock().unlock();
			else
				fileLock.writeLock().unlock();
		} else
			assert false;
	}
	
	private synchronized EventID getNewEventID() 
	{
		EventID newfileOpId = new EventID(hostname, 0, fileEventCounter);
		fileEventCounter++;
		return newfileOpId;
	}
	
	private EventID getLatestUpdate(String filePath)
	{
		return this.latestFileUpdates.get(filePath);
	}
	
	/**
	 * Returns the previous event id
	 */
	private EventID setLatestUpdate(String filePath, EventID latestUpdateEvent)
	{
		return this.latestFileUpdates.put(filePath, latestUpdateEvent);
	}

	
	/** Client File Requests */

	/* Files */

	public FileSystemReply<Boolean> createFile(String filePath, Label sLabelFile, Label iLabelFile, Label sLabelRequestor,
			Label iLabelRequestor, EventID opEid) throws RemoteException, IOException, InfoFlowControlException 
	{
		File file = new File(aeolusRootPath + filePath);
		File parentFile = file.getParentFile();

		System.out.println("Creating file...");
		
		if (filePath != null && (!filePath.isEmpty()) && parentFile != null) 
		{
			this.initiateClientRequest(parentFile.getPath(), FileMode.WRITE);
			FileSystemReply<Boolean> result = fs.createFile(file.getPath(), sLabelFile, iLabelFile, sLabelRequestor, iLabelRequestor);

			String clientHost = "";
			try {
				clientHost = getClientHost();
			}
			catch (ServerNotActiveException e) {
				e.printStackTrace();
			}
			
			if(result.opResult == true)
			{
				EventID eid = getNewEventID();
				EventID fsPred = this.setLatestUpdate(parentFile.getPath(), eid);
				this.setLatestUpdate(file.getPath(), eid);
				collector.submitLogRecord(new EventRecord(eid, opEid, fsPred, AeolusApiConstants.FS_CREATE_FILE, null));
				result.predecessorEventId = eid;
			}
			else
			{
				result.predecessorEventId = this.getLatestUpdate(parentFile.getPath());
			}
			this.terminateClientRequest(parentFile.getPath(), FileMode.WRITE);
			
			return result;

		} 
		else {
			throw new FileNotFoundException( "Cannot create file. Invalid file pathname: " + filePath);
		}
	}

	public FileSystemReply<Boolean> createDir(String filePath, Label sLabelFile, Label iLabelFile, Label sLabelRequestor,
			Label iLabelRequestor, EventID opEid) throws RemoteException, InfoFlowControlException, IOException 
	{
		File file = new File(aeolusRootPath + filePath);
		File parentFile = file.getParentFile();

		if (filePath != null && (!filePath.isEmpty()) && parentFile != null) 
		{
			this.initiateClientRequest(parentFile.getPath(), FileMode.WRITE);
			FileSystemReply<Boolean> result = fs.createDir(file.getPath(), sLabelFile, iLabelFile, sLabelRequestor, iLabelRequestor);
		
			String clientHost = "";
			try {
				clientHost = getClientHost();
			}
			catch (ServerNotActiveException e) {
				e.printStackTrace();
			}
			
			if(result.opResult == true)
			{
				EventID eid = getNewEventID();
				EventID fsPred = this.setLatestUpdate(parentFile.getPath(), eid);
				this.setLatestUpdate(file.getPath(), eid);
				collector.submitLogRecord(new EventRecord(eid, opEid, fsPred, AeolusApiConstants.FS_CREATE_DIR, null));
				result.predecessorEventId = eid;
			}
			else
			{
				result.predecessorEventId = this.getLatestUpdate(parentFile.getPath());
			}
			this.terminateClientRequest(parentFile.getPath(), FileMode.WRITE);
			
			return result;

		} 
		else {
			throw new FileNotFoundException( "Cannot create directory. Invalid file pathname: " + filePath);
		}
	}

	public FileSystemReply<Boolean> delete(String filePath, Label sLabelRequestor, Label iLabelRequestor, EventID opEid)
			throws RemoteException, InfoFlowControlException, IOException 
	{
		File file = new File(aeolusRootPath + filePath);
		File parentFile = file.getParentFile();

		if (filePath != null && (!filePath.isEmpty()) && parentFile != null) 
		{
			initiateClientRequest(parentFile.getPath(), FileMode.WRITE);

			FileSystemReply<Boolean> result = fs.delete(file.getPath(), sLabelRequestor, iLabelRequestor);
			
			String clientHost = "";
			try {
				clientHost = getClientHost();
			}
			catch (ServerNotActiveException e) {
				e.printStackTrace();
			}
			
			EventID eid = getNewEventID();
			EventID fsPred = this.setLatestUpdate(parentFile.getPath(), eid);
			this.setLatestUpdate(file.getPath(), eid);
			collector.submitLogRecord(new EventRecord(eid, opEid, fsPred, AeolusApiConstants.FS_DELETE_DIR, null));
			result.predecessorEventId = eid;
		
			terminateClientRequest(parentFile.getPath(), FileMode.WRITE);

			return result;

		} 
		else {
			throw new FileNotFoundException("Cannot remove directory. Invalid file pathname: " + filePath);
		}
	}
	
}
