package edu.mit.csail.aeolus.filesystem.tests;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.rmi.RemoteException;

import junit.framework.Assert;

import org.junit.After;
import org.junit.BeforeClass;
import org.junit.Test;

import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.filesystem.FileSystemClient;
import edu.mit.csail.aeolus.filesystem.FileSystemManager;
import edu.mit.csail.aeolus.filesystem.FileSystemReply;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.runtime.RuntimeConfig;


public class OneClientTest {

	public static String fsHost;
	public static String fsMountLocation;
	public static Label testLabel = new Label();
	public static Label empty = new Label();
	public static FileSystemManager fs;
	public static FileSystemClient fsClient;
	
	@BeforeClass
	public static void config() throws RemoteException, AeolusException{
		RuntimeConfig c = RuntimeConfig.get();
		c.setFsHostName("localhost");
		c.setFsPort(2020);
		c.setEnableFSLogging(false);
		c.setEnableLogging(false);
		c.setEnableLogShipment(false);
		c.setFS_enableLogShipment(false);
		c.finalize();
		fsHost = c.getFsHostName();
		fs = new FileSystemManager();
		fs.start();
		fsClient = new FileSystemClient(fsHost);
		fsClient.register();
	}
	
	@After
	public void shutdownRMI() throws InterruptedException{
		Thread.sleep(1000);
	}
	
	@Test
	public void testArray() throws AeolusException, FileNotFoundException, IOException
	{
		Assert.assertTrue(fsClient.isAlive());

		//CreateDir
		FileSystemReply<Boolean> reply = fsClient.createDir("/items", empty, testLabel, empty, testLabel, new EventID("1"));
		Assert.assertTrue(reply.opResult);
		
		//CreateFile
		reply = fsClient.createFile("/items/a", testLabel, testLabel, empty, testLabel, new EventID("1"));
		Assert.assertTrue(reply.opResult);

		//Write
		byte[] content = new String("hello").getBytes();
		fsClient.writeFile("/items/a", content, false, testLabel, testLabel, new EventID("2"));
		
		//Read
		byte[] readContent = new byte[5];
		fsClient.readFile("/items/a", readContent, testLabel, empty, new EventID("3"));
		Assert.assertEquals("hello", new String(readContent));

		//IFC Violation
		try
		{
			fsClient.readFile("/items/a", readContent, empty, empty, new EventID("4"));
			Assert.fail();
		}
		catch (Exception e)
		{
			Assert.assertTrue(e instanceof edu.mit.csail.aeolus.exceptions.InfoFlowControlException);
		}
		
		//Get Labels
		FileSystemReply<Label> gLreply = fsClient.getSecrecyLabel("/items/a", empty, empty, new EventID("4"));
		Assert.assertEquals(testLabel.toString(), gLreply.opResult.toString());
		
		gLreply = fsClient.getIntegrityLabel("/items/a", empty, empty, new EventID("4"));
		Assert.assertEquals(testLabel.toString(), gLreply.opResult.toString());

		//ListDir
		FileSystemReply<String[]> sreply = fsClient.listDir("/items", empty, empty, new EventID("4"));
		
		Assert.assertEquals(1, sreply.opResult.length);
		
		//Remove File
		reply = fsClient.delete("/items/a", empty, testLabel, new EventID("4"));
		Assert.assertTrue(reply.opResult);
		
		
		//ListDir
		sreply = fsClient.listDir("/items", empty, empty, new EventID("4"));
		Assert.assertEquals(0, sreply.opResult.length);
		
		
		//Remove Directory
		reply = fsClient.delete("/items", empty, testLabel, new EventID("4"));
		Assert.assertTrue(reply.opResult);
		
		sreply = fsClient.listDir("/", empty, empty, new EventID("4"));
		Assert.assertEquals(0, sreply.opResult.length);
		
		
		try
		{
			// file should not be found since invalidation had to propagate
			fsClient.readFile("/items/a", readContent, testLabel, empty, new EventID("4"));
			Assert.fail();
		}
		catch (Exception e)
		{
			Assert.assertTrue(e instanceof FileNotFoundException);

		}
		fsClient.shutdown();
		fs.shutdown();
		
	}
	
	//@Test
	public void testServerMultithreadedClient() throws AeolusException, IOException, InterruptedException
	{
		final Label empty = new Label();
		
		FileSystemManager fs = new FileSystemManager();
		fs.start();
		
		final FileSystemClient fsClient = new FileSystemClient(fsHost);
		fsClient.register();
		Assert.assertTrue(fsClient.isAlive());
		

		FileSystemReply<Boolean> reply = fsClient.createFile("/\\a", empty, empty, empty, empty, new EventID("1"));
		Assert.assertTrue(reply.opResult);

		
		Runnable reader = new Runnable()
		{
			public void run()
			{
				try 
				{
					@SuppressWarnings("unused")
					byte[] content = new String("hello").getBytes();
					byte[] readContent = new byte[10];
					
					for(int i = 0; i < 100; i++)
					{
						System.out.println("READ");
						fsClient.readFile("/\\a", readContent, empty, empty, new EventID("3"));
					}
					System.out.println("READER DONE");
				} 
				catch (Exception e) {
					e.printStackTrace();
					Assert.fail();
				} 
				
			}
		};
		
		Runnable writer = new Runnable()
		{
			public void run()
			{
				byte[] content = new String("hello").getBytes();
				try 
				{
					for(int i = 0; i < 100; i++)
					{
						System.out.println("WRITE");
						fsClient.writeFile("/\\a", content, false, empty, empty, new EventID("2"));
					}
					
					System.out.println("WRITER DONE");
				} 
				catch (Exception e) {
					e.printStackTrace();
					Assert.fail();
				} 
			}
		};
		
		Thread t1 = new Thread(reader);
		t1.start();
		
		Thread t2 = new Thread(writer);
		t2.start();
		
		// FIXME wait for the threads to finish...
		Thread.sleep(10000);
		
		fsClient.delete("/\\a", empty, empty, new EventID("4"));
		
		fs.shutdown();
		fsClient.shutdown();
		
	}
	
}
