package edu.mit.csail.aeolus.infoflow;
import edu.mit.csail.aeolus.api.AeolusLabel;


/** 
 * INFORMATION FLOW MODEL
 * 
 * Defines when information can flow from some given A to some given B
 * 
 * viq
 */

public class InfoFlowRules {

	/**
	 * Returns true if A can flow to B
	 * Checks if:
	 * (1) A's secrecy label is a subset of B's secrecy label
	 *     AND
	 * (2) B's integrity label is a subset of A's integrity label
	 * 
	 * If conditions are satisfied, we say A is no more restrictive than B
	 * 
	 * @return true if information can flow from A to B
	 */
	public static boolean canFlow(final Label aSLabel, final Label aILabel, final Label bSLabel, final Label bILabel) 
	{
		return (aSLabel.isSubsetOf(bSLabel) && bILabel.isSubsetOf(aILabel));
	}

	public static boolean canFlow(final AeolusLabel aSLabel, final AeolusLabel aILabel, final AeolusLabel bSLabel, final AeolusLabel bILabel) 
	{
		return (aSLabel.isSubsetOf(bSLabel) && bILabel.isSubsetOf(aILabel));
	}

	/** 
	 * Returns true if A can flow to B
	 * Checks if:
	 * (1) A's secrecy label is a subset of B's secrecy label
	 *     AND
	 * (2) B's integrity label is a subset of A's integrity label
	 * NOTE: this is just a label-to-label check (useful only if all priv is reflected in the label)
	 * @return true if information can flow from A to B
	 */
	public static Boolean canFlow(InfoFlowState aState, InfoFlowState bState)
	{
		return (aState.SLabel.isSubsetOf(bState.SLabel) && bState.ILabel.isSubsetOf(aState.ILabel));
	}

	/**
	 * Checks if writer can write to obj (i.e whether information can flow from writer to obj)
	 * @return true if information can flow from writer to obj
	 */
	public static boolean canWrite(Label wSLabel, Label wILabel, Label objSLabel, Label objILabel)
	{
		return canFlow(wSLabel, wILabel, objSLabel, objILabel);
	}
	public static boolean canWrite(AeolusLabel wSLabel, AeolusLabel wILabel, AeolusLabel objSLabel, AeolusLabel objILabel)
	{
		return canFlow(wSLabel, wILabel, objSLabel, objILabel);
	}

	/**
	 * Checks if writer can write to obj (i.e whether information can flow from writer to obj)
	 * @return true if information can flow from writer to obj
	 */
	public static boolean canWrite(InfoFlowState writer, InfoFlowState obj)
	{
		return canFlow(writer, obj);
	}
	
	/**
	 * Checks if reader can read from obj (i.e whether information can flow from obj to reader)
	 * @return true if information can flow from obj to reader
	 */
	public static boolean canRead(Label rSLabel, Label rILabel, Label objSLabel, Label objILabel)
	{
		return canFlow(objSLabel, objILabel, rSLabel, rILabel);
	}
	public static boolean canRead(AeolusLabel rSLabel, AeolusLabel rILabel, AeolusLabel objSLabel, AeolusLabel objILabel)
	{
		return canFlow(objSLabel, objILabel, rSLabel, rILabel);
	}
	
	/**
	 * Checks if reader can read from obj (i.e whether information can flow from obj to reader)
	 * @return true if information can flow from obj to reader
	 */
	public static boolean canRead(InfoFlowState reader, InfoFlowState obj)
	{
		return canFlow(obj, reader);
	}

	/**
	 * Checks if:
	 * (1) secrecy and integrity labels are of same size
	 * (2) each secrecy tag of sLabel1 is in the secrecy label sLabel2
	 * (3) each integrity tag in iLabel1 is in the integrity label of iLabel2
	 */
	public static boolean canReadWrite(Label sLabel1, Label iLabel1, Label sLabel2, Label iLabel2)
	{
		if ((sLabel1.size() != sLabel2.size()) || (iLabel1.size() != iLabel2.size()))
		{
			return false;
		}

		if ((!sLabel1.isSubsetOf(sLabel2)) || (!iLabel1.isSubsetOf(iLabel2)))
		{
			return false;
		}
		
		return true;
	}
	
	public static boolean canReadWrite(AeolusLabel sLabel1, AeolusLabel iLabel1, AeolusLabel sLabel2, AeolusLabel iLabel2)
	{
		if ((sLabel1.size() != sLabel2.size()) || (iLabel1.size() != iLabel2.size()))
		{
			return false;
		}

		if ((!sLabel1.isSubsetOf(sLabel2)) || (!iLabel1.isSubsetOf(iLabel2)))
		{
			return false;
		}
		
		return true;
	}
	
	/**
	 * Checks if:
	 * (1) secrecy and integrity labels are of same size
	 * (2) each secrecy tag of state1 is in the secrecy label of state2
	 * (3) each integrity tag in state1 is in the integrity label of state2
	 */
	public static boolean canReadWrite(InfoFlowState state1, InfoFlowState state2)
	{
		return canReadWrite(state1.SLabel, state1.ILabel, state2.SLabel, state2.ILabel);
	}

	/**
	 * Return a new InfoFlowState object with:
	 * SLabel = union of the SLabels of a and b
	 * Ilabel = intersection of ILabels of a and b
	 */
	public static InfoFlowState mostRestrictive(InfoFlowState aState, InfoFlowState bState)
	{
		return new InfoFlowState(aState.SLabel.union(bState.SLabel), aState.ILabel.intersection(bState.ILabel));
	}
}
