package edu.mit.csail.aeolus.rpc;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;

import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.exceptions.ServiceNotFoundException;
import edu.mit.csail.aeolus.infoflow.PlatformInstanceID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.runtime.SystemDomain;
import edu.mit.csail.aeolus.runtime.SystemNode;

/**
 * 
 * Aeolus RPC mechanism (uses Java RMI)
 * 
 * We must ensure that the remote user object method is executed with the appropriate principal 
 * and process labels
 * 
 * We need the following information to execute the request:
 * - info flow state
 * 
 */

public class AeolusRpcService implements AeolusRpcInterface {
	
	/** Node running this service */
	private SystemNode platformNode;
	
	public AeolusRpcService(SystemNode node) 
	{
		this.platformNode = node;
	}

	/**
	 * Asks the local SYS-D to invoke the service method inside a new User-D thread with 
	 * the appropriate starting info flow state (included into the RpcToken)
	 */
	@SuppressWarnings("unchecked")
	public RpcReply invokeRemoteMethod(final String serviceName, final String methodName, 
		final Class[] parameterTypes, final Serializable[] parameterValues,
		final RpcToken token) throws RemoteException, RpcException
	{
	
		// TODO: check that the caller is trusted
		
		// catch up the current authority client
		platformNode.authClient.getUpToInvalidation(token.invalidationNumber);
		
		// find the PI where the service was registered
		SystemDomain sysd = platformNode.resolvePI(serviceName);
		
		if(sysd == null)
		{
			throw new RpcException("No such service: " + serviceName, 
									new ServiceNotFoundException("No binding for sevice " + serviceName));
		}
		
		return sysd.receiveRPC(serviceName, methodName, parameterTypes, parameterValues, token); 
		
	}
	
	/**
	 * Enables a remote PI launch
	 * @throws ClassNotFoundException - application class not found
         * @throws NoSuchMethodException - main method was not found
         * @throws InvocationTargetException - method main throws an exception (declared or runtime exception)
         * @throws AeolusException - thrown while launching the application for any other reason
	 */
	public RpcReply launchRemotePI(PrincipalID callerPid, String appName,
			String appParams, PrincipalID pid, EventID predEid, RpcToken rpcToken) 
			throws RemoteException, InvocationTargetException, ClassNotFoundException,
			       NoSuchMethodException, AeolusException, RpcException
	{
		// TODO: check that the caller is trusted
		return this.platformNode.launchPI(callerPid, pid, appName, appParams, predEid, rpcToken);
	}
	
	/**
	 * Enables a remote PI shutdown
	 * @throws AuthorityException - if caller not authorized to shutdown this platform instance
	 */
	public void shutdownRemotePI(PlatformInstanceID platformInstanceID, PrincipalID callerPid, boolean shutdownImmediately, EventID predEid) 
	throws RemoteException, RpcException, AuthorityException
	{
		// TODO: check that the caller is trusted
		this.platformNode.shutdownPI(platformInstanceID, callerPid, shutdownImmediately, predEid);
	}
	
	public int testCall(int a) throws RemoteException
	{
		System.out.println("AeolusRpcService test call, value =" + a);
		return a;
	}

}
