package edu.mit.csail.aeolus.rpc;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.runtime.UserProxy;

/**
 * This handler is passed to the constructor of the dynamically generated proxy 
 * (on the client side) and handles the invocations of all proxy methods
 * 
 * When any proxy method is invoked, we want to invoke the same method
 * at the remote server through the Aeolus platform
 * 
 * This handler sends a request to the SYS-D (through the proxy object) 
 * to make the appropriate rpc call
 * 
 * @author viq
 *
 */

public class RpcInvocationHandler implements InvocationHandler {
	
	// preloaded Method objects for the methods in java.lang.Object
    private static Method hashCodeMethod;
    private static Method equalsMethod;
    private static Method toStringMethod;
    static {
    	try {
    		hashCodeMethod = Object.class.getMethod("hashCode");
    		equalsMethod = Object.class.getMethod("equals", new Class[] { Object.class });
    		toStringMethod = Object.class.getMethod("toString");
        }catch (NoSuchMethodException e) {
    		throw new NoSuchMethodError(e.getMessage());
    	}
    }

    
	private UserProxy po;
	
	public RpcInvocationHandler(UserProxy po){
		this.po = po;
	}
	
	/**
	 * Invokes the proxy object sendRPC method
	 * which compiles and submits a request to the SYS-D
	 */
	public Object invoke(Object proxy, Method method, Object[] parameters) throws RpcException, AeolusException{
		Class<?> declaringClass = method.getDeclaringClass();
		if (declaringClass == Object.class) {
		    if (method.equals(hashCodeMethod)) {
		    	return proxyHashCode(proxy);
		    } else if (method.equals(equalsMethod)) {
		    	return proxyEquals(proxy, parameters[0]);
		    } else if (method.equals(toStringMethod)) {
		    	return proxyToString(proxy);
		    } else {
		    	throw new InternalError(
			    "unexpected Object method dispatched: " + method);
		    }
		} else {
			return po.sendRPCRequest(proxy, method, parameters);
		}
	}
	
	protected Integer proxyHashCode(Object proxy) {
		return new Integer(System.identityHashCode(proxy));
    }

    protected Boolean proxyEquals(Object proxy, Object other) {
    	return (proxy == other ? Boolean.TRUE : Boolean.FALSE);
    }

    protected String proxyToString(Object proxy) {
    	return proxy.getClass().getName() + '@' +
	    	Integer.toHexString(proxy.hashCode());
    }
}
