package edu.mit.csail.aeolus.runtime;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.List;
import java.util.HashMap;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import edu.mit.csail.aeolus.api.AeolusShared;
import edu.mit.csail.aeolus.api.AeolusLabel;
import edu.mit.csail.aeolus.api.AeolusLib;
import edu.mit.csail.aeolus.auditing.AeolusApiConstants;
import edu.mit.csail.aeolus.auditing.Collector;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.auditing.EventRecord;
import edu.mit.csail.aeolus.auditing.LoggingConstants;
import edu.mit.csail.aeolus.authority.client.api.InterfaceAuthorityClient;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.exceptions.ServiceNotFoundException;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.infoflow.PlatformInstanceID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.rpc.AeolusRpcInterface;
import edu.mit.csail.aeolus.rpc.RegisteredService;
import edu.mit.csail.aeolus.rpc.RpcReply;
import edu.mit.csail.aeolus.rpc.RpcToken;
import edu.mit.csail.aeolus.security.AeolusClassLoader;
import edu.mit.csail.aeolus.utilities.Logger;
import edu.mit.csail.aeolus.utilities.LoggerLevel;

/**
 * The SYS-D and its User-D pool constitute a platform instance (PI)
 * 
 * SYS-D has permissions to resources (files, remote calls, I/Os).
 * User applications must use SYS-D to obtain access to various resources. 
 * 
 * SYS-D receives requests from user domains and processes them via its
 * worker thread pool
 * 
 * SYS-D provides:
 * - launcher that starts up user domains
 * - Aeolus internal state for each user domain
 * - access to shared state
 * - access to resources such as the file system and authority client
 * 
 * 
 * @author viq
 * 
 */
public class SystemDomain {

	/* Node on which this platform instance runs */
	public SystemNode platformNode; 

	/* Principal of this platform instance */
	public PrincipalID pid;

	/* ID of platform instance represented by this SYS-D */
	public PlatformInstanceID platformInstanceID;

	/* ID of the principal that launched this platform instance */
	public PrincipalID launcherPid;

	/* Thread pool to launch new User-Ds */
	private UserDomainPool userDomainPool;

	/* Thread pool to handle proxy object requests */
	private ScheduledThreadPoolExecutor workerThreadPool;

	/* RPC: registered services on this platform instance - map of service name to its instance */
	private HashMap<String, RegisteredService> registeredServices;

	/* SYS-D resources */
	public InterfaceAuthorityClient authClient;
	public AeolusShared sharedRoot;
	//public AeolusFSInterface fileSystemManager;

	/* Audit trail collector */
	public Collector collector;
	public EventID piLaunchEid;
	private EventID predEid;

	public boolean shuttingDown;

	private Logger logger;

	private boolean enableLogging = true;

	private int rpcPort;

	private String aeolusRpcServiceName;

	private boolean shippingLogs;

	public SystemDomain(PlatformInstanceID platformInstanceID)
	{   
		this(null, null, platformInstanceID, null, null);
	}

	/**
	 * @param platformNodeID - node on which this PI runs
	 * @param platformInstanceID - unique id assigned to the PI
	 * @param useDummy -- TRUE to use a dummy authority client (DEFAULT.) 
	 * 	False to initialize a client connection to an authority server.
	 * @param enableLogging - True to enable the audit trail collection
	 */
	public SystemDomain(SystemNode node, PrincipalID pid, PlatformInstanceID id, PrincipalID callerPid, EventID opEid){	
		RuntimeConfig c = RuntimeConfig.get();
		// Read in config...
		this.enableLogging = c.isEnableLogging();
		this.shippingLogs = c.isEnableLogShipment();
		this.rpcPort = c.getRpcPort();
		this.aeolusRpcServiceName = c.getAeolusRpcServiceName();
		
		this.platformNode = node;
		this.pid = pid;
		this.platformInstanceID = id;
		this.launcherPid = callerPid;
		this.predEid = opEid;
		this.userDomainPool = new UserDomainPool(c.getAppDomainPoolSize(), 
				c.getAppDomainPoolSize(), 
				10, TimeUnit.MILLISECONDS, 
				new LinkedBlockingQueue<Runnable>(), 
				new UserDomainFactory(this));
		this.workerThreadPool = new ScheduledThreadPoolExecutor(c.getMaxWorkerThreads());
		this.registeredServices = new HashMap<String, RegisteredService>();

		this.shuttingDown = false;

		this.logger = new Logger(LoggerLevel.DEBUG, "LocalRuntime");

		this.initialize();
	}

	/**
	 *  SYS-D SETUP
	 */

	/**
	 *  This function allocates any necessary structures before the platform instance can be used 
	 */
	private void initialize(){
	    /*
	     * XXX Is this needed?
		EventID sharedRootEID = new EventID("" + platformInstanceID + "-sharedroot");
		this.sharedRoot = new AeolusShared(sharedRootEID, this);
	    */
		this.sharedRoot = null;
		if(platformNode == null)
			this.authClient = null;
		else
			this.authClient = platformNode.authClient; 
		this.collector = new Collector(this.platformInstanceID);
		if(enableLogging)
			this.enableLogging();

		logger.log("Initialized system domain with ID = " + this.platformInstanceID + " on " + platformNode);
	}

	/**
	 * Schedules log shipment to the central service
	 */
	private void enableLogging()
	{
		Runnable sendLogs = new Runnable() {
			public void run()
			{
				collector.sendLogs();
				workerThreadPool.schedule(this, LoggingConstants.sendLogsPeriod, TimeUnit.MILLISECONDS);
			}
		};
		if(shippingLogs)
			this.workerThreadPool.schedule(sendLogs, LoggingConstants.sendLogsPeriod, TimeUnit.MILLISECONDS);

		// if this is the starter PI, it will log the node launch event
		if(this.predEid == null)
		{
			Serializable[] params = {this.platformNode.hostname, this.platformNode.rootPid}; //TODO
			this.collector.submitLogRecord(0, new EventRecord(this.platformNode.nodeLaunchEid, null, null, 
					AeolusApiConstants.CREATE_NODE, params));
		}

	}

	/** SYS-D FUNCTIONALITY */


	//// HANDLE PROXY OBJECT REQUESTS /////

	/**
	 * Add a new request to the SYS-D worker thread pool
	 */
	public <T> void submitRequest(UserdRequest<T> req)
	{
		final Callable<T> reqCode = req.getCallable();
		final UserdRequest<T> _req = req;

		workerThreadPool.execute(new Runnable()
		{
			public void run()
			{
				try 
				{
					_req.setReturnValue(reqCode.call());
				}	 
				catch (Exception e) 
				{
					handleException(_req, e);
				}
				finally
				{
					finishExecution(_req);
				}
			}
		});
	}

	private void handleException(UserdRequest<?> req, Exception e)
	{
		e.printStackTrace();
		req.setException(e);		
	}

	private void finishExecution(UserdRequest<?> req)
	{	
		synchronized(req)
		{
			// notify the request sender thread that the request has been processed
			req.setDone();	
			req.notify();
		}
	}

	//// LAUNCH PI/USER-Ds /////

	/**   
	 * Launches the platform instance by running the specified application in a user domain 
	 * with the authority of this instance's pid and null labels
	 * 
	 * @throws ClassNotFoundException - application class not found
	 * @throws NoSuchMethodException - main method was not found
	 * @throws InvocationTargetException - method main throws an exception (declared or Runtime)
	 * @throws AeolusException - thrown while launching the application for any other reason
	 */
	public InfoFlowState launch(final String appName, final String arg, final List<Integer> basis)
	throws InvocationTargetException, ClassNotFoundException, NoSuchMethodException, AeolusException 
	{				
		logger.log("LAUNCH APP " + appName);

		// log the PI launch event
		piLaunchEid = new EventID(platformInstanceID.getValue(), 0, 0, 0);
		System.out.println("creating launch log event");
		Serializable[] params = { appName, arg, platformInstanceID , pid };
		collector.submitLogRecord(0, new EventRecord(piLaunchEid, predEid, platformNode.nodeLaunchEid, 
				AeolusApiConstants.PI_LAUNCH, params));
		final SystemDomain par = this;
		Callable<InfoFlowState> newUserD = new Callable<InfoFlowState>()
		{	
			public InfoFlowState call() throws ClassNotFoundException, NoSuchMethodException, 
			InvocationTargetException, IllegalAccessException
			{
				UserProxy po = (UserProxy)UserProxyMap.getMyProxyObject();
				if(po == null) throw new InternalError("User-D Proxy Object was null");

				AeolusClassLoader classLoader = new AeolusClassLoader(par);
				Class<?> app_class = classLoader.loadClass(appName, true);
				String[] argv = {arg};
				Class<?>[] argType = {argv.getClass()};
				Method m = app_class.getMethod("main", argType);
				
				InfoFlowState newState = new InfoFlowState(pid, new Label(), new Label());
				newState.basis = basis;
				po.invoke(newState, m, argv);

				return po.getIFState();
			}
		};

		Future<InfoFlowState> future = this.userDomainPool.submit(newUserD);

		// get the results after execution is done
		try {
			return future.get();
		} catch (ExecutionException e) {
			if(e.getCause() instanceof ClassNotFoundException){
				throw (ClassNotFoundException) e.getCause();
			}if(e.getCause() instanceof NoSuchMethodException){
				throw (NoSuchMethodException) e.getCause();
			}if(e.getCause() instanceof InvocationTargetException){
				throw (InvocationTargetException) e.getCause();
			}

			throw new AeolusException("Could not run the application ", e.getCause());
		} catch (Exception e) { // task was canceled or interrupted
			throw new AeolusException("Could not run the application: " + e.getMessage(), e);
		}
	}

	/**
	 * Shuts down the platform instance
	 * @param shutdownImmediately - if true, all running threads will be interrupted immediately;
	 * otherwise, will allow already submitted requests to run to completion but will not accept any new requests 
	 */
	public void shutdown(boolean shutdownImmediately, EventID predEid)
	{
		// log the PI shutdown event
		if(enableLogging){
			EventID pishutdownEid = new EventID(this.platformInstanceID.getValue(), 0, 0, 1);
			Serializable[] params = {this.platformInstanceID};
			this.collector.submitLogRecord(0, new EventRecord(pishutdownEid, predEid, this.piLaunchEid, 
					AeolusApiConstants.PI_SHUTDOWN, params));
		}

		this.shuttingDown = true;

		if(shutdownImmediately)
			this.userDomainPool.shutdownNow();
		else
			this.userDomainPool.shutdown();
		// want to finish sending the logs operation
		this.workerThreadPool.shutdown();
		// send any remaining records
		if(enableLogging)
			this.collector.sendLogs();

		// remove itself from the node service registry
		{
			for (String service :  this.registeredServices.keySet())
			this.platformNode.removeRegisteredPIService(service);
		}
	}

	/* SYS-D FUNCTIONALIY */

	/**
	 * Launches a platform instance on a remote node specified by the hostname
	 * The platform instance starts running the given application
	 * @throws RpcException, AeolusException 
	 * @throws NoSuchMethodException 
	 * @throws ClassNotFoundException 
	 * @throws InvocationTargetException 
	 */
	// public void remoteLaunch(PrincipalID callerPid, String hostname, PrincipalID pid, String appName, String params, EventID opEid) 
	public RpcReply remoteLaunch(PrincipalID callerPid, InfoFlowState CalleeState, String hostname, PrincipalID pid, String appName, String params, EventID opEid) 
	throws RpcException, AeolusException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException
	{
		try 
		{
			Registry registry;
			try
			{
				registry = LocateRegistry.getRegistry(hostname, rpcPort);
			}
			catch (RemoteException e)
			{
				throw new RpcException("Could not locate registry on host = " + hostname, e);
			}

			AeolusRpcInterface service = (AeolusRpcInterface) 
			registry.lookup(aeolusRpcServiceName);
 			RpcToken rpcToken = new RpcToken(CalleeState, opEid,
                                   this.platformNode.authClient.getCurrentInvalidationNumber(),
                                   this.platformNode.nodeId);
			RpcReply reply = service.launchRemotePI(callerPid, appName, params, pid, opEid, rpcToken);
                        this.authClient.getUpToInvalidation(reply.token.invalidationNumber);
			return reply;

		} 
		catch (RemoteException e) 
		{
			// e is actually a sub-type of RemoteException (e.g. ServerException)
			// only propagate the defined RpcExceptions
			if(e.getCause() instanceof RpcException)
			{
				throw (RpcException) e.getCause();
			}	
			throw new AeolusException(e);
		}
		catch (NotBoundException e)
		{
			throw new AeolusException(e);
		}
	}

	/**
	 * Shuts down a platform instance on a remote node specified by the hostname
	 * Any processes running in the PI will be terminated immediately
	 * @throws AeolusException 
	 */
	public void remoteShutdown(String hostname, PlatformInstanceID platformInstanceID, PrincipalID callerPid, 
			boolean shutdownImmediately, EventID opEid) throws RpcException, AeolusException{
		try {
			Registry registry;
			try
			{
				registry = LocateRegistry.getRegistry(hostname, rpcPort);         
			}
			catch (RemoteException e)
			{
				throw new RpcException("Could not locate registry on host = " + hostname, e);
			}

			AeolusRpcInterface service = (AeolusRpcInterface) registry.lookup(aeolusRpcServiceName);

			service.shutdownRemotePI(platformInstanceID, callerPid, shutdownImmediately, opEid);

		} 
		catch (RemoteException e) 
		{
			if(e.getCause() instanceof RpcException)
			{
				throw (RpcException) e.getCause();
			}	
			throw new AeolusException(e);
		}
		catch (NotBoundException e)
		{
			throw new AeolusException(e);
		}
			}

	//// FORKS  ////

	/**   
	 * Launch code in a new user domain thread with the provided info flow state state
	 */
	public void fork(final InfoFlowState startingState, final Runnable codeObj, final EventID userPredEid)
	{
		Runnable newUserD = new Runnable()
		{	
			public void run()
			{
				UserProxy po = (UserProxy)UserProxyMap.getMyProxyObject();
				if(po == null) throw new InternalError("User-D Proxy Object was null");
				try 
				{
					po.invoke(startingState, codeObj, userPredEid);
				} 
				catch (Exception e) 
				{
					e.printStackTrace();
				}
			}
		};

		this.userDomainPool.execute(newUserD);
	}

	/** 
	 * Launch code in a new user domain
	 * Returns the final IFC state and AeolusCallable object state
	 * @throws ExecutionException - any exception thrown while executing the code (declared or runtime)
	 * 								the actual exception can be accessed through getCause()
	 */
	public InfoFlowState callClosure(final InfoFlowState startingState, final Runnable codeObj,
			final EventID userPredEid) throws ExecutionException
			{
		Callable<InfoFlowState> newUserD = new Callable<InfoFlowState>()
		{	
			public InfoFlowState call() throws Exception
			{
				UserProxy po = (UserProxy)UserProxyMap.getMyProxyObject();
				if(po == null) throw new InternalError("User-D Proxy Object was null");
				po.invoke(startingState, codeObj, userPredEid);
				return po.getIFState();
			}
		};

		Future<InfoFlowState> future = this.userDomainPool.submit(newUserD);
		try 
		{
			return future.get();
		}
		catch (ExecutionException e) // exceptions from code invocation
		{
			throw e;
		}
		catch (Exception e) // task canceled/ interrupted
		{
			throw new RuntimeException("Could not call closure: " + e.getMessage(), e);
		}
			}

	//// RPCs ////

	/**
	 * Register a new service with the SYS-D
	 * If the service with same name existed, it will be overwritten
	 */
	public synchronized void addEntryToServiceMap(String serviceName, Class<?> service, EventID registerEid, PrincipalID pid, List<Integer> basis)
	{	
		this.registeredServices.put(serviceName, new RegisteredService(serviceName, service, registerEid, pid, basis));
		this.platformNode.registerServiceAtPI(serviceName, this);
	}

	/**
	 * Outgoing
	 * Sends an RPC to a different Aeolus node using Java RMI
	 * @throws RpcException - any exception thrown by the invoked service method 
	 * 						  or caused by host/service/method missing
	 */
	public RpcReply sendRPC(String remoteHostName, String serviceName,
			Method method, Serializable[] parameters,
			InfoFlowState callerState, EventID opEid) throws RpcException {
		try {
			Registry registry;
			try {
				registry = LocateRegistry.getRegistry(remoteHostName, rpcPort);
			} catch (RemoteException e) {
				throw new RpcException("Could not locate registry on host = "
						+ remoteHostName, e);
			}

			AeolusRpcInterface service = (AeolusRpcInterface) registry
					.lookup(aeolusRpcServiceName);

			return (RpcReply) service.invokeRemoteMethod(serviceName, method
					.getName(), method.getParameterTypes(), parameters,
					new RpcToken(callerState, opEid,
							this.platformNode.authClient
									.getCurrentInvalidationNumber(),
							this.platformNode.nodeId));
		} catch (RemoteException e) {
			// e is actually a sub-type of RemoteException (e.g.
			// ServerException)
			// only propagate the defined RpcExceptions
			if (e.getCause() instanceof RpcException) {
				throw (RpcException) e.getCause();
			}
			throw new RuntimeException(e);
		} catch (NotBoundException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Incoming
	 * Receive an RPC from a different Aeolus node
	 * Invokes a given method of a given service inside a new User-D with the given starting info flow state
	 * Returns the result of the invoked method and the final info flow state
	 * 
	 * @param className - name of user service
	 * @param methodName - name of method to invoke
	 * @param parameterTypes - parameter types to identify the method
	 * @param parameterValues - values to invoke the method with
	 * @throws RpcException - no service was found, service does not support the method, 
	 * 						  exception thrown during specified method invocation;
	 * 						  the cause can be retrieved using getCause()	
	 */
	@SuppressWarnings("unchecked")
	public synchronized RpcReply receiveRPC(String serviceName, String methodName, Class[] parameterTypes, 
			final Object[] parameterValues, final RpcToken rpcToken) throws RpcException
			{	
		// Check if the caller is trusted..
		try {
			if(! authClient.isSystemNodeRegistered(rpcToken.callerNodeID).value)
				throw new RpcException("Caller not registered.");
		} catch (AeolusException e1) {
			throw new RpcException("Cannot check registration of the caller.");
		}

		// we need to get the actual instance of the service class
		// this service must be registered with the SYS-D
		final RegisteredService regService = this.registeredServices.get(serviceName);
		if(regService == null)
			throw new RpcException("No such service: " + serviceName, new ServiceNotFoundException("No binding for sevice " + serviceName));

		
		final Method m;
		try 
		{
			m = regService.service.getMethod(methodName, parameterTypes);
		}
		catch (NoSuchMethodException e) 
		{
			throw new RpcException("No such method: " + methodName, e);
		}

		Callable<RpcReply> newUserD = new Callable<RpcReply>()
		{
			public RpcReply call() throws InvocationTargetException, IllegalAccessException, InstantiationException
			{
				UserProxy po = (UserProxy)UserProxyMap.getMyProxyObject();

				if(po == null) throw new InternalError("User-D Proxy Object was null");

				return po.receiveRpc(rpcToken, regService, m, parameterValues);
			}
		};

		Future<RpcReply> future = this.userDomainPool.submit(newUserD);

		try 
		{
			return future.get();
		}	 
		catch (ExecutionException e)
		{
			if(e.getCause() instanceof InvocationTargetException) // method code
			{
				throw new RpcException("Exception was thrown while executing the service call", ((InvocationTargetException) e.getCause()).getCause());
			}
			if(e.getCause() instanceof InstantiationException) // method code
			{
				throw new RpcException("Could not create a new instance of the service", ((InstantiationException) e.getCause()).getCause());
			}
			else // illegal access (should not occur)
			{
				throw new RuntimeException("Could not complete receive RPC call", e);
			}
		}
		catch (Exception e) // task was canceled/interrupted
		{
			UserProxy po = (UserProxy)UserProxyMap.getMyProxyObject();
			RpcReply rr = new RpcReply(e, new RpcToken(po.getIFState(), null,
				authClient.getCurrentInvalidationNumber(),
                        	platformNode.nodeId));
			return rr;
			// throw new RuntimeException("Could not complete receive RPC call", e);
		}	
			}

	long sharedStateCount = 0;
	// public SharedUID getUID() {
	// 	return new SharedUID(sharedStateCount++);
	// }
}
