package edu.mit.csail.aeolus.runtime;

import java.lang.reflect.InvocationTargetException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Random;
import java.util.ArrayList;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;

import edu.mit.csail.aeolus.api.SystemNodeID;
import edu.mit.csail.aeolus.rpc.RpcReply;
import edu.mit.csail.aeolus.rpc.RpcToken;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.auditing.graphdb.AeolusGraphLoggingService;
import edu.mit.csail.aeolus.authority.client.AuthorityClient;
import edu.mit.csail.aeolus.authority.client.api.InterfaceAuthorityClient;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.filesystem.FileSystemClient;
import edu.mit.csail.aeolus.filesystem.FileSystemManager;
import edu.mit.csail.aeolus.infoflow.PlatformInstanceID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.rpc.AeolusRpcInterface;
import edu.mit.csail.aeolus.rpc.AeolusRpcService;
import edu.mit.csail.aeolus.stream.server.api.RegisterReturn;
import edu.mit.csail.aeolus.utilities.RMIUtil;

/**
 * Represents a node running the Aeolus platform and registered with Aeolus
 * 
 * Each node runs the authority client and possibly the file manager and the
 * logging service
 * 
 * Each node also stores a registry of RPC services running on a platform
 * instance of this node
 * 
 * @author viq
 * 
 */

public class SystemNode {

	/** Root principal Id of this node */
	public PrincipalID rootPid;
	public SystemNodeID nodeId;

	/**
	 * Map of RPC services registered at this node to the platform instances on
	 * which they run
	 */
	private ConcurrentHashMap<String, SystemDomain> registeredServicePIs;

	/** Map of all platform instances running on this node */
	private ConcurrentHashMap<PlatformInstanceID, SystemDomain> launchedPIs;

	private Registry registry;

	/** Remote service object which enables RPCs */
	private AeolusRpcInterface rpcService;

	/** Aeolus authority client responsible for talks to AS */
	public InterfaceAuthorityClient authClient;

	/** File clients */
	public ConcurrentHashMap<String, FileSystemClient> fsClients;

	public AeolusGraphLoggingService loggingService;

	// Default should never actually get used in practice but
	// it helps during testing
	public EventID nodeLaunchEid = new EventID("def-node-launch"); 

	public String hostname;

	// private String certificate;
	// private String authorityServerPublicKey;

	/**
	 * Registers the node with Aeolus and sets up Aeolus structures
	 */
	public SystemNode() throws AeolusException {
		this(true, false, false);
	}

	public SystemNode(boolean isDataNode) throws AeolusException {
		this(true, isDataNode, false);
	}

	public SystemNode(boolean useDummy, boolean isDataNode,
			boolean isLoggingNode) throws AeolusException {
		initialize(useDummy, isDataNode, isLoggingNode);
	}

	/**
	 * Launches the first PI on this node running the specified application The
	 * PI starts running with the authority of the node's root PID
	 * 
	 * @throws AeolusException
	 */
	public void start(String starterAppName, String appParams)
			throws ExecutionException, InvocationTargetException,
			ClassNotFoundException, NoSuchMethodException, AeolusException, RpcException {
		try {
			hostname = InetAddress.getLocalHost().getHostName();
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
		this.nodeLaunchEid = new EventID(this.hostname, 0, 0);
		launchPI(rootPid, rootPid, starterAppName, appParams, null, null);
	}

	/**
	 * Sets up Aeolus structures (authority client, file manager) and registers
	 * the node with the AS
	 * 
	 * @throws AeolusException
	 */
	private void initialize(boolean useDummy, boolean isDataNode,
			boolean isLoggingNode) throws AeolusException {
		this.launchedPIs = new ConcurrentHashMap<PlatformInstanceID, SystemDomain>();
		this.fsClients = new ConcurrentHashMap<String, FileSystemClient>();
		setupAuthorityClient(useDummy);
		setupRpcService();
		setupServiceRegistry();
		if (isDataNode)
			setupFileManager();

		try {
		    if (isLoggingNode)
			setupLoggingService();
		} catch (RemoteException e) {
		    throw new AeolusException(e);
		}

		registerNode();
	}

	/**
	 * Initializes the authority state client
	 */
	private void setupAuthorityClient(boolean useDummy) {
		if (useDummy)
			authClient = new AuthorityProxyStub();
		else
			authClient = new AuthorityClient();
	}

	private void setupLoggingService() throws RemoteException {
		this.loggingService = new AeolusGraphLoggingService();
		loggingService.initialize();
	}

	/**
	 * Creates an rmiregistry and binds the Aeolus service to be accessible
	 * remotely by Aeolus client running on a different Aeolus node
	 */
	private void setupRpcService() {
		try {
			registry = RMIUtil.getOrCreate(RuntimeConfig.get().getRpcPort());
		} catch (RemoteException ex) {
			ex.printStackTrace();
			throw new RuntimeException("Could not set-up the RPC service");
		}
				
		try {
			rpcService = new AeolusRpcService(this);
			AeolusRpcInterface serviceStub = (AeolusRpcInterface) UnicastRemoteObject
					.exportObject(rpcService, 0);
			registry.rebind(RuntimeConfig.get().getAeolusRpcServiceName(),
					serviceStub);
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not set-up the RPC service", e);
		}
	}

	/**
	 * Sets up the local registry for services The registry will bind the
	 * service name with the PI providing the service
	 */
	private void setupServiceRegistry() {
		registeredServicePIs = new ConcurrentHashMap<String, SystemDomain>();
	}

	/**
	 * Initializes the file manager to be running on this node
	 * 
	 * @throws RemoteException
	 */
	private void setupFileManager() throws AeolusException {
		FileSystemManager fs;
		try {
			fs = new FileSystemManager();
			fs.start();
		} catch (RemoteException e) {
			e.printStackTrace();
			throw new AeolusException(
					"Could not initialize a file system on this node", e);
		}

	}

	/**
	 * Gets the file system client for this server hostname If the client does
	 * not exist, creates a new file client for this server node
	 * 
	 * @throws AeolusException
	 */
	public FileSystemClient getFSClient(String serverHostname)
			throws AeolusException {
		FileSystemClient fsClient = this.fsClients.get(serverHostname);
		if (fsClient == null) {
			fsClient = new FileSystemClient(serverHostname);
			fsClient.register();
			this.fsClients.put(serverHostname, fsClient);
		}

		return fsClient;
	}

	/**
	 * Registers this node with the authority state server (via the authority
	 * client)
	 * 
	 * @throws AeolusException
	 */
	private void registerNode() throws AeolusException {
		try {
			// TODO: registration should return the event ID of latest authority
			// update
			RegisterReturn rr = authClient.register();
			rootPid = rr.rootPID;
			nodeId = rr.systemNodeID;
			
		} catch (AeolusException e) {
			e.printStackTrace();
			throw e;
		}
	}

	/**
	 * Shutdown the node Shuts down the running services
	 * (RPC/Logging/FileManager)
	 */
	public void shutdown() {
		if (rpcService != null) {
			try {
				UnicastRemoteObject.unexportObject(rpcService, true);
			} catch (NoSuchObjectException e) {
				System.err.println("During shutdown: no such object in unexport");
			}
		}

		if (this.loggingService != null)
			this.loggingService.shutdownService();

		// shut down all the PIs

	}

	/* PI MANAGEMENT */

	/**
	 * Create a new platform instance and launch the given application in this
	 * instance
	 */
	public RpcReply launchPI(PrincipalID callerPid, PrincipalID pid,
			String appName, String appParams, EventID opEid, RpcToken rpcToken)
			throws InvocationTargetException, ClassNotFoundException,
			NoSuchMethodException, AeolusException, RpcException {
		Random rand = new Random();
		int piId = rand.nextInt(20); // FIXME
		// int piId = authClient.getNextPlatformInstanceID();
		PlatformInstanceID id = new PlatformInstanceID(piId);
		if (rpcToken != null) {
		   try {
                           if(! authClient.isSystemNodeRegistered(rpcToken.callerNodeID).value)
                                   throw new RpcException("Caller not registered." + " " + rpcToken.callerNodeID);
                   } catch (AeolusException e1) {
                           throw new RpcException("Cannot check registration of the caller.");
                   }
                }
		SystemDomain newPI = new SystemDomain(this, pid, id, callerPid, opEid);
		this.launchedPIs.put(id, newPI);
		InfoFlowState ifs = null;
		if (rpcToken == null) ifs = newPI.launch(appName, appParams, new ArrayList<Integer>());
		else {
			ifs = newPI.launch(appName, appParams, rpcToken.state.basis);
		}
		RpcToken RpcT = new RpcToken(ifs, null,
				newPI.authClient.getCurrentInvalidationNumber(),
                        	newPI.platformNode.nodeId);
		return new RpcReply(null, RpcT);

	}

	/**
	 * Shuts down a platform instance (only one request will actually succeed in
	 * calling shutdown on a given instance)
	 */
	public void shutdownPI(PlatformInstanceID platformInstanceID,
			PrincipalID callerPid, boolean shutdownImmediately, EventID opEid)
			throws RpcException, AuthorityException {
		SystemDomain pi = this.launchedPIs.remove(platformInstanceID);
		if (pi == null) {
			throw new RpcException("The platform instance with ID = "
					+ platformInstanceID + " does not exist on hostname = "
					+ this.hostname);
		}

		if (!this.authClient.isActee(callerPid, pi.pid).value) {
			throw new AuthorityException("Principal " + callerPid
					+ " is not authorized to shutdown the platform instance "
					+ platformInstanceID.getValue());
		}

		pi.shutdown(shutdownImmediately, opEid);
	}

	/**
	 * Associate a service with the given platform instance at the node registry
	 * If the service with same name existed, it will be overwritten
	 */
	public void registerServiceAtPI(String serviceName, SystemDomain pi) {
		registeredServicePIs.put(serviceName, pi);
	}

	public void removeRegisteredPIService(String serviceName) {
		registeredServicePIs.remove(serviceName);
	}

	/**
	 * Return reference to the platform instance running this service Returns
	 * null if no such platform instance or if the platform instance is shutting
	 * down
	 */
	public SystemDomain resolvePI(String serviceName) {
		SystemDomain pi = registeredServicePIs.get(serviceName);
		if ((pi != null) && (!pi.shuttingDown)) {
			return pi;
		}
		return null;
	}
}
