package edu.mit.csail.aeolus.runtime;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.Callable;

import edu.mit.csail.aeolus.api.AeolusClosure;
import edu.mit.csail.aeolus.api.AeolusLib;
import edu.mit.csail.aeolus.api.AeolusLabel;
import edu.mit.csail.aeolus.api.AeolusShared;
import edu.mit.csail.aeolus.auditing.AeolusApiConstants;
import edu.mit.csail.aeolus.auditing.CompletionStatus;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.auditing.EventRecord;
import edu.mit.csail.aeolus.authority.api.AuthReply;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.filesystem.FileSystemClient;
import edu.mit.csail.aeolus.filesystem.FileSystemReply;
import edu.mit.csail.aeolus.infoflow.AuthStateConstants;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.InfoFlowRules;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.infoflow.PlatformInstanceID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.rpc.RegisteredService;
import edu.mit.csail.aeolus.rpc.RpcProxyInfo;
import edu.mit.csail.aeolus.rpc.RpcReply;
import edu.mit.csail.aeolus.rpc.RpcToken;
import edu.mit.csail.aeolus.security.AeolusCloner;

/**
 * Proxy object provides methods to a User-d to interact with Aeolus.
 * 
 * It exposes the Aeolus API, namely it provides methods for all Aeolus
 * functions and all external methods: modifications of authority state,
 * manipulations of labels, access to shared state, file use, I/O devices. For
 * I/O devices, PO provides a wrapper code that ensures information flow rules
 * are obeyed.
 * 
 * The PO stores the process' information flow state (principal ID, process
 * labels)
 * 
 * The PO interacts with the SYS-d (which has access to the shared state manager
 * and the authority state client).
 * 
 * API subsections: principals, tags, delegations and revocations, label
 * manipulations, authority closure, call, fork, shared state, shared object,
 * shared lock, shared queue, files, filestreams
 * 
 * viq
 * 
 */
public class UserLoggingProxy extends UserProxy {

	// Stores the latest logged user event
	private EventID lastEid;

	// Keeps a count of the number of logged user events
	private int eventCounter;

	// //// PO initialization //////

	/**
	 * Creates the proxy object
	 * 
	 * @param sysD
	 *            - the system domain associated with this proxy object
	 */
	public UserLoggingProxy(SystemDomain sysD, long myUserDomainId) {
		super(sysD, myUserDomainId);
		this.lastEid = EventID.systemEvent;
		this.eventCounter = 1;
	}

	// INTERNAL FUNCTIONALITY

	/**
	 * This is intended to clean up any state in the proxy and make it ready to
	 * be re-used (This is called when the User-d thread finishes executing and
	 * is returned to the thread pool - afterExecute() call in the AppDomainPool
	 * class)
	 */
	public void flush() {
		super.flush();
		this.lastEid = EventID.systemEvent;
		this.eventCounter = 1;
	}

	/*
	 * TODO: THESE SHOULD ONLY BE INVOKED BY AEOLUS CODE!
	 */
	/**
	 * This method creates a new EID and sets this user thread's last event
	 * pointer to that EID.
	 */
	public EventID getEID() {
		int cur = this.eventCounter++;
		EventID retValue = new EventID(sysd.platformInstanceID.getValue(),
				this.userDomainId, this.generation, cur);
		this.lastEid = retValue;
		return retValue;
	}

	public EventID getLastEID() {
		return this.lastEid;
	}

	public void submitLogRecord(EventRecord record) {
		sysd.collector.submitLogRecord(userDomainId, record);
	}

	/**
	 * Sets the initial information flow state and invokes the provided method
	 * with the given arguments To be used for static method invocation (e.g.
	 * "main" for app launch)
	 */
	public void invoke(InfoFlowState startingState, Method m, String[] argv)
			throws IllegalArgumentException, IllegalAccessException,
			InvocationTargetException {
		EventID userDLaunchEid = new EventID(
				sysd.platformInstanceID.getValue(), this.userDomainId,
				this.generation, 0);
		Serializable[] params = { startingState.pid, startingState.ILabel,
				startingState.SLabel };
		EventRecord record = new EventRecord(userDLaunchEid, null,
				sysd.piLaunchEid, AeolusApiConstants.USER_THREAD_LAUNCH, params);

		sysd.collector.submitLogRecord(this.userDomainId, record);
		this.lastEid = userDLaunchEid;

		super.invoke(startingState, m, argv);
	}

	/**
	 * Sets the info flow state and invokes the code object (e.g. used with a
	 * fork call)
	 */
	public void invoke(InfoFlowState startingState, Runnable code,
			EventID userPredEid) throws Exception {
		EventID userDLaunchEid = new EventID(
				sysd.platformInstanceID.getValue(), this.userDomainId,
				this.generation, 0);
		Serializable[] params = { startingState.pid, startingState.ILabel,
				startingState.SLabel };
		sysd.collector.submitLogRecord(this.userDomainId, new EventRecord(
				userDLaunchEid, userPredEid, sysd.piLaunchEid,
				AeolusApiConstants.USER_THREAD_LAUNCH, params));
		this.lastEid = userDLaunchEid;

		super.invoke(startingState, code, userPredEid);
	}

	// USER API

	/**
	 * Returns a copy of the current info flow state
	 */
	public InfoFlowState getInfoFlowState() {
		EventID opEid = getNextEventId();
		sysd.collector.submitLogRecord(this.userDomainId, new EventRecord(
				opEid, this.lastEid, AeolusApiConstants.GET_IFC));
		this.lastEid = opEid;

		return super.getInfoFlowState();
	}

	// AUTHORITY API

	// PRINCIPALS

	/**
	 * Creates a new principal in the process principals core and returns the
	 * new principal ID Process principal acts for the new principal
	 */
	public PrincipalID createPrincipal() throws AuthorityException,
			AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		// XXX should log the failure
		if (!this.state.SLabel.isEmpty())
			throw new InfoFlowControlException("Cannot create a new principal with non-null secrecy label");

		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid,
				AeolusApiConstants.CREATE_PRINCIPAL);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CREATE_PRINCIPAL_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<PrincipalID> reply = sysd.authClient.createPrincipal(
					state.pid, opEid);
			lr.setReturnValue(reply.value);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			return reply.value;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Creates a new principal in a new core and returns the new principal ID
	 * Process principal acts for the new principal
	 */
	public PrincipalID createPrincipalInNewCore() throws AuthorityException,
			AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		// XXX should log the failure
		if (!this.state.SLabel.isEmpty())
			throw new InfoFlowControlException("Cannot create a new principal with non-null secrecy label");

		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid,
				AeolusApiConstants.CREATE_PRINCIPAL_NEWCORE);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CREATE_PRINCIPAL_NEWCORE_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<PrincipalID> reply = sysd.authClient
					.createPrincipalInNewCore(state.pid, opEid);
			lr.setReturnValue(reply.value);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			return reply.value;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Creates a new principal in the same core as principal p and returns the
	 * new principal ID Process principal acts for the new principal
	 */
	public PrincipalID createPrincipal(PrincipalID p)
			throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] paramValues = { p };
		EventRecord lr = new EventRecord(opEid, this.lastEid,
				AeolusApiConstants.CREATE_PRINCIPAL);
		lr.setParams(paramValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CREATE_PRINCIPAL_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<PrincipalID> reply = sysd.authClient.createPrincipal(p,
					opEid);
			lr.setReturnValue(reply.value);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			return reply.value;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	// TAGS API

	/**
	 * Creates a new top-level tag and returns its ID Process principal has
	 * authority for the new tag
	 */
	public TagID createTag() throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid,
				AeolusApiConstants.CREATE_TAG);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CREATE_TAG_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<TagID> reply = sysd.authClient.createTaint(state.pid,
					opEid);
			lr.setReturnValue(reply.value);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			return reply.value;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Creates a sub-tag of tag t and returns the new sub-tag ID Tag t must be a
	 * top-level tag Process principal has authority for the new tag
	 */
	public TagID createSubTag(TagID t) throws AuthorityException,
			AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { t };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.CREATE_SUBTAG, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CREATE_SUBTAG_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<TagID> reply = sysd.authClient.createSubTag(t, state.pid,
					opEid);
			TagID tag = new TagID(reply.value.getValue(), t.getValue());
			lr.setReturnValue(reply.value);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			return reply.value;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	// DELEGATIONS AND REVOCATIONS API

	/**
	 * Adds an act-for link from actor principal p2 to actee principal p1
	 * Process principal must act for principal p1
	 */
	public void actFor(PrincipalID p1, PrincipalID p2)
			throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { p1, p2 };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.ACT_FOR, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.ACT_FOR_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<Void> reply = sysd.authClient.delegateActsFor(state.pid,
					p2, p1, opEid);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Removes the act-for link from principal p2 to principal p1 Process
	 * principal must act for principal p1
	 */
	public void revokeActFor(PrincipalID p1, PrincipalID p2)
			throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { p1, p2 };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.REVOKE_ACT_FOR, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.REVOKE_ACT_FOR_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<Void> reply = sysd.authClient.revokeActsFor(state.pid,
					p2, p1, opEid);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Gives authority for tag t from grantor principal p1 to grantee principal
	 * p2 Process principal must act for the grantor principal p1, the grantor
	 * principal p1 must have authority for tag t
	 */
	public void delegate(TagID t, PrincipalID p1, PrincipalID p2)
			throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { t, p1, p2 };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.DELEGATE, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.DELEGATE_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<Void> reply = sysd.authClient.grant(state.pid, p1, p2, t,
					opEid);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Revokes authority for tag t from grantor principal p1 to grantee
	 * principal p2 Process principal must act for the grantor principal p1 and
	 * the tag t must have been previously delegated from p1 to p2
	 */
	public void revokeDelegate(TagID t, PrincipalID p1, PrincipalID p2)
			throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { t, p1, p2 };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.REVOKE_DELEGATE, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.REVOKE_DELEGATE_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<Void> reply = sysd.authClient.revokeGrant(state.pid, p1,
					p2, t, opEid);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	// LABEL MANIPULATIONS API

	/**
	 * Adds given tag id to the state secrecy label
	 */
	public void addSecrecy(TagID tid) throws AeolusException {
		ArrayList<TagID> tids = new ArrayList<TagID>();
		tids.add(tid);
		addSecrecy(tids);
	}

	/**
	 * Adds given tag ids to the state secrecy label
	 * 
	 * @param tids
	 *            : list of tag ids to add
	 */
	public void addSecrecy(List<TagID> tids) throws AeolusException {

		if (tids == null)
			return;
		EventID opEid = getNextEventId();
		submitLogRecord(new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.ADD_SECRECY, parametrizeTags(tids)));
		this.lastEid = opEid;

		for (TagID tid : tids)
			// this.state.SLabel.addTag(tid);
			super.addSecrecy(tid);
	}

	/**
	 * Removes given tag id from the state integrity label
	 */
	public void removeIntegrity(TagID tid) throws AeolusException {
		List<TagID> list = new ArrayList<TagID>();
		list.add(tid);
		removeIntegrity(list);
	}

	/**
	 * Make sure to use this method when changing tags for removal/addition.
	 * Reason: to make sure that the type of the parameters in the returned
	 * Serializable[] is the same/easy to change in consistency with the
	 * auditing specification (DeclassifyEvent, EndorseEvent, ...).
	 * 
	 * 
	 * @param tids
	 * @return
	 */
	private Serializable[] parametrizeTags(List<TagID> tids) {
		if (!(tids instanceof Serializable))
			tids = new ArrayList<TagID>(tids);
		return new Serializable[] { (Serializable) tids };
	}

	/**
	 * Removes given tag ids from the state integrity label
	 * 
	 * @param tids
	 *            : list of tag ids to remove
	 */
	public void removeIntegrity(List<TagID> tids) throws AeolusException {
		// try {
			if (tids == null)
				return;

			EventID opEid = getNextEventId();
			submitLogRecord(new EventRecord(opEid, this.lastEid, null,
					AeolusApiConstants.REMOVE_INTEGRITY, parametrizeTags(tids)));
			this.lastEid = opEid;

			for (TagID tid : tids)
				super.removeIntegrity(tid);
	}

	/**
	 * Removes given tag ID from the secrecy label if process is authorized
	 * Authority is checked via a call to the SYS-d Will not declassify if the
	 * process is running as Ppublic
	 */
	public void declassify(TagID tid) throws AeolusException {
		ArrayList<TagID> tags = new ArrayList<TagID>();
		tags.add(tid);
		declassify(tags);
	}

	/**
	 * Removes given tag ids from the secrecy label if process is authorized
	 * Authority is checked via a call to the SYS-d Will not declassify if the
	 * process is running as Ppublic
	 * 
	 * @param tids
	 *            : list of tag ids to remove
	 */
	public void declassify(List<TagID> tids) throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.DECLASSIFY, parametrizeTags(tids));
		this.lastEid = opEid;

		if (tids == null)
			return;

		AuthReply<Boolean> reply = sysd.authClient.hasAuthority(state.basis,
				state.pid, tids);
		lr.setCommunicationPred(reply.event);
		if (reply.value) {
			for (TagID tid : tids)
				state.SLabel.removeTag(tid);

			this.checkReadFileStreams();
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} else {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			throw new AuthorityException("No authority to declassify as PID = "
					+ state.pid.getValue());
		}
	}

	/**
	 * Adds given tag ID to the integrity label if process is authorized
	 * Authority is checked via a call to the SYS-d Will not endorse if the
	 * process is running as Ppublic
	 */
	public void endorse(TagID tid) throws AuthorityException, AeolusException {
		ArrayList<TagID> tids = new ArrayList<TagID>();
		tids.add(tid);
		try {
			endorse(tids);
		} catch (AuthorityException e) {
			throw e;
		}
	}

	/**
	 * Adds given tag ids to the integrity label if process is authorized
	 * Authority is checked via a call to the SYS-d Will not endorse if the
	 * process is running as Ppublic
	 * 
	 * @param tids
	 *            : list of tag ids to add
	 */
	public void endorse(final List<TagID> tids) throws AuthorityException, AeolusException {
		if (tids == null)
			return;

		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { new ArrayList(tids) };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.ENDORSE, parameterValues);
		this.lastEid = opEid;

		AuthReply<Boolean> reply = sysd.authClient.hasAuthority(state.basis,
				state.pid, tids);
		lr.setCommunicationPred(reply.event);
		if (reply.value) {
			for (TagID tid : tids)
				this.state.ILabel.addTag(tid);

			this.checkReadFileStreams();
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} else {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			throw new AuthorityException("No authority to endorse as PID = "
					+ state.pid.getValue());
		}
	}

	// AUTHORITY CLOSURE API

	/**
	 * Creates a new authority closure and returns its closure ID The closure ID
	 * is bound to the provided interface A new anonymous principal PA is
	 * generated and also bound to this closure ID, This principal has no
	 * authority
	 */
	// public ClosureID createClosure(
	// Class<? extends AeolusCallable> closureInterface)
	// throws AeolusException {
	// // make sure this is an interface
	// if (!closureInterface.isInterface())
	// throw new AeolusException("Requires an interface");
	//
	// EventID opEid = getNextEventId();
	// Serializable[] parameterValues = { closureInterface.getName() };
	// EventRecord lr = new EventRecord(opEid, this.lastEid, null,
	// AeolusApiConstants.CREATE_CLOSURE, parameterValues);
	// this.lastEid = opEid;
	// sysd.collector.submitLogRecord(this.userDomainId, lr);
	//
	// EventID replyEid = getNextEventId();
	// lr = new EventRecord(replyEid, this.lastEid,
	// AeolusApiConstants.CREATE_CLOSURE_REPLY);
	// this.lastEid = replyEid;
	// try {
	// AuthReply<ClosureReply> reply = sysd.authClient.createClosure(
	// closureInterface.getName(), opEid);
	// lr.setReturnValue(reply.value.cid);
	// lr.setCommunicationPred(reply.event);
	// sysd.collector.submitLogRecord(this.userDomainId, lr);
	// return reply.value.cid;
	// } catch (Exception e) {
	// lr.setCompletionStatus(CompletionStatus.FAILED);
	// sysd.collector.submitLogRecord(this.userDomainId, lr);
	// throw new AeolusException(e);
	// }
	// }

	// public AeolusCallable getClosure(ClosureID cid, Class<?> closureClass)
	// throws AeolusException {
	// EventID opEid = getNextEventId();
	// Serializable[] parameterValues = { cid };
	// EventRecord lr = new EventRecord(opEid, this.lastEid, null,
	// AeolusApiConstants.GET_CLOSURE, parameterValues);
	// this.lastEid = opEid;
	//
	// AuthReply<ClosureTableEntry> reply = sysd.authClient
	// .getClosureInfo(cid);
	// lr.setCommunicationPred(reply.event);
	// ClosureTableEntry closureInfo = reply.value;
	// String interfaceName = closureInfo.className;
	// try {
	// Class<?> closureInterface = Class.forName(interfaceName);
	//
	// // need to check that this class implements the closure interface
	// Class<?>[] interfaces = closureClass.getInterfaces();
	// boolean found = false;
	// for (int i = 0; i < interfaces.length; i++) {
	// if (interfaces[i].equals(closureInterface)) {
	// found = true;
	// break;
	// }
	// }
	// if (!found)
	// throw new Exception(
	// "Cannot provide closure object: closureClass must implemented the interface bound to this closure");
	//
	// sysd.collector.submitLogRecord(this.userDomainId, lr);
	// return (AeolusCallable) closureClass.newInstance();
	// } catch (ClassNotFoundException e) {
	// lr.setCompletionStatus(CompletionStatus.FAILED);
	// sysd.collector.submitLogRecord(this.userDomainId, lr);
	// throw new AeolusException("Closure class was not found "
	// + interfaceName, e);
	// } catch (Exception e) {
	// lr.setCompletionStatus(CompletionStatus.FAILED);
	// sysd.collector.submitLogRecord(this.userDomainId, lr);
	// throw new AeolusException("Could not create closure with name "
	// + interfaceName, e);
	// }
	// }

	/**
	 * Adds an act-for link from the anonymous principal of closure cid to
	 * principal pid (the anonymous principal will be authoritative for all tags
	 * that principal pid has authority for) Closure cid must exist and process
	 * principal must act for principal pid
	 */
	public void closureActFor(PrincipalID pid, ClosureID cid)
			throws AuthorityException, AeolusException {
		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { pid, cid };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.CLOSURE_ACT_FOR, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CLOSURE_ACT_FOR_REPLY);
		this.lastEid = replyEid;
		try {
			AuthReply<Void> reply = sysd.authClient.closureActsFor(state.pid,
					cid, pid, opEid);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Gives authority for tag tid from grantor principal pid to grantee
	 * anonymous principal of closure cid Closure CL1must exist, process
	 * principal must act for the grantor principal pid, the grantor principal
	 * pid must have authority for tag tid
	 */
	public void closureDelegate(TagID tid, PrincipalID pid, ClosureID cid)
			throws AuthorityException, AeolusException {
		EventID opEid = getNextEventId();
		Serializable[] parameterValues = { tid, pid, cid };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.CLOSURE_DELEGATE, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CLOSURE_DELEGATE_REPLY);
		this.lastEid = replyEid;

		try {
			AuthReply<Void> reply = sysd.authClient.closureGrant(state.pid,
					cid, pid, tid, opEid);
			lr.setCommunicationPred(reply.event);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	// //// CALL, FORK API //////

	/**
	 * Runs code with different authority
	 * 
	 * Saves current state, then invokes the codeObj with the calleePid After
	 * invocation, the state is restored, but the process labels are updated to
	 * reflect any additional contamination picked up during the call
	 * 
	 * @param codeObj
	 *            - method to run
	 * @param calleePid
	 *            - pid to run the method with
	 */
	public <T> T call(Callable<T> codeObj, PrincipalID calleePid)
			throws AuthorityException, Exception {
		Serializable[] parameterValues = { calleePid };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.CALL, parameterValues);
		this.lastEid = opEid;

		if (calleePid.getValue() != AuthStateConstants.pidPublic) {
			AuthReply<Boolean> reply = sysd.authClient.canSwitch(state.basis,
					state.pid, calleePid);
			lr.setCommunicationPred(reply.event);
			if (!reply.value) {
				lr.setCompletionStatus(CompletionStatus.FAILED);
				sysd.collector.submitLogRecord(this.userDomainId, lr);
				throw new AuthorityException("Call cannot switch to PID = "
						+ calleePid);
			}

			sysd.collector.submitLogRecord(this.userDomainId, lr);
			this.state.basis.add(calleePid.getValue());
		} else {
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		}

		PrincipalID callerPid = this.state.pid;
		this.state.pid = calleePid;
		T t = null;
		try {
			t = codeObj.call();
		} finally {
			// restore state
			if(calleePid.getValue() != AuthStateConstants.pidPublic)
			    this.state.basis.remove(this.state.basis.size() - 1);
			this.state.pid = callerPid;

			EventID replyEid = getNextEventId();
			EventRecord rec = new EventRecord(replyEid, this.lastEid,
					AeolusApiConstants.CALL_REPLY);
			rec.setCommunicationPred(opEid);
			sysd.collector.submitLogRecord(this.userDomainId, rec);
			this.lastEid = replyEid;
		}
		return (t);
	}

	/**
	 * Checks that the current process' labels and PID allow the fork to be made
	 * If there are no violations, a new User-d is set up to run the specified
	 * method
	 * 
	 * The code object is copied to the new process
	 * 
	 * The actual fork request is handled in Sys-d
	 * 
	 * @param codeObj
	 *            - method to invoke
	 * @param pid
	 *            of - principal to switch to
	 */
	public void fork(Runnable codeObj, PrincipalID pid)
			throws AuthorityException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		Serializable[] parameterValues = { pid };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.FORK, parameterValues);
		this.lastEid = opEid;

		if (pid.getValue() != AuthStateConstants.pidPublic) {
			AuthReply<Boolean> reply = sysd.authClient.canSwitch(state.basis,
					state.pid, pid);
			lr.setCommunicationPred(reply.event);

			if (!reply.value) {
				lr.setCompletionStatus(CompletionStatus.FAILED);
				sysd.collector.submitLogRecord(this.userDomainId, lr);
				throw new AuthorityException("Fork cannot switch to PID = "
						+ pid);
			}

			sysd.collector.submitLogRecord(this.userDomainId, lr);
		}

		// Runnable codeObjCopy = (Runnable) HelperUtils.deepCopy(this,
		// codeObj);
		Runnable codeObjCopy = AeolusCloner.Clone(codeObj);
		InfoFlowState forkedState = new InfoFlowState(this.state);
		forkedState.pid = pid;
		forkedState.basis.add(pid.getValue());

		sysd.fork(forkedState, codeObjCopy, opEid);
	}

	/**
	 * Given a number of call frames up from the caller of this method, return
	 * the name of the method that many call frames up.
	 */
	private String getCallerMethodName(int frames) {
		return (new Throwable()).getStackTrace()[frames + 1].getMethodName();
	}

	public InfoFlowState aeolusStartClosureCall(AeolusClosure closure) {
		/* XXX need to ensure that only AeolusClosure can call this */
	        /* XXX the first elt in paramterValues is supposed to be the
		 * closure ID, but we don't have closure IDs anymore
		 */
		Serializable[] parameterValues = { 0, closure.getPID().getPrincipalID() };
		EventID opEid = getNextEventId();
		submitLogRecord(new EventRecord(opEid, this.lastEid, closure.closureLastEid,
				AeolusApiConstants.CALL_CLOSURE, parameterValues));
		this.lastEid = opEid;
		closure.closureLastEid = opEid;

		return (super.aeolusStartClosureCall(closure));
	}

	public void aeolusEndClosureCall(AeolusClosure closure,
			InfoFlowState oldState) {
		/* XXX need to ensure that only AeolusClosure can call this */

		// XXX original code didn't make a new eid here...should we?

		super.aeolusEndClosureCall(closure, oldState);

		// return event
		EventID opEid = getNextEventId();
		submitLogRecord(new EventRecord(opEid, this.lastEid, closure.closureLastEid,
				AeolusApiConstants.CALL_CLOSURE_REPLY, null));
		this.lastEid = opEid;
		closure.closureLastEid = opEid;
	}

	public InfoFlowState aeolusStartSharedCall(AeolusShared shared) throws InfoFlowControlException {
		/* XXX need to ensure that only AeolusShared can call this */
		EventID opEid = getNextEventId();
		submitLogRecord(new EventRecord(opEid, this.lastEid,
				shared.sharedLastEid, AeolusApiConstants.SHARED_CALL,
				getCallerMethodName(2), null));
		this.lastEid = opEid;
		shared.sharedLastEid = opEid;

		return (super.aeolusStartSharedCall(shared));
	}

	public void aeolusEndSharedCall(AeolusShared shared, InfoFlowState oldState) {
		EventID opEid = getNextEventId();

		super.aeolusEndSharedCall(shared, oldState);

		// return event
		submitLogRecord(new EventRecord(opEid, this.lastEid,
				shared.sharedLastEid, AeolusApiConstants.SHARED_RETURN,
				getCallerMethodName(2), null));
		this.lastEid = opEid;
		shared.sharedLastEid = opEid;
	}

	public InfoFlowState aeolusStartSharedRead(AeolusShared shared) throws InfoFlowControlException {
		/* XXX need to ensure that only AeolusShared can call this */
		EventID opEid = getNextEventId();
		submitLogRecord(new EventRecord(opEid, this.lastEid,
				shared.sharedLastEid, AeolusApiConstants.SHARED_CALL,
				getCallerMethodName(2), null));
		this.lastEid = opEid;
		shared.sharedLastEid = opEid;

		return (super.aeolusStartSharedRead(shared));
	}

	public void aeolusEndSharedRead(AeolusShared shared, InfoFlowState oldState) {
		EventID opEid = getNextEventId();

		super.aeolusEndSharedRead(shared, oldState);

		// return event
		submitLogRecord(new EventRecord(opEid, this.lastEid,
				shared.sharedLastEid, AeolusApiConstants.SHARED_RETURN,
				getCallerMethodName(2), null));
		this.lastEid = opEid;
		shared.sharedLastEid = opEid;
	}

	public InfoFlowState aeolusStartSharedWrite(AeolusShared shared) throws InfoFlowControlException {
		/* XXX need to ensure that only AeolusShared can call this */
		EventID opEid = getNextEventId();
		submitLogRecord(new EventRecord(opEid, this.lastEid,
				shared.sharedLastEid, AeolusApiConstants.SHARED_CALL,
				getCallerMethodName(2), null));
		this.lastEid = opEid;
		shared.sharedLastEid = opEid;

		return (super.aeolusStartSharedWrite(shared));
	}

	public void aeolusEndSharedWrite(AeolusShared shared, InfoFlowState oldState) {
		EventID opEid = getNextEventId();

		super.aeolusEndSharedWrite(shared, oldState);

		// return event
		sysd.collector
				.submitLogRecord(this.userDomainId, new EventRecord(opEid,
						this.lastEid, shared.sharedLastEid,
						AeolusApiConstants.SHARED_RETURN,
						getCallerMethodName(2), null));
		this.lastEid = opEid;
		shared.sharedLastEid = opEid;
	}

	// //// File API //////

	/**
	 * Creates a new file
	 */
	public boolean createFile(String hostname, String filePath,
			Label sLabelFile, Label iLabelFile) throws FileNotFoundException,
			IOException, InfoFlowControlException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		Serializable[] parameterValues = { hostname, filePath, sLabelFile,
				iLabelFile };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.CREATE_FILE, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CREATE_FILE_REPLY);
		this.lastEid = replyEid;
		try {
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<Boolean> reply = fsClient.createFile(filePath,
					sLabelFile, iLabelFile, AeolusLib.getSecrecy().getLabel(), AeolusLib.getIntegrity().getLabel(), opEid);

			lr.setCommunicationPred(reply.predecessorEventId);
			lr.setReturnValue(reply.opResult);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			return reply.opResult;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}

	}

	/**
	 * Returns the contents of the given file
	 */
	public int readFile(String hostname, String filePath, byte[] fileContent)
			throws FileNotFoundException, IOException,
			InfoFlowControlException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.READ_FILE, parameterValues);
		this.lastEid = opEid;

		try {
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<Integer> reply = fsClient.readFile(filePath,
					fileContent, state.SLabel, state.ILabel, opEid);
			lr.setFsBeforeTime(reply.beforeTime);
			lr.setFsAfterTime(reply.afterTime);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			return reply.opResult;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Writes to the given file
	 */
	public void writeFile(String hostname, String filePath,
			final byte[] fileContent, boolean appendMode)
			throws FileNotFoundException, IOException,
			InfoFlowControlException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.WRITE_FILE, parameterValues);
		this.lastEid = opEid;

		try {
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<Void> reply = fsClient.writeFile(filePath,
					fileContent, appendMode, AeolusLib.getSecrecy().getLabel(), AeolusLib.getIntegrity().getLabel(), opEid);
			lr.setFsBeforeTime(reply.beforeTime);
			lr.setFsAfterTime(reply.afterTime);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Creates a new directory
	 */
	public boolean createDir(String hostname, String filePath,
			Label sLabelFile, Label iLabelFile) throws FileNotFoundException,
			IOException, InfoFlowControlException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		Serializable[] parameterValues = { hostname, filePath, sLabelFile,
				iLabelFile };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.CREATE_DIRECTORY, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.CREATE_DIRECTORY_REPLY);
		this.lastEid = replyEid;

		try {
			//TODO state.slabel is not the same as aeoluslib.getsecrecy!!!
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<Boolean> reply = fsClient.createDir(filePath,
					sLabelFile, iLabelFile, AeolusLib.getSecrecy().getLabel(), AeolusLib.getIntegrity().getLabel(), opEid);

			lr.setReturnValue(reply.opResult);
			lr.setCommunicationPred(reply.predecessorEventId);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			return reply.opResult;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Removes a directory
	 */
	public boolean delete(String hostname, String filePath)
			throws FileNotFoundException, IOException,
			InfoFlowControlException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.REMOVE_DIRECTORY, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.REMOVE_DIRECTORY_REPLY);
		this.lastEid = replyEid;
		try {
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<Boolean> reply = fsClient.delete(filePath,
					state.SLabel, state.ILabel, null);

			lr.setReturnValue(reply.opResult);
			lr.setCommunicationPred(reply.predecessorEventId);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			return reply.opResult;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Returns the list of all files in a given directory
	 * 
	 * @return array of strings
	 */
	public String[] listDir(String hostname, String filePath)
			throws FileNotFoundException, IOException,
			InfoFlowControlException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		Serializable[] parameterValues = { filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.LIST_DIRECTORY, parameterValues);
		this.lastEid = opEid;
		try {
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<String[]> reply = fsClient.listDir(filePath,
					state.SLabel, state.ILabel, null);
			lr.setCommunicationPred(reply.predecessorEventId);
			lr.setFsBeforeTime(reply.beforeTime);
			lr.setFsAfterTime(reply.afterTime);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			return reply.opResult;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}
	}

	public Label getIntegrityLabel(String hostname, String filePath)
			throws FileNotFoundException, IOException,
			InfoFlowControlException, AeolusException {
		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.GET_INTEGRITY_LABEL, parameterValues);
		this.lastEid = opEid;
		try {
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<Label> reply = fsClient.getIntegrityLabel(filePath,
					state.SLabel, state.ILabel, null);

			lr.setCommunicationPred(reply.predecessorEventId);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			return reply.opResult;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}
	}

	public Label getSecrecyLabel(String hostname, String filePath)
			throws FileNotFoundException, IOException,
			InfoFlowControlException, AeolusException {
		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.GET_SECRECY_LABEL, parameterValues);
		this.lastEid = opEid;
		try {
			FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
			FileSystemReply<Label> reply = fsClient.getSecrecyLabel(filePath,
					state.SLabel, state.ILabel, null);

			lr.setCommunicationPred(reply.predecessorEventId);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			return reply.opResult;
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof FileNotFoundException)
				throw (FileNotFoundException) e;
			else if (e instanceof IOException)
				throw (IOException) e;
			else if (e instanceof InfoFlowControlException)
				throw (InfoFlowControlException) e;
			else
				throw new AeolusException(e);
		}
	}

	// //// Filestream API //////

	public FileSystemReply<Integer> openFileInputStream(String hostname,
			String filePath) throws AeolusException, RemoteException,
			FileNotFoundException, IOException {
		checkAuthorityEnabled(); // XXX should log on failure

		// TODO: fix the error scenario
		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);

		FileSystemReply<Integer> reply = fsClient.openFileInputStream(filePath,
				AeolusLib.getSecrecy().getLabel(), AeolusLib.getIntegrity().getLabel(), null);

		int handle = reply.opResult;
		filestreamGenerations.put(handle, labelGeneration - 1);
		readFilestreamHandles.add(handle);

		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.OPEN_WRITE_FILESTREAM, parameterValues);
		lr.setFsBeforeTime(reply.beforeTime);
		sysd.collector.submitLogRecord(this.userDomainId, lr);
		this.lastEid = opEid;

		return reply;
	}

	public FileSystemReply<Integer> openFileOutputStream(String hostname,
			String filePath, boolean appendMode) throws AeolusException,
			IOException {
		checkAuthorityEnabled(); // XXX should log on failure

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Integer> reply = fsClient.openFileOutputStream(
				filePath, appendMode, AeolusLib.getSecrecy().getLabel(), AeolusLib.getIntegrity().getLabel(),
				null);

		int handle = reply.opResult;
		filestreamGenerations.put(handle, labelGeneration - 1);
		writeFilestreamHandles.add(handle);

		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.OPEN_READ_FILESTREAM, parameterValues);
		lr.setFsBeforeTime(reply.beforeTime);
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		this.lastEid = opEid;

		return reply;
	}

	public void closeFilestream(int handle, String filePath, String hostname)
			throws IOException, AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

	    try {
		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Void> reply;
		this.filestreamGenerations.remove(handle);
		// handle is in one of these lists
		if (this.readFilestreamHandles.remove(new Integer(handle)))
			reply = fsClient.closeFileInputStream(handle, filePath);
		else {
			this.writeFilestreamHandles.remove(new Integer(handle));
			reply = fsClient.closeFileOutputStream(handle, filePath);
		}

		Serializable[] parameterValues = { hostname, filePath };
		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.CLOSE_FILESTREAM, parameterValues);
		lr.setFsAfterTime(reply.afterTime);
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		this.lastEid = opEid;
	    } catch (AeolusException e) {
                System.out.println("UserLoggingProxy: closeFilestream: Aeolus Exception occurred unexpectedly");
		e.printStackTrace();
                System.exit(-1);

	    }
	}

	// RPC

	/**
	 * This registers the user service to Aeolus Binds the service name to the
	 * service class in the PI and the node If a service by the same name
	 * existed, it will be replaced
	 * 
	 * @param serviceName
	 *            - name of the interface implemented by the service
	 * @param service
	 *            - class implementing the given service
	 */
	public void registerService(String serviceName, Class<?> service) 
		throws InfoFlowControlException {
		Serializable[] parameterValues = { serviceName };
		final EventID opEid = getNextEventId();
		sysd.collector.submitLogRecord(this.userDomainId, new EventRecord(
				opEid, this.lastEid, null, AeolusApiConstants.REGISTER_SERVICE,
				parameterValues));
		this.lastEid = opEid;

                if (!this.state.SLabel.isEmpty())
                        throw new InfoFlowControlException("Cannot register a service with non-null secrecy label");
		this.sysd.addEntryToServiceMap(serviceName, service, opEid, state.pid, state.basis);
	}

	/**
	 * Return a dynamically generated proxy service stub that implements the
	 * interface associated with the remote service object by serviceName
	 * 
	 * @param remoteHostName
	 *            - hostname of the node running the remote service
	 * @param serviceName
	 *            - name that the service is bound to on the remote host
	 * @param serviceClass
	 *            - local Class which defines the interface of the service
	 * @throws ClassNotFoundException
	 *             - if interface with serviceName cannot be found on the client
	 */
	public Object getService(String remoteHostName, String serviceName,
			Class<?> serviceClass) {
		Serializable[] parameterValues = { remoteHostName, serviceName,
				serviceClass.toString() };
		final EventID opEid = getNextEventId();
		sysd.collector.submitLogRecord(this.userDomainId, new EventRecord(
				opEid, this.lastEid, null, AeolusApiConstants.GET_SERVICE,
				parameterValues));
		this.lastEid = opEid;

		return super.getService(remoteHostName, serviceName, serviceClass);
	}

	// VN/NODE LAUNCH API

	/**
	 * Request to launch a local or remote platform instance
	 * 
	 * @param hostname
	 *            - hostname of the node to lauch the VN at
	 * @param state
	 *            - initial info flow state for the launched VN
	 * @param appName
	 *            - application to start running at the VN
	 * @param appParams
	 *            - parameters to supply to the application
	 */
	public void launchPlatformInstance(String hostname, PrincipalID pid,
			String appName, String appParams) throws RpcException, InfoFlowControlException,
			AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] params = { hostname, pid, appName };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.PI_LAUNCH_REQUEST, params);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.PI_LAUNCH_REPLY);
		this.lastEid = replyEid;
		try {
			if ((!this.state.SLabel.isEmpty()))
				throw new InfoFlowControlException(
						"Cannot launch a VN with non-null secrecy label");

			if (pid.getValue() != AuthStateConstants.pidPublic) {
				AuthReply<Boolean> reply = sysd.authClient.canSwitch(
						state.basis, state.pid, pid);
				lr.setCommunicationPred(reply.event);

				if (!reply.value) {
					throw new AuthorityException(
							"Not authorized to launch a VN with pid = "
									+ pid.getValue());
				}
			}

	                InfoFlowState ifs = new InfoFlowState(state);
                	if (!pid.equals(state.pid)) ifs.basis.add(pid.getValue());
                	ifs.pid = pid;
                	ifs.ILabel = new Label();
                	ifs.SLabel = new Label();
                // sysd.remoteLaunch(state.pid, hostname, pid, appName, appParams, null);
                //                 sysd.remoteLaunch(state.pid, ifs, hostname, pid, appName, appParams, null);
                //
			RpcReply reply = sysd.remoteLaunch(state.pid, ifs, hostname, pid, appName, appParams,
			 		opEid);
                	AeolusLib.addSecrecy(new AeolusLabel(reply.token.state.SLabel).members()); // add is the same as union
                	Label returnedILabel = reply.token.state.ILabel;
                	Label currentILabel = this.state.ILabel;
                	Label newILabel = currentILabel.intersection(returnedILabel);
                	this.state.ILabel = newILabel;
                	if (reply.exc != null) throw new Exception(reply.exc);

			sysd.collector.submitLogRecord(this.userDomainId, lr);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof RpcException)
				throw (RpcException) e;
			else if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	public void shutdownPlatformInstance(String hostname,
			PlatformInstanceID piId, boolean shutdownImmediately)
			throws AuthorityException, AeolusException, RpcException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] params = { hostname, piId, shutdownImmediately };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.PI_SHUTDOWN_REQUEST, params);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.PI_SHUTDOWN_REPLY);
		this.lastEid = replyEid;
		try {
			if ((!this.state.SLabel.isEmpty())
					|| (!this.state.ILabel.isEmpty()))
				throw new AuthorityException(
						"Cannot shutdown a VN with non-null labels");

			sysd.remoteShutdown(hostname, piId, state.pid, shutdownImmediately,
					opEid);
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof RpcException)
				throw (RpcException) e;
			else if (e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Shuts down the indicated system node given appropriate authority
	 */
	public void deleteNode(String hostname) throws AuthorityException,
			AeolusException {
		checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		Serializable[] params = { hostname };
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.DELETE_NODE_REQUEST, params);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		// TODO: sysd.authClient.deleteNode(hostname, state.pid);

		// predecessor should be the latest update from the auth server
		EventID replyEid = getNextEventId();
		lr = new EventRecord(replyEid, this.lastEid,
				AeolusApiConstants.DELETE_NODE_REPLY);
		this.lastEid = replyEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

	}

	// I/O Devices

	// for now just check that the labels are null

	/**
	 * Access to an output device can be allowed if the process secrecy is null
	 */
	public void writeToIODevice() throws InfoFlowControlException, AeolusException {
			checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid,
				AeolusApiConstants.WRITE_TO_IODEVICE);
		this.lastEid = opEid;

		if (!this.state.SLabel.isEmpty()) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			throw new InfoFlowControlException(
					"Cannot write to I/O device with non-null secrecy label");
		}

		sysd.collector.submitLogRecord(this.userDomainId, lr);
	}

	/**
	 * Access to input devices can be allowed if the process integrity label is
	 * null
	 * 
	 * @throws InfoFlowControlException
	 */
	public void readFromIODevice() throws InfoFlowControlException, AeolusException {
			checkAuthorityEnabled(); // XXX should log on failure

		EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid,
				AeolusApiConstants.READ_FROM_IODEVICE);
		this.lastEid = opEid;

		if (!this.state.ILabel.isEmpty()) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);
			throw new InfoFlowControlException(
					"Cannot read from I/O device with non-null integrity label");
		}

		sysd.collector.submitLogRecord(this.userDomainId, lr);
	}

	// INTERNAL USE

	/**
	 * Submit an RPC request to the SYS-D This method is not exposed to the
	 * user, it is invoked by the dynamic proxy generated on the client side
	 * 
	 * @throws RpcException
	 */
	public Serializable sendRPCRequest(Object proxy, Method method,
			Object[] parameters) throws RpcException, AeolusException {
		RpcProxyInfo info = this.rpcProxies.get(proxy);

		if (info == null)
			assert false;

		String remoteHostName = info.getRemoteHostName();

		// each proxy can only implement one interface
		String fullInterfaceName = proxy.getClass().getInterfaces()[0]
				.getName();

		// retrieve the interface name without package information since this
		// could be different on server
		int lastDotIndex = fullInterfaceName.lastIndexOf(".");
		String interfaceName = fullInterfaceName;
		if (lastDotIndex != -1) {
			interfaceName = fullInterfaceName.substring(lastDotIndex + 1);
		}
		String serviceName = info.getServiceName(); // what it was before,
													// changed by wjarjoui -
													// interfaceName;
		// String serviceName = interfaceName;
		InfoFlowState startingState = new InfoFlowState(this.state);

		Serializable[] params = null;
		if (parameters != null) {
			params = new Serializable[parameters.length];
			for (int i = 0; i < parameters.length; i++) {
				try {
					params[i] = (Serializable) parameters[i];
				} catch (Exception e) {
					throw new RpcException(
							"The supplied parameter is not a Serializable object");
				}
			}
		}
		Serializable[] parameterValues = { remoteHostName, serviceName,
				method.getName() };
		final EventID opEid = getNextEventId();
		EventRecord lr = new EventRecord(opEid, this.lastEid, null,
				AeolusApiConstants.SEND_RPC, parameterValues);
		this.lastEid = opEid;
		sysd.collector.submitLogRecord(this.userDomainId, lr);

		try {
			RpcReply reply = sysd.sendRPC(remoteHostName, serviceName, method,
					params, startingState, opEid);
			this.sysd.authClient
					.getUpToInvalidation(reply.token.invalidationNumber);
                        AeolusLib.addSecrecy(new AeolusLabel(reply.token.state.SLabel).members()); // add is the same as union
                        Label returnedILabel = reply.token.state.ILabel;
                        Label currentILabel = this.state.ILabel;
                        Label newILabel = currentILabel.intersection(returnedILabel);
                        this.state.ILabel = newILabel;

			EventID replyEid = getNextEventId();
			lr = new EventRecord(replyEid, opEid,
					AeolusApiConstants.RECEIVE_RPC_REPLY);
			this.lastEid = replyEid;

			lr.setCommunicationPred(reply.token.predecessorEventId);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

                        if (reply.exc == null) return reply.opResult;
                        else throw new Exception(reply.exc);
			
		} catch (Exception e) {
			lr.setCompletionStatus(CompletionStatus.FAILED);
			sysd.collector.submitLogRecord(this.userDomainId, lr);

			if (e instanceof RpcException)
				throw (RpcException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Logs the event of receiving an RPC (not user-accessible) sets the info
	 * flow state and invokes the method on the service
	 * 
	 * @throws InstantiationException
	 */
	public RpcReply receiveRpc(RpcToken rpcInfo, RegisteredService regService,
			Method m, Object[] parameterValues)
			throws IllegalArgumentException, IllegalAccessException,
			InvocationTargetException, InstantiationException {
		EventID receiveRpcEid = new EventID(sysd.platformInstanceID.getValue(),
				this.userDomainId, this.generation, 0);
		Serializable[] params = { rpcInfo.state.ILabel, rpcInfo.state.SLabel,
				m.toString() };
		EventRecord receipt = new EventRecord(receiveRpcEid,
				rpcInfo.predecessorEventId, regService.registrationEid,
				AeolusApiConstants.RECEIVE_RPC, params);
		this.lastEid = receiveRpcEid;
		sysd.collector.submitLogRecord(this.userDomainId, receipt);

		this.rpcCallerPid = rpcInfo.state.pid;
                this.state.SLabel = rpcInfo.state.SLabel;
                this.state.ILabel = rpcInfo.state.ILabel;
		this.state.pid = regService.pid;
		this.state.basis = regService.basis;
										// instance

		// TODO: catch exception to always send-reply
		Object serviceInstance = regService.service.newInstance();
		Serializable opResult = null;
		Exception exc = null;
		try {
		opResult = (Serializable) m.invoke(serviceInstance,
				parameterValues);
		} catch (Exception e) {
			exc = e;
		}

		EventID sendRpcReplyEid = this.getNextEventId();
		EventRecord reply = new EventRecord(sendRpcReplyEid, this.lastEid,
				AeolusApiConstants.SEND_RPC_REPLY);
		reply.setParams(new Serializable[] { rpcInfo.state });
		this.lastEid = sendRpcReplyEid;
		sysd.collector.submitLogRecord(this.userDomainId, reply);

		// event after service method invocation

		RpcToken replyToken = new RpcToken(this.state, this.lastEid,
				sysd.authClient.getCurrentInvalidationNumber(),
				sysd.platformNode.nodeId);
		if (exc == null) return new RpcReply(opResult, replyToken);
		else return new RpcReply(exc, replyToken);
	}

	// Auditing user interface

	public EventID createEvent(String methodName, Serializable[] parameters,
			Serializable result) {
		return createEvent(methodName, parameters, result, null, null, null);
	}

	public EventID createEvent(String methodName, Serializable[] parameters,
			Serializable result, Date timestamp) {
		return createEvent(methodName, parameters, result, timestamp, null,
				null);
	}

	public EventID createEvent(String methodName, List<EventID> preds,
			Serializable[] parameters, Serializable result) {
		return createEvent(methodName, parameters, result, null, preds, null);
	}

	public EventID createEvent(String methodName, List<EventID> preds,
			Serializable[] parameters, Serializable result, Date timestamp) {
		return createEvent(methodName, parameters, result, timestamp, preds,
				null);
	}

	public EventID createEvent(String methodName, List<EventID> preds,
			Serializable[] parameters, Serializable result,
			CompletionStatus status, Date timestamp) {
		return createEvent(methodName, parameters, result, null, preds, status);

	}

	private EventID createEvent(String methodName, Serializable[] parameters,
			Serializable result, Date timestamp, List<EventID> preds,
			CompletionStatus status) {
		EventID userEvent = getNextEventId();
		EventRecord rec = new EventRecord(userEvent, this.lastEid,
				AeolusApiConstants.APPLICATION_LVL);
		rec.setParams(parameters);
		rec.setReturnValue(result);
		rec.setInfoString(methodName);
		if (timestamp != null)
			rec.setTimestamp(timestamp);
		if (preds != null)
			rec.setOtherPreds(preds);
		if (status != null)
			rec.setCompletionStatus(status);

		this.lastEid = userEvent;
		sysd.collector.submitLogRecord(this.userDomainId, rec);
		return userEvent;

	}

	// Event ID generation

	private EventID getNextEventId() {
		EventID eid = new EventID(sysd.platformInstanceID.getValue(),
				this.getUserAppDomainId(), generation, eventCounter);
		eventCounter++;
		return eid;
	}
}
