package edu.mit.csail.aeolus.runtime;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.Callable;

import edu.mit.csail.aeolus.api.AeolusClosure;
import edu.mit.csail.aeolus.api.AeolusFileInputStream;
import edu.mit.csail.aeolus.api.AeolusFileOutputStream;
import edu.mit.csail.aeolus.api.AeolusLib;
import edu.mit.csail.aeolus.api.AeolusLabel;
import edu.mit.csail.aeolus.api.AeolusShared;
import edu.mit.csail.aeolus.auditing.CompletionStatus;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.authority.api.AuthReply;
import edu.mit.csail.aeolus.authority.api.ClosureReply;
import edu.mit.csail.aeolus.authority.api.ClosureTableEntry;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.filesystem.FileSystemClient;
import edu.mit.csail.aeolus.filesystem.FileSystemReply;
import edu.mit.csail.aeolus.infoflow.AuthStateConstants;
import edu.mit.csail.aeolus.infoflow.ClosureID;
import edu.mit.csail.aeolus.infoflow.InfoFlowRules;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.infoflow.PlatformInstanceID;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;
import edu.mit.csail.aeolus.rpc.RegisteredService;
import edu.mit.csail.aeolus.rpc.RpcInvocationHandler;
import edu.mit.csail.aeolus.rpc.RpcProxyInfo;
import edu.mit.csail.aeolus.rpc.RpcReply;
import edu.mit.csail.aeolus.rpc.RpcToken;
import edu.mit.csail.aeolus.security.AeolusCloner;
import edu.mit.csail.aeolus.utilities.HelperUtils;

/**
 * Proxy object provides methods to a User-d to interact with Aeolus.
 * 
 * It exposes the Aeolus API, namely it provides methods for all Aeolus functions
 * and all external methods: modifications of authority state, manipulations of labels,
 * access to shared state, file use, I/O devices.
 * For I/O devices, PO provides a wrapper code that ensures information flow rules are obeyed.
 * 
 * The PO stores the process' information flow state (principal ID, process labels)
 * 
 * The PO interacts with the SYS-d (which has access to the shared state manager and the authority
 * state client).
 * 
 * API subsections: principals, tags, delegations and revocations, label manipulations, 
 * 					authority closure, call, fork, shared state, shared object, shared lock, 
 * 					shared queue, files, filestreams 
 * 
 * viq
 *
 */
public class UserProxy implements UserProxyInterface{

	// SYS-d reference
	protected SystemDomain sysd;

	// Id of the associated UserDomain
	protected long userDomainId;

	// Number of times this proxy object has been reused
	protected int generation;

	// Information flow state
	protected InfoFlowState state;

	protected PrincipalID blockID;

	private int sharedStateRecursionCount = 0;
	// Generation number of the label, incremented on each label change
	protected int labelGeneration = 0;

	// File system state
	protected List<Integer> readFilestreamHandles;
	protected List<Integer> writeFilestreamHandles;
	// XXX Would be cleaner for this to be part of the filestream state.
	protected HashMap<Integer, Integer> filestreamGenerations;

	// Stores the RPC service proxy object stubs 
	// and any associated data to correctly direct the calls
	// every time a stub's method is invoked
	protected HashMap<Object, RpcProxyInfo> rpcProxies;
	public PrincipalID rpcCallerPid;

	////// PO initialization //////

	/**
	 * Creates the proxy object
	 * @param sysD - the system domain associated with this proxy object
	 */
	public UserProxy(SystemDomain sysD, long myUserDomainId)
	{
		this.sysd = sysD;
		this.userDomainId = myUserDomainId;
		this.generation = 1;
		this.state = new InfoFlowState();
		this.blockID = state.pid;
		this.rpcProxies = new HashMap<Object, RpcProxyInfo>();
		this.readFilestreamHandles = new ArrayList<Integer>();
		this.writeFilestreamHandles = new ArrayList<Integer>();
		this.filestreamGenerations = new HashMap<Integer, Integer>();
	}

	public void createBlock() {
		this.blockID = null;
	}

	public void setCurrentBlock(PrincipalID p) {
		this.blockID = p;
	}

	// INTERNAL FUNCTIONALITY

	/**
	 * This is intended to clean up any state in the proxy and make it ready to be re-used
	 * (This is called when the User-d thread finishes executing and is returned to the
	 * thread pool - afterExecute() call in the AppDomainPool class)
	 */  
	public void flush()
	{
		this.generation++;
		this.state = new InfoFlowState();
		this.rpcProxies.clear();

		this.readFilestreamHandles.clear();
		this.writeFilestreamHandles.clear();
		this.filestreamGenerations.clear();
		this.sharedStateRecursionCount = 0;
	}

	/**
	 * Sets the initial information flow state and invokes the provided method with the given arguments
	 * To be used for static method invocation (e.g. "main" for app launch)
	 */
	public void invoke(InfoFlowState startingState, Method m, String[] argv) 
	throws IllegalArgumentException, IllegalAccessException, InvocationTargetException
	{
		this.state = startingState;
		m.invoke(null, (Object) argv);
	}
	/**
	 * Sets the info flow state and invokes the code object
	 * (e.g. used with a fork call)
	 */
	public void invoke(InfoFlowState startingState, Runnable code, EventID userPredEid) throws Exception
	{   
		this.state = startingState;
		code.run();
	}

	// USER API

	/**
	 * Returns a copy of the current info flow state
	 */
	public InfoFlowState getInfoFlowState()
	{
		return new InfoFlowState(state);
	}
	
	public PlatformInstanceID getPlatformInstanceID()
	{
		return this.sysd.platformInstanceID;
	}

	/**
	 * If this is a remote method invocation,
	 * returns the pid of the caller to this method;
	 * otherwise returns null
	 */
	public PrincipalID getRpcCallerPid()
	{
		return this.rpcCallerPid;
	}
	
	// AUTHORITY API
        public boolean canSwitch(PrincipalID pid) {
            return sysd.authClient.canSwitch(state.basis, state.pid, pid).value;
        }

	// tests whether pid1 acts-for pid2 (per authority/server/AuthManager)
        public boolean isActee(PrincipalID pid1, PrincipalID pid2) {
            return sysd.authClient.isActee(pid1, pid2).value;
        }

	public boolean hasAuthority(PrincipalID pid, TagID tid) {
            return sysd.authClient.hasAuthority(pid, tid).value;
        }


	// PRINCIPALS

	/**
	 * Creates a new principal in the process principal core 
	 * and returns the new principal ID
	 * Process principal acts for the new principal
	 */
	public PrincipalID createPrincipal() throws AuthorityException, InfoFlowControlException, AeolusException
	{
		if (this.blockID == null) {
			PrincipalID p = createPrincipalInNewCore();
			this.blockID = p;
			return p;
		}
		checkAuthorityEnabled();

		if (!this.state.SLabel.isEmpty())
			throw new InfoFlowControlException("Cannot create a new principal with non-null secrecy label");

		try{
			AuthReply<PrincipalID> reply = sysd.authClient.createPrincipal(state.pid, null);
			return reply.value;
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Creates a new principal in a new core 
	 * and returns the new principal ID
	 * Process principal acts for the new principal
	 */
	public PrincipalID createPrincipalInNewCore() throws AuthorityException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		if (!this.state.SLabel.isEmpty()) throw new InfoFlowControlException("Cannot create a new principal with non-null secrecy label");
		try{
			AuthReply<PrincipalID> reply = sysd.authClient.createPrincipalInNewCore(state.pid, null);
			return reply.value;
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Creates a new principal in the same core as principal p
	 * and returns the new principal ID
	 * Process principal acts for the new principal
	 */
	public PrincipalID createPrincipal(PrincipalID p) throws AuthorityException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		if (!this.state.SLabel.isEmpty()) throw new InfoFlowControlException("Cannot create a new principal with non-null secrecy label");

		try{
			AuthReply<PrincipalID> reply = sysd.authClient.createPrincipal(p, null);
			return reply.value;
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	// TAGS API

	/**
	 * Creates a new top-level tag and returns its ID
	 * Process principal has authority for the new tag
	 */
	public TagID createTag() throws AuthorityException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		if (!this.state.SLabel.isEmpty()) throw new InfoFlowControlException("Cannot create a new tag with non-null secrecy label");

		try{
			AuthReply<TagID> reply = sysd.authClient.createTaint(state.pid, null);
			return reply.value;
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Creates a sub-tag of tag t and returns the new sub-tag ID
	 * Tag t must be a top-level tag
	 * Process principal has authority for the new tag
	 * @throws  
	 */
	public TagID createSubTag(TagID parent) throws InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		if (!this.state.SLabel.isEmpty()) throw new InfoFlowControlException("Cannot create a subtag with non-null secrecy label");

		AuthReply<TagID> reply = null;
		try {
			reply = sysd.authClient.createSubTag(parent, state.pid, null);
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		TagID tag = reply.value;
		return new TagID(tag.getValue(), parent.getValue());
	}

	/**
	 * Creates a sub-tag of tag t, stores the tag info in the core
	 * of principal p and returns the new sub-tag ID
	 * Tag t must be a top-level tag
	 * Process principal has authority for the new tag
	 */
	public TagID createSubTag(TagID t, PrincipalID p) throws AuthorityException, AeolusException
	{
		return null;
	}

	// DELEGATIONS AND REVOCATIONS API

	/**
	 * Adds an act-for link from actor principal p2 
	 * to actee principal p1
	 * Process principal must act for principal p1
	 */
	public void actFor(PrincipalID p1, PrincipalID p2) throws AuthorityException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		if (!this.state.SLabel.isEmpty()) throw new InfoFlowControlException("Cannot invoke actFor with non-null secrecy label");

		try{
			sysd.authClient.delegateActsFor(state.pid, p2, p1, null);
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Removes the act-for link from principal p2 to principal p1
	 * Process principal must act for principal p1
	 */
	public void revokeActFor(PrincipalID p1, PrincipalID p2) throws AuthorityException, AeolusException
	{
		checkAuthorityEnabled();

		try{
			sysd.authClient.revokeActsFor(state.pid, p2, p1, null);
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Gives authority for tag t from grantor principal p1 
	 * to grantee principal p2
	 * Process principal must act for the grantor principal p1, 
	 * the grantor principal p1 must have authority for tag t
	 */
	public void delegate(TagID t, PrincipalID p1, PrincipalID p2) throws AuthorityException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		if (!this.state.SLabel.isEmpty()) throw new InfoFlowControlException("Cannot delegate a tag with non-null secrecy label");

		try{
			sysd.authClient.grant(state.pid, p1, p2, t, null);
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Revokes authority for tag t from grantor principal p1 to grantee principal p2
	 * Process principal must act for the grantor principal p1 and the tag t must have been 
	 * previously delegated from p1 to p2 
	 */
	public void revokeDelegate(final TagID t, final PrincipalID p1, final PrincipalID p2) throws AuthorityException, AeolusException
	{
		checkAuthorityEnabled();

		try{
			sysd.authClient.revokeGrant(state.pid, p1, p2, t, null);
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}


	// LABEL MANIPULATIONS API

	/**
	 * Adds given tag id to the state secrecy label
	 */
	public void addSecrecy(TagID tid) throws AeolusException
	{
		checkAuthorityEnabled();

		//case where tid is already in the label is taken care of internally (add method for TreeSet only adds unique)
		TagID[] tags = this.state.SLabel.getTags();
		if(tid.getParent()==0){
			//if top level tag, remove all its subtags from the label then add it
			for(TagID tag : tags){
				if(tag.getParent()==tid.getValue())
					this.state.SLabel.removeTag(tag);
			}
			this.state.SLabel.addTag(tid);
			this.checkWriteFileStreams();
		}
		else if(!(this.state.SLabel.hasTag(new TagID(tid.getParent())))){
			this.state.SLabel.addTag(tid);
			this.checkWriteFileStreams();
		}
	}

	/**
	 * Adds given tag ids to the state secrecy label
	 * @param tids: list of tag ids to add
	 */
	public void addSecrecy(List<TagID> tids) throws AeolusException
	{
		if(tids == null)
			return;
		
		
		for (TagID tid : tids)
//			this.state.SLabel.addTag(tid);
			addSecrecy(tid);
	}

	/**
	 * Removes given tag id from the state integrity label
	 */
	public void removeIntegrity(TagID tid) throws AeolusException
	{
		checkAuthorityEnabled();

		this.state.ILabel.removeTag(tid);
		if(tid.getParent()==0){//if top tag-faster if we check its a negative number?
			TagID[] tags = this.state.ILabel.getTags();
			for(TagID tag : tags){
				if(tag.getParent()==tid.getValue())
					this.state.ILabel.removeTag(tag);
			}
		}
		this.checkWriteFileStreams();
	}

	/**
	 * Removes given tag ids from the state integrity label
	 * @param tids: list of tag ids to remove
	 */
	public void removeIntegrity(List<TagID> tids) throws AeolusException
	{
		if(tids == null)
			return;

		for (TagID tid : tids)
				removeIntegrity(tid);
	}

	/**
	 * Removes given tag ID from the secrecy label if process is authorized
	 * Authority is checked via a call to the SYS-d 
	 * Will not declassify if the process is running as Ppublic
	 */
	public void declassify(TagID tid) throws AuthorityException, AeolusException
	{
		checkAuthorityEnabled();

		ArrayList<TagID> tags = new ArrayList<TagID>();
		tags.add(tid);

		AuthReply<Boolean> reply = sysd.authClient.hasAuthority(state.basis, state.pid, tags);
		if (reply.value){
			state.SLabel.removeTag(tid);
			this.checkReadFileStreams();
		}
		else
			throw new AuthorityException("No authority to declassify as PID = " + state.pid.getValue());
	}

	/**
	 * Removes given tag ids from the secrecy label if process is authorized
	 * Authority is checked via a call to the SYS-d 
	 * Will not declassify if the process is running as Ppublic
	 * @param tids: list of tag ids to remove
	 */
	public void declassify(List<TagID> tids) throws AeolusException
	{
		if(tids == null)
			return;

		checkAuthorityEnabled();

		AuthReply<Boolean> reply = sysd.authClient.hasAuthority(state.basis, state.pid, tids);
		if (reply.value){
			for (TagID tid : tids)
				state.SLabel.removeTag(tid);
			
			this.checkReadFileStreams();
		}
		else
			throw new AuthorityException("No authority to declassify as PID = " + state.pid.getValue());
	}

	/**
	 * Adds given tag ID to the integrity label if process is authorized
	 * Authority is checked via a call to the SYS-d 
	 * Will not endorse if the process is running as Ppublic
	 */
	public void endorse(TagID tid) throws AeolusException
	{
		checkAuthorityEnabled();

		ArrayList<TagID> tids = new ArrayList<TagID>();
		tids.add(tid);
		AuthReply<Boolean> reply = sysd.authClient.hasAuthority(state.basis, state.pid, tids);
		if (reply.value){
			this.state.ILabel.addTag(tid);
			this.checkReadFileStreams();
		}
		else
			throw new AuthorityException("No authority to endorse as PID = " + state.pid.getValue());
	}

	/**
	 * Adds given tag ids to the integrity label if process is authorized
	 * Authority is checked via a call to the SYS-d 
	 * Will not endorse if the process is running as Ppublic
	 * @param tids: list of tag ids to add
	 */
	public void endorse(List<TagID> tids) throws AeolusException
	{
		if(tids == null)
			return;
		
		checkAuthorityEnabled();

		AuthReply<Boolean> reply = sysd.authClient.hasAuthority(state.basis, state.pid, tids);
		if (reply.value){
			for (TagID tid : tids)
				this.state.ILabel.addTag(tid);

			this.checkReadFileStreams();
		}
		else
			throw new AuthorityException("No authority to endorse as PID = " + state.pid.getValue());
	}

	// AUTHORITY CLOSURE API

	/**
	 * Creates a new closure with the proided interface and returns the closure ID
	 * A new anonymous principal PA is generated and bound to this key and closure ID,
	 * this principal has no authority
	 *
	 * XXX old API - can probably be deleted
	 */
//	public ClosureID createClosure(Class<? extends AeolusCallable> closureInterface) throws AeolusException
//	{
//		try{
//			AuthReply<ClosureReply> reply = sysd.authClient.createClosure(closureInterface.getName(), null);
//			return reply.value.cid;
//		}
//		catch (Exception e){
//			throw new AeolusException(e);
//		}
//	}

	/**
	 * Creates a new closure with the given key and stores the closure info in the same core 
	 * as principal pid
	 * A new anonymous principal is generated and bound to this key and new closure ID,
	 * this principal has no authority
	 * @return new closure ID
	 * TODO: add this function to AS
	 *
	 * XXX old API - can probably be deleted
	 */
//	public ClosureID createClosure(Class<? extends AeolusCallable> closureInterface, PrincipalID pid)
//	{
//		return null;
//	}
//
//	/* XXX old API - can probably be deleted */
//	public AeolusCallable getClosure(ClosureID cid, Class<?> closureClass) throws AeolusException
//	{
//		AuthReply<ClosureTableEntry> reply = sysd.authClient.getClosureInfo(cid);
//		ClosureTableEntry closureInfo = reply.value;
//		String interfaceName = closureInfo.className;
//		try
//		{
//			Class<?> closureInterface = Class.forName(interfaceName);
//			
//			// need to check that this class implements the closure interface
//			Class<?>[] interfaces = closureClass.getInterfaces();
//			boolean found = false;
//			for(int i = 0; i < interfaces.length; i++)
//			{
//				if(interfaces[i].equals(closureInterface))
//				{
//					found = true;
//					break;
//				}	
//			}
//			if(!found)
//				throw new AeolusException("closureClass must implemented the interface bound to this closure");
//			
//			return (AeolusCallable) closureClass.newInstance();
//		}
//		catch (ClassNotFoundException e) {
//			throw new AeolusException("Closure class was not found " + interfaceName, e);
//		}
//		catch(Exception e) {
//			throw new AeolusException("Could not create closure with name " + interfaceName, e);
//		}
//	}
	
    public InfoFlowState aeolusStartClosureCall(AeolusClosure closure) {
	/* XXX need to ensure that only AeolusClosure can call this */
	    InfoFlowState callerState = this.state;
	    this.state = new InfoFlowState(closure.getPID().getPrincipalID(), new Label(this.state.SLabel), new Label(this.state.ILabel), closure.getBasisCopy());
	    return (callerState);
    }

    public void aeolusEndClosureCall(AeolusClosure closure, InfoFlowState oldState) {
	/* XXX need to ensure that only AeolusClosure can call this */
	oldState.SLabel.unionInto(this.state.SLabel);
	oldState.ILabel.intersectInto(this.state.ILabel);
	this.checkAllFileStreams();
	this.state = oldState;
    }

    public InfoFlowState aeolusStartSharedCall(AeolusShared shared) throws InfoFlowControlException {
	/* XXX need to ensure that only AeolusShared can call this */
	if (!this.state.SLabel.equalsLabel(shared.getSLabel().getLabel()))
	    throw new InfoFlowControlException("shared object call with incompatible secrecy labels: process " + this.state.SLabel + " vs object " + shared.getSLabel().getLabel());
	if (!this.state.ILabel.equalsLabel(shared.getILabel().getLabel()))
	    throw new InfoFlowControlException("shared object call with incompatible integrity labels: process " + this.state.ILabel + " vs object " + shared.getILabel().getLabel());

	sharedStateRecursionCount++;
	return (null);
    }

    public void aeolusEndSharedCall(AeolusShared shared, InfoFlowState oldState) {
	/* XXX need to ensure that only AeolusShared can call this */

	sharedStateRecursionCount--;
    }


    /**
     * This hook should be called at the start of a method that does a
     * read from a shared state object.  It checks whether the labels of
     * the object allow the read.
     *
     * Note that user-defined shared objects will use aeolusStartSharedCall(),
     * which conservatively assumes that the method both reads and writes
     * the object.  This method is intended for built-in types such as queues
     * and locks.
     */
    public InfoFlowState aeolusStartSharedRead(AeolusShared shared) throws InfoFlowControlException {
	/* XXX need to ensure that only AeolusShared can call this */

	if (!InfoFlowRules.canRead(this.state.SLabel, this.state.ILabel, shared.getSLabel().getLabel(), shared.getILabel().getLabel()))
	    throw new InfoFlowControlException("shared object read with incompatible labels: process S=" + this.state.SLabel + " I=" + this.state.ILabel + " vs object S=" + shared.getSLabel() + " I=" + shared.getILabel());
	return (null);
    }

    public void aeolusEndSharedRead(AeolusShared shared, InfoFlowState oldState) {
    }

    /**
     * This hook should be called at the start of a method that does a
     * *blind* write to a shared state object.  It checks whether the
     * labels of the object allow the write.
     *
     * Note that user-defined shared objects will use aeolusStartSharedCall(),
     * which conservatively assumes that the method both reads and writes
     * the object.  This method is intended for built-in types such as queues.
     */
    public InfoFlowState aeolusStartSharedWrite(AeolusShared shared) throws InfoFlowControlException {
	/* XXX need to ensure that only AeolusShared can call this */

	if (!InfoFlowRules.canWrite(this.state.SLabel, this.state.ILabel, shared.getSLabel().getLabel(), shared.getILabel().getLabel()))
	    throw new InfoFlowControlException("shared object write with incompatible labels: process S=" + this.state.SLabel + " I=" + this.state.ILabel + " vs object S=" + shared.getSLabel() + " I=" + shared.getILabel());
	return (null);
    }

    public void aeolusEndSharedWrite(AeolusShared shared, InfoFlowState oldState) {
    }

	protected boolean isAuthorityEnabled() {
		return (sharedStateRecursionCount == 0);
	}

	//protected void checkAuthorityEnabled() throws AuthorityException {
	//	if (!isAuthorityEnabled())
	//		throw new AuthorityException("Attempt to use authority within shared state");
	//}
	protected void checkAuthorityEnabled() throws AeolusException {
		if (!isAuthorityEnabled())
			throw new AeolusException("Attempt to use authority within shared state");
	}

	/**
	 * Adds an act-for link from the anonymous principal of closure cid to principal pid
	 * (the anonymous principal will be authoritative for all tags that principal pid has authority for)
	 * Closure cid must exist and process principal must act for principal pid
	 */
	public void closureActFor(PrincipalID pid, ClosureID cid) throws AuthorityException, AeolusException
	{
		try{
			sysd.authClient.closureActsFor(state.pid, cid, pid, null);
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Gives authority for tag tid from grantor principal pid 
	 * to grantee anonymous principal of closure cid
	 * Closure CL1must exist, process principal must act for the grantor principal pid, 
	 * the grantor principal pid must have authority for tag tid 
	 */
	public void closureDelegate(TagID tid, PrincipalID pid, ClosureID cid) throws AuthorityException, AeolusException
	{
		try{
			sysd.authClient.closureGrant(state.pid, cid, pid, tid, null);
		}
		catch (Exception e){
			if(e instanceof AuthorityException)
				throw (AuthorityException) e;
			else
				throw new AeolusException(e);
		}
	}

	////// CALL, FORK API //////

	/**
	 *  Runs code with different authority
	 * 
	 *  Saves current state, then invokes the codeObj with the calleePid
	 *  After invocation, the state is restored, but the process labels 
	 *  are updated to reflect any additional contamination picked up during the call
	 *  
	 *  @param codeObj - method to run
	 *  @param calleePid - pid to run the method with
	 */
	public <T> T call(Callable<T> codeObj, PrincipalID calleePid) throws AuthorityException, Exception{
		PrincipalID callerPid = this.state.pid;
		if (calleePid.getValue() != AuthStateConstants.pidPublic){
			Boolean reply =
			    sysd.authClient.canSwitch(state.basis, state.pid, calleePid).value;
			if (!reply)
				throw new AuthorityException("Call cannot switch to PID = " + calleePid);
			this.state.basis.add(calleePid.getValue());
		}

		T t = null;
		try {
		    this.state.pid = calleePid;
		    t = codeObj.call();
		} finally {
		    // restore state
		    if(calleePid.getValue() != AuthStateConstants.pidPublic)
			this.state.basis.remove(this.state.basis.size() - 1);
		    this.state.pid = callerPid;
		}
		return t;
	}

	/**
	 * Checks that the current process' labels and PID allow the fork to be made
	 * If there are no violations, a new User-d is set up to run the specified method
	 * 
	 * The code object is copied to the new process
	 * 
	 * The actual fork request is handled in Sys-d
	 * @param codeObj - method to invoke
	 * @param pid of - principal to switch to
	 */
	public void fork(Runnable codeObj, PrincipalID pid) throws AuthorityException, AeolusException
	{
		checkAuthorityEnabled();

		if (pid.getValue() != AuthStateConstants.pidPublic)
		{
			Boolean reply = sysd.authClient.canSwitch(state.basis, state.pid, pid).value;
			if (!reply)
				throw new AuthorityException("Fork cannot switch to PID = " + pid);    		
		}

		Runnable codeObjCopy = AeolusCloner.Clone(codeObj);
		InfoFlowState forkedState = new InfoFlowState(this.state);
		forkedState.pid = pid;
		forkedState.basis.add(pid.getValue());

		sysd.fork(forkedState, codeObjCopy, null);
	}

	// SHARED STATE API

	// Root object (serves to bootstrap communication between processes)

	public AeolusShared getRoot() {
		return sysd.sharedRoot;
	}

        public void setRoot(AeolusShared root) throws InfoFlowControlException
        {
            /*
             * The root object has empty secrecy and integrity labels, so
             * writing it is allowed by processes with an empty secrecy
             * label and any integrity label.
             */
            if (!this.state.SLabel.isEmpty())
                        throw new InfoFlowControlException("setRoot() requires an empty secrecy label; process had label " + this.state.SLabel);

                sysd.sharedRoot = root;
        }

	
	////// File API //////

	/**
	 * Creates a file system on this node
	 * This just starts the file system manager on this node (if not already running)
	 */
	public void createFilesystem()
	{
		//TODO
	}

	public void removeFilesystem(){}

	
	public FileSystemClient getFSClient(String hostname) throws AeolusException
	{
		return sysd.platformNode.getFSClient(hostname);
	}
	
	/**
	 * Creates a new file
	 */
	public boolean createFile(String hostname, String filePath, Label sLabelFile, Label iLabelFile) 
			throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Boolean> reply = fsClient.createFile(filePath, sLabelFile, iLabelFile, state.SLabel, state.ILabel, null);
		return reply.opResult;
	}

	/**
	 * Creates a new directory
	 */
	public boolean createDir(String hostname, String filePath, Label sLabelFile, Label iLabelFile)
			throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();
		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Boolean> reply = fsClient.createDir(filePath, sLabelFile, iLabelFile, state.SLabel, state.ILabel, null);

		return reply.opResult;
	}

	/**
	 * Removes a directory
	 */
	public boolean delete(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Boolean> reply = fsClient.delete(filePath, state.SLabel, state.ILabel, null);
		
		return reply.opResult;
	}

	/**
	 * Returns the list of all files in a given directory
	 * @return array of strings
	 */
	public String[] listDir(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<String[]> reply = fsClient.listDir(filePath, state.SLabel, state.ILabel, null);
		return reply.opResult;
	}

	/**
	 * Returns the contents of the given file
	 */
	public int readFile(String hostname, String filePath, byte[] fileContent)
				throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Integer> reply = fsClient.readFile(filePath, fileContent, state.SLabel, state.ILabel, null);
		return reply.opResult;
	}

	/**
	 * Writes to the given file
	 */
	public void writeFile(String hostname, String filePath, byte[] fileContent, boolean appendMode)
				throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		fsClient.writeFile(filePath, fileContent, appendMode, state.SLabel, state.ILabel, null);		
	}
	
	/**
	 * Requests the Secrecy Label of a file
	 */
	public Label getSecrecyLabel(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Label> reply = fsClient.getSecrecyLabel(filePath, state.SLabel, state.ILabel, null);
		return reply.opResult;
	}
	
	/**
	 * Requests the Integrity Label of a file
	 */
	public Label getIntegrityLabel(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Label> reply = fsClient.getIntegrityLabel(filePath, state.SLabel, state.ILabel, null);
		return reply.opResult;
	}
	
	
	////// Filestream API //////

	public FileSystemReply<Integer> openFileInputStream(String hostname, String filePath)
		throws  RemoteException, FileNotFoundException, IOException, AuthorityException,
			AeolusException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		
		FileSystemReply<Integer> reply = fsClient.openFileInputStream(filePath, this.state.SLabel, this.state.ILabel, null);
		
		int handle = reply.opResult;
		/*
		 * XXX Setting the starting generation number to less than the
		 * current labelGeneration triggers a label check the first
		 * time the stream is accessed.  We do this because that's
		 * what the old code did, but I don't see why it's necessary.
		 */
		filestreamGenerations.put(handle, labelGeneration - 1);
		readFilestreamHandles.add(handle);
		
		return reply;
	}
	
	public FileSystemReply<Integer> openFileOutputStream(String hostname, String filePath, boolean appendMode) throws AeolusException, IOException
	{
		checkAuthorityEnabled();

		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		FileSystemReply<Integer> reply = fsClient.openFileOutputStream(filePath, appendMode, this.state.SLabel, this.state.ILabel, null);
		
		int handle = reply.opResult;
		/*
		 * XXX Setting the starting generation number to less than the
		 * current labelGeneration triggers a label check the first
		 * time the stream is accessed.  We do this because that's
		 * what the old code did, but I don't see why it's necessary.
		 */
		filestreamGenerations.put(handle, labelGeneration - 1);
		writeFilestreamHandles.add(handle);
		
		return reply;
	}
	
	/**
	 * Check that the user is authorized to read from the file-stream
	 */
	public boolean canReadFromStream(int handle, Label fileSLabel, Label fileILabel)
	{
		if (!isAuthorityEnabled())
			return false;

		if (this.filestreamGenerations.get(handle) < labelGeneration)
		{
			if (!InfoFlowRules.canRead(AeolusLib.getSecrecy().getLabel(), AeolusLib.getIntegrity().getLabel(), fileSLabel, fileILabel))
				return false;
			else // reset check bit
				this.filestreamGenerations.put(handle, labelGeneration);
		}
		return true;
	}
	
	/**
	 * Check that the user is authorized to write to the file-stream
	 */
	public boolean canWriteToStream(int handle, Label fileSLabel, Label fileILabel)
	{
		if (!isAuthorityEnabled())
			return false;

		if (this.filestreamGenerations.get(handle) < labelGeneration)
		{
			if (!InfoFlowRules.canWrite(AeolusLib.getSecrecy().getLabel(), AeolusLib.getIntegrity().getLabel(), fileSLabel, fileILabel)){
				return false;
			}
			else // reset check bit
				this.filestreamGenerations.put(handle, labelGeneration);
		}
		return true;
	}
	
	public void closeFilestream(int handle, String filePath, String hostname) throws RemoteException, FileNotFoundException, IOException, AeolusException
	{
		checkAuthorityEnabled();

	   try {
		FileSystemClient fsClient = sysd.platformNode.getFSClient(hostname);
		this.filestreamGenerations.remove(handle);
		// handle is in one of these lists
		if(this.readFilestreamHandles.remove(new Integer(handle)))
			fsClient.closeFileInputStream(handle, filePath);
		else 
		{
			this.writeFilestreamHandles.remove(new Integer(handle));
			fsClient.closeFileOutputStream(handle, filePath);
		}
	   } catch (AeolusException e) {
		System.out.println("UserProxy: closeFilestream: Aeolus Exception occurred unexpectedly");
		e.printStackTrace();
		System.exit(-1);
	   }
	}
	
	
	// RPC

	/**
	 * This registers the user service to Aeolus
	 * Binds the service name to the service class in the PI and the node
	 * If a service by the same name existed, it will be replaced
	 * @param serviceName - name of the interface implemented by the service
	 * @param service - class of the given service 
	 */
	public void registerService(String serviceName, Class<?> service) throws InfoFlowControlException
	{
		if (!this.state.SLabel.isEmpty())
			throw new InfoFlowControlException("Cannot register a service with non-null secrecy label");
		this.sysd.addEntryToServiceMap(serviceName, service, null, state.pid, state.basis);
	}

	/**
	 * Return service stub that implements the interface specified by serviceName.
	 * Remote methods should be invoked using the service stub by casting the returned
	 * object to the remote service interface; for example:
	 * Foo service = (Foo) getService("localhost", "path.to.Foo", serviceClass);
	 * service.methodX();
	 * A method invocation on the service stub will result in a remote procedure call
	 * to the remote host. 
	 * @param remoteHostName - node where the desired remote service was registered
	 * @param serviceName - name that the service is bound to on the remote host
	 * @param serviceClass - local Class which defines the interface of the service
	 * @return service stub, must be cast to the interface specified by serviceName by the caller
	 */
	public Object getService(String remoteHostName, String serviceName, Class<?> serviceClass) {
		RpcInvocationHandler handler = new RpcInvocationHandler(this);

		// create a proxy with this handler
		Object service = Proxy.newProxyInstance(serviceClass.getClassLoader(), new Class[] {serviceClass}, handler);
		this.rpcProxies.put(service, new RpcProxyInfo(remoteHostName, serviceName, 0));
		return service;
	}

	// VN/NODE LAUNCH API

	/**
	 * Request to launch a local or remote platform instance with the specified authority
	 * Caller labels must be null
	 * Caller must act for pid
	 * @param hostname - hostname of the node to lauch the VN at
	 * @param appName - application to start running at the VN
	 * @param appParams - parameters to supply to the application
	 * @throws AeolusException 
	 * @throws NoSuchMethodException 
	 * @throws ClassNotFoundException 
	 * @throws InvocationTargetException 
	 * @throws AuthorityException
	 * @throws InfoFlowControlException
	 */
	public void launchPlatformInstance(String hostname, PrincipalID pid, String appName, String appParams) 
				throws InfoFlowControlException, AuthorityException, AeolusException,
					RpcException, InvocationTargetException, ClassNotFoundException,
					NoSuchMethodException, Exception
	{
		checkAuthorityEnabled();

		if((!this.state.SLabel.isEmpty()))
			throw new InfoFlowControlException("Cannot launch a VN with non-null secrecy label");

		if (pid.getValue() != AuthStateConstants.pidPublic)
		{
			Boolean reply = sysd.authClient.canSwitch(state.basis, state.pid, pid).value;
			if (!reply)
				throw new AuthorityException("Not authorized to launch a VN with pid = " + pid.getValue());    		
		}
		// The only important piece of information to be sent is the basis.
		InfoFlowState ifs = new InfoFlowState(state);
		if (!pid.equals(state.pid)) ifs.basis.add(pid.getValue());
		ifs.pid = pid;
		ifs.ILabel = new Label();
		ifs.SLabel = new Label();
		// sysd.remoteLaunch(state.pid, hostname, pid, appName, appParams, null);
		RpcReply reply = sysd.remoteLaunch(state.pid, ifs, hostname, pid, appName, appParams, null);
		// Note: remoteLaunch itself throws some exceptions: at this time we are asuming
		// that those exceptions occur before main runs at the remote node
                AeolusLib.addSecrecy(new AeolusLabel(reply.token.state.SLabel).members()); // add is the same as union
                Label returnedILabel = reply.token.state.ILabel;
                Label currentILabel = this.state.ILabel;
                Label newILabel = currentILabel.intersection(returnedILabel);
                this.state.ILabel = newILabel;
                if (reply.exc != null) throw new Exception(reply.exc);
	}

	public void shutdownPlatformInstance(String hostname, PlatformInstanceID piId, boolean shutdownImmediately) 
						throws AuthorityException, AeolusException, RpcException
	{
		checkAuthorityEnabled();

		// if((!this.state.SLabel.isEmpty()) || (!this.state.ILabel.isEmpty()))
		// Barbara says only the secrecy label must be null.
		if((!this.state.SLabel.isEmpty()))
			throw new AuthorityException("Cannot shutdown a VN with non-null labels");

		sysd.remoteShutdown(hostname, piId, state.pid, shutdownImmediately, null);
	}

	public void shutdown() throws InfoFlowControlException, AeolusException
	{
		checkAuthorityEnabled();

		// Barbara says only the secrecy label must be null.
		// if((!this.state.SLabel.isEmpty()) || (!this.state.ILabel.isEmpty()))
		if((!this.state.SLabel.isEmpty()))
			throw new InfoFlowControlException("Cannot shutdown a PI with non-null labels");

		sysd.shutdown(true, null);
	}

	/**
	 * Shuts down the indicated system node given
	 * appropriate authority
	 */
	public void deleteNode(String hostname) throws AuthorityException, AeolusException
	{
		//sysd.authClient.deleteNode(hostname, state.pid);
	}

	// I/O Devices

	// for now just check that the labels are null

	/**
	 * Access to an output device can be allowed if the process secrecy is null
	 */
	public void writeToIODevice() throws InfoFlowControlException, AeolusException
	{
			checkAuthorityEnabled();

		if(!this.state.SLabel.isEmpty())
			throw new InfoFlowControlException("Cannot write to I/O device with non-null secrecy label");
	}

	/**
	 * Access to input devices can be allowed if the process integrity label is null
	 * @throws InfoFlowControlException
	 */
	public void readFromIODevice() throws InfoFlowControlException, AeolusException
	{
			checkAuthorityEnabled();

		if(!this.state.ILabel.isEmpty())
			throw new InfoFlowControlException("Cannot read from I/O device with non-null integrity label");
	}
	
	// Auditing user interface

	public EventID getLastEID() {
		return null;
	}	
	public EventID createEvent(String methodName, Serializable[] parameters, Serializable result)
	{
		return null;
	}
	public EventID createEvent(String methodName, Serializable[] parameters, Serializable result, Date timestamp)
	{
		return null;
	}
	public EventID createEvent(String methodName, List<EventID> preds, Serializable[] parameters, Serializable result)
	{
		return null;
	}
	public EventID createEvent(String methodName, List<EventID> preds, Serializable[] parameters, Serializable result, Date timestamp)
	{
		return null;
	}
	public EventID createEvent(String methodName, List<EventID> preds, Serializable[] parameters, Serializable result, CompletionStatus status, Date timestamp)
	{
		return null;
	}
	public EventID createEvent(int methodName, EventID pred, Serializable[] parameters, Serializable result)
	{
		return null;
	}
	public EventID createEvent(int methodName, EventID pred, Serializable[] parameters, Serializable result, CompletionStatus status)
	{
		return null;
	}
	
	// INTERNAL USE

	/**
	 * Submit an RPC request to the SYS-D
	 * This method should not be exposed to the user, it is invoked by the dynamic proxy generated on the client side
	 * @throws RpcException 
	 */
	//TODO: actually privatize this.
	public Serializable sendRPCRequest(Object proxy, Method method, Object[] parameters) throws RpcException, AeolusException{	
		RpcProxyInfo info = this.rpcProxies.get(proxy);

		if(info == null)
			assert false;   	

		String remoteHostName = info.getRemoteHostName();

		String serviceName = info.getServiceName();
		InfoFlowState startingState = new InfoFlowState(this.state);

		Serializable[] params = null;
		if(parameters != null){
			params = new Serializable[parameters.length];
			for(int i = 0; i < parameters.length; i++){
				try{
					params[i] = (Serializable) parameters[i];
				}catch (Exception e){
					throw new RpcException("The supplied parameter is not a Serializable object");
				}
			}
		}

		try{
			RpcReply reply = sysd.sendRPC(remoteHostName, serviceName, method, params, startingState, null);
			this.sysd.authClient.getUpToInvalidation(reply.token.invalidationNumber);
			AeolusLib.addSecrecy(new AeolusLabel(reply.token.state.SLabel).members()); // add is the same as union
			Label returnedILabel = reply.token.state.ILabel;
			Label currentILabel = this.state.ILabel;
			Label newILabel = currentILabel.intersection(returnedILabel);
			this.state.ILabel = newILabel;
			if (reply.exc == null) return reply.opResult;
			else throw new Exception(reply.exc);
		}catch(Exception e){
			if(e instanceof RpcException)
				throw (RpcException) e;
			else
				throw new AeolusException(e);
		}
	}

	/**
	 * Logs the event of receiving an RPC (not user-accessible)
	 * sets the info flow state and invokes the method on the service
	 */
	public RpcReply receiveRpc(RpcToken rpcInfo, RegisteredService regService, Method m, Object[] parameterValues) 
	throws IllegalArgumentException, IllegalAccessException, InvocationTargetException, InstantiationException{
		this.state.SLabel = rpcInfo.state.SLabel;
		this.state.ILabel = rpcInfo.state.ILabel;
		// the following statement is questionable!  rpc should run with pid per launch call
		// actually the next statement does the right thing.
		this.rpcCallerPid = rpcInfo.state.pid;
		this.state.pid = regService.pid;
		this.state.basis = regService.basis;

		// creates a new instance of the service class and invokes the method
		Object serviceInstance = regService.service.newInstance();
		
		Serializable opResult = null;
		Exception exc = null;
		try {
		opResult = (Serializable) m.invoke(serviceInstance, parameterValues);
		} catch (Exception e) {
			exc = e;
		}
		if (exc == null) return new RpcReply(opResult,
			new RpcToken(this.state, null, sysd.authClient.getCurrentInvalidationNumber(),
    			sysd.platformNode.nodeId));
		else return  new RpcReply(exc, 
			new RpcToken(this.state, null, sysd.authClient.getCurrentInvalidationNumber(),
    			sysd.platformNode.nodeId));
	}

	public long getUserAppDomainId(){
		return this.userDomainId;
	}

	public InfoFlowState getIFState(){
		return this.state;
	}

	/**
	 * Check all write file streams the next time they are used.
	 * This should be called every time the current label becomes
	 * more restrictive.
	 */
	protected void checkWriteFileStreams(){
		/*
		 * We no longer implement the optimization of handling read
		 * and write file streams separately, but the interface is
		 * left here so that it could be reimplemented easily later.
		 */
		labelGeneration++;
	}

	/**
	 * Check all read file streams the next time they are used.
	 * This should be called every time the current label becomes
	 * less restrictive.
	 */
	protected void checkReadFileStreams()
	{
		/*
		 * We no longer implement the optimization of handling read
		 * and write file streams separately, but the interface is
		 * left here so that it could be reimplemented easily later.
		 */
		labelGeneration++;
	}

	/**
	 * Equivalent to calling checkReadFileStreams() and
	 * checkWriteFileStreams() -- but possibly faster.
	 */
	protected void checkAllFileStreams()
	{
		labelGeneration++;
	}
	
	public Label getSLabel(){
	    /* XXX This doesn't copy the label, so it shouldn't be public. */
		return this.state.SLabel;
	}
	public Label getILabel(){
	    /* XXX This doesn't copy the label, so it shouldn't be public. */
		return this.state.ILabel;
	}

//	private boolean safeSerializingContext = false;
//	private SharedStateMap serializationMap;
//	public void setSafeSerializingContext() {
//		SecurityManager sm = System.getSecurityManager();
//		if(sm != null)
//			sm.checkPermission(new RuntimePermission("set-serialization-context"));
//		safeSerializingContext = true;
//	}
//	public void unsetSafeSerializingContext() {
//		SecurityManager sm = System.getSecurityManager();
//		if(sm != null)
//			sm.checkPermission(new RuntimePermission("set-serialization-context"));
//		safeSerializingContext = false;
//	}
//
//	
//	public void createSerializationMap() {
//		if(! safeSerializingContext || serializationMap != null)
//			throw new RuntimeException("Serializing Shared Objects requires a safe context.");
//		serializationMap = new SharedStateMap();
//	}
//
//	public SharedStateElement getSharedStateElement(SharedUID uid) {
//		if(! safeSerializingContext )
//			throw new RuntimeException("Serializing Shared Objects requires a safe context.");
//		return serializationMap.getSharedStateElement(uid);
//	}
//
//	public SharedUID getUID() {
//		return sysd.getUID();
//	}
//
//	public SharedStateMap popSerializationMap() {
//		if(! safeSerializingContext )
//			throw new RuntimeException("Serializing Shared Objects requires a safe context.");
//		SharedStateMap retVal = this.serializationMap;
//		this.serializationMap = null;
//		return retVal;
//	}
//
//	public void setSerializationMap(SharedStateMap map) {
//		if(! safeSerializingContext )
//			throw new RuntimeException("Serializing Shared Objects requires a safe context.");
//		this.serializationMap = map;
//	}
//
//	public void putSharedStateElement(SharedUID serializationUID,
//			SharedStateElement sharedStateElement) {
//		if(! safeSerializingContext )
//			throw new RuntimeException("Serializing Shared Objects requires a safe context.");
//		this.serializationMap.putSharedStateElement(serializationUID, sharedStateElement);
//	}
}
