package edu.mit.csail.aeolus.runtime;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;
import java.util.Date;
import java.util.List;
import java.util.concurrent.Callable;

import edu.mit.csail.aeolus.api.AeolusClosure;
import edu.mit.csail.aeolus.api.AeolusFileInputStream;
import edu.mit.csail.aeolus.api.AeolusFileOutputStream;
import edu.mit.csail.aeolus.api.AeolusShared;
import edu.mit.csail.aeolus.auditing.CompletionStatus;
import edu.mit.csail.aeolus.auditing.EventID;
import edu.mit.csail.aeolus.exceptions.AuthorityException;
import edu.mit.csail.aeolus.exceptions.AeolusException;
import edu.mit.csail.aeolus.exceptions.InfoFlowControlException;
import edu.mit.csail.aeolus.exceptions.RpcException;
import edu.mit.csail.aeolus.filesystem.FileSystemReply;
import edu.mit.csail.aeolus.filesystem.FileSystemClientInterface;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;
import edu.mit.csail.aeolus.infoflow.Label;
import edu.mit.csail.aeolus.infoflow.PrincipalID;
import edu.mit.csail.aeolus.infoflow.TagID;

/** 
 * Proxy object interface that is exposed to the user applications
 * 
 */
public interface UserProxyInterface {
	
	// PI LAUNCH
	
	public void launchPlatformInstance(String hostname, PrincipalID pids, String appName, String appParams)
		throws InfoFlowControlException, AuthorityException, AeolusException, RpcException,
			InvocationTargetException, ClassNotFoundException, NoSuchMethodException, Exception;
	public void shutdown() throws AuthorityException, AeolusException, RpcException;
	public void deleteNode(String hostname) throws AuthorityException, AeolusException;
	// PRINCIPALS

	public boolean canSwitch(PrincipalID pid);
	public boolean isActee(PrincipalID pid1, PrincipalID pid2);
	public boolean hasAuthority(PrincipalID pid, TagID tid);
	public long getUserAppDomainId();
	
	/**
	 * Creates a new principal in the process's principal's core 
	 * @return new principal ID
	 * @throws AuthorityException 
	 * effects process principal acts for the new principal
	 */
	public PrincipalID createPrincipal() throws AuthorityException, AeolusException;
	
	/**
	 * Creates a new principal in a new core 
	 * @return new principal ID
	 * @throws AuthorityException 
	 * effects process principal acts for the new principal
	 */
	public PrincipalID createPrincipalInNewCore() throws AuthorityException, AeolusException;
	
	/**
	 * Creates a new principal in the same core as principal p
	 * @return new principal ID
	 * @throws AuthorityException 
	 * effects process principal acts for the new principal
	 */
	public PrincipalID createPrincipal(PrincipalID p) throws AuthorityException, AeolusException;
	
	// TAGS
	
	/**
	 * Creates a new top-level tag 
	 * @return new tag ID
	 * @throws AuthorityException 
	 * effects process principal has authority for the new tag
	 */
	public TagID createTag() throws AuthorityException, AeolusException;
	
	/**
	 * Creates a sub-tag of tag t
	 * Tag t must be a top-level tag
	 * @return new sub-tag ID
	 * @throws AuthorityException 
	 * effects process's principal has authority for the new tag
	 */
	public TagID createSubTag(TagID t) throws AuthorityException, AeolusException;
	
	/**
	 * Creates a sub-tag of tag t and stores the tag info in the core
	 * of principal p
	 * Tag t must be a top-level tag
	 * @return new sub-tag ID
	 * @throws AuthorityException 
	 * effects process's principal has authority for the new tag
	 */
	public TagID createSubTag(TagID t, PrincipalID p) throws AuthorityException, AeolusException;
	
	
	// DELEGATIONS AND REVOCATIONS API
	
	/**
	 * Adds an act-for link from actor principal p2 to actee principal p1
	 * Process's principal must act for principal p1
	 */
	public void actFor(PrincipalID p1, PrincipalID p2) throws AuthorityException, AeolusException;
	
	/**
	 * Removes the act-for link from principal p2 to principal p1
	 * Process's principal must act for principal p1
	 */
	public void revokeActFor(PrincipalID p1, PrincipalID p2) throws AuthorityException, AeolusException;
	
	/**
	 * Gives authority for tag t from grantor principal p1 to grantee principal p2
	 * Process's principal must act for the grantor principal p1, 
	 * the grantor principal p1 must have authority for tag t
	 */
	public void delegate(TagID t, PrincipalID p1, PrincipalID p2) throws AuthorityException, AeolusException;
	
	/**
	 * Revokes authority for tag t from grantor principal p1 to grantee principal p2
	 * Process's principal must act for the grantor principal p1 and the tag t must have been 
	 * previously delegated from p1 to p2 
	 */
	public void revokeDelegate(TagID t, PrincipalID p1, PrincipalID p2) throws AuthorityException, AeolusException;

	
	// LABEL MANIPULATIONS API

	/**
     * Adds given tag ID to the secrecy label
     * @throws AeolusException if running in shared state
     */
	public void addSecrecy(TagID tid) throws AeolusException;

	/**
     * Removes given tag ID from the integrity label
     * @throws AeolusException if running in shared state
     */
	public void removeIntegrity(TagID tid) throws AeolusException;

	/**
     * Removes given tag ID from the secrecy label
     * @throws AeolusException if process is running with a public PID
     * or does not have authority to remove the tag
     */
	public void declassify(TagID tid) throws AuthorityException, AeolusException;

	 /**
     * Adds given tag ID to the integrity label
     * @throws AeolusException if process is running with a public PID
     * or does not have authority to add the tag
     */
	public void endorse(TagID tid) throws AeolusException;
	
	/**
     * Adds given tag IDs to the secrecy label
     * @param tids - list of tag IDs to add
     * @throws AeolusException if running in shared state
     */
	public void addSecrecy(List<TagID> tids) throws AeolusException;

	/**
     * Removes given tag IDs from the integrity label
     * @param tids - list of tag IDs to remove
     * @throws AeolusException if running in shared state
     */
	public void removeIntegrity(List<TagID> tids) throws AeolusException;

	/**
     * Removes given tag IDs from the secrecy label
     * @param tids - list of tag IDs to remove
     * @throws AeolusException if process is running with a public PID
     * or does not have authority to remove the tags
     */
	public void declassify(List<TagID> tids) throws AeolusException;

	 /**
     * Adds given tag IDs to the integrity label
     * @param tids - list of tag IDs to add
     * @throws AeolusException if process is running with a public PID
     * or does not have authority to add the tags
     */
	public void endorse(List<TagID> tids) throws AeolusException;

	public Label getSLabel();
	public Label getILabel();
	
	 /**
     * Returns a copy of the InfoFlowState object
     */
	public InfoFlowState getInfoFlowState();
	
	
	// AUTHORITY CLOSURE API
	
	/**
	 * Creates a new closure 
	 * A new anonymous principal is generated and bound to this key and new closure ID,
	 * this principal has no authority
	 * @returns new closure ID
	 */
	//public ClosureID createClosure(String key) throws AeolusException;
	
//	public ClosureID createClosure(Class<? extends AeolusCallable> closureInterface) throws AeolusException;
	
	/**
	 * Creates a new closure and stores the closure info in the same core 
	 * as principal pid
	 * A new anonymous principal is generated and bound to this key and new closure ID,
	 * this principal has no authority
	 * @return new closure ID
	 */
//	public ClosureID createClosure(Class<? extends AeolusCallable> closureInterface, PrincipalID pid);
	
//	public AeolusCallable getClosure(ClosureID cid, Class<?> closureClass) throws AeolusException;
	
	/**
	 * Adds an act-for link from the anonymous principal of closure cid to principal pid
	 * (the anonymous principal will be authoritative for all tags that principal pid has authority for)
	 * Closure cid must exist and process's principal must act for principal pid
	 */
//	public void closureActFor(PrincipalID pid, ClosureID cid) throws AuthorityException, AeolusException;
	
	/**
	 * Gives authority for tag tid from grantor principal pid to grantee anonymous principal of closure cid
	 * Closure cid must exist, process principal must act for the grantor principal pid, 
	 * the grantor principal pid must have authority for tag tid 
	 */
//	public void closureDelegate(TagID tid, PrincipalID pid, ClosureID cid) throws AuthorityException, AeolusException;

	// CALL, FORK API

	 /**
     *  Runs code with the specified principal ID in the same process
     *  @param codeObj - code object to invoke
     *  @param calleePid - PID to run the code with
     * 
     *  After invocation the principal is restored and the process labels are updated 
     *  to reflect any additional contamination picked up during the call
     * 
     *  @throws AuthorityException if requester principal does not act for calleePid or 
     *  process is running with public PID, or with same PID and the specified calleePid
	 * @throws Exception - exception thrown when executing the codeObj
     */
        public <T> T call(Callable<T> codeObj, PrincipalID calleePid) throws AuthorityException, Exception;

	/**
	 * Runs code with the specified principal ID in a different new process
	 * @param codeObj - code object to invoke
	 * @param pid - PID to run the code with
	 * @throws AuthorityException if requester principal does not act for pid or has the same PID
	 */
	public void fork(Runnable codeObj, PrincipalID pid) throws AuthorityException, AeolusException;

	
	// CLOSURE API

    public InfoFlowState aeolusStartClosureCall(AeolusClosure closure);
    public void aeolusEndClosureCall(AeolusClosure closure, InfoFlowState oldState);
    public InfoFlowState aeolusStartSharedCall(AeolusShared shared) throws InfoFlowControlException ;
    public void aeolusEndSharedCall(AeolusShared shared, InfoFlowState oldState);
    public InfoFlowState aeolusStartSharedRead(AeolusShared shared) throws InfoFlowControlException ;
    public void aeolusEndSharedRead(AeolusShared shared, InfoFlowState oldState);
    public InfoFlowState aeolusStartSharedWrite(AeolusShared shared) throws InfoFlowControlException ;
    public void aeolusEndSharedWrite(AeolusShared shared, InfoFlowState oldState);
	
		// RPC API
	
	/**
	 * Registers user service with Aeolus by binding the service name 
	 * with the provided service class
     * If a service by the same name already exists, it will be replaced
     * @param serviceName - name to bind the service to
     * @param service - service class implementing the serviceName interface 
     */
	public void registerService(String serviceName, Class<?> service) throws InfoFlowControlException;
	
	/**
     * Return a proxy stub to the remote service object which implements the serviceName interface
     * @param remoteHostName - service location
	 * @param serviceName - name that the service is bound to on the remote host
	 * @param serviceClass - local Class which defines the interface of the service
     */
	public Object getService(String remoteHostName, String serviceName, Class<?> serviceClass);
	
	
	// SHAREDSTATE API

    public AeolusShared getRoot() ;
    public void setRoot(AeolusShared root) throws InfoFlowControlException;

	// FILE API

	public boolean createFile(String hostname, String filePath, Label sLabelFile, Label iLabelFile) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;

	public int readFile(String hostname, String filePath, byte[] fileContent) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;
	
	public void writeFile(String hostname, String filePath, byte[] fileContent, boolean appendMode) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;
		
	public boolean createDir(String hostname, String filePath, Label sLabelFile, Label iLabelFile) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;

	public String[] listDir(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;

	public boolean delete(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;
	
	public Label getIntegrityLabel(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;
	
	public Label getSecrecyLabel(String hostname, String filePath) throws FileNotFoundException, IOException, InfoFlowControlException, AeolusException;
	
	// FILESTREAM API
	
	public FileSystemReply<Integer> openFileInputStream(String hostname, String filePath) throws AeolusException, RemoteException, FileNotFoundException, IOException;

	public FileSystemReply<Integer> openFileOutputStream(String hostname, String filePath, boolean appendMode) throws AeolusException, IOException;
	public FileSystemClientInterface getFSClient(String hostname) throws AeolusException;
	public boolean canWriteToStream(int filestreamHandle, Label sLabel, Label iLabel) throws IOException;
	public boolean canReadFromStream(int filestreamHandle, Label sLabel, Label iLabel) throws IOException;
	public void closeFilestream(int filestreamHandle, String filePath, String hostname) throws IOException, AeolusException;
	 
	 
	 // I/O DEVICES
	 
	 public void writeToIODevice() throws InfoFlowControlException, AeolusException;
	 
	 public void readFromIODevice() throws InfoFlowControlException, AeolusException;
	 
	 // AUDITING INTERFACE
	 
	 public EventID getLastEID();
	 public EventID createEvent(String methodName, Serializable[] parameters, Serializable result);
	 public EventID createEvent(String methodName, Serializable[] parameters, Serializable result, Date timestamp);
	 public EventID createEvent(String methodName, List<EventID> preds, Serializable[] parameters, Serializable result);
	 public EventID createEvent(String methodName, List<EventID> preds, Serializable[] parameters, Serializable result, Date timestamp);
	 public EventID createEvent(String methodName, List<EventID> preds, Serializable[] parameters, Serializable result, CompletionStatus status, Date timestamp);
	 public EventID createEvent(int methodName, EventID pred, Serializable[] parameters, Serializable result);
	 public EventID createEvent(int methodName, EventID pred, Serializable[] parameters, Serializable result, CompletionStatus status);
	 public void createBlock(); 
	 public void setCurrentBlock(PrincipalID p); 
}
