package edu.mit.csail.aeolus.runtime;

import java.util.concurrent.Callable;

/**
 * Request sent from a User-d to the Sys-d 
 * Constructed from a Callable object, stores the return value of the request
 * and any exceptions generated by the code inside the Callable object 
 * 
 * @author viq
 *
 */
public class UserdRequest<T>{

	/** True if request has a reply */
	private boolean reply = false;
	
	/** True if request blocks the SYS-D */
	private boolean blocking = false;
	
	/** True if request finished executing */
	private boolean done = false;
	
	/** Code to execute */
	private Callable<T> callable = null;
	
	/** Return value of the executed code */
	private T returnValue = null;

	/** Any exception thrown during request execution */
	private Exception exception = null;
	
	/**
	 * Assemble a new request with a return value
	 * @param codeToExecute - code to be executed by the SYS-d
	 */
	public UserdRequest(Callable<T> codeToExecute)
	{
		this.callable = codeToExecute;
		this.reply = true;
	}
	
	/**
	 * Returns the code to be executed
	 * @return Callable object with a return value of type Object
	 * 		   or null if request constructed from a Runnable 
	 */
	public Callable<T> getCallable()
	{
		return this.callable;
	}
	
	/** 
	 * Returns true if this request expects a return value
	 */
	public boolean withReply()
	{
		return this.reply;
	}
	
	/**
	 * Returns true if the request should block the SYS-D thread
	 */
	public boolean isBlocking()
	{
		return this.blocking;
	}
	
	/**
	 * Returns true if the request has completed
	 */
	public boolean isDone()
	{
		return this.done;
	}
	
	/**
	 * Returns the exception that occured during the execution of the request
	 * or null no exception was thrown
	 */
	public Exception getException()
	{
		return this.exception;
	}
	
	public T getReturnValue()
	{
		return this.returnValue;
	}
	
	/** Request has completed */
	public void setDone()
	{
		this.done = true;
	}
	
	/** 
	 * Set the request to block the SYS-D thread
	 */
	public void setBlocking()
	{
		this.blocking = true;
	}
	
	public void setReturnValue(T value)
	{
		this.returnValue = value;
	}
	
	/**
	 * Store the request execution exception
	 */
	public void setException(Exception e)
	{
		this.exception = e;
	}
}
