package edu.mit.csail.aeolus.security;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.lang.reflect.*;
import sun.misc.Unsafe;
import sun.reflect.FieldAccessor;

// The following imports are just so we can refer to the classes to build
// a list of things that are illegal to clone.
import edu.mit.csail.aeolus.runtime.*;

/**
 * This class handles deep copying of objects, using only a shallow copy when
 * the object is known to be immutable.
 *
 * @author das
 */
public final class AeolusCloner
{
    /**
     * A Cloner is an object that knows how to clone a particular
     * class.  Cloners are part of the internals of AeolusCloner,
     * which can clone any class.
     *
     * Using different cloners for different classes means that we can
     * do a quick dispatch to the cloner's Clone() method without doing
     * any extra tests to see if the object is immutable or an array or
     * whatnot.
     */
    protected static abstract class Cloner<T> {
	abstract public T Clone(T obj)
	    throws InstantiationException, IllegalAccessException;

     	abstract public T Clone(T obj, OAHashMap<Object, Object> alreadyCloned) throws InstantiationException, IllegalAccessException;
    }

    /**
     * An ImmutableCloner is registered for each immutable object.
     *
     * XXX This could be a singleton because it's not actually
     * specialized to any specific type T.
     */
    protected static final class ImmutableCloner<T> extends Cloner<T> {
	public T Clone(T obj) throws InstantiationException, IllegalAccessException
	{
	    return (obj);
	}

     	public T Clone(T obj, OAHashMap<Object, Object> alreadyCloned) throws InstantiationException, IllegalAccessException
	{
	    return (obj);
	}
    }

    /**
     * This cloner always throws an exception.  This is the appropriate cloner
     * to use for objects that should be illegal to clone.
     */
    protected static final class IllegalCloner<T> extends Cloner<T> {
	public T Clone(T obj) throws InstantiationException, IllegalAccessException
	{
	    throw new IllegalAccessException("It is illegal to clone " + obj);
	}

     	public T Clone(T obj, OAHashMap<Object, Object> alreadyCloned) throws InstantiationException, IllegalAccessException
	{
	    return Clone(obj);
	}
    }

    /**
     * The ObjectCloner class handles most ordinary, mutable objects.
     * It works by cloning each field of the object, including fields
     * of superclasses.
     */
    protected static final class ObjectCloner<T> extends Cloner<T> {
	private final Class<T> clazz;
	private final FieldAccessor[] fields;

	public ObjectCloner(Class<T> clazz) {
	    this.clazz = clazz;

	    try {
		Method acquireFieldAccessorMethod = Field.class.getDeclaredMethod("acquireFieldAccessor", boolean.class);
		acquireFieldAccessorMethod.setAccessible(true);
		List<FieldAccessor> fieldList = new LinkedList<FieldAccessor>();
		for (Class<?> c = clazz; c != Object.class; c = c.getSuperclass()) {
		    for (Field field : c.getDeclaredFields()) {
			if (Modifier.isStatic(field.getModifiers()))
			    continue;
			fieldList.add((FieldAccessor)acquireFieldAccessorMethod.invoke(field, true));
		    }
		}
		acquireFieldAccessorMethod.setAccessible(false);
		fields = fieldList.toArray(new FieldAccessor[0]);
	    } catch (NoSuchMethodException e) {	// Java needs to get a CLU.
		throw new RuntimeException(e);
	    } catch (IllegalAccessException e) {
		throw new RuntimeException(e);
	    } catch (InvocationTargetException e) {
		throw new RuntimeException(e);
	    }
	}

	public T Clone(T obj) throws InstantiationException, IllegalAccessException {
	    return (Clone(obj, new OAHashMap<Object, Object>()));
	}

     	public T Clone(T obj, OAHashMap<Object, Object> alreadyCloned) throws InstantiationException, IllegalAccessException
	{
	    Object existingCopy = alreadyCloned.get(obj);
	    if (existingCopy != null)
		return ((T)existingCopy);

	    T copy = newInstance(clazz);
	    alreadyCloned.put(obj, copy);
	    for (FieldAccessor field : fields)
		field.set(copy, CloneInternal(field.get(obj), alreadyCloned));
	    return (copy);
	}
    }

    /**
     * ArrayCloner handles cloning of arrays of mutable objects.
     */
    protected static final class ArrayCloner<T> extends Cloner<T> {
	private final Class<?> elementType;

	public ArrayCloner(Class<T> clazz) {
	    elementType = clazz.getComponentType();
	}

	public T Clone(T obj) throws InstantiationException, IllegalAccessException {
	    return (Clone(obj, new OAHashMap<Object, Object>()));
	}

     	public T Clone(T obj, OAHashMap<Object, Object> alreadyCloned) throws InstantiationException, IllegalAccessException
	{
	    Object existingCopy = alreadyCloned.get(obj);
	    if (existingCopy != null)
		return ((T)existingCopy);

	    int len = Array.getLength(obj);
	    T copy = (T)Array.newInstance(elementType, len);
	    alreadyCloned.put(obj, copy);
	    for (int i = 0; i < len; i++) {
		Array.set(copy, i, CloneInternal(Array.get(obj, i), alreadyCloned));
	    }
	    return (copy);
	}
    }

    /**
     * For some reason, the OpenJDK JVM's implementations of
     * JVM_GetArrayElement() and JVM_SetArrayElement() are incredibly
     * slow, so treating arrays of immutable elements specially turns
     * out to be a big win.
     */
    protected static final class ImmutableArrayCloner<T> extends Cloner<T> {
	private Class<?> elementType;

	public ImmutableArrayCloner(Class<T> clazz) {
	    elementType = clazz.getComponentType();
	}

	public T Clone(T obj) throws InstantiationException, IllegalAccessException
	{
	    return (Clone(obj, new OAHashMap<Object, Object>()));
	}

     	public T Clone(T obj, OAHashMap<Object, Object> alreadyCloned) throws InstantiationException, IllegalAccessException
	{
	    Object existingCopy = alreadyCloned.get(obj);
	    if (existingCopy != null)
		return ((T)existingCopy);

	    int len = Array.getLength(obj);
	    T copy = (T)Array.newInstance(elementType, len);
	    alreadyCloned.put(obj, copy);
	    System.arraycopy(obj, 0, copy, 0, len);
	    return (copy);
	}
    }

    /**
     * This factory method instantiates the appropriate Cloner for the
     * given class.
     */
    static <T> Cloner<T> makeCloner(Class<T> clazz) {

	if (clazz.isArray()) {
	    if (verifier.isSharableReferenceType(clazz.getComponentType()))
		return (new ImmutableArrayCloner<T>(clazz));
	    else
		return (new ArrayCloner<T>(clazz));
	}

	if (verifier.isSharableClass(clazz)) {
	    System.out.println(clazz + " is immutable");
	    return (new ImmutableCloner<T>());
	}

	return (new ObjectCloner<T>(clazz));
    }

    // XXX This needs to be converted to a concurrent hash map so that the
    // cloner is thread-safe.  If this slows down retrievals, though, we
    // can do something fancier like making sure that the hash map is
    // pre-populated with all the types it's ever going to see.
    protected static final Map<Class<?>, Cloner<?>> classInfo = new ConcurrentHashMap<Class<?>, Cloner<?>>(64, 0.5f);

    public static final SharabilityVerifier verifier = new SharabilityVerifier();

    // No need to have multiple copies of these, since they have no state.
    protected static final Cloner<?> theImmutableCloner = new ImmutableCloner<Class<?>>();
    protected static final Cloner<?> theIllegalCloner = new IllegalCloner<Class<?>>();

    
    /*
     * Sun doesn't want us to have access to sun.misc.Unsafe, but we
     * can foil them with reflection!  That's what the following hack
     * is about.
     */
    private static final Unsafe unsafe;
    static {
	try {
	    Field f = Unsafe.class.getDeclaredField("theUnsafe");
	    f.setAccessible(true);
	    unsafe = (Unsafe)f.get(null);
	    f.setAccessible(false);
	} catch (Exception e) {
	    throw new RuntimeException(e);
	}

	/*
	 * Some things should never be cloned.  It's probably a bit dubious
	 * that we clone everything that isn't explicitly disallowed, but
	 * it's impractical to patch a new marker interface into nearly
	 * every type in the JDK and third-party libraries.
	 */
	addIllegal(UserProxy.class);
	addIllegal(UserDomain.class);
	addIllegal(SystemDomain.class);
	addIllegal(Thread.class);
    }

    /**
     * Mark the given class as immutable (i.e., doesn't require cloning).
     */
    protected static void addImmutable(Class<?> clazz) {
	classInfo.put(clazz, theImmutableCloner);
    }

    /**
     * Mark the given class as illegal to clone.  An attempt to clone
     * an instance of this object will result in an exception.
     */
    protected static void addIllegal(Class<?> clazz) {
	classInfo.put(clazz, theIllegalCloner);
    }

    /**
     * Get the cloner for the given class. The common case is that we expect
     * the cloner to be cached in the classInfo map. Otherwise, we need to
     * figure out what kind of class it is and assign it a cloner.
     */
    @SuppressWarnings("unchecked")
    protected static <T> Cloner<T> getCloner(Class<T> clazz) {
	// This is the critical path.
	Cloner<T> cloner = ((Cloner<T>)classInfo.get(clazz));
	if (cloner != null)
	    return (cloner);

	// This is the slow path.
	cloner = makeCloner(clazz);
	//	System.out.println("making cloner for " + clazz + ": " + cloner);
	classInfo.put(clazz, cloner);
	return (cloner);
    }

    private AeolusCloner() {
    }

    /**
     * Allocate an *uninitialized* instance of a class. This will not work
     * on array types.
     */
    @SuppressWarnings("unchecked")
    protected static <T> T newInstance(Class<T> clazz)
	throws InstantiationException, IllegalAccessException
    {
	return ((T)unsafe.allocateInstance((java.lang.Class)clazz));
    }

    /**
     * This is the public interface to AeolusCloner. It does a deep copy of
     * the given object, but avoids copies for any known-immutable objects.
     *
     * XXX The method name should be lowercase to conform to the Java
     * standard naming scheme.
     */
    public static <T> T Clone(T obj)
    {
        try {
            if (obj == null)
                return (null);
            
            Class<T> clazz = (Class<T>)obj.getClass();
            Cloner<T> cloner = getCloner(clazz);
            
            /*
             * NB: We don't allocate the alreadyCloned hash map here;
             * that job is the responsibility of the first cloner that
             * needs it. Immutable/array objects won't require the
             * table at all, and we don't want to waste time creating
             * one here.
             */
            
            //	System.out.println("Clone " + clazz);
            return (cloner.Clone(obj));
        } catch (InstantiationException ex) {
            throw new RuntimeException(ex);
        } catch (IllegalAccessException ex) {
            throw new RuntimeException(ex);
	}
    }

    /**
     * This is the internal version of Clone(), which takes an extra
     * alreadyCloned table as its argument, to handle cycle and
     * duplicate detection.  This method handles all of the generic
     * parts of cloning an object, while the cloners handle the
     * type-specific parts.  This method is mutually recursive with
     * the cloners.
     */
    protected static <T> T CloneInternal(T obj, OAHashMap<Object, Object> alreadyCloned)
	throws InstantiationException, IllegalAccessException
    {
	if (obj == null)
	    return (null);

	Class<T> clazz = (Class<T>)obj.getClass();
	Cloner<T> cloner = getCloner(clazz);

	//	System.out.println("CloneInternal " + clazz);
	return (cloner.Clone(obj, alreadyCloned));
    }

    /**
     * A few sanity checks...
     */
    static void runTests() throws Exception {
	AeolusCloner cloner = new AeolusCloner();

	String t = cloner.Clone("foo");
	if (t != "foo")
	    throw new RuntimeException();

	MyTestClass o = new MyTestClass(42);
	MyTestClass oCopy = cloner.Clone(o);
	o.x = 43;
	if (oCopy.x != 42)
	    throw new RuntimeException();

	MyTestClass2 o2 = new MyTestClass2();
	o2.a[0] = 1;
	o2.a[1] = 2;
	MyTestClass2 o2Copy = cloner.Clone(o2);
	o2.a[0] = 3;
	if (o2Copy.a[0] != 1 || o2Copy.a[1] != 2)
	    throw new RuntimeException();
    }

    static void runBenchmarks() throws Exception {
	AeolusCloner cloner = new AeolusCloner();
	Object o = new MyTestClass(42);
	//Object o = "foo";
	long start, end;

	// These iterations are to warm up the JIT and get it to do the
	// appropriate optimizations. We don't want it recompiling our
	// methods during execution of the benchmark!
        for (int i = 0; i < 100000; i++) {
	    //o = cloner.Clone(o);
	    Object o2 = cloner.Clone(o);
	    //o = new MyTestClass(o);
        }

	start = System.currentTimeMillis();
	//	while(true) {
        for (int i = 0; i < 10000000; i++) {
	    //o = cloner.Clone(o);
	    Object o2 = cloner.Clone(o);
	    //o = new MyTestClass(o);
        }
	end = System.currentTimeMillis();
	System.out.println(end - start);
    }

    public static void main(String args[]) throws Exception {
	//	OAHashMap.runTests();
	//	AeolusCloner.runTests();
	AeolusCloner.runBenchmarks();
    }

}

class MyTestClass {
    public int x;
    
    public MyTestClass(int x) {
	this.x = x;
    }

    public MyTestClass(MyTestClass o) {
	this.x = o.x;
    }

}

class MyImmutableTestClass {
    public final int x;
    
    public MyImmutableTestClass(int x) {
	this.x = x;
    }

    public MyImmutableTestClass(MyTestClass o) {
	this.x = o.x;
    }
}

class MyTestClass2 {
    public int a[];

    public MyTestClass2() {
	a = new int[10];
    }
}

class MyTestClass3 {
    public Object a[];

    public MyTestClass3() {
	a = new Object[5];
    }
}

class MyTestClass4 {
    public Object a;
    public Object b;
    public Object c;
    public Object d;
    /*
    public Object e;
    public Object f;
    public Object g;
    public Object h;
    */

    public MyTestClass4() {
	a = new Object();
	b = new Object();
	c = new Object();
	d = new Object();
	/*
	e = new Object();
	f = new Object();
	g = new Object();
	h = new Object();
	*/
    }
}

/**
 * This class implements an open-addressed hash table. It is designed
 * to have especially low overhead, especially when the hash table is
 * going to be used minimally, then thrown away.  This is achieved
 * mainly by avoiding unnecessary memory allocation (via open
 * addressing).
 *
 * The interface differs from a standard Map in that the methods
 * require an additional parameter specifying the hash value.  This
 * avoids duplicate hash computations when a value must be looked up
 * in multiple hash tables.
 */
final class OAHashMap<K, V>
{
    // primes[i] = the largest prime less than or equal to 2**i
    static final int primes[] = {
	1, 2, 3, 7, 13, 31, 61, 127, 251, 509, 1021,
	2039, 4093, 8191, 16381, 32749, 65521, 131071, 262139, 524287, 1048573
    };

    // We use heuristics to adapt the initial table size to the workload.
    private static int initialSizeIndex = 2;
    private static int instances = 0;
    private static int expansions = 0;

    private final float fillFactor;	// Fill factor

    // All of these variables are initialized in initTable().
    private int sizeIndex;	// Index of the current size in the size table
    private int	size;		// Current table size
    private int numEntries;	// Current number of entries
    private int maxEntries;	// Max entries before expansion is needed
    private Object table[];

    @SuppressWarnings("unchecked")
    private void initTable(int log2Size) {
	sizeIndex = log2Size;
	size = primes[log2Size];
	maxEntries = (int)(fillFactor * size);
	numEntries = 0;

	/*
	 * It turns out that it's quite slow to allocate Java arrays, and we
	 * can save 30ns per clone by keeping the keys and the values in a
	 * single array.  The first half of the array stores the keys, while
	 * the second half stores the values.
	 */
	table = new Object[size * 2];
    }

    /**
     * Create a new hash table of size approximately 2**log2Size.  The
     * load factor must not exceed 0.5, since quadratic probing may
     * fail to find a free bucket if the table is more than half full.
     */
    public OAHashMap(int log2Size, float fillFactor) {
	assert(log2Size >= 0 && log2Size < primes.length);

	if (fillFactor <= 0.0f || fillFactor > 0.5f)
	    throw new IllegalArgumentException();
	this.fillFactor = fillFactor;

	initTable(log2Size);
	instances++;
    }

    public OAHashMap(int log2Size) {
	/*
	 * The choice of default fill factor is important.  Too large
	 * and we waste time making bigger tables.  Too small and we
	 * get hash collisions.  Some cursory benchmarks showed that
	 * 40% is a good choice.
	 */
	this(log2Size, 0.4f);
    }

    public OAHashMap() {
	this(initialSizeIndex);
    }

    private K getKey(int index) {
	return ((K)table[index]);
    }

    private V getValue(int index) {
	return ((V)table[size + index]);
    }

    private void setKey(int index, K key) {
	table[index] = key;
    }

    private void setValue(int index, V value) {
	table[size + index] = value;
    }

    /**
     * Look up the given key in the hash table.
     */
    @SuppressWarnings("unchecked")
    public V get(K key) {
	int hashIndex = System.identityHashCode(key);

	// Quadratic probing
	for (int i = 0; ; i++) {
	    int bucket = (hashIndex + (i * i)) % size;
	    K bucketKey = getKey(bucket);
	    if (bucketKey == null) {
		return (null);
	    } else if (bucketKey == key) {
		return (getValue(bucket));
	    }
	}

	// Not reached: control never exits the loop above except via return
    }

    /**
     * Insert the given key/value pair into the hash table.
     * Returns the old value, or null if one didn't exist.
     */
    @SuppressWarnings("unchecked")
    public V put(K key, V value) {
	int hashIndex = System.identityHashCode(key);

	// Expand table if needed.  Note that we might expand the table
	// unnecessarily if the entry is already in the table, but that's
	// okay.
	if (numEntries >= maxEntries)
	    expand();

	// Quadratic probing
	for (int i = 0; ; i++) {
	    int bucket = (hashIndex + (i * i)) % size;
	    K bucketKey = getKey(bucket);
	    if (bucketKey == null) {
		// Found an empty bucket, so insert here.
		setKey(bucket, key);
		setValue(bucket, value);
		numEntries++;
		return (null);
	    } else if (bucketKey == key) {
		// Found an existing entry, so update it.
		V oldValue = getValue(bucket);
		setValue(bucket, value);
		return (oldValue);
	    }
	}

	// Not reached: control never exits the loop above except via return
    }

    /**
     * Roughly double the size of the hash table.
     */
    @SuppressWarnings("unchecked")
    protected void expand() {

	// Save the old table and allocate a new one.
	Object oldTable[] = table;
	int oldSize = size;

	initTable(sizeIndex + 1);

	// Rehash into the new table.
	for (int i = 0; i < oldSize; i++) {
	    if (oldTable[i] != null)
		put((K)oldTable[i], (V)oldTable[oldSize + i]);
	}

	/*
	 * Update our heuristics to avoid having to resize hash tables in
	 * the future.  We start using a larger default table size whenever
	 * at least 1/3 of the tables we've made so far had to be expanded.
	 *
	 * Of course a real program would have many objects of varying
	 * sizes (and hence varying hash set sizes) to clone, so what
	 * we should really be doing is use a cleverer heuristic that
	 * takes more information from the caller.
	 *
	 * XXX also, there's no heuristic to decrease the default size
	 */
	expansions++;
	if (instances > 100 && expansions * 3 > instances) {
	    initialSizeIndex++;
	    expansions = instances = 0;
	}
    }

    /**
     * Unit tests for the class.
     */
    static void runTests() {
	OAHashMap<String, Integer> ht = new OAHashMap<String, Integer>(2);

	ht.put("foo", 42);
	ht.put("bar", 43);
	if (ht.get("foo") != 42)
	    throw new RuntimeException();
	if (ht.get("bar") != 43)
	    throw new RuntimeException();
	if (ht.get("baz") != null)
	    throw new RuntimeException();
	ht.put("foo1", 42);
	ht.put("foo2", 42);
	ht.put("foo3", 42);
	ht.put("foo4", 42);
	ht.put("foo5", 42);
	ht.put("foo6", 42);
	ht.put("foo7", 42);
	ht.put("foo8", 42);
	if (ht.get("foo") != 42)
	    throw new RuntimeException();
	if (ht.get("bar") != 43)
	    throw new RuntimeException();
	if (ht.get("baz") != null)
	    throw new RuntimeException();
    }
}
