package edu.mit.csail.aeolus.security;

import java.io.*;
import java.util.*;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.reflect.Method;
import java.security.ProtectionDomain;

import edu.mit.csail.aeolus.api.*;
import edu.mit.csail.aeolus.infoflow.InfoFlowState;

import org.objectweb.asm.*;
import org.objectweb.asm.tree.*;
import org.objectweb.asm.util.*;

/**
 * The ClassTransformer performs load-time transformations on classes
 * to implement Aeolus shared objects and closures.  Shared object
 * methods are transformed to copy all of their arguments, return
 * values, and thrown exceptions.  Closure methods are transformed to
 * change the principal for the duration of the closure, and merge the
 * information flow labels on return.  Additionally, synchronization
 * via MONITORENTER and MONITOREXIT JVM instructions is removed to
 * ensure that covert channels via object locks are not possible.
 *
 * @author das
 */
public final class ClassTransformer implements ClassFileTransformer
{
    private static enum TransformationType {
	NONE(true, false, null),
        CALLABLE(false, false, AeolusClosure.class),
	SHARED(true, true, AeolusShared.class);
	
	private final boolean copyArgs;
	private final boolean transformConstructor;
	private final Class classType;

	TransformationType(boolean copyArgs, boolean transformConstructor,
			   Class classType)
	{
	    this.copyArgs = copyArgs;
	    this.transformConstructor = transformConstructor;
	    this.classType = classType;
	}
	
	boolean shouldCopyArgs() {
	    return (copyArgs);
	}

	boolean shouldTransformConstructor() {
	    return (transformConstructor);
	}

	Class classType() {
	    return (classType);
	}
    };

    private static String cloneMethodDesc = Type.getMethodDescriptor(Type.getType(Object.class), new Type[] { Type.getType(Object.class) } );

    protected final AeolusClassLoader loader;

    public ClassTransformer(AeolusClassLoader loader) {
	this.loader = loader;
    }

    public byte[] transform(ClassLoader loader, String className, Class<?> classBeingRedefined, ProtectionDomain protectionDomain, byte[] classfileBuffer) throws IllegalClassFormatException {
	ClassReader reader = new ClassReader(classfileBuffer);
	ClassNode clazz = new ClassNode();
	reader.accept(clazz, 0);

	TransformationType transformationType;
	try {
	    transformationType = getTransformationType(Class.forName(getExternalName(clazz.superName), true, loader));
	    if (transformationType == TransformationType.NONE)
		return (null);
	} catch (ClassNotFoundException e) {
	    throw new RuntimeException("Could not find loaded superclass", e);
	}

	ClassWriter writer = new AeolusClassWriter(ClassWriter.COMPUTE_FRAMES, loader);
	try {
	    transform(clazz, transformationType);
	} catch (NoSuchMethodException e) {
	    throw new RuntimeException(e);
	}

	//	ClassVisitor trace = new TraceClassVisitor(new PrintWriter(System.out));
	//	clazz.accept(trace);

	clazz.accept(writer);
	return (writer.toByteArray());
    }

    protected String getExternalName(String internalName) {
	return (internalName.replace('/', '.'));
    }

    /**
     * Determine which transformation type to apply: the shared-object
     * transformation, the closure transformation, or neither.
     *
     * XXX We don't handle classes that are both shared objects and
     * closures.  Currently such classes can't exist because both
     * AeolusShared and AeolusClosure are abstract classes, and Java
     * doesn't support multiple inheritance.  But if that limitation
     * is removed in the future, this needs to be fixed.
     */
    protected TransformationType getTransformationType(Class clazz) {
	if (clazz == null)
	    return (TransformationType.NONE);
	if (clazz == TransformationType.CALLABLE.classType())
	    return (TransformationType.CALLABLE);
	if (clazz == TransformationType.SHARED.classType())
	    return (TransformationType.SHARED);
	return (getTransformationType(clazz.getSuperclass()));
    }

    /**
     * Transform a single method within a class.
     */
    protected void transform(ClassNode clazz, TransformationType transformationType) throws NoSuchMethodException {
	for (MethodNode method : new ArrayList<MethodNode>(clazz.methods)) {
	    removeSynchronization(method);
	    if (shouldTransformMethod(method, transformationType)) {
		System.out.println("Transforming method in " + clazz.name + ": " + method.name + method.desc);
		String wrappedMethodName = getWrappedMethodName(method.name);
		clazz.methods.add(generateWrapper(method, clazz, wrappedMethodName, transformationType));
		renameAndPrivatizeMethod(method, wrappedMethodName);
	    }
	}
    }

    protected String getWrappedMethodName(String methodName) {
	if (methodName.equals("<init>"))
	    return ("$init$aeolus");
	else
	    return (methodName + "$aeolus");
    }

    /**
     * Remove synchronization instructions (MONITORENTER and MONITOREXIT)
     * from the given method.  Aeolus prevents separate threads from sharing
     * mutable state, so the synchronization should not be needed.
     * Additionally, threads could leak information by acquiring locks on
     * immutable objects that Aeolus has allowed them to share, so we need
     * to disallow these instructions.
     *
     * If the method itself is synchronized, we don't currently remove
     * that bit.  For AeolusShared, synchronized methods are not
     * needed, but they don't pose covert channels because they can
     * only be used to lock the object itself, and not other,
     * arbitrary objects.  Synchronized methods will disqualify an
     * object from being AeolusSafe, however...
     */
    protected void removeSynchronization(MethodNode method) {
	ListIterator iterator = method.instructions.iterator();

	while (iterator.hasNext()) {
	    AbstractInsnNode insn = (AbstractInsnNode)iterator.next();
	    switch (insn.getOpcode()) {
	    case Opcodes.MONITORENTER:
	    case Opcodes.MONITOREXIT:
		/*
		 * The MONITORENTER and MONITOREXIT instructions take an
		 * argument, which we must consume. Popping it from the stack
		 * is easier than finding and removing the code that put it
		 * on the stack in the first place. Hopefully the JIT will
		 * find and eliminate the dead code.
		 */
		iterator.set(new InsnNode(Opcodes.POP));
		break;
	    default:
		break;
	    }
	}
    }

    private boolean shouldTransformMethod(MethodNode method, TransformationType transformationType) {
	// We can't transform static methods.  What authority would they
	// run with?
	if ((method.access & Opcodes.ACC_STATIC) != 0)
	    return (false);

	// We can't transform abstract methods, and there's no need.
	if ((method.access & Opcodes.ACC_ABSTRACT) != 0)
	    return (false);

	// We presume that private methods are only invoked after we have
	// already "entered" the closure through a public method, so there's
	// no need to transform them. Of course, a static method could
	// invoke a private method on an instance that is passed to it, but
	// we already know that static methods don't run with authority.
	//
	// Protected methods can't be handled in the same manner because
	// they can be invoked by other classes in the same package.
	if ((method.access & Opcodes.ACC_PRIVATE) != 0)
	    return (false);

	// We need to transform AeolusShared constructors to
	// ensure that references cannot be leaked via the
	// constructor, i.e., they are copied.  Constructors for
	// AeolusClosure, on the other hand, should not be
	// instrumented, because the constructor simply runs with the
	// authority of the caller.
	//
	// XXX Currently we don't handle AeolusShared constructors,
	// because it's too hard.  Java constructors have too much
	// magic, so our wrapping technique doesn't work for them.
	if (/* !transformationType.shouldTransformConstructor()
	       && */ method.name.equals("<init>"))
	    return (false);

	return (true);
    }
    
    protected void renameAndPrivatizeMethod(MethodNode method, String newName) {
	method.name = newName;
	method.access &= ~(Opcodes.ACC_PUBLIC | Opcodes.ACC_PROTECTED);
	method.access |= Opcodes.ACC_PRIVATE;
    }

    /**
     * Generate a wrapper method with the same signature as the given method.
     * The wrapper does various evil things, then calls the original method
     * by its new name, wrappedMethodName.
     *
     * The wrapper code we generate implements roughly the following
     * Java code:
     *
     * <PRE>
     * RETURNTYPE METHODNAME(ARGS...) throws EXCEPTIONS... {
     *     InfoFlowState state = callStart();
     *     for (Object arg : ARGS)
     *         arg = aeolusClone(arg);
     *     try {
     *         retval = aeolusClone(WRAPPEDMETHOD(ARGS...));
     *     } catch (Throwable t) {
     *         callEnd(state);
     *         throw aeolusClone(t);
     *     }
     *     callEnd(state);   
     *     return (retval);
     * }
     * </PRE>
     *
     * The invocation of the wrapped method is via invokespecial, which,
     * unlike the Java code above, is guaranteed to invoke the method called
     * WRAPPEDMETHOD in this class, and not some subclass.
     *
     * The code for copying arguments is omitted if static analysis shows that
     * the formal parameter type can never refer to an object that can access
     * mutable state directly.  The code is also omitted if the
     * transformationType parameter indicates that we don't need to for this
     * kind of object.
     *
     * Note that the JIT will happily inline and optimize the instrumented
     * method, so the performance penalty for the transformation is
     * virtually zero.  OpenJDK 1.7+ produces much better results than 1.6.
     *
     * The class must provide the following methods:
     * <PRE>
     *   protected final InfoFlowState callStart();
     *   protected final void callEnd();
     *   protected final &lt;T&gt; T aeolusClone(T obj); (if applicable)
     * </PRE>
     */
    protected MethodNode generateWrapper(MethodNode method, ClassNode clazz, String wrappedMethodName, TransformationType transformationType) throws NoSuchMethodException {
	MethodNode wrapperMethod = new MethodNode(method.access, method.name, method.desc, method.signature, (String[])method.exceptions.toArray(new String[0]));
	InsnList instructions = wrapperMethod.instructions;
	LabelNode startLabel = new LabelNode();
	LabelNode endLabel = new LabelNode();

	wrapperMethod.maxLocals = Type.getArgumentTypes(method.desc).length + 4;

	// Three local variables:
	//    $infoFlowState: holds the result of callStart();
	//    $result: holds the result of calling the wrapped method
	//    $exception: stores any exception thrown by the wrapped method
	LocalVariableNode infoFlowStateVar = new LocalVariableNode("$infoFlowState", Type.getDescriptor(InfoFlowState.class), null, startLabel, endLabel, wrapperMethod.maxLocals++);
	if (wrapperMethod.localVariables == null)
	    wrapperMethod.localVariables = new LinkedList<LocalVariableNode>();
	wrapperMethod.localVariables.add(infoFlowStateVar);

	LocalVariableNode resultVar = null;
	if (Type.getReturnType(method.desc) != Type.VOID_TYPE) {
	    resultVar = new LocalVariableNode("$result", Type.getReturnType(method.desc).getDescriptor(), null, startLabel, endLabel, wrapperMethod.maxLocals++);
	    wrapperMethod.localVariables.add(resultVar);
	}
	LocalVariableNode exceptionVar = new LocalVariableNode("$exception", Type.getDescriptor(Throwable.class), null, startLabel, endLabel, wrapperMethod.maxLocals++);
	wrapperMethod.localVariables.add(exceptionVar);

	wrapperMethod.instructions.add(startLabel);

	// Generate code to call callStart().
	wrapperMethod.instructions.add(codeToStartCall(clazz, infoFlowStateVar.index));

	// Copy arguments as needed.
	int argumentIndex = 1;
	if (transformationType.shouldCopyArgs()) {
	    for (Type type : Type.getArgumentTypes(method.desc)) {
		if (needsCopy(type)) {
		    wrapperMethod.instructions.add(codeToClone(clazz, type, argumentIndex));
		    wrapperMethod.instructions.add(new VarInsnNode(Opcodes.ASTORE, argumentIndex));
		}
		argumentIndex++;
	    }
	}

	LabelNode tryStart = new LabelNode();
	wrapperMethod.instructions.add(tryStart);

	// Generate code to call the wrapped method.
	wrapperMethod.instructions.add(new VarInsnNode(Opcodes.ALOAD, 0));
	argumentIndex = 1;
	for (Type type : Type.getArgumentTypes(method.desc))
	    wrapperMethod.instructions.add(new VarInsnNode(type.getOpcode(Opcodes.ILOAD), argumentIndex++));
	wrapperMethod.instructions.add(new MethodInsnNode(Opcodes.INVOKESPECIAL, clazz.name, wrappedMethodName, method.desc));
	if (resultVar != null) {
	    wrapperMethod.instructions.add(new VarInsnNode(Type.getReturnType(method.desc).getOpcode(Opcodes.ISTORE), resultVar.index));
	}
	LabelNode tryEnd = new LabelNode();
	wrapperMethod.instructions.add(tryEnd);

	// Generate code to call callEnd() and return.
	wrapperMethod.instructions.add(codeToEndCall(clazz, infoFlowStateVar.index));
	if (resultVar != null) {
	    if (transformationType.shouldCopyArgs() && needsCopy(Type.getReturnType(method.desc))) {
		wrapperMethod.instructions.add(codeToClone(clazz, Type.getReturnType(method.desc), resultVar.index));
	    } else {
		wrapperMethod.instructions.add(new VarInsnNode(Type.getReturnType(method.desc).getOpcode(Opcodes.ILOAD), resultVar.index));
	    }
	}
	wrapperMethod.instructions.add(new InsnNode(Type.getReturnType(method.desc).getOpcode(Opcodes.IRETURN)));

	// Generate code for the exception handler.
	LabelNode handlerStart = new LabelNode();
	wrapperMethod.instructions.add(handlerStart);
	wrapperMethod.instructions.add(new VarInsnNode(Opcodes.ASTORE, exceptionVar.index));
	wrapperMethod.instructions.add(codeToEndCall(clazz, infoFlowStateVar.index));
	if (transformationType.shouldCopyArgs()) {
	    wrapperMethod.instructions.add(codeToClone(clazz, Type.getType(Throwable.class), exceptionVar.index));
	} else {
	    wrapperMethod.instructions.add(new VarInsnNode(Type.getType(Throwable.class).getOpcode(Opcodes.ILOAD), exceptionVar.index));
	}
	wrapperMethod.instructions.add(new InsnNode(Opcodes.ATHROW));
	wrapperMethod.instructions.add(endLabel);

	wrapperMethod.tryCatchBlocks.add(new TryCatchBlockNode(tryStart, tryEnd, handlerStart, null));

	return (wrapperMethod);
    }

    /**
     * Return the code needed to load a variable of the given type, at
     * the given frame index, and clone it.  The type must be a reference
     * type.  The cloned object is left on top of the stack.
     */
    protected InsnList codeToClone(ClassNode clazz, Type type, int index) {
	InsnList instructions = new InsnList();
	instructions.add(new VarInsnNode(Opcodes.ALOAD, 0));
	instructions.add(new VarInsnNode(type.getOpcode(Opcodes.ILOAD), index));
	instructions.add(new MethodInsnNode(Opcodes.INVOKEVIRTUAL, clazz.name, "aeolusClone", cloneMethodDesc));
	instructions.add(new TypeInsnNode(Opcodes.CHECKCAST, type.getInternalName()));
	return (instructions);
    }

    /**
     * Return the code needed to invoke callStart(). The state object returned
     * by callStart() is stored in the frame at the specified index.
     */
    protected InsnList codeToStartCall(ClassNode clazz, int index) {
	InsnList instructions = new InsnList();
	instructions.add(new VarInsnNode(Opcodes.ALOAD, 0));
	instructions.add(new MethodInsnNode(Opcodes.INVOKEVIRTUAL, clazz.name, "callStart", Type.getMethodDescriptor(Type.getType(InfoFlowState.class), new Type[0])));
	instructions.add(new VarInsnNode(Opcodes.ASTORE, index));
	return (instructions);
    }

    /**
     * Return the code needed to invoke callEnd(). The state object to pass
     * to callEnd() is loaded from the given index.
     */
    protected InsnList codeToEndCall(ClassNode clazz, int index) {
	InsnList instructions = new InsnList();
	instructions.add(new VarInsnNode(Opcodes.ALOAD, 0));
	instructions.add(new VarInsnNode(Opcodes.ALOAD, index));
	instructions.add(new MethodInsnNode(Opcodes.INVOKEVIRTUAL, clazz.name, "callEnd", Type.getMethodDescriptor(Type.VOID_TYPE, new Type[] { Type.getType(InfoFlowState.class) })));
	return (instructions);
    }

    protected boolean needsCopy(Type type) {
	switch (type.getSort()) {
	case Type.ARRAY:
	case Type.OBJECT:
	    break;
	default:
	    System.out.println(type + " not copied: value type");
	    return (false);
	}

	/*
	 * XXX We currently have a problem here that's inconvenient to solve.
	 * The mutability checker uses reflection, which means that it only
	 * works on classes that are already loaded.  However, we also rewrite
	 * methods at load time, and the rewriting depends on knowing which
	 * method formal parameters are mutable.  The current kludge is to
	 * assume that anything that isn't loaded yet is mutable.
	 */
	Class clazz;
	try {
	    clazz = Class.forName(type.getClassName(), false, null);
	} catch (ClassNotFoundException e) {
	    System.out.println(type + " copied: not loaded yet");
	    return (true);
	}

	assert(clazz != null);
	if (AeolusCloner.verifier.isSharableReferenceType(clazz)) {
	    System.out.println(type + " not copied: is sharable");
	    return (false);
	} else {
	    System.out.println(type + " copied: not sharable");
	    return (true);
	}
    }
}

class AeolusClassWriter extends ClassWriter {
    private final ClassLoader loader;

    public AeolusClassWriter(int flags, ClassLoader loader) {
	super(flags);
	this.loader = loader;
    }

    protected String getCommonSuperClass(final String type1,
					 final String type2) {
	try {
	    return super.getCommonSuperClass(type1, type2);
	} catch (Exception e) {}

	ClassInfo c, d;
	try {
	    c = new ClassInfo(type1, loader);
	    d = new ClassInfo(type2, loader);
	} catch (Throwable e) {
	    throw new RuntimeException(e);
	}
	if (c.isAssignableFrom(d)) {
	    return type1;
	}
	if (d.isAssignableFrom(c)) {
	    return type2;
	}
	if (c.isInterface() || d.isInterface()) {
	    return "java/lang/Object";
	} else {
	    do {
		c = c.getSuperclass();
	    } while (!c.isAssignableFrom(d));
	    return c.getType().getInternalName();
	}
    }
}
