package edu.mit.csail.aeolus.security;

import java.util.*;
import java.lang.reflect.*;
import edu.mit.csail.aeolus.api.*;

/**
 * The SharabilityVerifier uses reflection to check which classes are sharable.
 * A sharable class is one that does not invoke stateful behavior, and can
 * therefore be shared among Aeolus threads.  For efficiency, the determination
 * is made (and cached) solely on the class and not on the instance, so an
 * object with a final field of type Object would be considered mutable, even
 * if the field of that particular instance pointed to a String or an Integer.
 *
 * It's frequently useful to know whether a particular reference could ever
 * refer to a mutable object, but subtype polymorphism makes this challenging.
 * Therefore, we introduced the AeolusSafe class, which is a marker interface
 * indicating that a class (and all of its subclasses) are sharable.  This
 * contract is enforced by the verifyClass() method of the SharabilityVerifier
 * at load time.
 *
 * @author das
 */
public class SharabilityVerifier
{

    // A record of which classes are sharable.
    protected HashMap<Class, Boolean> classSharabilityStatus = new HashMap<Class, Boolean>();

    public SharabilityVerifier() {
	addSharable(String.class);
	addSharable(Integer.class);
	addSharable(Long.class);
	addSharable(Boolean.class);
	addSharable(Float.class);
	addSharable(Double.class);
	addSharable(Character.class);
	addSharable(Byte.class);
	addSharable(Short.class);
	addSharable(Object.class);
	addSharable(Class.class);
    }

    /**
     * Return true if we can guarantee that an object of the given type
     * is sharable.
     */
    public boolean isSharableClass(Class clazz) {
	return (isSharable(clazz, true));
    }

    /**
     * Return true if we can guarantee that a reference of the given type
     * always refers to a sharable object (or null).
     */
    public boolean isSharableReferenceType(Class clazz) {
	return (isSharable(clazz, false));
    }

    /**
     * Register an exception, i.e., force the given class to be sharable.
     */
    protected void addSharable(Class clazz) {
	classSharabilityStatus.put(clazz, true);
    }

    /**
     * If mutableSubtypesOkay is false, this method determines whether a
     * <b>reference</b> of the given type always refers to a sharable object.
     * 
     * If mutableSubtypesOkay is false, this method determines whether an
     * <b>object</b> of the given type is sharable.
     */
    protected boolean isSharable(Class clazz, boolean mutableSubtypesOkay) {
	// Base case for recursion up the class hierarchy:
	// java.lang.Object has a null superclass.
	if (clazz == null)
	    return (mutableSubtypesOkay);

	if (clazz.isPrimitive())
	    return (true);

	if (clazz.isArray())
	    return (false);

	if (AeolusSafe.class.isAssignableFrom(clazz))
	    return (true);

	if (mutableSubtypesOkay || Modifier.isFinal(clazz.getModifiers()))
	    return (checkClassSharable(clazz));

	return (false);
    }

    /**
     * This method must be called when a class is loaded to ensure that
     * it satisfies the contract of AeolusSafe, if applicable.
     */
    public void verifyClass(Class clazz) throws ClassCastException {
	if (!AeolusSafe.class.isAssignableFrom(clazz))
	    return;
	if (AeolusShared.class.isAssignableFrom(clazz))
	    return;

	try {
	    if (!checkClassSharable(clazz)) {
		throw new ClassCastException("AeolusSafe class " + clazz + " isn't actually immutable!");
	    }
	} finally {
	    // There's no need to remember that the class is sharable,
	    // since it's AeolusSafe.
	    classSharabilityStatus.remove(clazz);
	}
    }

    private boolean checkClassSharable(Class clazz) {
	assert (!clazz.isArray());

	Boolean sharable = classSharabilityStatus.get(clazz);;
	if (sharable != null)
	    return (sharable);

	/*
	 * We temporarily mark the class as sharable so that any self-
	 * references will be considered okay during the recursive analysis.
	 */
	classSharabilityStatus.put(clazz, true);
	sharable = false;
	try {
	    for (Field field : clazz.getDeclaredFields()) {
		if (!Modifier.isFinal(field.getModifiers()))
		    return (false);
		if (!isSharableReferenceType(field.getType()))
		    return (false);
	    }

	    // A class with synchronized methods isn't sharable because
	    // the object's lock is technically a piece of mutable state.
	    // Persumably a class would only use synchronized methods if
	    // it actually had mutable state, though, so this restriction
	    // shouldn't be problematic!
	    // NB: The ClassTransformer must also ensure that the class has
	    // no MONITORENTER or MONITOREXIT instructions.
	    for (Method method : clazz.getDeclaredMethods()) {
		if (Modifier.isSynchronized(method.getModifiers())) {
		    System.out.println("SharabilityVerifier warning: class " + clazz + " has a synchronized method but no mutable fields");
		    return (false);
		}
	    }

	    if (!isSharableClass(clazz.getSuperclass()))
		return (false);

	    // Non-static inner classes may refer to a different outer class
	    // if the outer class is subclassed.
	    if (clazz.getEnclosingClass() != null) {
		if (Modifier.isStatic(clazz.getModifiers())) {
		    if (!isSharableClass(clazz.getEnclosingClass()))
			return (false);
		} else {
		    if (!isSharableReferenceType(clazz.getEnclosingClass()))
			return (false);
		}
	    }

	    // XXX handle local classes.  (Actually, they might be
	    // handled already if the local class has a reference to
	    // the surrounding class that we will notice, but we need
	    // to make sure this works.)

	    // If we get here, the class passed the check.
	    sharable = true;
	} finally {
	    classSharabilityStatus.put(clazz, sharable);
	}

	return (sharable);
    }
}
