package edu.mit.csail.aeolus.stream.server;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.RemoteServer;
import java.rmi.server.ServerNotActiveException;
import java.rmi.server.UnicastRemoteObject;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;
import java.util.concurrent.TimeUnit;

import edu.mit.csail.aeolus.authority.api.AuthorityConfiguration;
import edu.mit.csail.aeolus.authority.api.InterfaceAuthorityServer;
import edu.mit.csail.aeolus.stream.client.InvalidationPacket;
import edu.mit.csail.aeolus.stream.client.api.IStreamClient;
import edu.mit.csail.aeolus.stream.server.api.InterfaceStreamServer;
import edu.mit.csail.aeolus.stream.server.api.RegisterReturn;
import edu.mit.csail.aeolus.stream.server.api.SnapshotNumAndName;
import edu.mit.csail.aeolus.utilities.Pair;

public class StreamServer extends RemoteServer implements InterfaceStreamServer {
	private static final long serialVersionUID = 1136073118385895057L;

	public final static int CLIENT_CONNECTIONS_TIMEOUT_SECONDS = 5;
	
	final int maxNumStreamClients = 1000;

    Boolean initialized = false;
    Boolean verbose = false;
    List<Pair<String, String>> listOfStreamClients;
    InterfaceAuthorityServer authServer;

    // stream state
    int lastSeqNumStreamed = 0;

    // snapshot state
    private SnapshotNumAndName lastSnapshot;
    int lastSnapshotVersion = -1;
    String lastSnapshotDbName = "";
    int lastSnapshotSeqNum = -1;

	// ----------------- INTERFACE STREAM SERVER ----------------------
    public void initialize(){
        if (initialized) 
        	return;

        // Connect to authUpdater and authDb
        listOfStreamClients = new LinkedList<Pair<String,String>>();

        try{
        	Registry r = java.rmi.registry.LocateRegistry.getRegistry();
        	authServer = (InterfaceAuthorityServer) r.lookup(AuthorityConfiguration.AuthorityServiceName);
			InterfaceStreamServer export = (InterfaceStreamServer) UnicastRemoteObject.exportObject(this, 0);
        	r.rebind(AuthorityConfiguration.StreamServiceName, export);
        }catch(RemoteException e){
        	e.printStackTrace();
        	return;
        } catch (NotBoundException e) {
			e.printStackTrace();
			return;
		}
        
        // Create the initial snapshot
        createSnapshot();
        initialized = true;
    }
    
    
	public RegisterReturn register(
			String clientRMIName) throws RemoteException{
		if (initialized == false){
            throw new RuntimeException("Register() - Server not initialized");
        }
		
		String clientRMIString;
		try {
			clientRMIString = getClientHost();
		} catch (ServerNotActiveException e) {
			e.printStackTrace();
			return null;
		}
		
        if (listOfStreamClients.size() >= maxNumStreamClients) 
            throw new RuntimeException("Too many clients connected.");
        Pair<String, String> p = new Pair<String,String>
        	(clientRMIString, clientRMIName);

		RegisterReturn ret = authServer.registerClient(clientRMIString);
        ret.ss = lastSnapshot;

        if (listOfStreamClients.contains(p)){
        	return ret;
        }
        
        listOfStreamClients.add(p);

        return ret;
    }

    public void deregister(String clientRMIString, String clientRMIName){
        if (listOfStreamClients.remove(
        		new Pair<String,String>(clientRMIString, clientRMIName))){
        	System.out.println("Deregistered client for connection timeout.");
        }else{
        	System.out.println("Deregistered a non-existent client.");
        }
    }

    // StreamNow
    // - Invoked periodically to multicast stream messages.
    // - Creates a snapshot and retires old snapshot at the end of this.
    /**
     * @return lastSeqNum
     */
    public int streamNow(){
        if (initialized == false){
//            ErrWriteLine("StreamNow() - Server not initialized");
        	throw new RuntimeException("streamNow(): server not initialized.");
        }

        // Create a snapshot to mark the end of the stream interval
        createSnapshot();
        
        int startingEntrySeqNum = lastSeqNumStreamed + 1;
        int endingEntrySeqNum = lastSnapshot.getSnapshotSeqNum();
        
        InvalidationPacket invalidations;
		try {
			invalidations = authServer.getInvalidationPacket(startingEntrySeqNum, endingEntrySeqNum);
		} catch (RemoteException e) {
        	throw new RuntimeException(e);
		}

        int count = invalidations.invalidations.size();
        int expected = endingEntrySeqNum - startingEntrySeqNum + 1;

        if (verbose) 
        	System.out.println("StreamNow() from  " + startingEntrySeqNum
        			+ " - " + endingEntrySeqNum + ", " + count + " messages");

        // Sanity check
        if (count != expected){
            throw new RuntimeException("ERROR: Counted " + count + " msg, expected " + expected);
        }

        // Broadcast the stream messages
        broadcastInvalidations(invalidations);
        
        // Update stream state
        lastSeqNumStreamed = endingEntrySeqNum;

        // Clean up old snapshots and truncate log
        try {
			authServer.removeOldSnapshots();
		} catch (RemoteException e) {
        	throw new RuntimeException(e);
		}
        
        return endingEntrySeqNum;
    }

    // ------------------------ PRIVATE --------------------------
    private void createSnapshot(){
        int snapshotVersion = lastSnapshotVersion + 1;

        SnapshotNumAndName snapshot;
		try {
			snapshot = authServer.createSnapshot(snapshotVersion);
		} catch (RemoteException e) {
        	throw new RuntimeException(e);
		}
        // update stream server state
        lastSnapshotVersion = snapshotVersion;
        lastSnapshot = snapshot;
    }

    private void broadcastInvalidations(final InvalidationPacket ip){
    	for (Pair<String,String> clientConn : new LinkedList<Pair<String,String>>(listOfStreamClients)){
    		// inform client
    		final String hostname = clientConn.first;
    		final String serviceName = clientConn.second;
    		FutureTask<Boolean> t = new FutureTask<Boolean>(
    				new Callable<Boolean>() {
    					public Boolean call() {
    						try {
    							Registry registry = LocateRegistry.getRegistry(hostname);
    							final IStreamClient streamClient = (IStreamClient) registry.lookup(serviceName);
    							streamClient.handleInvalidationPacket(ip);
    							return true;
    						} catch (RemoteException e) {
    							return false;
    						} catch (NotBoundException e) {
    							return false;
    						}
    					}
    				});
			Thread thread = new Thread(t);
    		try{
    			thread.start();
    			boolean b = t.get(CLIENT_CONNECTIONS_TIMEOUT_SECONDS, TimeUnit.SECONDS);
    			if(!b)
    				throw new Exception();
    		}catch(Exception e){
    			t.cancel(true);
    			deregister(clientConn.first, clientConn.second);
    		}
    	}
    }

}
