package edu.mit.csail.aeolus.utilities;

import java.io.Serializable;
import java.util.*;
import java.lang.reflect.Array;

/**
 * This is a Set implementation based on sorted arrays. It is faster than a
 * HashSet when there are few elements (e.g., 16 or fewer).
 */
public final class ArraySet<E extends Comparable> implements Set<E>,
		Serializable {
	private ArrayList<E> elts;

	private void init() {
		assert (elts == null);
		elts = new ArrayList<E>(2);
	}

	public ArraySet() {
	}

	public ArraySet(ArraySet<? extends E> set) {
		if (set.elts != null)
			elts = new ArrayList(set.elts);
	}

	public ArraySet(List<? extends E> list) {
		// XXX could be better
		elts = new ArrayList<E>(list.size());
		for (E e : list)
			add(e);
	}

	public boolean add(E e) {
		int i;

		if (elts == null) {
			init();
			elts.add(e);
			return (true);
		}

		for (i = elts.size(); i > 0; i--) {
			int cmp = elts.get(i - 1).compareTo(e);
			if (cmp < 0)
				break;
			if (cmp == 0)
				return (false);
		}

		elts.add(i, e);
		return (true);
	}

	public boolean addAll(Collection<? extends E> c) {
		boolean changed = false;
		for (E e : c) {
			if (add(e))
				changed = true;
		}
		return (changed);
	}

	public void clear() {
		if (elts != null)
			elts.clear();
	}

	public boolean contains(Object o) {
		if (elts == null)
			return (false);

		for (int i = elts.size() - 1; i >= 0; i--) {
			int cmp = elts.get(i).compareTo(o);
			if (cmp < 0){
				return (false);
			}
			if (cmp == 0)
				return (true);
		}
		return (false);
	}

	public boolean containsAll(Collection c) {
		if (elts == null)
			return (c.isEmpty());

		for (Object o : c) {
			if (!contains(o))
				return (false);
		}
		return (true);
	}

	public boolean equals(Object o) {
		if (o == null || !(o instanceof ArraySet))
			return (false);

		ArraySet<E> set = (ArraySet<E>) o;
		if (elts == null)
			return (set.elts == null);
		if(isEmpty())
			return set.isEmpty();
		return (elts.equals(set.elts));
	}

	public boolean isEmpty() {
		if (elts == null)
			return (true);
		else
			return (elts.isEmpty());
	}

	public Iterator<E> iterator() {
		if (elts == null)
			return ((new LinkedList()).iterator()); // laziness
		return (elts.iterator());
	}

	public boolean remove(Object o) {
		if (elts == null)
			return (false);
		return (elts.remove(o));
	}

	public boolean removeAll(Collection c) {
		boolean changed = false;
		for (Object o : c) {
			if (remove(o))
				changed = true;
		}
		return (changed);
	}

	public boolean retainAll(Collection c) {
		if (elts == null)
			return (false);

		// XXX could be much better
		boolean changed = false;
		for (int i = 0; i < size();) {
			if (!c.contains(elts.get(i))) {
				remove(elts.get(i));
				changed = true;
			} else {
				i++;
			}
		}
		return (changed);
	}

	public int size() {
		if (elts == null)
			return (0);

		return (elts.size());
	}

	public Object[] toArray() {
		if (elts == null)
			return (new Object[0]);
		return (elts.toArray());
	}

	public <T> T[] toArray(T[] a) {
		if (elts == null)
			return ((T[]) Array.newInstance(a.getClass().getComponentType(), 0));
		return (elts.toArray(a));
	}

	public String toString() {
		if (elts == null)
			return "[]";
		return (elts.toString());
	}

}
